package provide DS9 1.0

# capture general errors
# this only captures gui errors, not xpa errors

proc bgerror {err} {
    global cursor

    tk_messageBox -message "An internal error has been detected: $err" \
	-type ok -icon error

    # reset cursor
    set cursor(count) 0
    UnsetWatchCursor
}

# unwind xpa errors
# requires catch {} to allow a check to take place

proc InitXPAError {} {
    global errorInfo

    set errorInfo {}
}

proc CatchXPAError {xpa} {
    global errorInfo

    if {$errorInfo != {}} {
	# reset cursor
	set cursor(count) 0
	UnsetWatchCursor

	xpaerror $xpa "$errorInfo"
	set errorInfo {}
    }
}

# force capture xpa/interactive errors

proc Error {message} {
    global ds9

    if {$ds9(xpa) == 0} {
	tk_messageBox -message $message -type ok -icon error
    } else {
	xpaerror $ds9(xpa) $message
    }

    # reset cursor
    set cursor(count) 0
    UnsetWatchCursor
}

proc Debug {which varname} {
    upvar $varname var
    global current

    if {$current(frame) != ""} {
	$current(frame) debug $which $var
    }
}

proc ToYesNo {value} {
    if {$value == 1} {
	return "yes\n"
    } else {
	return "no\n"
    }
}

proc FromYesNo {value} {
    set v [string tolower $value]

    if {$v == "no" || $v == "false" || $v == "off" || $v == 0} {
	return 0
    } else {
	return 1
    }
}

proc RealizeDS9 {} {
    global current

    # make sure the frame is realized
    # this looks funny, but UpdateGraphXAxis causes a
    # series of events that we need
    UpdateGraphXAxis
    $current(frame) update now
}

proc Sex2D {str} {
    scan $str "%d:%d:%f" h m s
    return [expr $h+($m/60.)+($s/60./60.)]
}

proc SetCursor {cursor} {
    global ds9
    global iis
    global current
    global canvas

    # if init phase, don't change cursor
    if {$ds9(init)} {
	return
    }

    # if iis cursor mode, don't change cursor
    if {$iis(state)} {
	return
    }

    if {$ds9(cursor) && ($current(cursor) != $cursor)} {
	set current(cursor) $cursor

	if {$cursor != ""} {
	    $canvas(name) configure -cursor $cursor
	} else {
	    $canvas(name) configure -cursor {}
	}
    }
}

# NOTE
# This is where everything gets done. In particular, <Configure> events are
# processed here. Also, any screen updates are done. In short, if it changes
# the geometry or changes the contents, you need to call Set/UnsetWatchCursor

proc SetWatchCursor {} {
    global ds9
    global cursor
    global canvas

    global debug
    if {$debug(tcl,watch)} {
	puts "SetWatchCursor Start $cursor(count)"
    }

    # if init phase, don't change cursor

    if {$ds9(init)} {
	return
    }

    if {$cursor(count) == 0} {
	set cursor(save) [$canvas(name) cget -cursor]
	$canvas(name) configure -cursor {}
	.ds9 configure -cursor watch
	update idletasks
    }
    incr cursor(count)

    if {$debug(tcl,watch)} {
	puts "SetWatchCursor End $cursor(count)"
    }
}

proc UnsetWatchCursor {} {
    global ds9
    global cursor
    global canvas

    global debug
    if {$debug(tcl,watch)} {
	puts "UnsetWatchCursor Start $cursor(count)"
    }

    # if init phase, don't change cursor

    if {$ds9(init)} {
	return
    }

    if {$cursor(count)>0} {
	incr cursor(count) -1
    }

    if {$cursor(count) == 0} {
	.ds9 configure -cursor {}
	$canvas(name) configure -cursor $cursor(save)
	update idletasks
    }    

    if {$debug(tcl,watch)} {
	puts "UnsetWatchCursor End $cursor(count)"
    }
}

proc AboutBox {} {
    global about
    global ed

    set w ".about"
    set width 340
    set height 320

    set ed(ok) 0

    DialogCreate $w "About SAOImage DS9" -borderwidth 2
    frame $w.c -relief groove -borderwidth 2
    frame $w.buttons -relief groove -borderwidth 2
    pack $w.c $w.buttons -fill x -ipadx 4 -ipady 4

    canvas $w.c.c -background white -height $height -width $width
    pack $w.c.c
    
    set ed(sun) [image create photo -format gif -file sun.gif]

    $w.c.c create image 0 0 -image $ed(sun) -anchor nw
    $w.c.c create text 140 20 -text $about(message) -anchor nw

    button $w.buttons.ok -text "OK" -default active -command {set ed(ok) 1}
    grid $w.buttons.ok -padx 10

    bind $w <Return> {set ed(ok) 1}

    DialogCenter $w
    DialogWait $w ed(ok)
    DialogDismiss $w

    if {$ed(ok)} {
	image delete $ed(sun)
    }

    unset ed
}

proc QuitDS9 {} {
    catch {IISClose}
    focus {}
    exit
}

proc OpenSource {} {
    set filename [OpenFileDialog tclfbox]
    if {$filename != {}} {
	uplevel #0 "source \{$filename\}"
    }
}

proc UpdateView {} {
    global debug
    if {$debug(tcl,layout)} {
	puts "UpdateView"
    }

    SetWatchCursor
    LayoutView
    UnsetWatchCursor
}

proc LayoutView {} {
    global current
    global panel
    global info
    global pan
    global mag
    global buttons
    global canvas
    global colorbar
    global graph
    global graphHorz
    global graphVert
    global ds9

    global debug
    if {$debug(tcl,layout)} {
	puts "LayoutView"
    }

    grid forget $panel(name)
    pack forget $info(name)
    pack forget $pan(name)
    pack forget $mag(name)
    grid forget $buttons(name)
    grid forget $canvas(name)
    grid forget $graph(spacer)
    grid forget $graphVert(name)
    grid forget $graphHorz(name)
    grid forget $colorbar(name)

    if {$info(view) || $mag(view) || $pan(view)} {
	grid $panel(name) -row 0 -sticky ew -columnspan 3
    }

    if {$info(view)} {
	pack $info(name) -side left -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
    }

    if {$mag(view)} {
	pack $mag(name) -side right -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
    }

    if {$pan(view)} {
	pack $pan(name) -side right -padx $ds9(internalpad) \
	    -pady $ds9(internalpad)
    }

    if {$buttons(view)} {
	grid $buttons(name) -row 1 -sticky ew -columnspan 3
    }

    if {$canvas(auto)} {
	grid $canvas(name) -row 2 -sticky news
    } else {
	$canvas(name) configure -width $canvas(size) -height $canvas(size)
	grid $canvas(name) -row 2
    }

    if {$graphVert(view)} {
	switch $ds9(displaymode) {
	    blink -
	    single {grid $graphVert(name) -row 2 -column 2 -rowspan 2 \
			-sticky ns}
	    tile {}
	}
    }

    if {$graphHorz(view) || $graphVert(view)} {
	switch $ds9(displaymode) {
	    blink -
	    single {grid $graph(spacer) -row 3 -column 1}
	    tile {}
	}
    }

    if {$graphHorz(view)} {
	switch $ds9(displaymode) {
	    blink -
	    single {grid $graphHorz(name) -row 4 -columnspan 2 -sticky ew}
	    tile {}
	}
    }

    if {$colorbar(view)} {
	grid $colorbar(name) -row 5 -sticky ew
    }
}

proc ToggleFreeze {} {
    global ds9

    if {$ds9(freeze)} {
	set ds9(freeze) 0
	UnFreezeCanvas
	UnFreezePanner
    } else {
	set ds9(freeze) 1
	FreezeCanvas
	FreezePanner
    }
}

proc ChangeMode {} {
    global ds9
    global canvas

    bind $canvas(name) <Button-1> {}
    bind $canvas(name) <B1-Motion> {}
    bind $canvas(name) <ButtonRelease-1> {}

    foreach f $ds9(frames) {
	$f crosshair off
    }
    UpdateMarkerMenu

    UnselectAllMarkers
    ClearGraphData
    ClearInfoBox
    ClearPixelTable

    switch $ds9(mode) {
	pointer {SetCursor {}}
	crosshair {
	    foreach f $ds9(frames) {
		$f crosshair on
	    }
	    ShowGraphData
	    SetCursor crosshair
	}
	colorbar {
	    bind $canvas(name) <Button-1> {ButtonColormap %x %y}
	    bind $canvas(name) <B1-Motion> {MotionColormap %x %y}
	    bind $canvas(name) <ButtonRelease-1> {ReleaseColormap %x %y}
	    SetCursor center_ptr
	}
	zoom {SetCursor sizing}
	pan {SetCursor fleur}
	rotate {SetCursor exchange}
	examine {SetCursor target}
	illustrate {SetCursor arrow}
    }
}

proc HTTPLog {token} {
    global http

    if {$http(log)} {
	upvar 0 $token t
	DisplayLog "$t(http)\n"
	DisplayLog "$t(type)\n"
	DisplayLog "$t(totalsize)\n"
	DisplayLog [BreakUp $t(meta)]
	DisplayLog "$t(status)\n"
    }
}

proc InPath {which} {
    global env
    global tcl_platform

    if {$tcl_platform(platform) != "windows"} {
	set target ${which}
	set paths [split $env(PATH) :]
    } else {
	set target ${which}.exe
	set paths [split $env(PATH) \;]
    }

    foreach p $paths {
	if {[file executable [file join $p $target]]} {
	    return 1
	}
    }
    return 0
}
