/*
 *
 * filter.c -- hi-level management of filters
 *
 */

/*
 *      Copyrights:
 *
 *    	      Copyright (c) 1996 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <sys/types.h>
#include <file.h>
#include <find.h>
#include <macro.h>
#include <word.h>
#include <column.h>
#include <filter.h>
#include <dl.h>
#include <zprocess.h>

/*
 *
 * private routines
 *
 */

/* this is needed by lex */
static Filter default_filter = NULL;
static Filter null_filter = NULL;

#ifdef ANSI_FUNC
static void
FiltInitRegions(void)
#else
static void FiltInitRegions()
#endif
{
  initevregions();
  initimregions();
}

/*
 *
 * public routines
 *
 */

/*
 *
 * FilterClip -- save the filter without the enclosing brackets
 *
 */
#ifdef ANSI_FUNC
char *
FilterClip(char *filter)
#else
char *FilterClip(filter)
     char *filter;
#endif
{
  char *tbuf;
  char *tptr;
  char *ofilter;
  int len;
  int i;

  /* check for the obvious */
  if( !filter || (*filter == '\0') )
    return(NULL);

  /* save filter, but without the enclosing brackets */
  tbuf = NewString(filter);
  nowhite(tbuf, tbuf);
  tptr = tbuf;
  if( *tbuf == '[' ){
    tptr++;
    len = strlen(tptr);
    for(i=len-1; i>=0; i-- ){
      if( tptr[i] == ']' ){
	tptr[i] = '\0';
	break;
      }
    }
  }
  if( tptr && (*tptr != '\0') ){
    /* we extend the buffer because flex has problems in filt_init_buffer 
       if its smaller than sizeof(FILE *) */
    ofilter = calloc(strlen(tptr)+SZ_LINE, sizeof(char));
    nowhite(tptr, ofilter);
  }
  else{
    ofilter = NULL;
  }
  free(tbuf);
  return(ofilter);
}

/*
 *
 * FilterDefault -- get the default filter handle for the parser
 *
 */
#ifdef ANSI_FUNC
Filter
FilterDefault(void)
#else
Filter FilterDefault()
#endif
{
  return(default_filter);
}

/*
 *
 * FilterNull -- get the null filter pointer
 *
 */
#ifdef ANSI_FUNC
Filter
FilterNull(void)
#else
Filter FilterNull()
#endif
{
  /* allocate the null filter record, if necessary */
  if( !null_filter )
    null_filter = (Filter)calloc(1, sizeof(FilterRec));
  return(null_filter);
}

/*
 *
 * FilterNum -- get a unique number associated with openfilter
 *
 */
static int filter_num=0;
#ifdef ANSI_FUNC
int
FilterNum(void)
#else
int FilterNum()
#endif
{
  filter_num++;
  return(filter_num);
}

/*
 *
 * FilterPath -- get the path through which we look for things
 *
 */
#ifdef ANSI_FUNC
char *
FilterPath(void)
#else
char *FilterPath()
#endif
{
  static char *filter_path=NULL;
  if( filter_path == NULL ){
    if( (filter_path = (char *)getenv("path")) == NULL )
      filter_path = (char *)getenv("PATH");
  }
  return(filter_path);
}

/*
 *
 * FilterOpen -- open a filter and initialize the symbol table
 *
 */
#ifdef ANSI_FUNC
Filter
FilterOpen(FITSHead header, char *string, char *mode)
#else
Filter FilterOpen(header, string, mode)
     FITSHead header;		/* fitsy header struct */
     char *string;		/* filter string */
     char *mode;		/* "C", "f77", "awk", etc. */
#endif
{
  char *s;
  char tbuf[SZ_LINE];
  char *argv[3];
  Filter filter;

  /* return if we have no filter */
  if( (string == NULL) || (*string == '\0') ||
	!strcmp(string, "[]") || !strcmp(string, "\n") )
    return(NOFILTER);

  /* allocate a new filter record */
  if( (filter = (Filter)calloc(1, sizeof(FilterRec))) == NULL )
    return(NULL);

  /* this is the default filter handle for the parser */
  default_filter = filter;

  /* replace file with contents, if possible */
  if( (*string != '@') || !(s = FileContents(&(string[1]), 0, NULL)) )
    s = NewString(string);
  /* set the input filter string for the parser */
  filter->string = FilterClip(s);
  if( s ) free(s);
  /* return if we have no filter */
  if( !filter->string || (*filter->string == '\0') )
    goto error;

  /* save mode */
  filter->mode=NewString(mode);

  /* process the mode string */
  if( (s = NewString(filter->mode)) ){
    /* set global convert flag */
    if( keyword(s, "convert", tbuf, SZ_LINE) )
      filter->convert = istrue(tbuf);
    /* set the event section */
    if( keyword(s, "evsect", tbuf, SZ_LINE) )
      filter->evsect = NewString(tbuf);
    if( keyword(s, "debug", tbuf, SZ_LINE) )
      filter->debug = istrue(tbuf);
  }
  if( s ) free(s);

  /* determine which method of program we use */
  filter->method = DEFAULT_FILTER_METHOD;
  *tbuf = '\0';
  if( (s=(char *)getenv("FILTER_METHOD")) ){
    strcpy(tbuf, s);
  }
  if( !*tbuf && (s=NewString(filter->mode)) ){
    keyword(s, "method", tbuf, SZ_LINE);
    if( s ) free(s);
  }
  if( *tbuf ){
    if( !strcasecmp(tbuf, "c") )
      filter->method = METHOD_C;
  }

  /* determine which type of filtering we do */
  filter->type = DEFAULT_FILTER_TYPE;
  *tbuf = '\0';
  if( (s=(char *)getenv("FILTER_TYPE")) ){
    strcpy(tbuf, s);
  }
  if( !*tbuf && (s=NewString(filter->mode)) ){
    keyword(s, "type", tbuf, SZ_LINE);
    if( s ) free(s);
  }
  if( *tbuf ){
    if( !strncasecmp(tbuf, "ev", 2) )
      filter->type = TYPE_EVENTS;
    else if( !strncasecmp(tbuf, "im", 2) )
      filter->type = TYPE_IMAGE;
  }

  /* determine which type of process execution we do */
  filter->ptype = DEFAULT_FILTER_PTYPE;
  *tbuf = '\0';
  if( (s=(char *)getenv("FILTER_PTYPE")) ){
    strcpy(tbuf, s);
  }
  if( !*tbuf && (s=NewString(filter->mode)) ){
    keyword(s, "ptype", tbuf, SZ_LINE);
    if( s ) free(s);
  }
  if( *tbuf ){
    if( *tbuf == 'p' )
      filter->ptype = PTYPE_PROCESS;
    else if( *tbuf == 'c' )
      filter->ptype = PTYPE_CONTAINED;
#ifdef USE_DL
    else if( *tbuf == 'd' )
      filter->ptype = PTYPE_DYNAMIC;
#endif
  }

  /* method-specific processing */
  switch(filter->method){
  case METHOD_C:
    if( ((filter->cc = NewString(getenv("FILTER_CC"))) == NULL) 	&&
	((filter->cc = NewString(getenv("CC"))) == NULL) 		&&
	((filter->cc = Find("gcc", "x", NULL, FilterPath())) == NULL)   &&
	((filter->cc = Find("cc", "x", NULL, ".")) == NULL)  		&&
	((filter->cc = Find("cc", "x", NULL, CC_PATH)) == NULL)         ){
      fprintf(stderr, "ERROR: no compiler for filter compilation\n");
      goto error;
    }
    /* if we wanted dynamic but had no gcc, use process */
    if( (filter->ptype == PTYPE_DYNAMIC) && !strstr(filter->cc, "gcc") )
      filter->ptype = PTYPE_PROCESS;
    /* get default file names */
    sprintf(tbuf, "$FILTER_OBJDIR:$FILTER_LIBDIR:%s:$HOME/.funtools:$SAORD_ROOT/lib:/usr/lib:/usr/local/lib:/opt/local/lib", OBJPATH);
    switch( filter->type ){
    case TYPE_EVENTS:
      /* normally, we filter events analytically using evregions.o */
      if( !filter->evsect )
	filter->objs = Find("evregions.o", "r", NULL, tbuf);
      /* if evsect=xxx is specified, we filter by image pixels */
      else
	filter->objs = Find("imregions.o", "r", NULL, tbuf);
      break;
    case TYPE_IMAGE:
      /* image are filtered by image pixels */
      filter->objs = Find("imregions.o", "r", NULL, tbuf);
      break;
    }
    /* if we wanted to a process, but have no objects, use self-contained */
    if( (filter->ptype == PTYPE_PROCESS) && !filter->objs )
      filter->ptype = PTYPE_CONTAINED;
    /* allow extra stuff on the command line */
    if( !(filter->extra = NewString(getenv("FILTER_EXTRA"))) )
      filter->extra = NewString(" ");
    /* set shared switches for gcc */
    if( strstr(filter->cc, "gcc") )
      filter->shflags = NewString(GCC_SHARED_FLAGS);
    break;
  default:
    break;
  }

  /* save the fits header */
  filter->fhd = header;

  /* init the symbol table */
  if( !FilterSymbolInit(filter) )
    goto error;

  /* init parser variables */
  FiltInitParser();

  /* init regions */
  FiltInitRegions();

  /* open the program file that we will compile from the filter */
  if( !FilterProgOpen(filter) )
    goto error;

  /* set up to parse the input filter */
  FiltScanString(filter->string);

  /* parse filter and generate the filter-specific part of the C program */
  filtlex();

  /* check for errors while parsing the string */
  if( FilterParseError() )
    goto error;

  /* make sure we have something to filter */
  if( !_FilterString() ){
    FilterClose(filter);
    return(NOFILTER);
  }

  /* write the symbols to the appropriate program */
  if( !FilterProgWrite(filter) )
    goto error;

  /* append the program body to C program */
  if( !FilterProgAppend(filter) )
    goto error;

  /* that is all we need to output */
  if( !FilterProgClose(filter) )
    goto error;

  /* compile the filter program */
  if( !FilterProgCompile(filter) )
    goto error;

  if( !filter->debug ){
    switch(filter->ptype){
    case PTYPE_PROCESS:
    case PTYPE_CONTAINED:
      /* now start the filter process we just readied */
      argv[0] = filter->prog;
      argv[1] = filter->args;
      argv[2] = NULL;
      if( !ProcessOpen(filter->prog, argv, 
		       &(filter->ichan), &(filter->ochan), &(filter->pid)) )
	goto error;
      break;
#ifdef USE_DL
    case PTYPE_DYNAMIC:
     if( !(filter->dl=DLOpen(filter->prog)) )
       goto error;
     break;
#endif
    default:
      goto error;
    }
  }

  /* return the good news */
  return(filter);

error:
  FilterClose(filter);
  return(NULL);

}

/*
 *
 * FilterEvents -- filter events by sending them to a co-process,
 * which returns a char array of good event flags
 *
 */
#ifdef ANSI_FUNC
int
FilterEvents(Filter filter, char *ebuf, int esize, int n, int *vbuf)
#else
int FilterEvents(filter, ebuf, esize, n, vbuf)
     Filter filter;
     char *ebuf;
     int esize;
     int n;
     int *vbuf;
#endif
{
  int i;
  int get, got, flag;
  char *eptr;
  char *etop;
  char *obuf;
  char *optr;
#ifdef USE_DL
  FilterTableCall evrtn;
#endif
  FilterSymbols sp;

  /* make sure we have a valid filter and a valid table */
  if( !filter || (filter == NOFILTER) || !filter->fhd || !filter->fhd->table )
    return(0);

#ifdef LATER
  /* if its just field, we realy don't need to do much */
  if( !strcasecmp(filter->string, "field()") ){
    for(i=0; i<n; i++)
      vbuf[i] = 1;
    return 1;
  }
#endif

  /* we only send the necessary parts of each event record */
  if( (obuf = (char *)calloc(n*filter->evsize, sizeof(char))) == NULL )
    return(0);
  optr = obuf;

  /* extract (and convert) the necessary columns of the event */
  for(eptr=ebuf, etop=ebuf+(n*esize); eptr<etop; eptr += esize){
    /* loop through the symbol table and process columns */
    for(i=0; i<filter->nsyms; i++){
      sp = &(filter->symtab[i]);
      /* skip accidentally empty ones */
      if( (sp->name == NULL) || (*sp->name == '\0') )
	continue;
      /* process this type of symbol */
      switch(sp->type){
      case SYM_COL:
	ColumnLoad(eptr+filter->fhd->table->col[sp->idx].offset,
		   filter->fhd->table->col[sp->idx].size,
		   filter->fhd->table->col[sp->idx].n,
		   filter->convert, optr+sp->offset);
	break;
      default:
	break;
      }
    }
    /* bump to next output record */
    optr += filter->evsize;
  }

  switch(filter->ptype){
  case PTYPE_PROCESS:
  case PTYPE_CONTAINED:
    /* write events to the filter process */
    ProcessWrite(filter->ochan, obuf, n*filter->evsize);
    /* read back result flags */
    get = n * sizeof(unsigned int);
    ProcessRead(filter->ichan, vbuf, get, &got);
    flag = (get==got);
    break;
#ifdef USE_DL
  case PTYPE_DYNAMIC:
    if( (evrtn=(FilterTableCall)DLSym(filter->dl, filter->pname)) ){
      filter->g = (*evrtn)(filter->g, obuf, n, filter->evsize, vbuf);
      flag = 1;
    }
    else
      flag = -1;
    break;
#endif
  default:
    flag = -1;
    break;
  }

  /* free up malloc'ed space */
  if( obuf ) free(obuf);

  /* return the news */
  return flag;
}	

/*
 *
 * FilterImage -- filter image by sending sections to a co-process,
 * which returns segments of valid image data
 *
 */
#ifdef ANSI_FUNC
int
FilterImage(Filter filter, int x0, int x1, int y0, int y1, int block,
	    FilterMask *masks, int *nreg)
#else
int FilterImage(filter, x0, x1, y0, y1, block, masks, nreg)
     Filter filter;
     int x0, x1, y0, y1, block;
     FilterMask *masks;
     int *nreg;
#endif
{
  char tbuf[SZ_LINE];
  int got;
#ifdef USE_DL
  FilterImageCall imrtn;
#endif

  /* make sure we have a valid filter */
  if( !filter || (filter == NOFILTER) )
    return(0);

  switch(filter->ptype){
  case PTYPE_PROCESS:
  case PTYPE_CONTAINED:
    /* write command to process */
    sprintf(tbuf, "%d %d %d %d %d\n", x0, x1, y0, y1, block);
    ProcessWrite(filter->ochan, tbuf, strlen(tbuf));
    /* read back mask records */
    *masks = ProcessRead(filter->ichan, NULL, -1, &got);
    got /=  sizeof(FilterMaskRec);
    break;
#ifdef USE_DL
  case PTYPE_DYNAMIC:
    if( (imrtn=(FilterImageCall)DLSym(filter->dl, filter->pname)) ){
      *masks = (*imrtn)(x0, x1, y0, y1, block, &got);
    }
    else
      return(-1);
    break;
#endif
  default:
    return(-1);
  }

  /* how many separate regions? */
  if( nreg ) *nreg = FilterRegionCount(TOK_IREG);

  /* return the news */
  return(got);
}

/*
 *
 * FilterClose -- close a filter and free up memory
 *
 */
#ifdef ANSI_FUNC
int FilterClose(Filter filter)
#else
int FilterClose(filter)
     Filter filter;
#endif
{
  int status=0;
  int i, j;
  Scan scan, tscan;

  if( !filter || (filter == NOFILTER) )
    return(0);

  switch(filter->ptype){
  case PTYPE_PROCESS:
  case PTYPE_CONTAINED:
    /* close the filter process */
    if( filter->pid > 0 )
      ProcessClose(filter->pid, &status);
    break;
#ifdef USE_DL
  case PTYPE_DYNAMIC:
    if( filter->dl )
      DLClose(filter->dl);
    break;
#endif
  default:
    break;
  }

  /* call any method-specific cleanup routines */
  FilterProgClose(filter);
  FilterProgEnd(filter);

  /* free alloc'ed memory */
  FilterSymbolFree(filter);
  if( filter->pname )   free(filter->pname);
  if( filter->mode )    free(filter->mode);
  if( filter->cc )      free(filter->cc);
  if( filter->objs )    free(filter->objs);
  if( filter->extra )   free(filter->extra);
  if( filter->shflags ) free(filter->shflags);
  if( filter->code )    free(filter->code);
  if( filter->prog )    free(filter->prog);
  if( filter->args )    free(filter->args);
  if( filter->string )  free(filter->string);
  if( filter->xbin )    free(filter->xbin);
  if( filter->ybin )    free(filter->ybin);
  if( filter->symtab )  free(filter->symtab);
  if( filter->evsect )  free(filter->evsect);
  /* free region information */
  if( filter->g ){
    for(i=0; i<filter->g->maxshapes; i++){
      if( filter->g->shapes[i].scanlist ){
	for(j=0; j<=filter->g->ymax; j++){
	  if( filter->g->shapes[i].scanlist[j] ){
	    for(scan=filter->g->shapes[i].scanlist[j]; scan; ){
	      tscan = scan->next;
	      free(scan);
	      scan = tscan;
	    }
	  }
	}
	free(filter->g->shapes[i].scanlist);
      }
    }
    if( filter->g->shapes ) free(filter->g->shapes);
    if( filter->g->ybuf )   free(filter->g->ybuf);
    if( filter->g->x0 )     free(filter->g->x0);
    if( filter->g->x1 )     free(filter->g->x1);
    if( filter->g )         free(filter->g);
  }
  free(filter);
  /* return the status from closing the sub-process */
  return(status);
}


/*
 *
 * FilterConcats -- concat n filters, removing enclosing brackets as necessary
 *
 */
#ifdef ANSI_FUNC
char *
FilterConcats(char **filters, int n)
#else
char *FilterConcats(filters, n)
  char **filters;
  int n;
#endif
{
  char *ofilter;
  char *optr;
  char *clip;
  int i;
  int len;

  /* need at least 1 string */
  if( n == 0 )
    return(NULL);

  /* get max length of output filter */
  for(len=0, i=0; i<n; i++){
    if( filters[i] && (*filters[i] != '\0') ){
      len += strlen(filters[i]);
    }
  }

  /* need something to return */
  if( len == 0 )
    return(NULL);

  /* allocate space for the output filter */
  ofilter = (char *)calloc(len+1, sizeof(char));
  optr = ofilter;

  /* take each input filter */
  for(i=0; i<n; i++){
    if( filters[i] && (*filters[i] != '\0') ){
      /* clip off the brackets */
      clip = FilterClip(filters[i]);
      /* copy the clipped filter to the output */
      if( clip && (*clip != '\0') ){
	strcpy(optr, clip);
	free(clip);
	while( *optr )
	  optr++;
      }
    }
  }

  /* this is the concat'ed filter */
  return(ofilter);
}

/*
 *
 * FilterConcat -- concat 2 filters, with an optional operation in between
 *
 */
#ifdef ANSI_FUNC
char *
FilterConcat(char *f1, char *f2, char *op)
#else
char *FilterConcat(f1, f2, op)
     char *f1;
     char *f2;
     char *op;
#endif
{
  char *nf1, *nf2, *nop;
  char *result;
  char *filters[4];
  int n=0;

  /* clip the filters and look for null extensions */
  nf1 = FilterClip(f1);
  nf2 = FilterClip(f2);
  nop = FilterClip(op);

  if( nf1 && (*nf1 != '\0') ){
    filters[n++] = nf1;
  }
  if( nop && (*nop != '\0') &&
      nf1 && (*nf1 != '\0') &&
      nf2 && (*nf2 != '\0') ){
    filters[n++] = nop;
  }
  if( nf2 && (*nf2 != '\0') ){
    filters[n++] = nf2;
  }

  /* get the concated filters */
  result = FilterConcats(filters, n);

  /* free up space */
  if( nf1 )
    free(nf1);
  if( nf2 )
    free(nf2);
  if( nop )
    free(nop);

  /* return result */
  return(result);
}

/*
 *
 * FilterString -- return the filter string used for this filter
 *
 */
#ifdef ANSI_FUNC
char *
FilterString(Filter filter)
#else
char *FilterString(filter)
     Filter filter;
#endif
{
  if( !filter || (filter == NOFILTER) )
    return NULL;
  else
    return filter->string;
}
