#include "map.h"
#include "head.h"
#include "util.h"

FitsMap::FitsMap()
{
  mapdata_ = NULL;
  mapsize_ = 0;
}

char* FitsMap::enddata()
{
  // return the address of the first data byte pass the end of data
  return head_ ? (char*)data_+head_->databytes() : NULL;
}

int FitsMap::endsize()
{
  // return the size - (header+data)
  return head_ ? 
    (mapsize_ - (int)((char*)data_+head_->databytes()-mapdata_)) : 0;
}

FitsFitsMap::FitsFitsMap(ScanMode mode)
{
  if (!valid_)
    return;

  if (mode == EXACT || pExt_ || pIndex_)
    processExact();
  else
    processRelax();
}

void FitsFitsMap::processExact()
{
  // find head and data for specified unit

  char* here = mapdata_;
  int size = mapsize_;

  // simple check for fits file

  if (strncmp(mapdata_,"SIMPLE  ",8) && strncmp(mapdata_,"XTENSION",8))
    goto error;

  if (!(pExt_ || (pIndex_>0))) {

    // we are only looking for a primary image

    head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
    if (head_->isValid()) {
      goto found;
    }
  }
  else {

    // we are looking for an extension
    // keep the primary header

    primary_ = new FitsHead(here, size, FitsHead::EXTERNAL);
    managePrimary_ = 1;
    if (!primary_->isValid())
      goto error;
    here += primary_->headbytes() + primary_->databytes();
    size -= primary_->headbytes() + primary_->databytes();

    if (pExt_) {
      while (size > 0) {
	head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
	if (!head_->isValid())
	  goto error;
	ext_++;

	if (head_->extname()) {
	  char* a = toUpper(head_->extname());
	  char* b = toUpper(pExt_);
	  if (!strcmp(a,b)) {
	    delete [] a;
	    delete [] b;
	    goto found;
	  }
	  delete [] a;
	  delete [] b;
	}

	here += head_->headbytes() + head_->databytes();
	size -= head_->headbytes() + head_->databytes();
	delete head_;
	head_ = NULL;
      }
    }
    else {
      for (int i=1; i<pIndex_ && size>0; i++) {
	head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
	if (!head_->isValid())
	  goto error;
	ext_++;

	here += head_->headbytes() + head_->databytes();
	size -= head_->headbytes() + head_->databytes();
	delete head_;
	head_ = NULL;
      }

      head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
      if (head_->isValid()) {
	ext_++;
	goto found;
      }
    }
  }

  // Must have an error

error:
  if (head_)
    delete head_;
  head_ = NULL;
  if (primary_ && managePrimary_)
    delete primary_;
  primary_ = NULL;

  data_ = NULL;
  valid_ = 0;
  return;

  // found it!

found:
  data_ = here + head_->headbytes();
  valid_ = 1;
  return;
}

void FitsFitsMap::processRelax()
{
  char* here = mapdata_;
  int size = mapsize_;

  // simple check for fits file

  if (strncmp(mapdata_,"SIMPLE  ",8) && strncmp(mapdata_,"XTENSION",8))
    goto error;

  // check to see if there is an image in the primary

  head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
  if (head_->isValid() && 
      head_->naxes() > 0 && 
      head_->naxis(0) > 0 && 
      head_->naxis(1) > 0)
    goto found;

  // ok, no image, save primary and lets check extensions

  here += head_->headbytes() + head_->databytes();
  size -= head_->headbytes() + head_->databytes();
  primary_ = head_;
  managePrimary_ = 1;
  head_ = NULL;

  while (size > 0) {
    head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
    if (!head_->isValid())
      goto error;
    ext_++;

    // check for image

    if (head_->isImage() && 
	head_->naxes() > 0 && 
	head_->naxis(0) > 0 && 
	head_->naxis(1) > 0)
      goto found;

    // else, check for bin table named STDEVT, EVENTS, RAYEVENT

    if (head_->isTable() && head_->extname()) {
      char* a = toUpper(head_->extname());
      if (!strcmp("STDEVT", a) ||
	  !strcmp("EVENTS", a) ||
	  !strcmp("RAYEVENT", a)) {
	delete [] a;
	goto found;
      }
      else
	delete [] a;
    }

    here += head_->headbytes() + head_->databytes();
    size -= head_->headbytes() + head_->databytes();
    delete head_;
    head_ = NULL;
  }

  // did not find anything, bail out

 error:
  if (head_)
    delete head_;
  head_ = NULL;
  if (primary_ && managePrimary_)
    delete primary_;
  primary_ = NULL;

  data_ = NULL;
  valid_ = 0;
  return;

  // found it!

 found:
  data_ = here + head_->headbytes();
  valid_ = 1;
  return;
}

FitsFitsNextMap::FitsFitsNextMap(FitsFile* p)
{
  FitsMap* prev = (FitsMap*)p;

  head_ = prev->head();
  manageHead_ = 0;
  primary_ = prev->primary();
  managePrimary_ = 0;
  ext_ = prev->ext();

  mapdata_ = prev->mapdata();
  mapsize_ = prev->mapsize();

  data_ = (char*)prev->data() + head_->pixbytes();
  valid_ = 1;
  return;
}

FitsArrMap::FitsArrMap()
{
  if (!valid_)
    return;

  // reset
  valid_ = 0;

  // check to see if we have a nonzero width, height, and bitpix
  if (!validArrayParams())
    return;

  // check to see if dimensions equal mapped space
  if ((pWidth_ * pHeight_ * abs(pBitpix_)/8)+pSkip_ > mapsize_)
    return;

  // skip to start of data
  data_ = mapdata_ + pSkip_;

  // new header
  head_ = new FitsHead(pWidth_, pHeight_, pBitpix_);
  if (!head_->isValid())
    return;

  // do we byteswap?
  setArrayByteSwap();

  // made it this far, must be valid
  valid_ = 1;
  orgFits_ = 0;
}

FitsMosaicMap::FitsMosaicMap()
{
  if (!valid_)
    return;

  char* here = mapdata_;
  int size = mapsize_;

  // keep the primary header

  primary_ = new FitsHead(here, size, FitsHead::EXTERNAL);
  managePrimary_ = 1;
  if (!primary_->isValid())
    goto error;

  here += primary_->headbytes() + primary_->databytes();
  size -= primary_->headbytes() + primary_->databytes();

  // first extension

  head_ = new FitsHead(here, size, FitsHead::EXTERNAL);
  if (!head_->isValid())
    goto error;
  ext_++;
  data_ = here + head_->headbytes();
  valid_ = 1;
  return;

error:
  if (head_)
    delete head_;
  head_ = NULL;
  if (primary_ && managePrimary_)
    delete primary_;
  primary_ = NULL;

  data_ = NULL;
  valid_ = 0;
  return;
}

FitsMosaicNextMap::FitsMosaicNextMap(FitsFile* p)
{
  FitsMap* prev = (FitsMap*)p;

  primary_ = prev->primary();
  managePrimary_ = 0;
  ext_ = prev->ext();

  mapdata_ = prev->enddata();
  mapsize_ = prev->endsize();

  head_ = new FitsHead(mapdata_, mapsize_, FitsHead::EXTERNAL);
  if (head_->isValid()) {
    data_ = mapdata_ + head_->headbytes();
    ext_++;
    valid_ = 1;
  }
  else {
    delete head_;
    head_ = NULL;
  
    data_ = NULL;
    valid_ = 0;
  }
}
