%{
#define YYPARSE_PARAM fr
#define YYDEBUG 1

#define FR ((Frame*)fr)
#define FITS (FR->findFits(1))

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "vector.h"
#include "list.h"
#include "basemarker.h"
#include "coord.h"

extern int ciaolex(void);
extern void ciaoerror(const char*);

int ciaosign;
int ciaosign2;

static char *color = "green";
static char *font = "helvetica 10 normal";
static char *text = "";
static unsigned short props;

static List<Vertex> polylist;

static void setProps(unsigned short* props, unsigned short prop, int value);

%}

%union {
  double real;
  int integer;
  char str[256];
  double vector[3];
}

%type <real> numeric

%type <real> angle
%type <real> value
%type <real> sexagesimal
%type <vector> coord

%token <integer> INT
%token <real> REAL

%token <integer> HOUR
%token <integer> MINUTE
%token <real> SECOND
%token <real> ARCMINUTE
%token <real> ARCSECOND

%token <str> SEXSTR

%token ANNULUS_
%token BOX_
%token CIRCLE_
%token DEBUG_
%token ELLIPSE_
%token FALSE_
%token OFF_
%token ON_
%token PIE_
%token POINT_
%token POLYGON_
%token ROTBOX_
%token VERSION_

%%

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "CIAO Regions File 1.0" << endl;}
	| init include shape
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

init:	{
	  props =
	    Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE | Marker::SOURCE;
	}

angle	: numeric {$$ = degToRad($1);}
	;

value	: numeric {$$ = FITS->mapLenToRef($1, PHYSICAL);}
	| ARCMINUTE {$$ = FITS->mapLenToRef($1, EQUATORIAL_FK5, ARCMIN);}
	| ARCSECOND {$$ = FITS->mapLenToRef($1, EQUATORIAL_FK5, ARCSEC);}
	;

sexagesimal: SEXSTR {$$ = parseDMS($1);}

coord	: sexagesimal sp sexagesimal
	{
	  Vector r = FITS->map(Vector($1*360./24.,$3), EQUATORIAL_FK5, REF);

	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| HOUR MINUTE SECOND {ciaosign2 = ciaosign;} sp INT ARCMINUTE ARCSECOND
        {
	  Vector r = FITS->map(
	    Vector(hmsToDegree(ciaosign2,$1,$2,$3), 
	    dmsToDegree(ciaosign,$6,$7,$8)),
	    EQUATORIAL_FK5, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| INT ARCMINUTE ARCSECOND {ciaosign2 = ciaosign;} sp 
	  INT ARCMINUTE ARCSECOND
        {
	  Vector r = FITS->map(
	    Vector(dmsToDegree(ciaosign2,$1,$2,$3),
	    dmsToDegree(ciaosign,$6,$7,$8)),
	    EQUATORIAL_FK5, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| numeric sp numeric 
	{
	  Vector r = FITS->map(Vector($1,$3), PHYSICAL, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	;

include	: /* empty */
	| '+' {setProps(&props, Marker::INCLUDE, 1);}
	| '-' {setProps(&props, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep
	  {FR->createCircleCmd(Vector($3),$5,color,1,font,text,props,NULL);}

	| ELLIPSE_ bp coord sp value sp value sp angle ep
	  {FR->createEllipseCmd(Vector($3),Vector($5,$7),$9,
	    color,1,font,text,props,NULL);}

	| ANNULUS_ bp coord sp value sp value ep
	  {FR->createAnnulusCmd(Vector($3),$5,$7,1,color,1,font,
	    text,props,NULL);}

	| PIE_ bp coord sp value sp value sp angle sp angle ep
	  {FR->createPandaCmd(Vector($3),($9-M_PI/2),($11-M_PI/2),1,
	    $5,$7,1,color,1,font,text,props,NULL);}

	| BOX_ bp coord sp value sp value ep
	  {FR->createBoxCmd(Vector($3),Vector($5,$7),0,
	    color,1,font,text,props,NULL);}

	| ROTBOX_ bp coord sp value sp value sp angle ep
	  {FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	     color,1,font,text,props,NULL);}

	| POINT_ bp coord ep 
	  {FR->createBoxCirclePointCmd(Vector($3),color,1,font,
	    text,props,NULL);}

	| POLYGON_ {polylist.deleteAll();} bp polyNodes ep
	  {FR->createPolygonCmd(polylist,color,1,font,text,props,NULL);} 
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist.append(new Vertex($1));}
	;

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}
