#include <strstream.h>

#include "circle.h"
#include "frame.h"
#include "fitsimage.h"
#include "util.h"

Circle::Circle(const Circle& a) : BaseCircle(a) {}

Circle::Circle(Frame* p, const Vector& ctr, double r, 
	       const char* clr, int w, const char* f, const char* t, 
	       unsigned short prop, const char* c)
  : BaseCircle(p, ctr, 0, clr, w, f, t, prop, c)
{
  annuli = 1;
  radii = new double[1];
  radii[0] = r;

  strcpy(type, "circle");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Circle::updateBBox()
{
  // bound marker

  Vector c = center * parent->refToCanvas;
  Vector r = Vector(radii[0],radii[0]);
  if (!(properties & FIXED))
    r *= parent->getZoom();
  bbox = BBox(c-r, c+r);

  // generate handles

  handle[0] = bbox.ll;
  handle[1] = bbox.lr();
  handle[2] = bbox.ur;
  handle[3] = bbox.ul();

  // bound handles

  for (int i=0; i<numHandle; i++)
    bbox.bound(handle[i]);

  // make room for handles

  bbox.expand(3);

  // calculate overall bbox

  calcAllBBox();
}

void Circle::edit(const Vector& v, int h)
{
  Vector r = (v-center).abs();
  radii[0] = r[0] > r[1] ? r[0] : r[1];

  updateBBox();
  doCallBack(&editCB);
}

void Circle::setRadius(double r)
{
  radii[0] = r;

  updateBBox();
  doCallBack(&editCB);
}

// list

void Circle::list(ostream& str, CoordSystem sys, 
		  CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(center);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      listPre(str,sys,ptr);

      Vector v = ptr->map(center,REF,sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << ptr->mapLenFromRef(radii[0], sys) << ')';

      listPost(str,delim);
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listPre(str,sys,ptr);

      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->map(center,REF,sys);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	      << ptr->mapLenFromRef(radii[0], sys, ARCSEC) << '"' << ')';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapToWCSStr(center, REF, sys, format, buf, 64);
	  char ra[16];
	  char dec[16];
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << type << '(' << ra << ',' << dec << ',' 
	      << ptr->mapLenFromRef(radii[0], sys, ARCSEC) 
	      << '"' << ')';
	}
	break;
      }

      listPost(str,delim);
    }
    else
      str << "";
    break;
  }
}

void Circle::listCiao(ostream& str, CoordSystem sys, 
		      CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      listCiaoPre(str);

      sys = PHYSICAL;
      Vector v = ptr->map(center, REF, sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << ptr->mapLenFromRef(radii[0], sys) << ')'
	  << delim;
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listCiaoPre(str);

      sys = EQUATORIAL_FK5;

      char buf[64];
      ptr->mapToWCSStr(center, REF, sys, SEXAGESIMAL, buf, 64);
      char ra[16];
      char dec[16];
      istrstream wcs(buf,64);
      wcs >> ra >> dec;
      str << type << '(' << ra << ',' << dec << ',' 
	  << ptr->mapLenFromRef(radii[0], sys, ARCMIN) << '\'' << ')'
	  << delim;
    }
    else
      str << "";
    break;
  }
}

void Circle::listPros(ostream& str, CoordSystem sys, 
		      CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  switch (sys) {
  case IMAGE:
  case DETECTOR:
  case AMPLIFIER:
    sys = IMAGE;
  case PHYSICAL:
    {
      listProsCoordSystem(str, sys);
      str << "; ";
      Vector v = ptr->map(center,REF,sys);
      str << type << ' ' << setprecision(8) << v 
	  << ptr->mapLenFromRef(radii[0], IMAGE)
	  << delim;
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listProsCoordSystem(str,sys);
      str << "; ";

      // display wcs radius in ARCSEC

      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->map(center,REF,sys);
	  str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] << "d " 
	      << ptr->mapLenFromRef(radii[0], sys, ARCSEC) << '"';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapToWCSStr(center, REF, sys, format, buf, 64);
	  char ra[16];
	  char dec[16];
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << type << ' ' << ra << ' ';
	  if (dec[0]=='+')
	    str << dec+1 << ' ';
	  else
	    str << dec << ' ';

	  str << ptr->mapLenFromRef(radii[0], sys, ARCSEC) << '"';
	}
	break;
      }

      str << delim;
    }
    else
      str << "";
    break;
  }
}

void Circle::listSAOtng(ostream& str, CoordSystem sys, 
			CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      listSAOtngPre(str,delim);

      Vector v = ptr->map(center, REF, IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << ptr->mapLenFromRef(radii[0], IMAGE) << ')';

      listSAOtngPost(str,delim);
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listSAOtngPre(str,delim);

      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->map(center, REF, sys);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	      << ptr->mapLenFromRef(radii[0], IMAGE) << ')';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapToWCSStr(center, REF, sys, format, buf, 64);
	  char ra[16];
	  char dec[16];
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << type << '(' << ra << ',' << dec << ',' 
	      << ptr->mapLenFromRef(radii[0], IMAGE) << ')';
	}
	break;
      }

      listSAOtngPost(str,delim);
    }
    else
      str << "";
    break;
  }
}

void Circle::listSAOimage(ostream& str, CoordSystem sys, 
			  CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = ptr->map(center,REF,IMAGE);
  str << "circle(" << setprecision(8) << v[0] << ',' << v[1] 
      << ',' << radii[0] << ')';

  str << delim;
}
