%{
#define YYPARSE_PARAM fr
#define YYDEBUG 1

#define FR ((Frame*)fr)
#define FITS (FR->findFits(globalTile))

#define DISCARD_(x) {yyclearin; mkDiscard(x);}

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "vector.h"
#include "list.h"
#include "basemarker.h"
#include "coord.h"

extern int mklex(void);
extern void mkerror(const char*);
extern void mkDiscard(int);

int mksign;
int mksign2;

static CoordSystem globalSystem;
static CoordSystem localSystem;

static int globalTile;

static unsigned short globalProps;
static unsigned short localProps;

static int globalWidth;
static int localWidth;

static char globalColor[16];
static char localColor[16];

static char globalFont[32];
static char localFont[32];

static char globalText[80];
static char localText[80];

static char localComment[80];

static int globalLine1;
static int localLine1;
static int globalLine2;
static int localLine2;

static int globalPoint;
static int localPoint;

static double globalTextAngle;
static double localTextAngle;

static CoordSystem globalRulerCoordSystem;
static CoordSystem globalRulerDistSystem;
static CoordFormat globalRulerDistFormat;

static CoordSystem localRulerCoordSystem;
static CoordSystem localRulerDistSystem;
static CoordFormat localRulerDistFormat;

static int localPanda;

static List<Vertex> polylist;

static double aAnnuli[MAXANNULI];
static Vector aVector[MAXANNULI];
static int aNum;
static int aStatus;
static Vector aCenter;
static double aAngle[MAXANGLES];
static int aAngNum;
static unsigned short aProps;
static char aColor[16];
static int aWidth;
static char aFont[32];
static char aText[80];
static char aComment[80];

static void setProps(unsigned short* props, unsigned short prop, int value);
static CoordSystem checkWCSSystem();

enum {CIRCLE,BOX,DIAMOND,CROSS,XPT,ARROW,BOXCIRCLE};
	
%}

%union {
  double real;
  int integer;
  char str[256];
  double vector[3];
}

%type <real> numeric
%type <integer> yesno

%type <real> angle
%type <real> optangle
%type <real> value
%type <real> sexagesimal
%type <vector> coord
%type <integer> coordSystem
%type <integer> coordFormat
%type <integer> property
%type <integer> pointProp
%type <integer> numberof

%token <integer> INT
%token <real> REAL
%token <str> STRING

%token <integer> HOUR
%token <integer> MINUTE
%token <real> SECOND
%token <real> DEGREE
%token <real> ARCMINUTE
%token <real> ARCSECOND
%token <real> RADIAN
%token <real> PHYCOORD
%token <real> IMGCOORD

%token <str> SEXSTR

%token AMPLIFIER_
%token ANNULUS_
%token ARCMIN_
%token ARCSEC_
%token ARROW_
%token B1950_
%token BACKGROUND_
%token BOX_
%token BOXCIRCLE_
%token CCD_
%token CIRCLE_
%token COLOR_
%token CROSS_
%token DEBUG_
%token DEGREES_
%token DELETE_
%token DETECTOR_
%token DIAMOND_
%token ECLIPTIC_
%token EDIT_
%token ELLIPSE_
%token FALSE_
%token FIELD_
%token FIXED_
%token FK4_
%token FK5_
%token FONT_
%token GALACTIC_
%token GLOBAL_
%token ICRS_
%token IGNORE_
%token IMAGE_
%token INCLUDE_
%token J2000_
%token LINE_
%token LINEAR_
%token LOGICAL_
%token MOVE_
%token N_
%token NO_
%token OFF_
%token ON_
%token PANDA_
%token PHYSICAL_
%token PIE_
%token PIXELS_
%token POINT_
%token POLYGON_
%token PROJECTION_
%token ROTATE_
%token ROTBOX_
%token RULER_
%token SELECT_
%token SOURCE_
%token TEXT_
%token TEXTANGLE_
%token TILE_
%token TRUE_
%token VERSION_
%token WIDTH_
%token X_
%token Y_
%token YES_

%%

start	: initGlobal commands processAnnulus
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "DS9 Regions File 3.0" << endl;}
	| GLOBAL_ global comment
	| TILE_ INT {globalTile = $2;}
	| coordSystem {globalSystem = (CoordSystem)$1;} comment
	| initLocal include shape
	| generalComment
	;

generalComment: '#' {DISCARD_(0)} STRING
	;

comment : /* empty */
	| '#' {DISCARD_(1)} STRING
	;

shapeComment : /* empty */ processAnnulus
	| '#' {DISCARD_(1)} STRING processAnnulus 
	  {strncpy(localComment,$3,80);}
	| '#' local processAnnulus
	| '#' local {DISCARD_(1)} STRING processAnnulus
	  {strncpy(localComment,$4,80);}
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

yesno	: INT {$$=($1 ? 1 : 0);}

	| YES_ {$$=1;}
	| Y_ {$$=1;}
	| ON_ {$$=1;}
	| TRUE_ {$$=1;}

	| NO_ {$$=0;}
	| N_ {$$=0;}
	| OFF_ {$$=0;}
	| FALSE_ {$$=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

optangle: /* empty */ {$$ = FR->mapAngleToRef(0,localSystem);}
	| angle {$$ = $1;}
	;

angle	: numeric {$$ = FR->mapAngleToRef(degToRad($1),localSystem);}
	| DEGREE {$$ = FR->mapAngleToRef(degToRad($1),localSystem);}
	| RADIAN {$$ = FR->mapAngleToRef($1,localSystem);}
	;

value	: numeric
	  {$$ = FITS->mapLenToRef($1, localSystem);}
	| PHYCOORD
	  {$$ = FITS->mapLenToRef($1, PHYSICAL);}
	| IMGCOORD
	  {$$ = FITS->mapLenToRef($1, IMAGE);}
	| DEGREE
	  {$$ = FITS->mapLenToRef($1, checkWCSSystem(), DEGREES);}
	| ARCMINUTE 
	  {$$ = FITS->mapLenToRef($1, checkWCSSystem(), ARCMIN);}
	| ARCSECOND 
	  {$$ = FITS->mapLenToRef($1, checkWCSSystem(), ARCSEC);}
	;

numberof: N_ '=' INT {$$ = $3;}
	;

sexagesimal: SEXSTR {$$ = parseDMS($1);}

coord	: sexagesimal sp sexagesimal
	{
	  localSystem = checkWCSSystem();
	  Vector r;
	  if (localSystem == GALACTIC ||
	      localSystem == ECLIPTIC) 
	    r = FITS->map(Vector($1,$3), localSystem, REF);
	  else
	    r = FITS->map(Vector($1*360./24.,$3), localSystem, REF);

	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| HOUR MINUTE SECOND {mksign2 = mksign;} sp INT ARCMINUTE ARCSECOND
        {
	  localSystem = checkWCSSystem();
	  Vector r = FITS->map(
	    Vector(hmsToDegree(mksign2,$1,$2,$3), 
	    dmsToDegree(mksign,$6,$7,$8)),
	    localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| INT ARCMINUTE ARCSECOND {mksign2 = mksign;} sp 
	  INT ARCMINUTE ARCSECOND
        {
	  localSystem = checkWCSSystem();
	  Vector r = FITS->map(
	    Vector(dmsToDegree(mksign2,$1,$2,$3),
	    dmsToDegree(mksign,$6,$7,$8)),
	    localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| numeric sp numeric 
	{
	  Vector r = FITS->map(Vector($1,$3), localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| DEGREE sp DEGREE
	{
	  Vector r = FITS->map(Vector($1,$3), checkWCSSystem(), REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| IMGCOORD sp IMGCOORD
	{
	  Vector r = FITS->map(Vector($1,$3), IMAGE, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| PHYCOORD sp PHYCOORD
	{
	  Vector r = FITS->map(Vector($1,$3), PHYSICAL, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	}
	| IMGCOORD sp PHYCOORD
	{
	  Vector rx = FITS->map(Vector($1,$1), IMAGE, REF);
	  Vector ry = FITS->map(Vector($3,$3), PHYSICAL, REF);
	  $$[0] = rx[0];
	  $$[1] = ry[0];
	  $$[2] = 1;
	}
	| PHYCOORD sp IMGCOORD
	{
	  Vector rx = FITS->map(Vector($1,$1), PHYSICAL, REF);
	  Vector ry = FITS->map(Vector($3,$3), IMAGE, REF);
	  $$[0] = rx[0];
	  $$[1] = ry[0];
	  $$[2] = 1;
	}
	;

coordSystem : IMAGE_ {$$ = IMAGE;}
	| LOGICAL_ {$$ = IMAGE;}
	| PHYSICAL_ {$$ = PHYSICAL;}
	| CCD_ {$$ = PHYSICAL;}
	| AMPLIFIER_ {$$ = AMPLIFIER;}
	| DETECTOR_ {$$ = DETECTOR;}
	| FK4_ {$$ = EQUATORIAL_FK4;}
	| B1950_  {$$ = EQUATORIAL_FK4;}
	| FK5_ {$$ = EQUATORIAL_FK5;}
	| J2000_  {$$ = EQUATORIAL_FK5;}	
	| ICRS_ {$$ = ICRS;}
	| GALACTIC_ {$$ = GALACTIC;}
	| ECLIPTIC_ {$$ = ECLIPTIC;}
	| LINEAR_ {$$ = LINEAR;}
	;

coordFormat : PIXELS_ {$$=PIXELS;}
	| DEGREES_ {$$=DEGREES;}
	| ARCMIN_ {$$=ARCMIN;}
	| ARCSEC_ {$$=ARCSEC;}
	;

property : SELECT_ {$$ = Marker::SELECT;}
	| EDIT_ {$$ = Marker::EDIT;}
	| MOVE_  {$$ = Marker::MOVE;}
	| ROTATE_ {$$ = Marker::ROTATE;}
	| DELETE_ {$$ = Marker::DELETE;}
	| INCLUDE_ {$$ = Marker::INCLUDE;}
	| SOURCE_ {$$ = Marker::SOURCE;}
	| FIXED_ {$$ = Marker::FIXED;}
	;

global	: global sp globalProperty
	| globalProperty
	;

globalProperty : property '=' yesno 
	{
	  setProps(&globalProps,$1,$3);
	  setProps(&localProps,$1,$3);
	}
	| COLOR_ '=' STRING 
	{
	  strncpy(globalColor,$3,16);
	  strncpy(localColor,$3,16);
	}
	| WIDTH_ '=' INT {globalWidth = localWidth = $3;}
	| FONT_ '=' STRING 
	{
	  strncpy(globalFont,$3,32);
	  strncpy(localFont,$3,32);
	}
	| TEXT_ '=' STRING 
	{
	  strncpy(globalText,$3,80);
	  strncpy(localText,$3,80);
	}
	| SOURCE_
	{
	  setProps(&globalProps,Marker::SOURCE,1);
	  setProps(&localProps,Marker::SOURCE,1);
	}
	| BACKGROUND_
	{
	  setProps(&globalProps,Marker::SOURCE,0);
	  setProps(&localProps,Marker::SOURCE,0);
	}
	| POINT_ '=' pointProp 
	{
	  globalPoint = localPoint = $3;
	}
	| LINE_ '=' INT INT 
	{
	  globalLine1 = localLine1 = $3;
	  globalLine2 = localLine2 = $4;
	}
	| RULER_ '=' globalRuler
	{
	}
	| TEXTANGLE_ '=' angle
	{
	  globalTextAngle = localTextAngle = $3;
	}
	;

globalRuler : coordFormat
	{
	  globalRulerCoordSystem = localRulerCoordSystem = IMAGE;
	  switch ((CoordFormat)$1) {
	  case PIXELS:
	    globalRulerDistSystem = localRulerDistSystem = IMAGE;
	    globalRulerDistFormat = localRulerDistFormat = PIXELS;
	    break;
	  case DEGREES:
	  case ARCMIN:
	  case ARCSEC:
	    globalRulerDistSystem = localRulerDistSystem = EQUATORIAL_FK5;
	    globalRulerDistFormat = localRulerDistFormat = (CoordFormat)$1;
	    break;
	  }
	}
	| coordSystem coordSystem
	{
	  globalRulerCoordSystem = localRulerCoordSystem = (CoordSystem)$1;
	  switch ((CoordSystem)$2) {
	  case IMAGE:
	  case PHYSICAL:
	  case DETECTOR:
	  case AMPLIFIER:
	    globalRulerDistSystem = localRulerDistSystem = (CoordSystem)$2;
	    globalRulerDistFormat = localRulerDistFormat = PIXELS;
	    break;
	  case EQUATORIAL_FK4:
	  case EQUATORIAL_FK5:
	  case ICRS:
	  case GALACTIC:
	  case ECLIPTIC:
	  case LINEAR:
	    globalRulerDistSystem = localRulerDistSystem = (CoordSystem)$2;
	    globalRulerDistFormat = localRulerDistFormat = DEGREES;
	    break;
	  }
	}
	| coordSystem coordFormat
	{
	  globalRulerCoordSystem = localRulerCoordSystem = (CoordSystem)$1;
	  switch ((CoordFormat)$2) {
	  case PIXELS:
	    globalRulerDistSystem = localRulerDistSystem = (CoordSystem)$2;
	    globalRulerDistFormat = localRulerDistFormat = PIXELS;
	    break;
	  case DEGREES:
	  case ARCMIN:
	  case ARCSEC:
	    globalRulerDistSystem = localRulerDistSystem = EQUATORIAL_FK5;
	    globalRulerDistFormat = localRulerDistFormat = (CoordFormat)$2;
	    break;
	  }
	}
	;

initGlobal:{
	  // global properties
	  globalSystem = PHYSICAL;
	  globalTile = 1;
	  globalProps =
	    Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE | Marker::SOURCE;
	  strcpy(globalColor,"green");
          globalWidth = 1;
	  strcpy(globalFont,"helvetica 10 normal");
	  strcpy(globalText,"");

	  // unique properties
	  globalLine1 = 0;
	  globalLine2 = 0;
	  globalRulerCoordSystem = PHYSICAL;
	  globalRulerDistSystem = PHYSICAL;
	  globalRulerDistFormat = PIXELS;
	  globalPoint = BOXCIRCLE;
	  globalTextAngle=0;
	} 
	;

local	: local sp localProperty
	| localProperty
	;

localProperty : property '=' yesno {setProps(&localProps,$1,$3);}
	| COLOR_ '=' STRING {strncpy(localColor,$3,16);}
	| WIDTH_ '=' INT {localWidth = $3;}
	| FONT_ '=' STRING {strncpy(localFont,$3,32);}
	| TEXT_ '=' STRING {strncpy(localText,$3,80);}
	| SOURCE_ {setProps(&localProps,Marker::SOURCE,1);}
	| BACKGROUND_ {setProps(&localProps,Marker::SOURCE,0);}
	| POINT_ '=' pointProp {localPoint = $3;}
	| LINE_ '=' INT INT {localLine1=$3; localLine2=$4;}
	| RULER_ '=' localRuler
	| TEXTANGLE_ '=' angle {localTextAngle=$3;}
	| PANDA_ '=' localPanda
	;

localRuler : coordFormat
	{
	  localRulerCoordSystem = IMAGE;
	  switch ((CoordFormat)$1) {
	  case PIXELS:
	    localRulerDistSystem = IMAGE;
	    localRulerDistFormat = PIXELS;
	    break;
	  case DEGREES:
	  case ARCMIN:
	  case ARCSEC:
	    localRulerDistSystem = EQUATORIAL_FK5;
	    localRulerDistFormat = (CoordFormat)$1;
	    break;
	  }
	}
	| coordSystem coordSystem
	{
	  localRulerCoordSystem = (CoordSystem)$1;
	  switch ((CoordSystem)$2) {
	  case IMAGE:
	  case PHYSICAL:
	  case DETECTOR:
	  case AMPLIFIER:
	    localRulerDistSystem = (CoordSystem)$2;
	    localRulerDistFormat = PIXELS;
	    break;
	  case EQUATORIAL_FK4:
	  case EQUATORIAL_FK5:
	  case ICRS:
	  case GALACTIC:
	  case ECLIPTIC:
	  case LINEAR:
	    localRulerDistSystem = (CoordSystem)$2;
	    localRulerDistFormat = DEGREES;
	    break;
	  }
	}
	| coordSystem coordFormat
	{
	  localRulerCoordSystem = (CoordSystem)$1;
	  switch ((CoordFormat)$2) {
	  case PIXELS:
	    localRulerDistSystem = (CoordSystem)$2;
	    localRulerDistFormat = PIXELS;
	    break;
	  case DEGREES:
	  case ARCMIN:
	  case ARCSEC:
	    localRulerDistSystem = EQUATORIAL_FK5;
	    localRulerDistFormat = (CoordFormat)$2;
	    break;
	  }
	}
	;

localPanda: {aNum=0; aAngNum=0;} '(' aAngs ')' '(' aRads ')' {localPanda = 2;}
	| IGNORE_ {localPanda=0;}
	;

initLocal : {
	  // global properties
	  localSystem = globalSystem;
	  localProps = globalProps;
	  strcpy(localColor,globalColor);
	  localWidth = globalWidth;
	  strcpy(localFont,globalFont);
	  strcpy(localText,globalText);
	  strcpy(localComment,"");

	  // unique properties
	  localLine1 = globalLine1;
	  localLine2 = globalLine2;
	  localPoint = globalPoint;
	  localRulerCoordSystem = globalRulerCoordSystem;
	  localRulerDistSystem = globalRulerDistSystem;
	  localRulerDistFormat = globalRulerDistFormat;
	  localTextAngle = globalTextAngle;
	  localPanda = 1;
	}
	;

pointProp : CIRCLE_ {$$ = CIRCLE;}
	| BOX_ {$$ = BOX;}
	| DIAMOND_ {$$ = DIAMOND;}
	| CROSS_ {$$ = CROSS;}
	| X_ {$$ = XPT;}
	| ARROW_ {$$ = ARROW;}
	| BOXCIRCLE_ {$$ = BOXCIRCLE;}
	;

include	: /* empty */ {setProps(&localProps, Marker::INCLUDE, 1);}
	| '+' {setProps(&localProps, Marker::INCLUDE, 1);}
	| '-' {setProps(&localProps, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep shapeComment
	  {FR->createCircleCmd(Vector($3),$5,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| ELLIPSE_ bp coord sp value sp value sp optangle ep shapeComment
	{
	  // for ellipse annulus
	  aStatus = 1;
	  aCenter = Vector($3);
	  aAngle[0] = $9;
	  aVector[0] = Vector($5,$7);
	  aNum = 1;
	  strncpy(aColor,localColor,16);
	  aWidth = localWidth;
	  strncpy(aFont,localFont,32);
	  strncpy(aText,localText,80);
	  strncpy(aComment,localComment,80);
	  aProps = localProps;

	  FR->createEllipseCmd(Vector($3),Vector($5,$7),$9,
	  localColor,localWidth,localFont,localText,localProps,localComment);
	}

	| BOX_ bp coord sp value sp value sp optangle ep shapeComment
	{
	  // for box annulus
	  aStatus = 3;
	  aCenter = Vector($3);
	  aAngle[0] = $9;
	  aVector[0] = Vector($5,$7);
	  aNum = 1;
	  strncpy(aColor,localColor,16);
	  aWidth = localWidth;
	  strncpy(aFont,localFont,32);
	  strncpy(aText,localText,80);
	  strncpy(aComment,localComment,80);
	  aProps = localProps;

	  FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    localColor,localWidth,localFont,localText,localProps,localComment);
	}
	| ROTBOX_ bp coord sp value sp value sp optangle ep shapeComment
	  {FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| POLYGON_ {polylist.deleteAll();} bp polyNodes ep shapeComment
	  {FR->createPolygonCmd(polylist, localColor,localWidth,localFont,
	      localText,localProps,localComment);} 

	| LINE_ bp coord sp coord ep shapeComment
	  {FR->createLineCmd(Vector($3),Vector($5),localLine1,localLine2,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| TEXT_ bp coord ep shapeComment
	  {FR->createTextCmd(Vector($3),localTextAngle,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| TEXT_ bp coord sp STRING ep {strncpy(localText,$5,80);} shapeComment
	  {FR->createTextCmd(Vector($3),localTextAngle,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| RULER_ bp coord sp coord ep shapeComment
	  {FR->createRulerCmd(Vector($3),Vector($5),
	    localRulerCoordSystem, localRulerDistSystem, localRulerDistFormat,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| PROJECTION_ bp coord sp coord sp value ep shapeComment
	{
	  FR->createProjectionCmd(Vector($3),Vector($5),$7,
	  "ProjectionPlot", "DeleteProjectionPlot",
	  localColor,localWidth,localFont,
	  localText,localProps,localComment);
	}

	| POINT_ bp coord ep shapeComment 
	{
	  switch (localPoint) {
	  case CIRCLE:
	    FR->createCirclePointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case BOX:
	    FR->createBoxPointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case DIAMOND:
	    FR->createDiamondPointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case CROSS:
	    FR->createCrossPointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case XPT:
	    FR->createExPointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case ARROW:
	    FR->createArrowPointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case BOXCIRCLE:
	    FR->createBoxCirclePointCmd(Vector($3),
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  }
	}
	| CIRCLE_ POINT_ bp coord ep shapeComment
	  {FR->createCirclePointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| BOX_ POINT_ bp coord ep shapeComment
	  {FR->createBoxPointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| DIAMOND_ POINT_ bp coord ep shapeComment
	  {FR->createDiamondPointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| CROSS_ POINT_ bp coord ep shapeComment
	  {FR->createCrossPointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| X_ POINT_ bp coord ep shapeComment
	  {FR->createExPointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| ARROW_ POINT_ bp coord ep shapeComment
	  {FR->createArrowPointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| BOXCIRCLE_ POINT_ bp coord ep shapeComment
	  {FR->createBoxCirclePointCmd(Vector($4),
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| ANNULUS_ bp coord sp value sp value ep shapeComment
	  {FR->createAnnulusCmd(Vector($3),$5,$7,1,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| ANNULUS_ bp coord sp value sp value sp {aNum=2;} aRads ep
	   shapeComment
	{
	  aAnnuli[0] = $5;
	  aAnnuli[1] = $7;
	  FR->createAnnulusCmd(Vector($3),aAnnuli,aNum,
	  localColor,localWidth,localFont,localText,localProps,localComment);
	}
	| ANNULUS_ bp coord sp value sp value sp numberof ep shapeComment
	  {FR->createAnnulusCmd(Vector($3),$5,$7,$9,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}

	| ELLIPSE_ ANNULUS_ bp coord sp angle sp
	    value sp value sp value sp value sp
	    {aNum=2;} vRads ep shapeComment
	{
	  aVector[0] = Vector($8,$10);
	  aVector[1] = Vector($12,$14);
	  FR->createEllipseAnnulusCmd(Vector($4),$6,aNum,
	  aVector,localColor,localWidth,localFont,
	  localText,localProps,localComment);
	}
	| ELLIPSE_ ANNULUS_ bp coord sp angle sp
	  value sp value sp value sp value sp
	  numberof ep shapeComment
	  {FR->createEllipseAnnulusCmd(Vector($4),
	    Vector($8,$10),Vector($12,$14),$16,$6,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| ELLIPSE_ bp coord sp value sp value sp optangle ep '&' '!' 
	  ELLIPSE_ bp coord sp value sp value sp optangle ep
	{	
	  aStatus = 2;
	  aVector[aNum++] = Vector($5,$7);
	}

	| BOX_ ANNULUS_ bp coord sp angle sp
	    value sp value sp value sp value sp
	    {aNum=2;} vRads ep shapeComment
	{
	  aVector[0] = Vector($8,$10);
	  aVector[1] = Vector($12,$14);
	  FR->createBoxAnnulusCmd(Vector($4),$6,aNum,
	  aVector,localColor,localWidth,localFont,
	  localText,localProps,localComment);
	}
	| BOX_ ANNULUS_ bp coord sp angle sp
	  value sp value sp value sp value sp
	  numberof ep shapeComment
	  {FR->createBoxAnnulusCmd(Vector($4),
	    Vector($8,$10),Vector($12,$14),$16,$6,
	    localColor,localWidth,localFont,
	    localText,localProps,localComment);}
	| BOX_ bp coord sp value sp value sp optangle ep '&' '!' 
	  BOX_ bp coord sp value sp value sp optangle ep
	{	
	  aStatus = 4;
	  aVector[aNum++] = Vector($5,$7);
	}

	| PANDA_ bp coord sp angle sp angle sp INT sp 
	    value sp value sp INT ep shapeComment
	{
	  switch (localPanda) {
	  case 0: /* ignore it */
	    break;
	  case 1: /* normal panda */
	    FR->createPandaCmd(Vector($3),$5,$7,$9,$11,$13,$15,
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  case 2: /* one of our special pandas */
	    FR->createPandaCmd(Vector($3),aAngle,aAngNum,aAnnuli,aNum,
	      localColor,localWidth,localFont,
	      localText,localProps,localComment);
	    break;
	  }
	}

	| PIE_ bp coord sp angle sp angle ep shapeComment
	| PIE_ bp coord sp angle sp angle sp aAngs ep shapeComment
	| PIE_ bp coord sp angle sp angle sp numberof ep shapeComment
	| FIELD_ bp ep shapeComment
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist.append(new Vertex($1));}
	;

aRads	: aRads sp aRad
	| aRad
	;

aRad	: value 
	{
	  if (aNum < MAXANNULI)
	    aAnnuli[aNum++] = $1;
	}
	;

aAngs	: aAngs sp aAng
	| aAng
	;

aAng	: angle 
	{
	  if (aAngNum < MAXANGLES)
	    aAngle[aAngNum++] = $1;
	}
	;

vRads	: vRads sp vRad
	| vRad
	;

vRad	: value sp value {aVector[aNum++] = Vector($1,$3);}
	;

processAnnulus : /* empty */
	{
	  switch (aStatus) {
	  case 0: // do nothing
	    break;
	  case 1: // we found just an ellipse, do nothing
	    break;
	  case 2: // ok we have an ellipse annulus
	    FR->markerDeleteLastCmd(); // delete the previous ellipse
	    FR->createEllipseAnnulusCmd(aCenter,aAngle[0],aNum,aVector,
	      aColor,aWidth,aFont,aText,aProps,aComment);
	    break;
	  case 3: // we found just a box, do nothing
	    break;
	  case 4: // ok, we have a box annulus
	    FR->markerDeleteLastCmd(); // delete the previous box
	    FR->createBoxAnnulusCmd(aCenter,aAngle[0],aNum,aVector,
	      aColor,aWidth,aFont,aText,aProps,aComment);
	    break;
	  }
	  aStatus = 0;
	}

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}

static CoordSystem checkWCSSystem()
{
  // if we have not defined a specific WCS, set to Native
  switch (localSystem) {
    case (IMAGE):
    case (PHYSICAL):
    case (DETECTOR):
    case (AMPLIFIER):
      return (NATIVEWCS);
    default:
      return localSystem;
  }
}
