#include "ellipse.h"
#include "frame.h"
#include "fitsimage.h"
#include "util.h"

Ellipse::Ellipse(const Ellipse& a) : BaseEllipse(a) {}

Ellipse::Ellipse(Frame* p, const Vector& ctr, const Vector& r, double a, 
		 const char* clr,  int w, const char* f, const char* t, 
		 unsigned short prop, const char* c)
  : BaseEllipse(p, ctr, a, clr, w, f, t, prop, c)
{
  annuli = 1;
  radii = new Vector[1];
  radii[0] = r;

  strcpy(type,"ellipse");
  handle = new Vector[4];
  numHandle = 4;

  updateBBox();
}

void Ellipse::edit(const Vector& v, int h)
{
  radii[0] = (v * Translate(-center) * flip * Rotate(-angle)).abs();

  updateBBox();
  doCallBack(&editCB);
}

void Ellipse::setRadius(const Vector& r)
{
  radii[0] = r;
  updateBBox();

  doCallBack(&editCB);
}

// private

void Ellipse::updateHandles()
{
  Vector r = radii[0];
  if (properties & FIXED)
    r /= parent->getZoom();

  Matrix m = Rotate(angle) * flip * Translate(center) * parent->refToCanvas;

  handle[0] = Vector(-r[0],-r[1]) * m;
  handle[1] = Vector( r[0],-r[1]) * m;
  handle[2] = Vector( r[0], r[1]) * m;
  handle[3] = Vector(-r[0], r[1]) * m;
}

// list

void Ellipse::list(ostream& str, CoordSystem sys, 
		   CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(center);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      listPre(str,sys,ptr);

      Vector r = ptr->mapLenFromRef(radii[0], sys);
      Vector v = ptr->map(center, REF, sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' 
	  << radToDeg(parent->mapAngleFromRef(angle,sys)) << ')';

      listPost(str,delim);
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listPre(str,sys,ptr);

      Vector r = ptr->mapLenFromRef(radii[0], sys, ARCSEC);
      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->map(center, REF, sys);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	      << r[0] << "\"" << ',' << r[1] << "\"" << ',' 
	      << radToDeg(parent->mapAngleFromRef(angle, sys)) << ')';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapToWCSStr(center, REF, sys, format, buf, 64);
	  char ra[16];
	  char dec[16];
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << type << '(' << ra << ',' << dec << ',' 
	      << r[0] << "\"" << ',' << r[1] << "\"" << ',' 
	      << radToDeg(parent->mapAngleFromRef(angle, sys)) << ')';
	}
	break;
      }

      listPost(str,delim);
    }
    else
      str << "";
    break;
  }
}

void Ellipse::listCiao(ostream& str, CoordSystem sys, 
		       CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  // radius is always in image coords

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      listCiaoPre(str);

      sys = PHYSICAL;
      Vector r = ptr->mapLenFromRef(radii[0], sys);
      Vector v = ptr->map(center, REF, sys);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')'
	  << delim;
    }
    break;
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listCiaoPre(str);

      sys = EQUATORIAL_FK5;

      Vector r = ptr->mapLenFromRef(radii[0], sys, ARCMIN);
      char buf[64];
      ptr->mapToWCSStr(center, REF, sys, SEXAGESIMAL, buf, 64);
      char ra[16];
      char dec[16];
      istrstream wcs(buf,64);
      wcs >> ra >> dec;
      str << type << '(' << ra << ',' << dec << ',' 
	  << r[0] << '\'' << ',' << r[1] << '\'' << ',' 
	  << radToDeg(angle) << ')'
	  << delim;
    }
    break;
  }
}

void Ellipse::listPros(ostream& str, CoordSystem sys, 
		       CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  switch (sys) {
  case IMAGE:
  case DETECTOR:
  case AMPLIFIER:
    sys = IMAGE;
  case PHYSICAL:
    {
      listProsCoordSystem(str, sys);
      str << "; ";
      Vector r = ptr->mapLenFromRef(radii[0], IMAGE);
      Vector v = ptr->map(center, REF, sys); 
      str << type << ' ' << setprecision(8) << v << r << radToDeg(angle)
	  << delim;
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listProsCoordSystem(str,sys);
      str << "; ";

      // display wcs radius in ARCSEC
      Vector r = ptr->mapLenFromRef(radii[0], sys, ARCSEC);

      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->map(center, REF, sys);
	  str << type << ' ' << setprecision(8) << v[0] << "d " << v[1] <<"d " 
	      << r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapToWCSStr(center, REF, sys, format, buf, 64);
	  char ra[16];
	  char dec[16];
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  if (dec[0]=='+')
	    str << type << ' ' << ra << ' ' << dec+1 << ' ' 
		<< r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
	  else
	    str << type << ' ' << ra << ' ' << dec << ' ' 
		<< r[0] << "\" " << r[1] << "\" " << radToDeg(angle);
	}
	break;
      }

      str << delim;
    }
    else
      str << "";
    break;
  }
}

void Ellipse::listSAOtng(ostream& str, CoordSystem sys, 
			 CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  // radius is always in image coords

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      listSAOtngPre(str,delim);

      Vector v = ptr->map(center,REF,IMAGE);
      Vector r = ptr->mapLenFromRef(radii[0],IMAGE);
      str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ',' 
	  << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';

      listSAOtngPost(str,delim);
    }
    break;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    if (parent->hasWCS()) {
      listSAOtngPre(str,delim);

      switch (format) {
      case DEGREES:
	{
	  Vector v = ptr->map(center, REF, sys);
	  Vector r = ptr->mapLenFromRef(radii[0], IMAGE);
	  str << type << '(' << setprecision(8) << v[0] << ',' << v[1] << ','
	      << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  ptr->mapToWCSStr(center, REF, sys, format, buf, 64);
	  Vector r = ptr->mapLenFromRef(radii[0], IMAGE);
	  char ra[16];
	  char dec[16];
	  istrstream wcs(buf,64);
	  wcs >> ra >> dec;
	  str << type << '(' << ra << ',' << dec << ',' 
	      << r[0] << ',' << r[1] << ',' << radToDeg(angle) << ')';
	}
	break;
      }

      listSAOtngPost(str,delim);
    }
    else
      str << "";
    break;
  }
}

void Ellipse::listSAOimage(ostream& str, CoordSystem sys, 
			   CoordFormat format, char delim)
{
  FitsImage* ptr = parent->findFits(1);

  // all coords are in image coords

  if (!(properties&INCLUDE))
    str << '-';

  Vector v = ptr->map(center,REF,IMAGE);
  str << "ellipse(" << setprecision(8) << v[0] << ',' << v[1] << ',' 
      << radii[0][0] << ',' << radii[0][1] << ',' << radToDeg(angle) << ')';

  str << delim;
}
