#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <limits.h>

#include <iostream.h>
#include <strstream.h>

#include "fitsimage.h"
#include "frame.h"
#include "util.h"

#include "mmap.h"
#include "alloc.h"
#include "allocgz.h"
#include "channel.h"
#include "share.h"
#include "socket.h"
#include "socketgz.h"
#include "var.h"
#include "iis.h"
#include "hist.h"


// WCS Mosaic
// we use these structures to map from one wcs to another

Vector FitsImage::WCScrpix;
Vector FitsImage::WCScdelt;
double FitsImage::WCSrot;
Orientation FitsImage::WCSorientation;
BBox FitsImage::WCSbb;

FitsImage::FitsImage(Frame* p, Frame::MosaicType m, int fm)
{
  parent = p;
  valid = 0;
  objectName = NULL;
  baseRootFileName = NULL;
  baseFullFileName = NULL;
  rootFileName = NULL;
  fullFileName = NULL;
  fits = NULL;
  image = NULL;
  data = NULL;

  isHist_ = 0;

  width_ = 0;
  height_ = 0;
  depth_ = 0;
  bitpix_ = 0;

  rotation = 0;
  zoom = Vector(1,1);
  orientation = NORMAL;

  keyLTMV = 0;
  keyATMV = 0;
  keyDTMV = 0;
  keyDATASEC = 0;

  memset(iparams,0,5*sizeof(int));
  memset(dparams,0,5*sizeof(int));
  memset(imparams,0,5*sizeof(int));
  memset(irparams,0,5*sizeof(int));
  memset(dmparams,0,5*sizeof(int));
  memset(drparams,0,5*sizeof(int));

  mosaicType_ = m;
  mosaicNum = fm;

  wcs = NULL;
  next_ = NULL;

  iisMode = 0;
  iiszt = 0;

  mWidgetToData = widgetToData.mm();
  mPannerToData = pannerToData.mm();
  mMagnifierToData = magnifierToData.mm();
  mPSToData = PSToData.mm();
}

FitsImage::~FitsImage()
{
  if (objectName)
    delete objectName;

  if (baseRootFileName)
    delete baseRootFileName;

  if (baseFullFileName)
    delete baseFullFileName;

  if (rootFileName)
    delete rootFileName;

  if (fullFileName)
    delete fullFileName;

  if (data)
    delete data;

  if (isHist_ && image)
    delete image;

  if (wcs)
    wcsfree(wcs);
}

FitsImageMaster::~FitsImageMaster()
{
  if (fits)
    delete fits;
}

FitsImageSlave::FitsImageSlave(Frame* p, const char* fn, FitsFile* f) 
  : FitsImage(p, Frame::NOMOSAIC, 0)
{
  fits = f;
  process(fn);
}

// Fits

FitsImageFitsAlloc::FitsImageFitsAlloc(Frame* p, const char* fn, 
				       Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsAlloc(fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsAllocGZ::FitsImageFitsAllocGZ(Frame* p, const char* fn, 
					   Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsAllocGZ(fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsChannel::FitsImageFitsChannel(Frame* p, Tcl_Interp* interp, 
					   const char* ch, const char* fn, 
					   Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsChannel(interp, ch, fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsMMap::FitsImageFitsMMap(Frame* p, const char* fn, 
				     Frame::MosaicType mt, int fm) 
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsMMap(fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsShare::FitsImageFitsShare(Frame* p, Frame::ShmType type, 
				       int id, int size, const char* fn,
				       Frame::MosaicType mt, int fm) 
  : FitsImageMaster(p, mt, fm)
{
  switch (type) {
  case Frame::SHMID:
    fits = new FitsFitsShareID(id, size, fn, FitsFile::RELAX);
    break;
  case Frame::KEY:
    fits = new FitsFitsShareKey(id, size, fn, FitsFile::RELAX);
    break;
  }
  process(fn);
}

FitsImageFitsSocket::FitsImageFitsSocket(Frame* p, int s, const char* fn, 
					 Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsSocket(s, fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsSocketGZ::FitsImageFitsSocketGZ(Frame* p, int s, const char* fn, 
					     Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsSocketGZ(s, fn, FitsFile::RELAX);
  process(fn);
}

FitsImageFitsVar::FitsImageFitsVar(Frame* p, Tcl_Interp* interp, 
				   const char* var, const char* fn, 
				   Frame::MosaicType mt, int fm) 
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsVar(interp, var, fn, FitsFile::RELAX);
  process(fn);
}

// Fits Next

FitsImageFitsNextAlloc::FitsImageFitsNextAlloc(Frame* p, const char* fn,
					       FitsFile* prev,
					       Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextAlloc(prev);
  process(fn);
}

FitsImageFitsNextAllocGZ::FitsImageFitsNextAllocGZ(Frame* p, const char* fn,
						   FitsFile* prev,
						   Frame::MosaicType mt,
						   int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextAllocGZ(prev);
  process(fn);
}

FitsImageFitsNextChannel::FitsImageFitsNextChannel(Frame* p, const char* fn,
						   FitsFile* prev,
						   Frame::MosaicType mt,
						   int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextChannel(prev);
  process(fn);
}

FitsImageFitsNextMMap::FitsImageFitsNextMMap(Frame* p, const char* fn,
					     FitsFile* prev,
					     Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextMMap(prev);
  process(fn);
}

FitsImageFitsNextShare::FitsImageFitsNextShare(Frame* p, const char* fn,
					       FitsFile* prev,
					       Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextShare(prev);
  process(fn);
}

FitsImageFitsNextSocket::FitsImageFitsNextSocket(Frame* p, const char* fn, 
						 FitsFile* prev,
						 Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextSocket(prev);
  process(fn);
}

FitsImageFitsNextSocketGZ::FitsImageFitsNextSocketGZ(Frame* p, const char* fn, 
						     FitsFile* prev,
						     Frame::MosaicType mt,
						     int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextSocketGZ(prev);
  process(fn);
}

FitsImageFitsNextVar::FitsImageFitsNextVar(Frame* p, const char* fn,
					   FitsFile* prev,
					   Frame::MosaicType mt, int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsFitsNextVar(prev);
  process(fn);
}

// Array

FitsImageArrAlloc::FitsImageArrAlloc(Frame* p, const char* fn)
  : FitsImageMaster(p)
{
  fits = new FitsArrAlloc(fn);
  process(fn);
}

FitsImageArrAllocGZ::FitsImageArrAllocGZ(Frame* p, const char* fn)
  : FitsImageMaster(p)
{
  fits = new FitsArrAllocGZ(fn);
  process(fn);
}

FitsImageArrChannel::FitsImageArrChannel(Frame* p, Tcl_Interp* interp, 
					 const char* ch, const char* fn)
  : FitsImageMaster(p)
{
  fits = new FitsArrChannel(interp, ch, fn);
  process(fn);
}

FitsImageArrMMap::FitsImageArrMMap(Frame* p, const char* fn) 
  : FitsImageMaster(p)
{
  fits = new FitsArrMMap(fn);
  process(fn);
}

FitsImageArrShare::FitsImageArrShare(Frame* p, Frame::ShmType type, int id,
				     int size, const char* fn) 
  : FitsImageMaster(p)
{
  switch (type) {
  case Frame::SHMID:
    fits = new FitsArrShareID(id, size, fn);
    break;
  case Frame::KEY:
    fits = new FitsArrShareKey(id, size, fn);
    break;
  }
  process(fn);
}

FitsImageArrSocket::FitsImageArrSocket(Frame* p, int s, const char* fn)
  : FitsImageMaster(p)
{
  fits = new FitsArrSocket(s, fn);
  process(fn);
}

FitsImageArrSocketGZ::FitsImageArrSocketGZ(Frame* p, int s, const char* fn)
  : FitsImageMaster(p)
{
  fits = new FitsArrSocketGZ(s, fn);
  process(fn);
}

FitsImageArrVar::FitsImageArrVar(Frame* p, Tcl_Interp* interp, 
				 const char* var, const char* fn)
  : FitsImageMaster(p)
{
  fits = new FitsArrVar(interp, var, fn);
  process(fn);
}

// Mosaic

FitsImageMosaicAlloc::FitsImageMosaicAlloc(Frame* p, const char* fn, 
					   Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicAlloc(fn);
  processMosaic(fn);
}

FitsImageMosaicAllocGZ::FitsImageMosaicAllocGZ(Frame* p, const char* fn,
					       Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicAllocGZ(fn);
  processMosaic(fn);
}

FitsImageMosaicChannel::FitsImageMosaicChannel(Frame* p, Tcl_Interp* interp,
					       const char* ch, const char* fn,
					       Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicChannel(interp, ch);
  processMosaic(fn);
}

FitsImageMosaicMMap::FitsImageMosaicMMap(Frame* p, const char* fn,
					       Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicMMap(fn);
  processMosaic(fn);
}

FitsImageMosaicShare::FitsImageMosaicShare(Frame* p, Frame::ShmType type,
					   int id, int size, const char* fn,
					       Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  switch (type) {
  case Frame::SHMID:
    fits = new FitsMosaicShareID(id, size);
    break;
  case Frame::KEY:
    fits = new FitsMosaicShareKey(id, size);
    break;
  }
  processMosaic(fn);
}

FitsImageMosaicSocket::FitsImageMosaicSocket(Frame* p, int s, const char* fn,
					       Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicSocket(s);
  processMosaic(fn);
}

FitsImageMosaicSocketGZ::FitsImageMosaicSocketGZ(Frame* p, int s, 
						 const char* fn,
						 Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicSocketGZ(s);
  processMosaic(fn);
}

FitsImageMosaicVar::FitsImageMosaicVar(Frame* p, Tcl_Interp* interp,
				       const char* var, const char* fn,
				       Frame::MosaicType mt)
  : FitsImageMaster(p, mt, 1)
{
  fits = new FitsMosaicVar(interp, var, fn);
  processMosaic(fn);
}

// Mosaic Next

FitsImageMosaicNextAlloc::FitsImageMosaicNextAlloc(Frame* p, const char* fn,
						   FitsFile* prev, 
						   Frame::MosaicType mt, 
						   int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextAlloc(prev);
  processMosaic(fn);
}

FitsImageMosaicNextAllocGZ::FitsImageMosaicNextAllocGZ(Frame* p,
						       const char* fn,
						       FitsFile* prev, 
						       Frame::MosaicType mt, 
						       int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextAllocGZ(prev);
  processMosaic(fn);
}

FitsImageMosaicNextChannel::FitsImageMosaicNextChannel(Frame* p, 
						       const char* fn, 
						       FitsFile* prev, 
						       Frame::MosaicType mt, 
						       int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextChannel(prev);
  processMosaic(fn);
}

FitsImageMosaicNextMMap::FitsImageMosaicNextMMap(Frame* p, const char* fn, 
						 FitsFile* prev, 
						 Frame::MosaicType mt, 
						 int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextMMap(prev);
  processMosaic(fn);
}

FitsImageMosaicNextShare::FitsImageMosaicNextShare(Frame* p, const char* fn,
						   FitsFile* prev, 
						   Frame::MosaicType mt, 
						   int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextShare(prev);
  processMosaic(fn);
}

FitsImageMosaicNextSocket::FitsImageMosaicNextSocket(Frame* p, const char* fn, 
						     FitsFile* prev, 
						     Frame::MosaicType mt, 
						     int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextSocket(prev);
  processMosaic(fn);
}

FitsImageMosaicNextSocketGZ::FitsImageMosaicNextSocketGZ(Frame* p,
							 const char* fn, 
							 FitsFile* prev,
							 Frame::MosaicType mt, 
							 int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextSocketGZ(prev);
  processMosaic(fn);
}

FitsImageMosaicNextVar::FitsImageMosaicNextVar(Frame* p, 
					       const char* fn, 
					       FitsFile* prev, 
					       Frame::MosaicType mt, 
					       int fm)
  : FitsImageMaster(p, mt, fm)
{
  fits = new FitsMosaicNextVar(prev);
  processMosaic(fn);
}

// IIS

FitsImageIIS::FitsImageIIS(Frame* p, int w, int h) 
  : FitsImageMaster(p, Frame::NOMOSAIC, 0)
{
  fits = new FitsIIS(w, h);
  process("");
  iisMode = 1;
}

void FitsImageIIS::iisErase()
{
  if (fits)
    ((FitsIIS*)fits)->erase();
}

const char* FitsImageIIS::iisGet(int x, int y)
{
  if (fits)
    return ((FitsIIS*)fits)->get(x,y);
}

void FitsImageIIS::iisSet(int x, int y, const char* d, int l)
{
  if (fits)
    ((FitsIIS*)fits)->set(x,y,d,l);
}

void FitsImageIIS::iisWCS(const Matrix& m, const Vector& z, int zt)
{
  dataToImage = Translate(-.5,-.5) * m;
  imageToData = dataToImage.invert();
  iisz = z;
  iiszt = zt;
}

void FitsImage::process(const char* fn)
{
  setFileName(fn);

  if (!fits->isValid())
    goto error;

  if (fits->isImage()) {
    image = fits;
    isHist_ = 0;

    FitsHead* head = image->head();
    width_ = head->naxis(1);
    height_ = head->naxis(2);
    depth_ = head->naxis(3);
    bitpix_ = head->bitpix();
    center_ = Vector(width_, height_)/2;

    valid = load();
    return;
  }
  else if (fits->isTable()) {
    isHist_ = 1;
    initHist();
    return;
  }

error:
  delete fits;
  fits = NULL;
  image = NULL;
  isHist_ = 0;
  valid = 0;
}

void FitsImage::processMosaic(const char* fn)
{
  setFileName(fn);

  if (!fits->isValid())
    goto error;

  if (fits->isImage()) {
    image = fits;
    isHist_ = 0;

    FitsHead* head = image->head();
    width_ = head->naxis(1);
    height_ = head->naxis(2);
    depth_ = 1;
    bitpix_ = head->bitpix();
    center_ = Vector(width_, height_)/2;

    valid = load();
    return;
  }
  else if (fits->isTable()) {
    isHist_ = 1;
    initHist();
    return;
  }

error:
  delete fits;
  fits = NULL;
  image = NULL;
  isHist_ = 0;
  valid = 0;
}

void FitsImage::setFileName(const char* fn)
{
  // strip any '[]'
  char* ffn = strip(fn);

  char lbuf[256];
  if (fits->find("EXTNAME")) {
    char* c = fits->getString("EXTNAME");
    {
      ostrstream str(lbuf,256);
      str << ffn << '[' << c << ']' << ends;
      baseFullFileName = dupstr(lbuf);
    }
    {
      char* m = root(ffn);
      ostrstream str(lbuf,256);
      str << m << '[' << c << ']' << ends;
      delete m;
      baseRootFileName = dupstr(lbuf);
    }
    delete c;
  }
  else if (fits->ext()) {
    {
      ostrstream str(lbuf,256);
      str << ffn << '[' << fits->ext() << ']' << ends;
      baseFullFileName = dupstr(lbuf);
    }
    {
      char* m = root(ffn);
      ostrstream str(lbuf,256);
      str << m << '[' << fits->ext() << ']' << ends;
      delete m;
      baseRootFileName = dupstr(lbuf);
    }
  }
  else {
    baseFullFileName = dupstr(ffn);
    baseRootFileName = root(ffn);
  }
  
  delete ffn;
  updateFileName();
}

void FitsImage::updateFileName()
{
  char lbuf[256];
  char* filter = (char*)fits->pFilter();
  {
    if (fullFileName)
      delete fullFileName;

    ostrstream str(lbuf,256);
    if (filter && *filter) {
      str << baseFullFileName << '[' << filter << ']' << ends;
      fullFileName = dupstr(lbuf);
    }
    else
      fullFileName = dupstr(baseFullFileName);
  }

  {
    if (rootFileName)
      delete rootFileName;

    ostrstream str(lbuf,256);
    if (filter && *filter) {
      str << baseRootFileName << '[' << filter << ']' << ends;
      rootFileName = dupstr(lbuf);
    }
    else
      rootFileName = dupstr(baseRootFileName);
  }
}

char* FitsImage::strip(const char* fn)
{
  if (fn) {
    char* r = dupstr(fn);           // dup the string
    char* ptr = r;                  // init the ptr
    while(*ptr != '[' && *ptr)      // walk it forward til '[' or end
      ptr++;
    *ptr = '\0';                    // zero out rest

    return r;                       // got it!
  }
  else
    return NULL;
}

char* FitsImage::root(const char* fn)
{
  if (fn) {
    const char* ptr = fn;           // init the ptr
    while(*ptr++);                  // walk it forward to end of string
    ptr--;                          // backup one
    while(*ptr != '/' && ptr != fn) // walk it backward til last / or beginning
      ptr--;
    if (*ptr == '/')                // step it over the last '/'
      ptr++;
    return dupstr(ptr);             // got it!
  }
  else
    return NULL;
}

void FitsImage::updateMatrices(Matrix& refToUser, Matrix& userToWidget, 
			       Matrix& widgetToCanvas, Matrix& userToPanner)
{
  // basic matrices

  // if in mosaic mode, we need to flip the image, based on orientation
  // which is set via the ATM/V keywords when the file is parsed.

  switch (mosaicType_) {
  case Frame::NOMOSAIC:
    dataToRef.identity();
    refToData.identity();
    break;
  case Frame::AUTO: // should never see this
    dataToRef.identity();
    refToData.identity();
    break;
  case Frame::IRAF:
    {
      // this is kluge, put it works ;-)
      // if the segment is flipped, we can have a discontinuity at
      // the edges, due to round off errors, so we 'nudge' it

      Vector oo = origin;
      Matrix flip;
      switch (orientation) {
      case XX:
	flip = FlipX();
	oo -= Vector(FLT_EPSILON,0);
	break;
      case YY:
	flip = FlipY();
	oo -= Vector(0,FLT_EPSILON);
	break;
      case XY:
	flip = FlipXY();
	oo -= Vector(FLT_EPSILON,FLT_EPSILON);
	break;
      }

      Vector c = datasec.center() * imageToData;
      dataToRef =
	Translate(-c) *
	flip *
	Rotate(rotation) *
	Scale(zoom) *
	Translate(c) *
	Translate(oo);

      refToData = dataToRef.invert();
    }
    break;
  case Frame::WCS:
  case Frame::WFPC2:
      Matrix flip;
      switch (orientation) {
      case XX:
	flip = FlipX();
	break;
      case YY:
	flip = FlipY();
	break;
      case XY:
	flip = FlipXY();
	break;
      }

      Vector c = getWCScrpix();
      dataToRef =
	Translate(-c) *
	flip *
	Rotate(rotation) *
	Scale(zoom) *
	Translate(c) *
	Translate(origin);

      refToData = dataToRef.invert();
    break;
  }

  // derived matrices

  refToCanvas = refToUser * userToWidget * widgetToCanvas;
  canvasToRef = refToCanvas.invert();

  physicalToData = physicalToImage * imageToData;
  dataToPhysical = physicalToData.invert();

  physicalToRef = physicalToImage * imageToData * dataToRef;
  refToPhysical = physicalToRef.invert();

  physicalToUser = physicalToImage * imageToData * dataToRef * refToUser;
  userToPhysical = physicalToUser.invert();

  physicalToCanvas = physicalToImage * imageToData * dataToRef * refToUser * 
    userToWidget * widgetToCanvas;
  canvasToPhysical = physicalToCanvas.invert();

  physicalToPanner = physicalToImage * imageToData * dataToRef * refToUser * 
    userToPanner;
  pannerToPhysical = physicalToPanner.invert();

  imageToRef = imageToData * dataToRef;
  refToImage = imageToRef.invert();

  imageToUser = imageToData * dataToRef * refToUser;
  userToImage = imageToUser.invert();

  imageToWidget = imageToData * dataToRef *  refToUser * userToWidget;
  widgetToImage = imageToWidget.invert();

  imageToCanvas = imageToData * dataToRef *  refToUser *
    userToWidget *  widgetToCanvas;
  canvasToImage = imageToCanvas.invert();

  imageToPanner = imageToData * dataToRef * refToUser * userToPanner;
  pannerToImage = imageToPanner.invert();

  dataToUser = dataToRef * refToUser;
  userToData = dataToUser.invert();

  dataToWidget = dataToRef *  refToUser *  userToWidget;
  widgetToData = dataToWidget.invert();

  dataToCanvas = dataToRef *  refToUser *  userToWidget * widgetToCanvas;
  canvasToData = dataToCanvas.invert();

  dataToPanner = dataToRef * refToUser * userToPanner;
  pannerToData = dataToPanner.invert();

  amplifierToRef = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef;
  refToAmplifier = amplifierToRef.invert();

  amplifierToUser = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser;
  userToAmplifier = amplifierToUser.invert();

  amplifierToCanvas = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser *
    userToWidget * 
    widgetToCanvas;
  canvasToAmplifier = amplifierToCanvas.invert();

  amplifierToPanner = 
    amplifierToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser * 
    userToPanner;
  pannerToAmplifier = amplifierToPanner.invert();

  detectorToRef = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef;
  refToDetector = detectorToRef.invert();

  detectorToUser = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser;
  userToDetector = detectorToUser.invert();

  detectorToCanvas = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser *
    userToWidget * 
    widgetToCanvas;
  canvasToDetector = detectorToCanvas.invert();

  detectorToPanner = 
    detectorToPhysical * 
    physicalToImage * 
    imageToData * 
    dataToRef * 
    refToUser * 
    userToPanner;
  pannerToDetector = detectorToPanner.invert();
}

const char* FitsImage::getValue(const Vector& v)
{
  if (!isIIS() || (iiszt != 1))
    return data->getValue(v);
  else {
    double value = data->getValueDouble(v);

    ostrstream str(buf,32);
    if (value <= 1)
      str << '<' << iisz[0] << ends;
    else if (value >= 200)
      str << '>' << iisz[1] << ends;
    else
      str << ((value-1) * (iisz[1]-iisz[0]))/199 + iisz[0] << ends;

    return buf;
  }
}

void FitsImage::setHistCursor(const Vector& v)
{
  histCursor = v * userToPhysical;
}

Matrix FitsImage::updateHistCenter()
{
  if (isHist_)
    return nextHist(getHistCenter());
  else
    return Matrix();
}

Matrix FitsImage::updateHistCursor()
{
  if (isHist_)
    return nextHist(histCursor);
  else
    return Matrix();
}

Matrix FitsImage::updateHist(Vector v)
{
  if (isHist_)
    return nextHist(v * userToPhysical);
  else
    return Matrix();
}

void FitsImage::initHist()
{
  // just in case
  if (!fits->isTable())
    return;

  // make sure we have cols to bin on
  if (!fits->pBinX())
    fits->setpBinX("X");

  if (!fits->pBinY())
    fits->setpBinY("Y");

  // determine buffersize
  calcHistSize();
  center_ = Vector(width_, height_)/2;

  // remember where we point
  histCursor = getHistCenter();

  // if needed, force to a bin boundary, then translate to center of bin cell
  if (parent->binFactor<1)
    actualHistCursor = histCursor;
  else
    actualHistCursor = 
      (histCursor/parent->binFactor).floor() * parent->binFactor;

  // build the matrix

  Matrix newPhysicalToData = 
    Translate(-actualHistCursor) *
    Scale(1./parent->binFactor) *
    Translate(center_.floor()) *
    Translate(.5,.5);

  image = new FitsHist(fits, width_, height_, newPhysicalToData,
		       parent->binFunction, parent->binFactor);

  if (image->isValid())
    valid = load();
  else {
    delete fits;
    fits = NULL;
    if (image)
      delete image;
    image = NULL;
    isHist_ = 0;
    valid = 0;
  }
}

Matrix FitsImage::nextHist(Vector c)
{
  // cursor is in bin (physical) coords
  calcHistSize();
  center_ = Vector(width_, height_)/2;

  // remember where we are pointing
  histCursor = c;

  // if needed, force to a bin boundary, then translate to center of bin cell
  if (parent->binFactor<1)
    actualHistCursor = histCursor;
  else
    actualHistCursor = 
      (histCursor/parent->binFactor).floor() * parent->binFactor;

  // new physicalToData

  Matrix newPhysicalToData =
    Translate(-actualHistCursor) *
    Scale(1./parent->binFactor) *
    Translate(center_.floor()) *
    Translate(.5,.5);

  // always center on center of pixel
  parent->cursor = (getCenter()).floor() + Vector(.5,.5);

  if (image)
    delete image;
  image = new FitsHist(fits, width_, height_, newPhysicalToData,
		       parent->binFunction, parent->binFactor);

  if (image->isValid()) {
    valid = load();
    return refToPhysical * newPhysicalToData * dataToRef;
  }
  else {
    delete fits;
    fits = NULL;
    image = NULL;
    isHist_ = 0;
    valid = 0;
    return Matrix();
  }
}

void FitsImage::calcHistSize()
{
  Vector s = getHistDim()/parent->binFactor;

  width_ = (int)(s[0]<parent->binBufferSize? s[0] : parent->binBufferSize);
  height_= (int)(s[1]<parent->binBufferSize? s[1] : parent->binBufferSize);
}

Vector FitsImage::getHistDim()
{
  // assumes we aready have our columns

  FitsTableHDU* hdu = (FitsTableHDU*)(fits->head())->hdu();
  Vector xd = hdu->dimension(fits->pBinX());
  Vector yd = hdu->dimension(fits->pBinY());
  return Vector(xd[1]-xd[0],yd[1]-yd[0]);
}

Vector FitsImage::getHistCenter()
{
  // assumes we aready have our columns

  FitsTableHDU* hdu = (FitsTableHDU*)(fits->head())->hdu();
  Vector xd = hdu->dimension(fits->pBinX());
  Vector yd = hdu->dimension(fits->pBinY());
  return Vector(xd[1]-xd[0],yd[1]-yd[0])/2 + Vector(xd[0],yd[0]);
}

void FitsImage::setHistFilter(const char* filter)
{
  if (fits) {
    fits->setpFilter(filter);
    updateFileName();
  }
}

const char* FitsImage::getHistFilter()
{
  if (fits)
    return fits->pFilter();
  else
    return NULL;
}

void FitsImage::setHistX(const char* str)
{
  if (fits)
    fits->setpBinX(str);
}

const char* FitsImage::getHistX()
{
  if (fits)
    return fits->pBinX();
  else
    return NULL;
}

void FitsImage::setHistY(const char* str)
{
  if (fits)
    fits->setpBinY(str);
}

const char* FitsImage::getHistY()
{
  if (fits)
    return fits->pBinY();
  else
    return NULL;
}

char* FitsImage::getHistList()
{
  if (fits && fits->isTable()) {
    FitsHead* head = fits->head();
    return ((FitsTableHDU*)head->hdu())->list();
  }
  else
    return NULL;
}

void FitsImage::updatePS(Matrix ps)
{
  dataToPS = dataToRef * ps;
  PSToData = dataToPS.invert();
}

void FitsImage::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  dataToMagnifier = dataToRef * refToMagnifier;
  magnifierToData = dataToMagnifier.invert();
}

BBox FitsImage::getWCSbb()
{
  switch (mosaicType_) {
  case Frame::NOMOSAIC:
    return BBox(1,1,width_,height_);
  case Frame::AUTO: // should not see this
    return BBox(1,1,width_,height_);
  case Frame::IRAF:
    return detsize * Scale(ccdsum) * Translate(-origin);
  case Frame::WCS:
  case Frame::WFPC2:
    return WCSbb;
  }
}

Vector FitsImage::getCenter()
{
  switch (mosaicType_) {
  case Frame::NOMOSAIC:
    return center_;
  case Frame::AUTO: // should not see this
    return center_;
  case Frame::IRAF:
    return Vector(detsize.center() * Scale(ccdsum)).floor();
  case Frame::WCS:
  case Frame::WFPC2:
    return WCSbb.center().floor();
  }
}

Vector FitsImage::getSize()
{
  switch (mosaicType_) {
  case Frame::NOMOSAIC:
    return Vector(width_, height_);
  case Frame::AUTO: // should not see this
    return Vector(width_, height_);
  case Frame::IRAF:
    return BBox(detsize * Scale(ccdsum)).size();
  case Frame::WCS:
  case Frame::WFPC2:
    return WCSbb.size();
  }
}

int* FitsImage::getDataParams(Frame::ScanMode which)
{
  switch (which) {
  case Frame::NODATASEC:
    return iparams;
  case Frame::DATASEC:
    return dparams;
  case Frame::UNODATASEC:
    return irparams;
  case Frame::UDATASEC:
    return drparams;
  }
}

int* FitsImage::getMinMaxParams(Frame::ScanMode which)
{
  switch (which) {
  case Frame::NODATASEC:
    return iparams;
  case Frame::DATASEC:
    return dparams;
  case Frame::UNODATASEC:
    return imparams;
  case Frame::UDATASEC:
    return dmparams;
  }
}

void FitsImage::setDataParams(int x0, int y0, int x1, int y1)
{
  irparams[0] = drparams[0] = x1;
  irparams[1] = drparams[1] = x0;
  irparams[2] = drparams[2] = x1;
  irparams[3] = drparams[3] = y0;
  irparams[4] = drparams[4] = y1;

  // now clip by DATASEC

  if (dparams[0] < drparams[0])
    drparams[0] = dparams[0];
  if (dparams[1] > drparams[1])
    drparams[1] = dparams[1];
  if (dparams[2] < drparams[2])
    drparams[2] = dparams[2];
  if (dparams[3] > drparams[3])
    drparams[3] = dparams[3];
  if (dparams[4] < drparams[4])
    drparams[4] = dparams[4];
}

void FitsImage::setMinMaxParams(int x0, int y0, int x1, int y1)
{
  imparams[0] = dmparams[0] = x1;
  imparams[1] = dmparams[1] = x0;
  imparams[2] = dmparams[2] = x1;
  imparams[3] = dmparams[3] = y0;
  imparams[4] = dmparams[4] = y1;

  // now clip by DATASEC

  if (dparams[0] < dmparams[0])
    dmparams[0] = dparams[0];
  if (dparams[1] > dmparams[1])
    dmparams[1] = dparams[1];
  if (dparams[2] < dmparams[2])
    dmparams[2] = dparams[2];
  if (dparams[3] > dmparams[3])
    dmparams[3] = dparams[3];
  if (dparams[4] < dmparams[4])
    dmparams[4] = dparams[4];

  if (data)
    data->resetMinMax();
}

int FitsImage::load()
{
  resetWCS();

  memset(iparams,0,5*sizeof(int));
  memset(dparams,0,5*sizeof(int));
  memset(imparams,0,5*sizeof(int));
  memset(irparams,0,5*sizeof(int));
  memset(dmparams,0,5*sizeof(int));
  memset(drparams,0,5*sizeof(int));

  if (!processKeywords())
    return 0;

  // create a valid FitsData

  FitsHead* head = image->head();
  if (data)
    delete data;

  switch (head->bitpix()) {
  case 8: 
    data = new FitsDatam<unsigned char>(image);
    break;
  case 16:
    data = new FitsDatam<short>(image);
    break;
  case -16:
    data = new FitsDatam<unsigned short>(image);
    break;
  case 32:
    data = new FitsDatam<BITPIX32>(image);
    break;
  case -32:
    data = new FitsDatam<float>(image);
    break;
  case -64:
    data = new FitsDatam<double>(image);
    break;
  }

  return data ? 1 : 0;
}

void FitsImage::resetWCS()
{
  // Process OBJECT keyword

  if (objectName)
    delete objectName;
  objectName = image->getString("OBJECT");

  // Process WCS keywords

  FitsHead* head = image->head();
  FitsHead* primary = image->primary();

  if (!primary) {
    hlength(head->cards(), head->headbytes());
    initWCS(head->cards());
  }
  else {
    // we have a primary header also
    // be sure to skip the middle 'END' keyword
    int ll = head->headbytes()-80+primary->headbytes();

    char* cards = new char[ll];
    memcpy(cards, head->cards(), head->headbytes()-80);
    memcpy(cards+head->headbytes()-80, primary->cards(), primary->headbytes());

    hlength(cards, ll);
    initWCS(cards);
    delete cards;
  }
}

void FitsImage::replaceWCS(FitsHead* h)
{
  // don't include primary header

  hlength(h->cards(), h->headbytes());
  initWCS(h->cards());

  // Process OBJECT keyword

  if (objectName)
    delete objectName;
  objectName = h->getString("OBJECT");
}

void FitsImage::appendWCS(FitsHead* app)
{
  FitsHead* head = image->head();
  FitsHead* primary = image->primary();

  if (!primary) {
    // append in front
    int ll = app->headbytes()-80+head->headbytes();

    char* cards = new char[ll];
    memcpy(cards, app->cards(), app->headbytes()-80);
    memcpy(cards+app->headbytes()-80, head->cards(), head->headbytes());

    hlength(cards, ll);
    initWCS(cards);
    delete cards;
  }
  else {
    // we have a primary header also
    // be sure to skip the middle 'end' keyword
    int ll = app->headbytes()-80+head->headbytes()-80+primary->headbytes();

    char* cards = new char[ll];
    memcpy(cards, app->cards(), app->headbytes()-80);
    memcpy(cards+app->headbytes()-80, head->cards(), head->headbytes()-80);
    memcpy(cards+app->headbytes()-80+head->headbytes()-80, 
	   primary->cards(), primary->headbytes());

    hlength(cards, ll);
    initWCS(cards);
    delete cards;
  }

  // process OBJECT keyword

  char* obj = app->getString("OBJECT");
  if (obj) {
    if (objectName)
      delete objectName;
    objectName = obj;
  }
}

void FitsImage::initWCS(const char* ptr)
{
  if (wcs)
    wcsfree(wcs);

  wcs = wcsinit(ptr);

  if (DebugWCS && hasWCS()) {
    cerr << "wcs->rot=" << wcs->rot << endl;
    cerr << "wcs->equinox=" << wcs->equinox << endl;
    cerr << "wcs->epoch=" << wcs->epoch << endl;
    cerr << "wcs->longpole=" << wcs->longpole << endl;
    cerr << "wcs->latpole=" << wcs->latpole << endl;

    {
      for (int i=0; i<2; i++)
	cerr << "wcs->crpix[" << i << "]=" << wcs->crpix[i] << endl;
    }
    {
      for (int i=0; i<2; i++)
	cerr << "wcs->crval[" << i << "]=" << wcs->crval[i] << endl;
    }
    {
      for (int i=0; i<2; i++)
	cerr << "wcs->cdelt[" << i << "]=" << wcs->cdelt[i] << endl;
    }
    {
      for (int i=0; i<4; i++)
	cerr << "wcs->cd[" << i << "]=" << wcs->cd[i] << endl;
    }
    {
      for (int i=0; i<4; i++)
	cerr << "wcs->pc[" << i << "]=" << wcs->pc[i] << endl;
    }
    cerr << "wcs->imrot=" << wcs->imrot << endl;
    cerr << "wcs->pa_north=" << wcs->pa_north << endl;
    cerr << "wcs->pa_east=" << wcs->pa_east << endl;
    cerr << "wcs->imflip=" << wcs->imflip << endl;
    cerr << "wcs->prjcode=" << wcs->prjcode << endl;

    cerr << "wcs->coorflip=" << wcs->coorflip << endl;
    cerr << "wcs->wcsproj=" << wcs->wcsproj << endl;

    {
      for (int i=0; i<2; i++)
	cerr << "wcs->ctype[" << i << "]=" << wcs->ctype[i] << endl;
    }
    cerr << "wcs->c1type=" << wcs->c1type << endl;
    cerr << "wcs->c2type=" << wcs->c2type << endl;
    cerr << "wcs->ptype=" << wcs->ptype << endl;
    cerr << "wcs->radecsys=" << wcs->radecsys << endl;
    cerr << "wcs->radecout=" << wcs->radecout << endl;
    cerr << "wcs->syswcs=" << wcs->syswcs << endl;
  }
}

FitsHead* FitsImage::getHead()
{
  if (image)
    return image->head();
  else    
    return NULL;
}

char* FitsImage::getCards()
{
  if (image) {
    FitsHead* head = image->head();
    if (head)
      return head->cards();
  }
  return NULL;
}

int FitsImage::getNCards()
{
  if (image) {
    FitsHead* head = image->head();
    if (head)
      return head->ncard();
  }
  return 0;
}

char* FitsImage::getKeyword(const char* key)
{
  return fits->getKeyword(key);
}

char* FitsImage::displayHeader()
{
  // always return the header to the native fits ext, not the image
  FitsHead* head = fits->head();
  //  FitsHead* head = image->head();

  int size = head->ncard() * FTY_CARDLEN;
  char* lbuf = new char[size+1];
  memcpy(lbuf, head->cards(), size);

  for (int i=0; i<head->ncard(); i++)
    lbuf[i*FTY_CARDLEN+79] = '\n';

  lbuf[size] = '\0';
  return lbuf;
}

char* FitsImage::displayPrimary()
{
  // always return the header to the native fits ext, not the image
  FitsHead* head = fits->primary();
  //  FitsHead* head = image->primary();

  int size = head->ncard() * FTY_CARDLEN;
  char* lbuf = new char[size+1];
  memcpy(lbuf, head->cards(), size);

  for (int i=0; i<head->ncard(); i++)
    lbuf[i*FTY_CARDLEN+79] = '\n';

  lbuf[size] = '\0';
  return lbuf;
}

void FitsImage::parseSection(char* lbuf, Vector* v1, Vector* v2)
{
  double x1, y1, x2, y2;
  char d; // dummy char
  istrstream str(lbuf);

  str >> d >> (*v1)[0] >> d >> (*v2)[0] >> d >> (*v1)[1] >> d >> (*v2)[1] >> d;
}

int FitsImage::processKeywords()
{
  // Image to Data

  imageToData = Translate(-.5, -.5);
  dataToImage = imageToData.invert();

  // Physical to Image (LTM/LTV keywords)

  if (image->find("LTM1_1") ||
      image->find("LTM1_2") ||
      image->find("LTM2_1") ||
      image->find("LTM2_2") ||
      image->find("LTV1") ||
      image->find("LTV2"))
    keyLTMV = 1;

  double ltm11 = image->getReal("LTM1_1", 1);
  double ltm12 = image->getReal("LTM1_2", 0);
  double ltm21 = image->getReal("LTM2_1", 0);
  double ltm22 = image->getReal("LTM2_2", 1);

  double ltv1 = image->getReal("LTV1", 0);
  double ltv2 = image->getReal("LTV2", 0);

  physicalToImage = Matrix(ltm11, ltm12, ltm21, ltm22, ltv1, ltv2);
  imageToPhysical = physicalToImage.invert();

  // CDD to Detector (DTM/DTV keywords)

  if (image->find("DTM1_1") ||
      image->find("DTM1_2") ||
      image->find("DTM2_1") ||
      image->find("DTM2_2") ||
      image->find("DTV1") ||
      image->find("DTV2"))
    keyDTMV = 1;

  double dtm11 = image->getReal("DTM1_1", 1);
  double dtm12 = image->getReal("DTM1_2", 0);
  double dtm21 = image->getReal("DTM2_1", 0);
  double dtm22 = image->getReal("DTM2_2", 1);

  double dtv1 = image->getReal("DTV1", 0);
  double dtv2 = image->getReal("DTV2", 0);

  physicalToDetector = Matrix(dtm11, dtm12, dtm21, dtm22, dtv1, dtv2);
  detectorToPhysical = physicalToDetector.invert();

  // Physical to Amplifier (ATM/ATV keywords)

  if (image->find("ATM1_1") ||
      image->find("ATM1_2") ||
      image->find("ATM2_1") ||
      image->find("ATM2_2") ||
      image->find("ATV1") ||
      image->find("ATV2"))
    keyATMV = 1;

  double atm11 = image->getReal("ATM1_1", 1);
  double atm12 = image->getReal("ATM1_2", 0);
  double atm21 = image->getReal("ATM2_1", 0);
  double atm22 = image->getReal("ATM2_2", 1);

  double atv1 = image->getReal("ATV1", 0);
  double atv2 = image->getReal("ATV2", 0);

  physicalToAmplifier = Matrix(atm11, atm12, atm21, atm22, atv1, atv2);
  amplifierToPhysical = physicalToAmplifier.invert();

  if (DebugMosaic) {
    cerr << endl;
    cerr << baseRootFileName << endl;
    cerr << "ATM/V: " << physicalToAmplifier << endl;
    cerr << "ATM/V-1: " << amplifierToPhysical << endl;
    cerr << "DTM/V: " << physicalToDetector << endl;
    cerr << "DTM/V-1: " << detectorToPhysical << endl;
    cerr << "LTM/V: " << physicalToImage << endl;
    cerr << "LTM/V-1: " << imageToPhysical << endl;
  }

  // DATASEC

  iparams[0] = width_;
  iparams[1] = 0;
  iparams[2] = width_;
  iparams[3] = 0;
  iparams[4] = height_;

  char* datstr = image->getString("DATASEC");
  if (datstr && *datstr) {
    Vector v1,v2;
    parseSection(datstr,&v1,&v2);

    if (v1[0] < 1)
      v1[0] = 1;
    if (v1[1] < 1)
      v1[1] = 1;
    if (v2[0] > width_)
      v2[0] = width_;
    if (v2[1] > height_)
      v2[1] = height_;

    keyDATASEC = 1;
    datasec = BBox(v1,v2);

    v1 -= Vector(1,1);
    dparams[0] = width_;
    dparams[1] = v1[0];
    dparams[2] = v2[0];
    dparams[3] = v1[1];
    dparams[4] = v2[1];
  }
  else {
    keyDATASEC = 0;
    datasec = BBox(1,1,width_,height_);

    dparams[0] = width_;
    dparams[1] = 0;
    dparams[2] = width_;
    dparams[3] = 0;
    dparams[4] = height_;
  }
  if (datstr)
    delete datstr;

  switch (mosaicType_) {
  case Frame::NOMOSAIC:
    return 1;
  case Frame::AUTO:
    if (processKeywordsIRAF()) {
      mosaicType_ = Frame::IRAF;
      return 1;
    } 
    else if (processKeywordsWCS()) {
      mosaicType_ = Frame::WCS;
      return 1;
    }
    else
      return 0;
  case Frame::IRAF:
    return processKeywordsIRAF();
  case Frame::WCS:
    return processKeywordsWCS();
  case Frame::WFPC2:
    return 1;
  }
}

int FitsImage::processKeywordsIRAF() 
{
  // DETSIZE

  {
    char* sizestr = image->getString("DETSIZE");
    if (!(sizestr && *sizestr)) {
      if (sizestr)
	delete sizestr;
      return 0;
    }
    
    Vector v1,v2;
    parseSection(sizestr,&v1,&v2);
    delete sizestr;

    detsize = BBox(v1,v2);

    if (DebugMosaic)
      cerr << "DETSIZE: " << detsize << endl;
  }

  // DETSEC

  {
    char* detstr =  image->getString("DETSEC");
    if (!(detstr && *detstr)) {
      if (detstr)
	delete detstr;
      return 0;
    }

    Vector v1,v2;
    parseSection(detstr,&v1,&v2);
    delete detstr;

    int xx = (v1[0] < v2[0]);
    int yy = (v1[1] < v2[1]);

    if (xx && yy)
      orientation = NORMAL;
    else if (!xx && yy)
      orientation = XX;
    else if (!xx && !yy)
      orientation = XY;
    else if (xx && !yy)
      orientation = YY;

    detsec = BBox(v1,v2);

    if (DebugMosaic)
      cerr << "DETSEC : " << detsec << endl;
  }

  // CCDSUM

  ccdsum = Vector(1,1);
  {
    char* ccdstr = image->getString("CCDSUM");
    if (ccdstr && *ccdstr) {
      double Ns, Np, Ns1, Np1;
      istrstream str(ccdstr);

      str >> Ns >> Np >> Ns1 >> Np1;
      ccdsum = Vector(1/Ns, 1/Np);

      if (DebugMosaic)
	cerr << "CCDSUM : " << ccdsum << endl;
    }

    if (ccdstr)
      delete ccdstr;
  }

  origin = detsec.ll * Scale(ccdsum) * Translate(-datasec.ll);

  if (DebugMosaic) {
    cerr << "Type: IRAF" << endl;
    cerr << "datasec: " << datasec << endl;
    cerr << "ccdsum : " << ccdsum << endl;
    cerr << "detsize: " << detsize << endl;
    cerr << "detsec : " << detsec << endl;
    cerr << "orientation: " << orientation << endl;
    cerr << "origin: " << origin << endl;
  }

  return 1;
}

int FitsImage::processKeywordsWCS() 
{
  if (!hasWCSEqu())
    return 0;

  if (mosaicNum == 1) {
    WCScrpix = getWCScrpix();
    WCScdelt = getWCScdelt();
    WCSrot   = getWCSRotation();
    WCSorientation = getWCSOrientation();

    rotation = 0;
    zoom = Vector(1,1);
    orientation = NORMAL;
    origin = Vector();

    // init WCS bbox
    WCSbb = BBox(Vector(1,1),Vector(width_,height_));
  }
  else {

    // calculate rotation, zoom, orientation, center
    rotation = WCSrot - getWCSRotation();
    double zz = fabs(getWCScdelt()[0]/WCScdelt[0]);
    zoom = Vector(zz,zz);
    orientation = (getWCSOrientation() == WCSorientation) ? NORMAL : XX;

    // calculate origin
    Vector aa = parent->fits->wcs2pix(
      pix2wcs(getWCScrpix(),EQUATORIAL_FK5),EQUATORIAL_FK5);
    origin = aa-getWCScrpix();

    // calculate WCS bbox
    Matrix flip;
    switch (orientation) {
    case XX:
      flip = FlipX();
      break;
    case YY:
      flip = FlipY();
      break;
    case XY:
      flip = FlipXY();
      break;
    }

    Vector c = getWCScrpix();
    Matrix m =
      Translate(-c) *
      flip *
      Rotate(rotation) *
      Scale(zoom) *
      Translate(c) *
      Translate(origin);

    WCSbb.bound(Vector(1,1)*m);
    WCSbb.bound(Vector(width_,1)*m);
    WCSbb.bound(Vector(width_,height_)*m);
    WCSbb.bound(Vector(1,height_)*m);
  }

  if (DebugMosaic) {
    cerr << "WCS fits" << endl
	 << " crpix " << getWCScrpix() << endl
	 << " cdelt " << getWCScdelt() << endl
	 << " rot " << radToDeg(getWCSRotation()) << endl
	 << " orientation " << getWCSOrientation() << endl;

    cerr << "WCS segment " << endl
	 << " rotation " << radToDeg(rotation) << endl
	 << " zoom " << zoom << endl
	 << " origin " << origin << endl;

    cerr << "WCSbb " << WCSbb << endl;
  }

  return 1;
}

int FitsImage::hasWCS()
{
  return (wcs && iswcs(wcs)) ? 1 : 0;
}

int FitsImage::hasWCSEqu()
{
  if (hasWCS())
    switch (wcsToCoordSystem(getradecsys(wcs))) {
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case ICRS:
      return 1;
    default:
      return 0;
    }
  else
    return 0;
}

int FitsImage::hasWCSLinear()
{
  if (hasWCS()) {
    if (wcsToCoordSystem(getradecsys(wcs)) == LINEAR)
      return 1;
    else
      return 0;
  }
  else
    return 0;
}

char* FitsImage::pix2wcs(Vector in, CoordSystem sys, 
			 CoordFormat format, char* lbuf, int len)
{
  if (hasWCS()) {
    wcsoutinit(wcs, (char*)coordSystemToWCS(sys));
    switch (format) {
    case PIXELS:
    case ARCMIN:
    case ARCSEC:
    case DEGREES:
      setwcsdeg(wcs,1);
      wcs->ndec = 5;
      break;
    case SEXAGESIMAL:
      setwcsdeg(wcs,0);
      wcs->ndec = 3;
      break;
    }
    setwcslin(wcs,2);
    pix2wcst(wcs, in[0], in[1], lbuf, len);
  }

  return lbuf;
}

Vector FitsImage::pix2wcs(Vector in)
{
  if (hasWCS()) {
    wcsoutinit(wcs, getradecsys(wcs));
    double x,y;
    ::pix2wcs(wcs, in[0], in[1], &x, &y);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::pix2wcs(Vector in, CoordSystem sys)
{
  if (hasWCS()) {
    wcsoutinit(wcs, (char*)coordSystemToWCS(sys));
    double x,y;
    ::pix2wcs(wcs, in[0], in[1], &x, &y);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2pix(Vector in)
{
  if (hasWCS()) {
    wcsininit(wcs, getradecsys(wcs));
    double x,y;
    int off;
    ::wcs2pix(wcs, in[0], in[1], &x, &y, &off);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2pix(Vector in, CoordSystem sys)
{
  if (hasWCS()) {
    wcsininit(wcs, (char*)coordSystemToWCS(sys));
    double x,y;
    int off;
    ::wcs2pix(wcs, in[0], in[1], &x, &y, &off);

    return Vector(x,y);
  }
  else
    return in;
}

Vector FitsImage::wcs2wcs(Vector in, CoordSystem sysi, 
			  CoordSystem syso)
{
  Vector r = in;
  if (hasWCS())
    wcscon(wcscsys((char*)coordSystemToWCS(sysi)), 
	   wcscsys((char*)coordSystemToWCS(syso)), 
	   wcs->equinox, 0., &r[0], &r[1], wcs->epoch);

  return r;
}

Vector FitsImage::wcs2pixDist(Vector in)
{
  if (hasWCS()) {
    Vector cd = getWCScdelt();
    return Vector(in[0]/cd[0], in[1]/cd[1]).abs();
  }
  else
    return in;
}

Vector FitsImage::pix2wcsDist(Vector in)
{
  if (hasWCS()) {
    Vector cd = getWCScdelt();
    return Vector(in[0]*cd[0], in[1]*cd[1]).abs();
  }
  else
    return in;
}

double FitsImage::wcsdist(Vector a, Vector b)
{
  if (hasWCS())
    return ::wcsdist(a[0],a[1],b[0],b[1]);
  else
    return 0;
}

int FitsImage::getWCSFlip()
{
  if (hasWCS())
    return wcs->coorflip;
  else
    return 0;
}

const char* FitsImage::getWCSSystem()
{
  if (hasWCS())
    return coordSystemToWCS(wcsToCoordSystem(getradecsys(wcs)));
  else
    return NULL;
}

Vector FitsImage::getWCScrval()
{
  if (hasWCS()) {
    if (!wcs->coorflip)
      return Vector(wcs->crval[0], wcs->crval[1]);
    else
      return Vector(wcs->crval[1], wcs->crval[0]);
  }
  else
    return Vector();
}

Vector FitsImage::getWCScrpix()
{
  if (hasWCS())
      return Vector(wcs->crpix[0], wcs->crpix[1]);
  else
    return Vector();
}

Vector FitsImage::getWCScdelt()
{
  if (hasWCS()) {
    if (!wcs->coorflip)
      return Vector(wcs->cdelt[0], wcs->cdelt[1]);
    else
      return Vector(wcs->cdelt[1], wcs->cdelt[0]);
  }
  else
    return Vector();
}

const char* FitsImage::coordSystemToWCS(CoordSystem sys)
{
  switch(sys) {
  case EQUATORIAL_FK4:
    return wcsName[0];
  case EQUATORIAL_FK5:
    return wcsName[1];
  case ICRS:
    return wcsName[2];
  case GALACTIC:
    return wcsName[3];
  case ECLIPTIC:
    return wcsName[4];
  case LINEAR:
    return wcsName[5];
  default:
    return NULL;
  }
}

CoordSystem FitsImage::wcsToCoordSystem(const char* str)
{
  if (!strncmp("FK4", str, 3))
    return EQUATORIAL_FK4;
  else if (!strncmp("FK5", str, 3))
    return EQUATORIAL_FK5;
  else if (!strncmp("ICR", str, 3))
    return ICRS;
  else if (!strncmp("GAL", str, 3))
    return GALACTIC;
  else if (!strncmp("ECL", str, 3))
    return ECLIPTIC;
  else if (!strncmp("LIN", str, 3))
    return LINEAR;
  else {
    cerr << "Frame Internal Error: Unknown WCS "<< str 
	 << " Assuming LINEAR." << endl;
    return LINEAR;
  }
}

double FitsImage::getWCSRotation()
{
  if (hasWCS())
    return -degToRad(wcs->imrot);
  else
    return 0;
}

Orientation FitsImage::getWCSOrientation()
{
  if (hasWCS())
    return wcs->imflip ? XX : NORMAL;
  else
    return NORMAL;
}
