#ifndef __fitsimage_h__
#define __fitsimage_h__

#include "frame.h"
#include "vector.h"
#include "fitsdata.h"
#include "coord.h"

class FitsFile;

extern "C" {
  #include "wcs.h"
}

// wcs coordinate system strings (for use with wcssubs)
static char *wcsName[] = { 
  "fk4",
  "fk5",
  "icrs",
  "galactic",
  "ecliptic",
  "linear"
};

class FitsImage {
protected:
  Frame* parent;            // pointer to parent
  int valid;                // is load a success
  char* objectName;         // object name
  char* baseRootFileName;   // base root filename
  char* baseFullFileName;   // base full filename
  char* rootFileName;       // root filename
  char* fullFileName;       // full filename
  FitsFile* fits;           // native fits ext
  FitsFile* image;          // fits image
  FitsData* data;           // raw fits image data

  int isHist_;
  Vector histCursor;        // start point to bin about
  Vector actualHistCursor;  // actual point to bin about		 

  int width_;
  int height_;
  int depth_;
  int bitpix_;
  Vector center_;

  double rotation;          // local rotation
  Vector zoom;              // local scaling;
  Orientation orientation; // local orientation
  Vector origin;            // local translation in image space

  char buf[32];             // tmp storage for returning strings

  // Mosaic items
  int keyLTMV;
  int keyATMV;
  int keyDTMV;
  int keyDATASEC;

  Vector ccdsum;            // CCDSUM keyword

  BBox detsize;             // DETSIZE keyword
  BBox detsec;              // DETSEC keyword
  BBox datasec;             // DATASEC keyword

  Frame::MosaicType mosaicType_;    // flag to indicate if part of mosaic
  int mosaicNum;            // order of mosaic

  // base cd matrix and crval/crpix for WCS mosaic
  static Vector WCScrpix;
  static Vector WCScdelt;
  static double WCSrot;
  static Orientation WCSorientation;
  static BBox WCSbb;

  Matrix amplifierToCanvas;
  Matrix amplifierToPhysical;
  Matrix amplifierToRef;
  Matrix amplifierToPanner;
  Matrix amplifierToUser;
  Matrix canvasToAmplifier;
  Matrix canvasToPhysical;
  Matrix canvasToData;
  Matrix canvasToDetector;
  Matrix canvasToImage;
  Matrix canvasToRef;
  Matrix physicalToAmplifier;
  Matrix physicalToCanvas;
  Matrix physicalToData;
  Matrix physicalToDetector;
  Matrix physicalToImage;
  Matrix physicalToRef;
  Matrix physicalToUser;
  Matrix physicalToPanner;
  Matrix dataToCanvas;
  Matrix dataToPhysical;
  Matrix dataToImage;
  Matrix dataToMagnifier;
  Matrix dataToRef;
  Matrix dataToPanner;
  Matrix dataToUser;
  Matrix dataToWidget;
  Matrix detectorToCanvas;
  Matrix detectorToPhysical;
  Matrix detectorToPanner;
  Matrix detectorToRef;
  Matrix detectorToUser;
  Matrix imageToCanvas;
  Matrix imageToPhysical;
  Matrix imageToData;
  Matrix imageToRef;
  Matrix imageToPanner;
  Matrix imageToUser;
  Matrix imageToWidget;
  Matrix magnifierToData;
  Matrix pannerToAmplifier;
  Matrix pannerToPhysical;
  Matrix pannerToData;
  Matrix pannerToDetector;
  Matrix pannerToImage;
  Matrix refToPhysical;
  Matrix refToData;
  Matrix refToImage;
  Matrix refToAmplifier;
  Matrix refToDetector;
  Matrix refToCanvas;
  Matrix userToAmplifier;
  Matrix userToPhysical;
  Matrix userToData;
  Matrix userToDetector;
  Matrix userToImage;
  Matrix widgetToData;
  Matrix widgetToImage;
  Matrix dataToPS;
  Matrix PSToData;

  // this is for optimization of frame
  double* mWidgetToData;
  double* mPannerToData;
  double* mMagnifierToData;
  double* mPSToData;

  int iparams[5];           // image clip/render bbox
  int dparams[5];           // DATASEC clip/render bbox
  int imparams[5];          // user specified minmax scan bbox
  int irparams[5];          // user specified valid data bbox
  int dmparams[5];          // DATASEC user specified minmax scan bbox
  int drparams[5];          // DATASEC user specified valid data bbox

  int iisMode;              // flag to indicate if iis
  Vector iisz;              // z1 and z2
  int iiszt;                // 0-Uniary, 1-Linear, 2-Log

  WorldCoor* wcs;           // wcs

  FitsImage* next_;          // next in list

  void process(const char*);
  void processMosaic(const char*);
  int load();
  Matrix nextHist(Vector);
  Vector getHistCenter();

  char* root(const char*);
  char* strip(const char*);
  void parseSection(char*, Vector*, Vector*);
  const char* coordSystemToWCS(CoordSystem);
  CoordSystem wcsToCoordSystem(const char*);

public:
  FitsImage(Frame*, Frame::MosaicType, int);
  virtual ~FitsImage();

  int isValid() {return valid;}
  FitsFile* getFitsFile() {return fits;}

  int isHist() {return isHist_;}
  int isImage() {return fits ? fits->isImage() : 0;}
  int isTable() {return fits ? fits->isTable() : 0;}
  void initHist();
  Matrix updateHistCenter();
  Matrix updateHistCursor();
  Matrix updateHist(Vector);
  void calcHistSize();
  Vector getHistDim();
  void setHistCursor(const Vector&);
  Vector getHistCursor() {return actualHistCursor;}
  void setHistFilter(const char*);
  const char* getHistFilter();
  void setHistX(const char*);
  const char* getHistX();
  void setHistY(const char*);
  const char* getHistY();
  char* getHistList();

  FitsImage* next() {return next_;}
  void setNext(FitsImage* n) {next_ = n;}
  const char* getObjectName() {return objectName;}
  const char* getRootFileName() {return rootFileName;}
  const char* getFullFileName() {return fullFileName;}
  void setFileName(const char*);
  void updateFileName();

  int width() {return width_;}
  int height() {return height_;}
  int depth() {return depth_;}
  int bitpix() {return bitpix_;}
  int ext() {return fits->ext();}
  Vector getCenter();
  Vector getSize();

  int* getDataParams(Frame::ScanMode);
  int* getMinMaxParams(Frame::ScanMode);
  void setDataParams(int,int,int,int);
  void setMinMaxParams(int,int,int,int);

  int hasLTMV() {return keyLTMV;}
  int hasATMV() {return keyATMV;}
  int hasDTMV() {return keyDTMV;}
  int isMosaic() {return (mosaicType_ != Frame::NOMOSAIC);}
  Frame::MosaicType mosaicType() {return mosaicType_;}
  int isIIS() {return iisMode;}
  int isCube() {return depth_>1 ? 1 : 0;}

  char* pix2wcs(Vector, CoordSystem, CoordFormat, char*, int);
  Vector pix2wcs(Vector);
  Vector pix2wcs(Vector, CoordSystem);
  Vector wcs2pix(Vector);
  Vector wcs2pix(Vector, CoordSystem);

  Vector wcs2wcs(Vector, CoordSystem, CoordSystem);

  Vector wcs2pixDist(Vector);
  Vector pix2wcsDist(Vector);
  double wcsdist(Vector, Vector);

  void initWCS(const char*);
  void resetWCS();
  void replaceWCS(FitsHead*);
  void appendWCS(FitsHead*);

  int processKeywords();
  int processKeywordsIRAF();
  int processKeywordsWCS();
  void setMosaicParams(int n, Frame::MosaicType t)
    {mosaicNum = n; mosaicType_ = t;}

  Orientation getWCSOrientation();
  double getWCSRotation();

  Orientation getOrientation() {return orientation;}

  int hasWCS();
  int hasWCSEqu();
  int hasWCSLinear();
  const char* getWCSSystem();
  Vector getWCScrval();
  Vector getWCScrpix();
  Vector getWCScdelt();
  int getWCSFlip();
  WorldCoor* getWCS() {return wcs;}

  BBox getWCSbb();

  void updateMatrices(Matrix&, Matrix&, Matrix&, Matrix&);
  void updateMagnifierMatrices(Matrix&);

  void updatePS(Matrix);

  double* getmWidgetToData() {return mWidgetToData;}
  double* getmPannerToData() {return mPannerToData;}
  double* getmMagnifierToData() {return mMagnifierToData;}
  double* getmPSToData() {return mPSToData;}

  Matrix& getAmplifierToPhysical() {return amplifierToPhysical;}
  Matrix& getAmplifierToRef() {return amplifierToRef;}
  Matrix& getAmplifierToUser() {return amplifierToUser;}
  Matrix& getCanvasToAmplifier() {return canvasToAmplifier;}
  Matrix& getCanvasToPhysical() {return canvasToPhysical;}
  Matrix& getCanvasToData() {return canvasToData;}
  Matrix& getCanvasToDetector() {return canvasToDetector;}
  Matrix& getCanvasToImage() {return canvasToImage;}
  Matrix& getCanvasToRef() {return canvasToRef;}
  Matrix& getPhysicalToAmplifier() {return physicalToAmplifier;}
  Matrix& getPhysicalToDetector() {return physicalToDetector;}
  Matrix& getPhysicalToImage() {return physicalToImage;}
  Matrix& getPhysicalToRef() {return physicalToRef;}
  Matrix& getPhysicalToUser() {return physicalToUser;}
  Matrix& getDataToCanvas() {return dataToCanvas;}
  Matrix& getDataToPhysical() {return dataToPhysical;}
  Matrix& getDataToImage() {return dataToImage;}
  Matrix& getDataToPanner() {return dataToPanner;}
  Matrix& getDataToRef() {return dataToRef;}
  Matrix& getDataToUser() {return dataToUser;}
  Matrix& getDataToWidget() {return dataToWidget;}
  Matrix& getDetectorToPhysical() {return detectorToPhysical;}
  Matrix& getDetectorToRef() {return detectorToRef;}
  Matrix& getDetectorToUser() {return detectorToUser;}
  Matrix& getImageToCanvas() {return imageToCanvas;}
  Matrix& getImageToPhysical() {return imageToPhysical;}
  Matrix& getImageToData() {return imageToData;}
  Matrix& getImageToRef() {return imageToRef;}
  Matrix& getImageToUser() {return imageToUser;}
  Matrix& getImageToWidget() {return imageToWidget;}
  Matrix& getMagnifierToData() {return magnifierToData;}
  Matrix& getRefToAmplifier() {return refToAmplifier;}
  Matrix& getRefToPhysical() {return refToPhysical;}
  Matrix& getRefToData() {return refToData;}
  Matrix& getRefToDetector() {return refToDetector;}
  Matrix& getRefToImage() {return refToImage;}
  Matrix& getRefToCanvas() {return refToCanvas;}
  Matrix& getPannerToAmplifier() {return pannerToAmplifier;}
  Matrix& getPannerToPhysical() {return pannerToPhysical;}
  Matrix& getPannerToData() {return pannerToData;}
  Matrix& getPannerToDetector() {return pannerToDetector;}
  Matrix& getPannerToImage() {return pannerToImage;}
  Matrix& getPhysicalToData() {return physicalToData;}
  Matrix& getUserToAmplifier() {return userToAmplifier;}
  Matrix& getUserToPhysical() {return userToPhysical;}
  Matrix& getUserToData() {return userToData;}
  Matrix& getUserToDetector() {return userToDetector;}
  Matrix& getUserToImage() {return userToImage;}
  Matrix& getWidgetToData() {return widgetToData;}
  Matrix& getWidgetToImage() {return widgetToImage;}

  Vector map(const Vector&, CoordSystem, CoordSystem);
  Vector mapWCS(const Vector&, CoordSystem, CoordSystem);
  Vector mapImage(const Vector&, CoordSystem);
  Vector mapPhysical(const Vector&, CoordSystem);
  Vector mapAmplifier(const Vector&, CoordSystem);
  Vector mapDetector(const Vector&, CoordSystem);
  Vector mapRef(const Vector&, CoordSystem);
  Vector mapUser(const Vector&, CoordSystem);
  Vector mapCanvas(const Vector&, CoordSystem);
  Vector mapPanner(const Vector&, CoordSystem);

  double mapLenToRef(double, CoordSystem, CoordFormat =DEGREES);
  Vector mapLenToRef(const Vector&, CoordSystem, CoordFormat =DEGREES);
  double mapLenFromRef(double, CoordSystem, CoordFormat =DEGREES);
  Vector mapLenFromRef(const Vector&, CoordSystem, CoordFormat =DEGREES);

  double mapLenFromRef(const Vector&, const Vector&, CoordSystem, CoordFormat);

  void mapToWCSStr(const Vector&, CoordSystem, CoordSystem, CoordFormat,
		   char*, int);

// we assume that 'data' is valid

  const char* getValue(const Vector& v);
  float getValueFloat(const Vector& v) {return data->getValueFloat(v);}
  float getValueFloat(int i) {return data->getValueFloat(i);}
  double getValueDouble(const Vector& v) {return data->getValueDouble(v);}
  double getValueDouble(int i) {return data->getValueDouble(i);}
  void setClip(double l, double h) {data->setClip(l,h);}

  const char* getMin() {return data ? data->getMin() : NULL;}
  const char* getMax() {return data ? data->getMax() : NULL;}

  double getMinDouble() {return data ? data->getMinDouble() : 0;}
  double getMaxDouble() {return data ? data->getMaxDouble() : 0;}

  const char* getLow() {return data ? data->getLow() : NULL;}
  const char* getHigh() {return data ? data->getHigh() : NULL;}
  double getLowDouble() {return data ? data->getLowDouble() : 0;}
  double getHighDouble() {return data ? data->getHighDouble() : 0;}

  void updateClip(Frame::ClipMode cm, 
		  Frame::MinMaxMode mm, int mmi,
		  float zc, int zs, int zl, 
		  float per,
		  double ul, double uh,
		  Frame::ScanMode ds)
    {if (data) 
      data->updateClip(cm,mm,mmi,zc,zs,zl,per,ul,uh,ds,getMinMaxParams(ds));}

  int hasDATAMIN() {return data ? data->hasDATAMIN() : 0;}
  int hasDATASEC() {return keyDATASEC;}
  int hasIRAFMIN() {return data ? data->hasIRAFMIN() : 0;}

  void bin(double* b, int l, double mn, double mx, Frame::ScanMode ds)
    {data->bin(b,l,mn,mx,getDataParams(ds));}

  char* displayHeader();
  char* displayPrimary();
  FitsHead* getHead();
  char* getCards();
  int getNCards();
  char* getKeyword(const char*);

  int FitsImage::saveFitsImageFile(const char* fn, int compress)
    {return image ? image->saveFitsImageFile(fn, compress) : 0;}

  int FitsImage::saveFitsImageChannel(Tcl_Interp* interp, const char* ch, 
				      int compress)
    {return image ? image->saveFitsImageChannel(interp, ch, compress) : 0;}

  int FitsImage::saveFitsImageSocket(int s, int compress)
    {return image ? image->saveFitsImageSocket(s, compress) : 0;}

  int FitsImage::saveFitsTableFile(const char* fn, int compress)
    {return fits ? fits->saveFitsTableFile(fn, compress) : 0;}

  int FitsImage::saveFitsTableChannel(Tcl_Interp* interp, const char* ch, 
				      int compress)
    {return fits ? fits->saveFitsTableChannel(interp, ch, compress): 0;}

  int FitsImage::saveFitsTableSocket(int s, int compress)
    {return fits ? fits->saveFitsTableSocket(s, compress) : 0;}
};

class FitsImageMaster : public FitsImage {
public:
  FitsImageMaster(Frame* p, Frame::MosaicType m =Frame::NOMOSAIC, int fm =0) 
    : FitsImage(p,m,fm) {}
  virtual ~FitsImageMaster();
};

class FitsImageSlave : public FitsImage {
public:
  FitsImageSlave(Frame*, const char*, FitsFile*);
  virtual ~FitsImageSlave() {}
};

// Fits

class FitsImageFitsAlloc : public FitsImageMaster {
public:
  FitsImageFitsAlloc(Frame*, const char*, 
		     Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsAllocGZ : public FitsImageMaster {
public:
  FitsImageFitsAllocGZ(Frame*, const char*, 
		       Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsChannel : public FitsImageMaster {
public:
  FitsImageFitsChannel(Frame*, Tcl_Interp*, const char*, const char*,
		       Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsMMap : public FitsImageMaster {
public:
  FitsImageFitsMMap(Frame*, const char*, 
		    Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsShare : public FitsImageMaster {
public:
  FitsImageFitsShare(Frame*, Frame::ShmType, int, int, const char*, 
		     Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsSocket : public FitsImageMaster {
public:
  FitsImageFitsSocket(Frame*, int, const char*, 
		      Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsSocketGZ : public FitsImageMaster {
public:
  FitsImageFitsSocketGZ(Frame*, int, const char*, 
			Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

class FitsImageFitsVar : public FitsImageMaster {
public:
  FitsImageFitsVar(Frame*, Tcl_Interp*, const char*, const char*,
		   Frame::MosaicType =Frame::NOMOSAIC, int =0);
};

// Fits Next

class FitsImageFitsNextAlloc : public FitsImageMaster {
public:
  FitsImageFitsNextAlloc(Frame*, const char*, FitsFile*, 
			 Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextAllocGZ : public FitsImageMaster {
public:
  FitsImageFitsNextAllocGZ(Frame*, const char*, FitsFile*, 
			   Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextChannel : public FitsImageMaster {
public:
  FitsImageFitsNextChannel(Frame*, const char*, FitsFile*,
			   Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextMMap : public FitsImageMaster {
public:
  FitsImageFitsNextMMap(Frame*, const char*, FitsFile*,
			Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextShare : public FitsImageMaster {
public:
  FitsImageFitsNextShare(Frame*, const char*, FitsFile*,
			 Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextSocket : public FitsImageMaster {
public:
  FitsImageFitsNextSocket(Frame*, const char*, FitsFile*,
			  Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextSocketGZ : public FitsImageMaster {
public:
  FitsImageFitsNextSocketGZ(Frame*, const char*, FitsFile*,
			    Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

class FitsImageFitsNextVar : public FitsImageMaster {
public:
  FitsImageFitsNextVar(Frame*, const char*, FitsFile*,
		       Frame::MosaicType= Frame::NOMOSAIC, int= 0);
};

// Array

class FitsImageArrAlloc : public FitsImageMaster {
public:
  FitsImageArrAlloc(Frame*, const char*);
};

class FitsImageArrAllocGZ : public FitsImageMaster {
public:
  FitsImageArrAllocGZ(Frame*, const char*);
};

class FitsImageArrChannel : public FitsImageMaster {
public:
  FitsImageArrChannel(Frame*, Tcl_Interp*, const char*, const char*);
};

class FitsImageArrMMap : public FitsImageMaster {
public:
  FitsImageArrMMap(Frame*, const char*);
};

class FitsImageArrShare : public FitsImageMaster {
public:
  FitsImageArrShare(Frame*, Frame::ShmType, int, int, const char*);
};

class FitsImageArrSocket : public FitsImageMaster {
public:
  FitsImageArrSocket(Frame*, int, const char*);
};

class FitsImageArrSocketGZ : public FitsImageMaster {
public:
  FitsImageArrSocketGZ(Frame*, int, const char*);
};

class FitsImageArrVar : public FitsImageMaster {
public:
  FitsImageArrVar(Frame*, Tcl_Interp*, const char*, const char*);
};

// Mosaic

class FitsImageMosaicAlloc : public FitsImageMaster {
public:
  FitsImageMosaicAlloc(Frame*, const char*, Frame::MosaicType);
};

class FitsImageMosaicAllocGZ : public FitsImageMaster {
public:
  FitsImageMosaicAllocGZ(Frame*, const char*, Frame::MosaicType);
};

class FitsImageMosaicChannel : public FitsImageMaster {
public:
  FitsImageMosaicChannel(Frame*, Tcl_Interp*, const char*, const char*,
			 Frame::MosaicType);
};

class FitsImageMosaicMMap : public FitsImageMaster {
public:
  FitsImageMosaicMMap(Frame*, const char*, Frame::MosaicType);
};

class FitsImageMosaicShare : public FitsImageMaster {
public:
  FitsImageMosaicShare(Frame*, Frame::ShmType, int, int, const char*,
		       Frame::MosaicType);
};

class FitsImageMosaicSocket : public FitsImageMaster {
public:
  FitsImageMosaicSocket(Frame*, int, const char*, Frame::MosaicType);
};

class FitsImageMosaicSocketGZ : public FitsImageMaster {
public:
  FitsImageMosaicSocketGZ(Frame*, int, const char*, Frame::MosaicType);
};

class FitsImageMosaicVar : public FitsImageMaster {
public:
  FitsImageMosaicVar(Frame*, Tcl_Interp*, const char*, const char*,
		     Frame::MosaicType);
};

// Mosaic Next

class FitsImageMosaicNextAlloc : public FitsImageMaster {
public:
  FitsImageMosaicNextAlloc(Frame*, const char*, FitsFile*, 
			   Frame::MosaicType, int);
};

class FitsImageMosaicNextAllocGZ : public FitsImageMaster {
public:
  FitsImageMosaicNextAllocGZ(Frame*, const char*, FitsFile*, 
			     Frame::MosaicType, int);
};

class FitsImageMosaicNextChannel : public FitsImageMaster {
public:
  FitsImageMosaicNextChannel(Frame*, const char*, FitsFile*,
			     Frame::MosaicType, int);
};

class FitsImageMosaicNextMMap : public FitsImageMaster {
public:
  FitsImageMosaicNextMMap(Frame*, const char*, FitsFile*,
			  Frame::MosaicType, int);
};

class FitsImageMosaicNextShare : public FitsImageMaster {
public:
  FitsImageMosaicNextShare(Frame*, const char*, FitsFile*,
			   Frame::MosaicType, int);
};

class FitsImageMosaicNextSocket : public FitsImageMaster {
public:
  FitsImageMosaicNextSocket(Frame*, const char*, FitsFile*,
			    Frame::MosaicType, int);
};

class FitsImageMosaicNextSocketGZ : public FitsImageMaster {
public:
  FitsImageMosaicNextSocketGZ(Frame*, const char*, FitsFile*,
			      Frame::MosaicType, int);
};

class FitsImageMosaicNextVar : public FitsImageMaster {
public:
  FitsImageMosaicNextVar(Frame*, const char*, FitsFile*,
			 Frame::MosaicType, int);
};

// IIS

class FitsImageIIS : public FitsImageMaster {
public:
  FitsImageIIS(Frame*, int, int);
  void iisErase();
  const char* iisGet(int x, int y);
  void iisSet(int x, int y, const char* d, int l);
  void iisWCS(const Matrix&, const Vector&, int);
};

#endif
