#include "fitsimage.h"
#include "util.h"

Vector FitsImage::map(const Vector& v, CoordSystem in, CoordSystem out)
{
  if (in==out)
    return v;

  switch (in) {
  case NATIVEWCS:
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    return mapWCS(v,in,out);

  case IMAGE:
    return mapImage(v,out);
  case PHYSICAL:
    return mapPhysical(v,out);
  case AMPLIFIER:
    return mapAmplifier(v,out);
  case DETECTOR:
    return mapDetector(v,out);
  case REF:
    return mapRef(v,out);
  case USER:
    return mapUser(v,out);
  case CANVAS:
    return mapCanvas(v,out);
  case PANNER:
    return mapPanner(v,out);

  default:
    cerr << "FitsImage Internal Error: unable to map coord system" 
	 << endl;
  }

  // if we get here, we have an error condition

  return v;
}

Vector FitsImage::mapWCS(const Vector& v, CoordSystem in, CoordSystem out)
{
  if (in!=NATIVEWCS) {
    switch (out) {
    case IMAGE:
      return wcs2pix(v, in);
    case DATA:
      return wcs2pix(v, in) * imageToData;
    case REF:
      return wcs2pix(v, in) * imageToRef;
    case USER:
      return wcs2pix(v, in) * imageToUser;
    default:
      cerr << "FitsImage Internal Error: unable to map WCS." << endl;
    }
  }
  else {
    switch (out) {
    case IMAGE:
      return wcs2pix(v);
    case DATA:
      return wcs2pix(v) * imageToData;
    case REF:
      return wcs2pix(v) * imageToRef;
    case USER:
      return wcs2pix(v) * imageToUser;
    default:
      cerr << "FitsImage Internal Error: unable to map WCS." << endl;
    }
  }

  return v;
}

Vector FitsImage::mapImage(const Vector& v, CoordSystem out)
{
  switch (out) {
  case DATA:
    return v * imageToData;
  case REF:
    return v * imageToRef;
  case USER:
    return v * imageToUser;

  default:
    cerr << "FitsImage Internal Error: unable to map IMAGE." << endl;
  }

  return v;
}

Vector FitsImage::mapPhysical(const Vector& v, CoordSystem out)
{
  switch (out) {
  case DATA:
    return v * physicalToData;
  case REF:
    return v * physicalToRef;
  case USER:
    return v * physicalToUser;

  default:
    cerr << "FitsImage Internal Error: unable to map Physical" << endl;
  }

  return v;
}

Vector FitsImage::mapAmplifier(const Vector& v, CoordSystem out)
{
  switch (out) {
  case DATA:
    return v * amplifierToPhysical * physicalToData;
  case REF:
    return v * amplifierToRef;
  case USER:
    return v * amplifierToUser;

  default:
    cerr << "FitsImage Internal Error: unable to map AMPLIFIER." << endl;
  }

  return v;
}

Vector FitsImage::mapDetector(const Vector& v, CoordSystem out)
{
  switch (out) {
  case DATA:
    return v * detectorToPhysical * physicalToData;
  case REF:
    return v * detectorToRef;
  case USER:
    return v * detectorToUser;

  default:
    cerr << "FitsImage Internal Error: unable to map DETECTOR." << endl;
  }

  return v;
}

Vector FitsImage::mapRef(const Vector& v, CoordSystem out)
{
  switch (out) {
  case USER:
    return v * parent->refToUser;
  case WIDGET:
    return v * parent->refToWidget;
  case CANVAS:
    return v * parent->refToCanvas;
  case WINDOW:
    return v * parent->refToWindow;

  case DATA:
    return v * refToData;
  case IMAGE:
    return v * refToImage;
  case PHYSICAL:
    return v * refToPhysical;
  case AMPLIFIER:
    return v * refToAmplifier;
  case DETECTOR:
    return v * refToDetector;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (hasWCS())
      return pix2wcs(v * refToImage, out);
    break;
  
  default:
    cerr << "FitsImage Internal Error: unable to map REF." << endl;
  }

  return v;
}      

Vector FitsImage::mapUser(const Vector& v, CoordSystem out)
{
  switch (out) {
  case REF:
    return v * parent->userToRef;
  case WIDGET:
    return v * parent->userToWidget;
  case CANVAS:
    return v * parent->userToCanvas;

  case DATA:
    return v * userToData;
  case IMAGE:
    return v * userToImage;
  case PHYSICAL:
    return v * userToPhysical;
  case AMPLIFIER:
    return v * userToAmplifier;
  case DETECTOR:
    return v * userToDetector;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (hasWCS())
      return pix2wcs(v * userToImage, out);
    break;

  default:
    cerr << "FitsImage Internal Error: unable to map USER." << endl;
  }

  return v;
}

Vector FitsImage::mapCanvas(const Vector& v, CoordSystem out)
{
  switch (out) {
  case REF:
    return v * parent->canvasToRef;
  case USER:
    return v * parent->canvasToUser;
  case WIDGET:
    return v * parent->canvasToWidget;
  case WINDOW:
    return v * parent->canvasToWindow;

  case DATA:
    return v * canvasToData;
  case IMAGE:
    return v * canvasToImage;
  case PHYSICAL:
    return v * canvasToPhysical;
  case AMPLIFIER:
    return v * canvasToAmplifier;
  case DETECTOR:
    return v * canvasToDetector;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (hasWCS())
      return pix2wcs(v * canvasToImage, out);
    break;

  default:
    cerr << "FitsImage Internal Error: unable to map CANVAS." << endl;
  }

  return v;
}

Vector FitsImage::mapPanner(const Vector& v, CoordSystem out)
{
  switch (out) {
  case REF:
    return v * parent->pannerToRef;
  case USER:
    return v * parent->pannerToUser;

  case DATA:
    return v * pannerToData;
  case IMAGE:
    return v * pannerToImage;
  case PHYSICAL:
    return v * pannerToPhysical;
  case AMPLIFIER:
    return v * pannerToAmplifier;
  case DETECTOR:
    return v * pannerToDetector;

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    if (hasWCS())
      return pix2wcs(v * pannerToImage, out);
    break;

  default:
    cerr << "FitsImage Internal Error: unable to map PANNER." << endl;
  }

  return v;
}

void FitsImage::mapToWCSStr(const Vector& v, CoordSystem in, CoordSystem out, 
			    CoordFormat format, char* buf, int length)
{
  if (hasWCS()) {
    switch (in) {
    case REF:
      pix2wcs(v * refToImage, out, format, buf, length);
      return;
    case USER:
      pix2wcs(v * userToImage, out, format, buf, length);
      return;
    case CANVAS:
      pix2wcs(v * canvasToImage, out, format, buf, length);
      return;
    case PANNER:
      pix2wcs(v * pannerToImage, out, format, buf, length);
      return;

    default:
      cerr << "FitsImage Internal Error: unable to mapToWCSStr." << endl;
    }
  }

  // if we get here, we have an error condition

  strcpy(buf,"");
}

// mapLen

double FitsImage::mapLenToRef(double d, CoordSystem sys, CoordFormat format)
{
  Vector r = mapLenToRef(Vector(d,0), sys, format);
  return r[0];
}

Vector FitsImage::mapLenToRef(const Vector& v, CoordSystem sys, 
			      CoordFormat format)
{
  switch (sys) {
  case CANVAS:
    return v; // special case

  case IMAGE:
    {
      Vector v1 = Vector() * imageToRef;
      Vector v2 = v * imageToRef;
      return (v2-v1).abs();
    }
  case PHYSICAL:
    {
      Vector v1 = Vector() * physicalToRef;
      Vector v2 = v * physicalToRef;
      return (v2-v1).abs();
    }
  case AMPLIFIER:
    {
      Vector v1 = Vector() * amplifierToPhysical * physicalToRef;
      Vector v2 = v * amplifierToPhysical * physicalToRef;
      return (v2-v1).abs();
    }
  case DETECTOR:
    {
      Vector v1 = Vector() * detectorToPhysical * physicalToRef;
      Vector v2 = v * detectorToPhysical * physicalToRef;
      return (v2-v1).abs();
    }

  case NATIVEWCS:
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS: 
    switch (format) {
    case PIXELS:
    case DEGREES:
      return wcs2pixDist(v);
    case ARCMIN:
      return wcs2pixDist(v/60);
    case ARCSEC:
      return wcs2pixDist(v/(60*60));
    }
    break;
  }

  // error
  cerr << "FitsImage Internal Error: mapLenToRef." << endl;
  return v;
}

double FitsImage::mapLenFromRef(double d, CoordSystem sys, CoordFormat format)
{
  Vector r = mapLenFromRef(Vector(d,0),sys,format);
  return r[0];
}

Vector FitsImage::mapLenFromRef(const Vector& v, CoordSystem sys,
				CoordFormat format)
{
  // really from image coords

  switch (sys) {
  case IMAGE:
    {
      Vector v1 = Vector() * refToImage;
      Vector v2 = v * refToImage;
      return (v2-v1).abs();
    }
  case PHYSICAL:
    {
      Vector v1 = Vector() * refToPhysical;
      Vector v2 = v * refToPhysical;
      return (v2-v1).abs();
    }
  case AMPLIFIER:
    {
      Vector v1 = Vector() * refToPhysical * physicalToAmplifier;
      Vector v2 = v * refToPhysical * physicalToAmplifier;
      return (v2-v1).abs();
    }
  case DETECTOR:
    {
      Vector v1 = Vector() * refToPhysical * physicalToDetector;
      Vector v2 = v * refToPhysical * physicalToDetector;
      return (v2-v1).abs();
    }

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    switch (format) {
    case PIXELS:
    case DEGREES:
      return pix2wcsDist(v);
    case ARCMIN:
      return pix2wcsDist(v*60);
    case ARCSEC:
      return pix2wcsDist(v*60*60);
    }
    break;
  }

  // error
  cerr << "FitsImage Internal Error: mapLenFromRef" << endl;
  return v;
}

double FitsImage::mapLenFromRef(const Vector& vv1, const Vector& vv2, 
			    CoordSystem sys, CoordFormat format)
{
  switch (sys) {
  case IMAGE:
    {
      Vector v1 = vv1 * refToImage;
      Vector v2 = vv2 * refToImage;
      return (v2-v1).length();
    }
  case PHYSICAL:
    {
      Vector v1 = vv1 * refToPhysical;
      Vector v2 = vv2 * refToPhysical;
      return (v2-v1).length();
    }
  case AMPLIFIER:
    {
      Vector v1 = vv1 * refToPhysical * physicalToAmplifier;
      Vector v2 = vv2 * refToPhysical * physicalToAmplifier;
      return (v2-v1).length();
    }
  case DETECTOR:
    {
      Vector v1 = vv1 * refToPhysical * physicalToDetector;
      Vector v2 = vv2 * refToPhysical * physicalToDetector;
      return (v2-v1).length();
    }

  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    switch (format) {
    case PIXELS:
    case DEGREES:
      {
	Vector v1 = pix2wcs(vv1 * refToImage);
	Vector v2 = pix2wcs(vv2 * refToImage);
	return wcsdist(v1,v2);
      }
    case ARCMIN:
      {
	Vector v1 = pix2wcs(vv1 * refToImage);
	Vector v2 = pix2wcs(vv2 * refToImage);
	return wcsdist(v1,v2)*60;
      }
    case ARCSEC:
      {
	Vector v1 = pix2wcs(vv1 * refToImage);
	Vector v2 = pix2wcs(vv2 * refToImage);
	return wcsdist(v1,v2)*60*60;
      }
    }
  }

  // error
  cerr << "FitsImage Internal Error: mapLenFromRef" << endl;
  return 0;
}

