#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <iomanip.h>

#include <Xlib.h>
#include <Xutil.h>

#include "frame.h"
#include "util.h"
#include "fitsimage.h"
#include "nan.h"

#include "contour.h"
#include "grid.h"

#include "hist.h"

#define BLT 1

#if BLT == 1
#include "blt.h"
#endif

// Debug

int DebugAST= 0;
int DebugMosaic= 0;
int DebugPerf= 0;
int DebugWCS= 0;
int DebugGZ= 0;

// Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer frFlexLexer
#include <FlexLexer.h>

frFlexLexer* frlexx = NULL;    // used by frlex
Frame* fr = NULL;           // used by frerror
extern int frparse(void*);

int frlex()
{
  return (frlexx ? frlexx->yylex() : 0);
}

void frerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = frlexx ? frlexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// Frame Member Functions

Frame::Frame(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) : Widget(i, c, item)
{
  name = NULL;
  fits = NULL;
  currentFits = NULL;
  fitsCount = 0;
  slice = 1;
  basePixmap = 0;
  baseXImage = NULL;

  displayMode = SINGLE;

  needsUpdate = NOUPDATE;

  zoom = 1;
  rotation = 0;
  orientation = NORMAL;

  wcsAlignment = 0;
  wcsZoom = 1;
  wcsRotation = 0;
  wcsOrientation = NORMAL;
  
  mosaicFast = 1;

  panPM = 0;
  panGCXOR = NULL;

  rotateRotation = 0;
  rotateSrcXM = NULL;
  rotateDestXM = NULL;
  rotatePM = 0;
  rotateGCXOR = NULL;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;
  invert = 0;
  cmapArea = 0;
  cmapMode = CENTER;

  colorCount = 0;

  colorScale = NULL;
  colorScaleType = LINEARSCALE;
  colorCells = NULL;
  indexCells = NULL;

  clipScope = LOCAL;
  clipMode = MINMAX;
  mmMode = SCAN;

  min = low = uLow = 1;
  max = high = uHigh = 100;

  zContrast = .25;
  zSampleSize = 600;
  zSampleLine = 120;

  mmIncr = 10;

  scanMode = DATASEC;

  autoCutPer = 1;

  useCrosshair = 0;
  crosshairGC = NULL;
  crosshairGCXOR = NULL;

  highliteGC = NULL;

  markerGC = NULL;
  selectGCXOR = NULL;
  markerDefaultSystem = IMAGE;
  editMarker = NULL;
  rotateMarker = NULL;
  undoMarkerType = NONE;

  pannerPixmap = 0;
  pannerXImage = NULL;
  pannerWidth = 0;
  pannerHeight = 0;
  pannerName[0] = '\0';
  usePanner = 0;

  magnifierPixmap = 0;
  magnifierXImage = NULL;
  magnifierWidth = 0;
  magnifierHeight = 0;
  magnifierName[0] = '\0';
  magnifierMode = ACCURATE;
  magnifierZoom = 4;
  useMagnifier = 0;
  useMagnifierGraphics = 1;
  useMagnifierCursor = 1;
  magnifierGC = NULL;

  useHighlite = 0;

  contour = NULL;
  contourGC = NULL;

  grid = NULL;
  gridGC = NULL;

  psLevel = 2;
  psResolution = 300;
  psInterpolate = 0;
  psColorSpace = RGB;

  binFunction = FitsHist::AVERAGE;
  binFactor = 1;
  binBufferSize = 1024;

  iisSaveRoot = 0;
  iisInteractive = 0;

  histequ = NULL;
  histogramX = NULL;
  histogramY = NULL;
}

Frame::~Frame()
{
  // delete fitsimages
  while (fits) {
    FitsImage* tmp = fits->next();
    delete fits;
    fits = tmp;
  }

  // delete markers
  markers.deleteAll();
  undoMarkers.deleteAll();

  // delete allocated storage
  if (name)
    delete name;

  if (basePixmap)
    Tk_FreePixmap(display, basePixmap);

  if (baseXImage)
    XDestroyImage(baseXImage);

  if (panPM)
    Tk_FreePixmap(display, panPM);

  if (crosshairGC)
    XFreeGC(display, crosshairGC);

  if (crosshairGCXOR)
    XFreeGC(display, crosshairGCXOR);

  if (highliteGC)
    XFreeGC(display, highliteGC);

  if (markerGC)
    XFreeGC(display, markerGC);

  if (selectGCXOR)
    XFreeGC(display, selectGCXOR);

  if (panGCXOR)
    XFreeGC(display, panGCXOR);

  if (rotateGCXOR)
    XFreeGC(display, rotateGCXOR);

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);

  if (pannerXImage)
    XDestroyImage(pannerXImage);

  if (magnifierPixmap)
    Tk_FreePixmap(display, magnifierPixmap);

  if (magnifierXImage)
    XDestroyImage(magnifierXImage);

  if (magnifierGC)
    XFreeGC(display, magnifierGC);

  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;

  if (contour)
    delete contour;
  auxcontours.deleteAll();

  if (contourGC)
    XFreeGC(display, contourGC);

  if (grid)
    delete grid;

  if (gridGC)
    XFreeGC(display, gridGC);

  if (histequ)
    delete histequ;
}

int Frame::parse(istrstream& istr)
{
  result = TCL_OK;
  frlexx = new frFlexLexer(&istr);
  fr = this;
  frparse(this);

  delete frlexx;
  frlexx = NULL;
  return result;
}

void Frame::reset()
{
  zoom = 1;
  rotation = 0;
  orientation = NORMAL;
  orientationMatrix.identity();

  wcsAlignment = 0;
  wcsZoom = 1;
  wcsRotation = 0;
  wcsOrientation = NORMAL;
  wcsOrientationMatrix.identity();
  
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;
  invert = 0;

  centerImage();

  unselectMarkers();

  crosshair = cursor;

  update(MATRIX);
}

void Frame::centerImage()
{
  // always center to center of pixel
  cursor = getCenter().floor() + Vector(.5,.5);
}

double Frame::getWCSRotation()
{
  return currentFits ? currentFits->getWCSRotation() : 0;
}

Orientation Frame::getWCSOrientation()
{
  return currentFits ? currentFits->getWCSOrientation() : NORMAL;
}

int Frame::hasLTMV()
{
  return currentFits && currentFits->hasLTMV();
}

int Frame::hasATMV() 
{
  return currentFits && currentFits->hasATMV();
}

int Frame::hasDTMV() 
{
  return currentFits && currentFits->hasDTMV();
}

int Frame::hasWCS() 
{
  return currentFits && currentFits->hasWCS();
}

int Frame::hasWCSEqu()
{
  return currentFits && currentFits->hasWCS() && currentFits->hasWCSEqu();
}

int Frame::hasWCSLinear()
{
  return currentFits && currentFits->hasWCS() && currentFits->hasWCSLinear();
}

// Rendering Routines

void Frame::invalidPixmap()
{
  Widget::invalidPixmap();

  if (basePixmap)
    Tk_FreePixmap(display, basePixmap);
  basePixmap = 0;

  if (baseXImage)
    XDestroyImage(baseXImage);
  baseXImage = NULL;

  needsUpdate = MATRIX;
}

void Frame::update(UpdateType flag)
{
  if (DebugPerf)
    cerr << "update" << endl;

  // Be careful, someone may have already set the flag at a lower level
  // therefor, only change the flag if we need more to be done

  if (flag < needsUpdate)
    needsUpdate = flag;
  redraw();
}

void Frame::update(UpdateType flag, BBox bb)
{
  if (DebugPerf)
    cerr << "update BB" << endl;

  // bb is in canvas coords

  if (flag < needsUpdate)
    needsUpdate = flag;
  redraw(bb);
}

void Frame::updateNow(UpdateType flag)
{
  if (DebugPerf)
    cerr << "updateNow" << endl;

  if (flag < needsUpdate)
    needsUpdate = flag;
  redrawNow();
}

void Frame::updateNow(UpdateType flag, BBox bb)
{
  if (DebugPerf)
    cerr << "updateNow BB" << endl;

  // bb is in canvas coords

  if (flag < needsUpdate)
    needsUpdate = flag;
  redrawNow(bb);
}

int Frame::updatePixmap(const BBox& bb)
{

  // Note: lack of breaks-- on purpose. 
  // If Matrices are update, both Base and Pixmap
  // also need to be updated. Same for Base-- ie, pixmap is also updated.
  // bb is in canvas coords

  switch (needsUpdate) {
  case MATRIX:
    updateMatrices();
    updateMarkers();
  case BASE:
    updateBase(bb);
  case PIXMAP:
    updatePM(bb);
    break;
  }

  needsUpdate = NOUPDATE;
  return TCL_OK;
}

void Frame::updateBase(const BBox& bbox)
{
  //  XSync(display,0);
  if (DebugPerf)
    cerr << "updateBase... ";

  int& width = options->width;
  int& height = options->height;

  if (!basePixmap) {
    if (!(basePixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				    width, height, depth))){
      cerr << "Frame Internal Error: Unable to Create Pixmap" << endl;
      exit(1);
    }

    // Geometry has changed, redefine our marker GCs including clip regions

    updateGCs();
  }

  if (!baseXImage) {
    if (!(baseXImage = XGETIMAGE(display, basePixmap, 0, 0, width, height, 
				 AllPlanes, ZPixmap))) {
      cerr << "Frame Internal Error: Unable to Create XImage" << endl;
      exit(1);
    }

    // we have a race condition. Some Truecolor ColorScales need to know the 
    // bytes per pixel, RGB masks, and byte order, from XImage struct.
    // yet, we receive colormap commands well before we render to the 
    // screen and have a valid XImage. 
    // So, we put off creating a colorscale until we are ready to render.

    if (!colorScale)
      updateColorScale();
  }

  if (!bbox.isEmpty()) {
    BBox bb = bbox * canvasToWidget;
    int x0 = (int)bb.ll[0] > 0 ? (int)bb.ll[0] : 0;
    int y0 = (int)bb.ll[1] > 0 ? (int)bb.ll[1] : 0;
    int x1 = (int)bb.ur[0] < width ? (int)bb.ur[0] : width;
    int y1 = (int)bb.ur[1] < height ? (int)bb.ur[1] : height;

    if (DebugPerf)
      cerr << ' ' << x0 << ' ' << y0 << ' ' << x1 << ' ' << y1 << ' ';
    
    if (currentFits)
      ximageToPixmap(basePixmap, baseXImage, x0, y0, x1, y1,
		     &FitsImage::getmWidgetToData);
    else {
      XSetForeground(display, gc, getWhiteColor());
      XFillRectangle(display, basePixmap, gc, x0, y0, x1-x0, y1-y0);
    }
  }
  
  renderContours(basePixmap, refToWidget, bbox*canvasToRef);

  if (grid)
    if (!(grid->render()))
	cerr << "Frame Internal Error: Unable to render Grid" << endl;

  updatePanner();

  if (DebugPerf)
    cerr << "end" << endl;
}

void Frame::updatePM(const BBox& bbox)
{
  // bbox is in Canvas Coords

  if (DebugPerf)
    cerr << "updatePM...";

  int& width = options->width;
  int& height = options->height;

  if (!pixmap) {
    if (!(pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				width, height, depth))) {
      cerr << "Frame Internal Error: Unable to Create Pixmap" << endl;
      exit(1);
    }
  }

  if (!bbox.isEmpty()) {
    /*
    BBox bb = bbox * canvasToWidget;
    Vector size = bb.size();
    XCopyArea(display, basePixmap, pixmap, gc, 
	      (int)bb.ll[0], (int)bb.ll[1], (int)size[0], (int)size[1],
	      (int)bb.ll[0], (int)bb.ll[1]);
    */

    BBox bb = bbox * canvasToWidget;
    int x0 = (int)bb.ll[0] > 0 ? (int)bb.ll[0] : 0;
    int y0 = (int)bb.ll[1] > 0 ? (int)bb.ll[1] : 0;
    int x1 = (int)bb.ur[0] < width ? (int)bb.ur[0] : width;
    int y1 = (int)bb.ur[1] < height ? (int)bb.ur[1] : height;
    int sx = x1-x0;
    int sy = y1-y0;

    if (DebugPerf)
      cerr << ' ' << x0 << ' ' << y0 << ' ' << x1 << ' ' << y1 << ' ';

    XCopyArea(display, basePixmap, pixmap, gc, x0, y0, sx, sy, x0, y0);
  }

  renderMarkers(bbox); // render the markers into pixmap
  renderCrosshair(); // render crosshair into pixmap
  renderHighLite();  // render a hightlite bbox into pixmap

  if (DebugPerf)
    cerr << "end" << endl;
}

Vector Frame::getCenter()
{
  return currentFits ? currentFits->getCenter() : Vector();
}

void Frame::updateMatrices()
{
  if (DebugPerf)
    cerr << "updateMatrices...";

  // These are the basic tranformation matrices

  Vector center = getCenter();

  if (!isIIS()) {
    refToUser = 
      Translate(-center) *                // translate to center image
      FlipY() *                           // flip y axis for X Windows
      Translate(center);                  // translate back from center
    userToRef = refToUser.invert();
  }
  else {
    refToUser.identity();
    userToRef.identity();
  }

  userToWidget =
    Translate(-cursor) *                // translate to cursor position
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(zoom) *                       // scale about cursor position
    Scale(wcsZoom) *                    // scale about wcs scale
    Translate(cursor) *                 // translate back from cursor

    Translate(-cursor) *                // translate to cursor positon

 // translate to center of widget. needs to be integer

    Translate((int)(options->width/2.), (int)(options->height/2.));

  widgetToUser = userToWidget.invert();

  widgetToCanvas = Translate(originX, originY); // translate by widget location
  canvasToWidget = widgetToCanvas.invert();

  short x, y;
  Tk_CanvasWindowCoords(canvas, 0, 0, &x, &y);
  canvasToWindow = Translate(x, y);
  windowToCanvas = canvasToWindow.invert();

  // These are derived Transformation Matrices

  refToWidget = refToUser * userToWidget;
  widgetToRef = refToWidget.invert();

  refToCanvas = refToUser * userToWidget * widgetToCanvas;
  canvasToRef = refToCanvas.invert();

  refToWindow = refToUser * userToWidget * widgetToCanvas * canvasToWindow;
  windowToRef = refToWindow.invert();

  userToCanvas = userToWidget * widgetToCanvas;
  canvasToUser = userToCanvas.invert();

  widgetToWindow = widgetToCanvas * canvasToWindow;
  windowToWidget = widgetToWindow.invert();

    // calculate panner matrices

  userToPanner =
    Translate(-center) *
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(calcPannerZoom()) *
    Translate(pannerWidth/2., pannerHeight/2.);
  pannerToUser = userToPanner.invert();

  refToPanner = refToUser * userToPanner;
  pannerToRef = refToPanner.invert();

  updateMagnifierMatrix();

  // update all image matrices

  FitsImage* ptr = fits;
  while (ptr) {
    ptr->updateMatrices(refToUser, userToWidget, widgetToCanvas, userToPanner);
    ptr = ptr->next();
  }

  if (DebugPerf)
    cerr << "end" << endl;
}

int Frame::isIIS()
{
  return currentFits && currentFits->isIIS();
}

void Frame::ximageToPixmap(Pixmap pmap, XImage* xmap, 
			   int x0, int y0, int x1, int y1, 
			   double* (FitsImage::*getMatrix)())
{
  // generate new image

  switch (displayMode) {
  case SINGLE:
    fillXImage(xmap, x0, y0, x1, y1, getMatrix);
    break;
  case MOSAIC:
    if (mosaicFast && fits->mosaicType() == IRAF)
      fillXImageMosaicFast(xmap, x0, y0, x1, y1, getMatrix);
    else
      fillXImageMosaic(xmap, x0, y0, x1, y1, getMatrix);
    break;
  }

  // update

  XPutImage(display, pmap, gc, xmap, x0, y0, x0, y0, x1-x0, y1-y0);
}

void Frame::renderCrosshair()
{
  Vector r = crosshair * refToWidget;

  if (useCrosshair) {
    XDRAWLINE(display, pixmap, crosshairGC, 
	      (int)r[0], 1, (int)r[0], options->height);
    XDRAWLINE(display, pixmap, crosshairGC,
	      1, (int)r[1], options->width, (int)r[1]);
  }
}

void Frame::drawCrosshair()
{
  if (visible) {
    Vector r = crosshair * refToWidget;

    // Vertical line

    BBox vv = BBox(Vector(r[0], 1), Vector(r[0], options->height)) * 
      widgetToWindow;
    XDRAWLINE(display, Tk_WindowId(tkwin), crosshairGCXOR, 
	      (int)vv.ll[0], (int)vv.ll[1], (int)vv.ur[0], (int)vv.ur[1]);

    // Horizontal line

    BBox hh = BBox(Vector(1, r[1]), Vector(options->width, r[1])) * 
      widgetToWindow;
    XDRAWLINE(display, Tk_WindowId(tkwin), crosshairGCXOR, 
	      (int)hh.ll[0], (int)hh.ll[1], (int)hh.ur[0], (int)hh.ur[1]);
  }
}

void Frame::eraseCrosshair()
{
  if (visible) {
    // This routine schedules an update for only the area written to previously
    // by drawCursor.

    Vector r = crosshair * refToWidget;

    // Vertical line

    BBox vv = BBox(Vector(r[0], 1), Vector(r[0], options->height)) * 
      widgetToCanvas;
    redrawNow(vv.expand(1));

    // Horizontal line

    BBox hh = BBox(Vector(1, r[1]), Vector(options->width, r[1])) * 
      widgetToCanvas;
    redrawNow(hh.expand(1));
  }
}

void Frame::psCrosshair(PSColorSpace mode)
{
  Vector r = crosshair * refToWidget;
  Vector a = Vector(r[0],1) * widgetToCanvas;
  Vector b = Vector(r[0],options->height) * widgetToCanvas;
  Vector c = Vector(1,r[1]) * widgetToCanvas;
  Vector d = Vector(options->width,r[1]) * widgetToCanvas;

  char buf[256];
  {
    ostrstream str(buf,256);
    switch ((Frame::PSColorSpace)mode) {
    case Frame::BW:
    case Frame::GRAY:
      psColorGray("green", str);
      str << " setgray";
      break;
    case Frame::RGB:
      psColorRGB("green", str);
      str << " setrgbcolor";
      break;
    case Frame::CMYK:
      psColorCMYK("green", str);
      str << " setcmykcolor";
      break;
    }
    str << endl << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }  

  {
    ostrstream str(buf,256);
    str << "newpath " 
	<< a.TkCanvasPs(canvas) << "moveto"
	<< b.TkCanvasPs(canvas) << "lineto"
	<< " stroke" << endl
	<< "newpath " 
	<< c.TkCanvasPs(canvas) << "moveto"
	<< d.TkCanvasPs(canvas) << "lineto"
	<< " stroke" << endl
	<< ends;
    Tcl_AppendResult(interp, buf, NULL);
  }
}

void Frame::renderContours(Pixmap pm, const Matrix& mm, const BBox& bb)
{
  // render from back to front
  // aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->render(pm, mm, bb);
    ptr=ptr->previous();
  }

  // main contour
  if (contour)
    contour->render(pm, mm, bb);
}

void Frame::updateContourCoords(const Matrix& mm)
{
  // main contour
  if (contour)
    contour->updateCoords(mm);

  // aux contours
  Contour* ptr=auxcontours.head();
  while (ptr) {
    ptr->updateCoords(mm);
    ptr=ptr->next();
  }
}

void Frame::renderHighLite()
{
  if (useHighlite)
    XDRAWRECTANGLE(display, pixmap, highliteGC, 1, 1, 
		   options->width-2, options->height-2);
}

// Other Private Member Functions

void Frame::wcsAlign()
{
  if (hasWCS() && hasWCSEqu() && wcsAlignment) {
    wcsZoom = 1;
    wcsRotation = currentFits->getWCSRotation();
    wcsOrientation = currentFits->getWCSOrientation();
  }
  else {
    wcsZoom = 1;
    wcsRotation = 0;
    wcsOrientation = NORMAL;
  }

  switch (wcsOrientation) {
  case NORMAL:
    wcsOrientationMatrix.identity();
    break;
  case XX:
    wcsOrientationMatrix = FlipX();
    break;
  case YY:
    wcsOrientationMatrix = FlipY();
    break;
  case XY:
    wcsOrientationMatrix = FlipXY();
    break;
  }

}

void Frame::initHistogram()
{
  // delete blt histogram
  if (histogramX)
    delete histogramX;
  histogramX = NULL;
  if (histogramY)
    delete histogramY;
  histogramY = NULL;

  // delete scale histogram
  if (histequ)
    delete histequ;
  histequ = NULL;

  if (colorScaleType == HISTEQUSCALE)
    updateColorScale();
}

void Frame::updateHistEqu()
{
  // if we don't have any data, bail
  if (!fits) {
    if (histequ)
      delete histequ;
    histequ = NULL;
    return;
  }

  // if we already have it, bail
  if (histequ)
    return;

  if (DebugPerf)
    cerr << "updateHistEqu...";

  // create pdf or histogram
  double* pdf = new double[HISTEQUSIZE];
  memset(pdf,0,HISTEQUSIZE*sizeof(double));

  FitsImage* ptr = fits;
  while (ptr) {
    ptr->bin(pdf,HISTEQUSIZE,min,max,scanMode);
    ptr = ptr->next();
  }

  // find a total/average
  double total, average;
  {
    total = 0;
    for (int i=0; i<HISTEQUSIZE; i++)
      total += pdf[i];
    average = total/HISTEQUSIZE;
  }

  // knock down peaks
  {
    for (int i=0; i<HISTEQUSIZE; i++)
      if (pdf[i] > average*3)
	pdf[i] = average*3;
  }

  // recalc total/average
  {
    total = 0;
    for (int i=0; i<HISTEQUSIZE; i++)
      total += pdf[i];
    average = total/HISTEQUSIZE;
  }

  // build transfer function (cdf)
  histequ = new double[HISTEQUSIZE];

  double bin = 0;
  int color,level;
  for (color=level=0; level<HISTEQUSIZE && color<HISTEQUSIZE; level++) {
    histequ[level] = (double)color/HISTEQUSIZE;
    bin += pdf[level];
    while (bin>=average && color<HISTEQUSIZE) {
      bin -= average;
      color++;
    }
  }
  while (level<HISTEQUSIZE)
    histequ[level++] = (double)(HISTEQUSIZE-1)/HISTEQUSIZE;

  delete pdf;

  if (DebugPerf)
    cerr << "end" << endl;
}

void Frame::updateClip()
{
  min = DBL_MAX;
  max = -DBL_MAX;

  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->updateClip(clipMode, mmMode, mmIncr, 
		      zContrast, zSampleSize, zSampleLine, 
		      autoCutPer, 
		      uLow, uHigh, 
		      scanMode);

      // find over-all min/max
      if (min > ptr->getMinDouble())
	min = ptr->getMinDouble();
      
      if (max <= ptr->getMaxDouble())
	max = ptr->getMaxDouble();

      ptr = ptr->next();
    }

    // find over-all low/high
    updateLowHigh(&low,&high);

    if (clipScope == GLOBAL) {
      ptr = fits;
      while (ptr) {
	ptr->setClip(low, high);
	ptr = ptr->next();
      }
    }
  }
  else {
    if (clipMode != USERCLIP) {
      low = 1;
      high = 100;
    }
    else {
      low = uLow;
      high = uHigh;
    }
  }
}

void Frame::updateLowHigh(double* ll, double* hh)
{
  // find over-all low high

  *ll = DBL_MAX;
  *hh = -DBL_MAX;
  FitsImage* ptr = fits;
  while (ptr) {
    if (*ll > ptr->getLowDouble())
      *ll = ptr->getLowDouble();

    if (*hh <= ptr->getHighDouble())
      *hh = ptr->getHighDouble();

    ptr = ptr->next();
  }
}

Vector Frame::getClip(ClipMode cm, float ac)
{
  // this routine will change current settings, calcucate, then reset

  double ll, hh;

  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->updateClip(cm, mmMode, mmIncr,
		      zContrast, zSampleSize,  zSampleLine, 
		      ac,
		      uLow, uHigh, 
		      scanMode);
      ptr = ptr->next();
    }

    updateLowHigh(&ll, &hh);
  }
  else {
    if (cm != USERCLIP) {
      ll = 1;
      hh = 100;
    }
    else {
      ll = uLow;
      hh = uHigh;
    }
  }

  // now reset everything
  updateClip();

  return Vector(ll,hh);
}

void Frame::unloadFits()
{
  // clean up from iis if needed
  if (isIIS()) {
    clipScope = LOCAL;
    clipMode = MINMAX;
    mmMode = SAMPLE;

    colorScaleType = LINEARSCALE;
    updateColorScale();
  }

  // delete markers
  markers.deleteAll();

  // delete any contour
  if (contour)
    delete contour;
  contour = NULL;
  auxcontours.deleteAll();

  // delete any grid
  if (grid)
    delete grid;
  grid = NULL;

  // now, delete any fitsimages
  while (fits) {
    FitsImage* tmp = fits->next();
    delete fits;
    fits = tmp;
  }
  fits = NULL;
  currentFits = NULL;
  fitsCount = 0;
  displayMode = SINGLE;

  // delete histograms (equalization,blt)
  initHistogram();
}

double Frame::calcZoom(int width, int height, Vector ur)
{
  // we need to calculate the width and height of the rotated image
  // so we can derived a zoom factor to shrink it to fit the requested size

  Vector ll(0,0);
  Vector ul(ll[0],ur[1]);
  Matrix m = Translate(-ur/2.) * Rotate(wcsRotation) * Rotate(rotation);

  Vector cc = ll * m;
  Vector dd = ul * m;

  float z1 = fabs(cc[0]) > fabs(cc[1]) ? fabs(cc[0]) : fabs(cc[1]);
  float z2 = fabs(dd[0]) > fabs(dd[1]) ? fabs(dd[0]) : fabs(dd[1]);

  return ((width<height ? width:height) / ((z1>z2 ? z1:z2)*2));
}

// Color support

void Frame::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    cerr << "Frame Internal Error: Unable to Alloc indexCells" << endl;
    exit(1);
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }
  memcpy(colorCells, cells, cnt*3);
}

// Panner Support

double Frame::calcPannerZoom()
{
  if (pannerPixmap && currentFits)
    return calcZoom(pannerWidth,pannerHeight,currentFits->getSize());
  else
    return 1;
}

void Frame::updatePanner()
{
  if (usePanner)
    if (currentFits)
      ximageToPixmap(pannerPixmap, pannerXImage, 
		     0, 0, pannerWidth, pannerHeight,
		     &FitsImage::getmPannerToData);
    else
      XFillRectangle(display, pannerPixmap, gc, 
		     0, 0, pannerWidth, pannerHeight);
  
  pannerPush();
}

void Frame::pannerPush()
{
  if (usePanner) {
    char buf[512];
    ostrstream str(buf, 512);
    if (currentFits) {

      // calculate bbox

      Matrix widgetToPanner = widgetToUser * userToPanner;
      BBox bb = BBox(Vector(0,0), Vector(options->width, options->height)) * 
	widgetToPanner;

      // calculate compass vectors

      Rotate orient = Rotate(wcsRotation) * Rotate(rotation) * 
	wcsOrientationMatrix * orientationMatrix;

      str << pannerName << " update " << pannerPixmap << ";";
      str << pannerName << " update bbox " << bb << ";";
      str << pannerName << " update image compass " << orient << ";";

      if (hasWCS()) {
	Matrix m = currentFits->getImageToUser() * userToPanner;

	// find origin of wcs-- we must take the ref point,
	// run it thru pix2wcs (which may convert to J2000 if needed)
	// and then run it back thru wcs2pix to find our ref point in J2000

	Vector orval = currentFits->pix2wcs(currentFits->getWCScrpix(),
					    EQUATORIAL_FK5);
	Vector orpix = currentFits->wcs2pix(orval,EQUATORIAL_FK5);
	Vector origin = orpix * m;

	// find normalized north & east vectors

	Vector delta = (currentFits->getWCScdelt() * 10).abs();

	Vector north = 
	  currentFits->wcs2pix(Vector(orval[0],orval[1]+delta[1]),
			       EQUATORIAL_FK5) * m;
	Vector ndiff = north - origin;

	Vector east = 
	  currentFits->wcs2pix(Vector(orval[0]+delta[0],orval[1]),
			       EQUATORIAL_FK5) * m;
	Vector ediff = east - origin;

	// and update the panner

	str << pannerName << " update wcs compass " << origin
	    << ndiff.normalize() << ediff.normalize() << ends;
      }
      else
	str << pannerName << " update wcs compass invalid" << ends;
    }
    else
      str << pannerName << " clear" << ends;

    Tcl_Eval(interp, buf);
  }
}

// Magnifier Support

void Frame::updateMagnifier()
{
  updateMagnifier(magnifierCursor);
}

void Frame::updateMagnifier(const Vector& v)
{
  // v is in CANVAS coords
  // save it, we may need it later
  magnifierCursor = v;

  if (useMagnifier && magnifierXImage && magnifierGC && currentFits) {

    // clear XImage data

    memset(magnifierXImage->data, 0, magnifierXImage->bytes_per_line * 
	   magnifierHeight);

    // fill XImage

    switch (magnifierMode) {
    case FAST:
      updateMagnifierReplicate(v);
      break;
    case ACCURATE:
      updateMagnifierZoom(v); // pixmap is updated inside
      break;
    }

    // render cursor
    
    if (useMagnifierCursor)
      renderMagnifierCursor(v);

    // notify the magnifier widget

    char buf[64];
    ostrstream str(buf, 64);
    str << magnifierName << " update " << magnifierPixmap << ends;

    Tcl_Eval(interp, buf);
  }
}

void Frame::updateMagnifierMatrix()
{
  // calculate partial magnifier matrix
  // This is a partial matix, it lacks the Translate at the begining

  userToMagnifier =                     
    Rotate(rotation) *                  // rotate about cursor position
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Scale(zoom) *                       // scale about cursor position
    Scale(wcsZoom) *                    // scale about wcs scale
    Scale(magnifierZoom) *              // scale
    Translate(magnifierWidth/2., magnifierHeight/2.);

  userToMagnifierCursor =                     
    Rotate(rotation) *
    Rotate(wcsRotation) *
    Scale(zoom) *
    Scale(wcsZoom) *
    Scale(magnifierZoom) *
    Translate(magnifierWidth/2., magnifierHeight/2.);

  Vector magCorrection = Vector(.5,-.5) *
    Rotate(rotation) *
    Rotate(wcsRotation) *
    orientationMatrix *
    wcsOrientationMatrix;

  ximageToMagnifier = 
    Translate(magCorrection) * // we need to offset by this correction
    Scale(magnifierZoom) *
    Translate(magnifierWidth/2.,magnifierHeight/2.);
}

void Frame::updateMagnifierZoom(const Vector& v)
{
  // v is in CANVAS coords
  Matrix refToMagnifier = refToUser * Translate(-(v * canvasToUser)) * 
    userToMagnifier;
  Matrix magnifierToRef = refToMagnifier.invert();

  // Update magnifier matrices

  FitsImage* ptr = fits;
  while (ptr) {
    ptr->updateMagnifierMatrices(refToMagnifier);
    ptr = ptr->next();
  }

  // Update ximage

  switch (displayMode) {
  case SINGLE:
    fillXImage(magnifierXImage, 0, 0, magnifierWidth, magnifierHeight, 
	       &FitsImage::getmMagnifierToData);
    break;
  case MOSAIC:
    if (mosaicFast && fits->mosaicType() == IRAF)
      fillXImageMosaicFast(magnifierXImage, 0, 0, 
			   magnifierWidth, magnifierHeight, 
			   &FitsImage::getmMagnifierToData);
    else
      fillXImageMosaic(magnifierXImage, 0, 0, 
		       magnifierWidth, magnifierHeight, 
		       &FitsImage::getmMagnifierToData);
    break;
  }

  // copy XImage to Pixmap

  XPutImage(display, magnifierPixmap, gc, magnifierXImage, 0, 0, 0, 0, 
	    magnifierWidth, magnifierHeight);

  if (useMagnifierGraphics) {

    // render markers
    // markers bounding box is in canvas coords
    // map the magnifier to a bounding box in canvas coords
    Matrix m = magnifierToRef * refToCanvas;
    BBox bb = BBox(v,v);
    bb.bound(Vector(0,0) * m);
    bb.bound(Vector(magnifierWidth,magnifierHeight) * m);
    renderMagnifierMarkers(refToMagnifier, bb);

    // render crosshair
    if (useCrosshair) {
      Vector r = crosshair * refToMagnifier;
      if (r[0]>=0 && r[0]<magnifierWidth)
	XDRAWLINE(display, magnifierPixmap, crosshairGC, 
		  (int)r[0], 1, (int)r[0], magnifierHeight);

      if (r[1]>=0 && r[1]<magnifierHeight)
	// bump Y by one, I don't know why, it just works
	XDRAWLINE(display, magnifierPixmap, crosshairGC,
		  1, (int)r[1], magnifierWidth, (int)r[1]);

    }

    // render contours
    renderContours(magnifierPixmap, refToMagnifier, bb*canvasToRef);
  }
}

void Frame::renderMagnifierCursor(const Vector& v)
{
    // first, the inner white box

    Vector u = v * canvasToUser;
    Matrix m = Translate(-u) * userToMagnifierCursor;

    Vector c[5];
    c[0] = (u + Vector(-.5,-.5)) * m;
    c[1] = (u + Vector( .5,-.5)) * m;
    c[2] = (u + Vector( .5, .5)) * m;
    c[3] = (u + Vector(-.5, .5)) * m;
    c[4] = c[0];
    
    XPoint pts[5];
    for (int i=0; i<5; i++) {
      Vector z = c[i].round();
      pts[i].x = (int)z[0];
      pts[i].y = (int)z[1];
    }
    XSetForeground(display, magnifierGC, whiteColor->pixel);
    XDRAWLINES(display, magnifierPixmap, magnifierGC, pts, 5, CoordModeOrigin);

    // now, the outer black box

    Matrix n =
      Translate(-(u * userToMagnifierCursor)) *
      Rotate(-rotation) *
      Rotate(-wcsRotation);
    Matrix o = n.invert();

    Vector d[5];
    {
      for (int i=0; i<5; i++)
	d[i] = c[i] * n;
    }

    d[0]+=Vector(-1,-1);
    d[1]+=Vector( 1,-1);
    d[2]+=Vector( 1, 1);
    d[3]+=Vector(-1, 1);
    d[4]=d[0];

    {
      for (int i=0; i<5; i++) {
	d[i] = d[i] * o;
	Vector z = d[i].round();
	pts[i].x = (int)z[0];
	pts[i].y = (int)z[1];
      }
    }
    XSetForeground(display, magnifierGC, blackColor->pixel);
    XDRAWLINES(display, magnifierPixmap, magnifierGC, pts, 5, CoordModeOrigin);
}

void Frame::updateGCs()
{
  // widget clip region

  BBox bbWidget = BBox(0, 0, options->width-1, options->height-1);
  Vector sizeWidget = bbWidget.size();

  rectWidget[0].x = (int)bbWidget.ll[0];
  rectWidget[0].y = (int)bbWidget.ll[1];
  rectWidget[0].width = (int)sizeWidget[0];
  rectWidget[0].height = (int)sizeWidget[1];

  // window clip region

  BBox bbWindow = bbWidget * widgetToWindow;
  Vector sizeWindow = bbWindow.size();

  rectWindow[0].x = (int)bbWindow.ll[0];
  rectWindow[0].y = (int)bbWindow.ll[1];
  rectWindow[0].width = (int)sizeWindow[0];
  rectWindow[0].height = (int)sizeWindow[1];

  // crosshairGC

  if (!crosshairGC) {
    crosshairGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, crosshairGC, greenColor->pixel);
    XSetLineAttributes(display, crosshairGC, 1, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, crosshairGC, 0, 0, rectWidget, 1, Unsorted);

  // crosshairGCXOR

  if (!crosshairGCXOR) {
    crosshairGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, crosshairGCXOR, greenColor->pixel);
    XSetLineAttributes(display, crosshairGCXOR, 1, LineSolid, 
		       CapButt, JoinMiter);
  }
  XSetClipRectangles(display, crosshairGCXOR, 0, 0, rectWindow, 1, Unsorted);

  if (!highliteGC) {
    highliteGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, highliteGC, blueColor->pixel);
    XSetLineAttributes(display, highliteGC, 2, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, highliteGC, 0, 0, rectWidget, 1, Unsorted);

  // contourGC

  if (!contourGC) {
    contourGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetLineAttributes(display, contourGC, 1, LineSolid, CapButt, JoinMiter);
  }
  XSetClipRectangles(display, contourGC, 0, 0, rectWidget, 1, Unsorted);

  // gridGC

  if (!gridGC) {
    gridGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  }
  XSetClipRectangles(display, gridGC, 0, 0, rectWidget, 1, Unsorted);

  // markerGC

  if (!markerGC)
    markerGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  if (!selectGCXOR) {
    selectGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
    XSetForeground(display, selectGCXOR, getWhiteColor());
    char dlist[] = {8,3};
    XSetLineAttributes(display, selectGCXOR, 1, LineOnOffDash, 
                     CapButt, JoinMiter);
    XSetDashes(display, selectGCXOR, 0, dlist, 2);
  }
  XSetClipRectangles(display, selectGCXOR, 0, 0, rectWindow, 1, Unsorted);

  // panGCXOR

  if (!panGCXOR)
    panGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, panGCXOR, 0, 0, rectWindow, 1, Unsorted);

  // rotateGCXOR

  if (!rotateGCXOR)
    rotateGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, rotateGCXOR, 0, 0, rectWindow, 1, Unsorted);
}

ContourScale* Frame::contourScale(int cnt, float l, float h, 
				  ColorScaleType type)
{
  switch (type) {
  case Frame::IISSCALE:
  case Frame::LINEARSCALE:
    return new LinearContourScale(cnt, l, h);
  case Frame::LOGSCALE:
    return new LogContourScale(cnt, l, h);
  case Frame::SQUAREDSCALE:
    return new SquaredContourScale(cnt, l, h);
  case Frame::SQRTSCALE:
    return new SqrtContourScale(cnt, l, h);
  case Frame::HISTEQUSCALE:
    updateHistEqu();
    return new HistEquContourScale(cnt, min, max, histequ, HISTEQUSIZE);
  }
}

void Frame::updateBin(const Matrix& m)
{
  crosshair *= m;
  updateMarkerCoords(m);
  updateContourCoords(m);
}

double Frame::mapAngleToRef(double angle, CoordSystem in)
{
  switch (in) {
  case NATIVEWCS:
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    {
      switch (getWCSOrientation()) {
      case NORMAL:
      case XY:
	return angle - getWCSRotation();
      case XX:
      case YY:
	return -angle - getWCSRotation();
      }
    }
  default:
    return angle;
  }
}

double Frame::mapAngleFromRef(double angle, CoordSystem out)
{
  switch (out) {
  case NATIVEWCS:
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR:
  case ICRS:
    {
      switch (getWCSOrientation()) {
      case NORMAL:
      case XY:
	return angle + getWCSRotation();
      case XX:
      case YY:
	return -angle - getWCSRotation();
      }
    }
  default:
    return angle;
  }
}

FitsImage* Frame::isInFits(const Vector& v, Matrix& (FitsImage::*getMatrix)(),
			   Vector* rv)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      Vector r = v * (currentFits->*getMatrix)();
      int* params = currentFits->getDataParams(scanMode);
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      if (r[0]>=xmin && r[0]<xmax && r[1]>=ymin && r[1]<ymax) {
	if (rv)
	  *rv = r;
	return currentFits;
      }
      else
	return NULL;
    }
    else
      return NULL;

  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      Vector r = v * (ptr->*getMatrix)();
      int* params = ptr->getDataParams(scanMode);
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      if (r[0]>=xmin && r[0]<xmax && r[1]>=ymin && r[1]<ymax) {
	if (rv)
	  *rv = r;
	return ptr;
      }
      ptr = ptr->next();
    }

    return NULL;
  }
}

FitsImage* Frame::findFits(const Vector& c)
{
  switch (displayMode) {
  case SINGLE:
    return currentFits;
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = c * ptr->getRefToData();
      int* params = ptr->getDataParams(scanMode);
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax)
	return ptr;

      ptr = ptr->next();
    }

    return currentFits;
    break;
  }
}

FitsImage* Frame::findFits(int which)
{
  switch (displayMode) {
  case SINGLE:
    return currentFits;
  case MOSAIC:
    FitsImage* ptr = fits;
    for (int i=1; i<which; i++)
      if (ptr)
	ptr = ptr->next();

    if (ptr)
      return ptr;
    else
      return currentFits;
  }
}

int Frame::findFits(FitsImage* p)
{
  switch (displayMode) {
  case SINGLE:
    return 1;
  case MOSAIC:
    FitsImage* ptr = fits;
    int r = 0;
    while (ptr) {
      r++;
      if (ptr == p)
	return r;
      ptr = ptr->next();
    }

    return r;
  }
}
