#ifndef __frame_h__
#define __frame_h__

#include <fstream.h>

#include <tk.h>

#include "widget.h"
#include "vector.h"
#include "colorscale.h"
#include "contourscale.h"
#include "marker.h"
#include "hist.h"
#include "list.h"
#include "coord.h"

#define SCALESIZE 16384
#define HISTEQUSIZE 16384

extern int DebugAST;
extern int DebugMosaic;
extern int DebugPerf;
extern int DebugWCS;
extern int DebugGZ;

class FitsImage;
class Contour;
class FVContour;
class Grid;
class Attribute;
class Marker;
class BaseBox;
class BaseCircle;
class BaseEllipse;
class BaseMarker;
class Circle;
class Annulus;
class Panda;
class Ellipse;
class EllipseAnnulus;
class Box;
class BoxAnnulus;
class Polygon;
class Point;
class Line;
class Text;
class Ruler;
class Projection;
class CirclePoint;
class BoxPoint;
class DiamondPoint;
class CrossPoint;
class ExPoint;
class ArrowPoint;
class BoxCirclePoint;

// Frame

class Frame : public Widget {
  friend FitsImage;
  friend Contour;
  friend FVContour;
  friend Grid;
  friend Attribute;
  friend Marker;
  friend BaseBox;
  friend BaseCircle;
  friend BaseEllipse;
  friend BaseMarker;
  friend Circle;
  friend Annulus;
  friend Panda;
  friend Ellipse;
  friend EllipseAnnulus;
  friend Box;
  friend BoxAnnulus;
  friend Polygon;
  friend Point;
  friend Line;
  friend Text;
  friend Ruler;
  friend Projection;
  friend CirclePoint;
  friend BoxPoint;
  friend DiamondPoint;
  friend CrossPoint;
  friend ExPoint;
  friend ArrowPoint;
  friend BoxCirclePoint;

public:
  enum Precision {DEFAULT, FIXED, SCIENTIFIC, INTEGER};
  enum ColorScaleType {LINEARSCALE, LOGSCALE, SQUAREDSCALE, 
		       SQRTSCALE, IISSCALE, HISTEQUSCALE};
  enum ClipScope {GLOBAL, LOCAL};
  enum ClipMode {MINMAX, ZSCALE, ZMAX, AUTOCUT, USERCLIP};
  enum MinMaxMode {SCAN, SAMPLE, DATAMIN, IRAFMIN};
  enum MagnifierMode {FAST, ACCURATE};
  enum MemType {ALLOC, ALLOCGZ, CHANNEL, MMAP, SHARE, SOCKET, SOCKETGZ, VAR};
  enum MarkerFormat {DS9, CIAO, SAOTNG, SAOIMAGE, PROS, RAWXY};
  enum FileNameType {BASIC, FULL};
  enum PSColorSpace {BW, GRAY, RGB, CMYK};
  enum ShmType {SHMID,KEY};
  enum ColormapMode {CENTER, CLICK};
  enum UndoMarkerType {NONE, MOVE, EDIT, DELETE};
  enum DisplayMode {SINGLE, MOSAIC};
  enum LoadMethod {LOADALL, INCR};
  enum ScanMode {NODATASEC, DATASEC, UNODATASEC, UDATASEC, RESET};
  enum MosaicType {NOMOSAIC, AUTO, IRAF, WCS, WFPC2};
  enum CompressType {NOCOMPRESS, GZ};

protected:
  FitsImage* fits;           // list of fits images
  int fitsCount;             // count of fits images
  int slice;                 // current slice
  ColorScale* colorScale;    // current color scale
  unsigned char* colorCells; // current color values
  unsigned short* indexCells; // current color indices

  XImage* baseXImage;        // ximage of image

  DisplayMode displayMode;       // display mode

  double rotation;           // current image rotation angle

  double rotateRotation;     // tmp cursor for interactive rotation
  XImage* rotateSrcXM;       // rotate src ximage
  XImage* rotateDestXM;      // rotate dest ximage
  Pixmap rotatePM;           // rotate pixmap
  GC rotateGCXOR;            // GC for interactive rotation

  int cmapID;                // current colormap id
  float bias;                // current colormap bias
  float contrast;            // current colormap contrast
  int invert;                // current colormap invert
  int cmapArea;              // current colormap update size
  ColormapMode cmapMode;     // current colormap update mode
  Vector cmapOffset;         // current colormap update offset

  int colorCount;            // number of dynamic colors
  ColorScaleType colorScaleType; // color scale type

  XImage* magnifierXImage;   // ximage for magnifier
  int magnifierWidth;        // magnifier width
  int magnifierHeight;       // magnifier height
  double magnifierZoom;      // magnifier zoom factor
  Vector magnifierCursor;     // we need to save the last cursor used

  int psInterpolate;         // postscript level 2 interpolation flag
  int psResolution;          // postscript resolution

  Vector iisLastCursor;      // iis cursor state info
  Vector iisSaveCursor;
  Window iisSaveRoot;
  int iisInteractive;
				  
  double* histogramX;      // scale histogram
  double* histogramY;      // scale histogram
  double* histequ;         // image histogram equalization xfer function

  Matrix canvasToRef;
  Matrix canvasToUser;
  Matrix canvasToWidget;
  Matrix canvasToWindow;
  Matrix refToCanvas;
  Matrix refToPanner;
  Matrix refToUser;
  Matrix refToWidget;
  Matrix refToWindow;
  Matrix pannerToRef;
  Matrix pannerToUser;
  Matrix userToCanvas;
  Matrix userToRef;
  Matrix userToMagnifier;
  Matrix userToMagnifierCursor;
  Matrix userToPanner;
  Matrix userToWidget;
  Matrix widgetToCanvas;
  Matrix widgetToRef;
  Matrix widgetToUser;
  Matrix widgetToWindow;
  Matrix windowToCanvas;
  Matrix windowToRef;
  Matrix windowToWidget;
  Matrix ximageToMagnifier;

protected:
  enum UpdateType {MATRIX, BASE, PIXMAP, NOUPDATE};

  char* name;                // object name
  FitsImage* currentFits;    // current fits image (for mosaic)
  Pixmap basePixmap;         // pixmap of image
				  
  UpdateType needsUpdate;           // updateFlags

  Vector cursor;             // current cursor position in widget coords
  double zoom;               // current image zoom value
  Orientation orientation;   // current image orientation
  Matrix orientationMatrix;  // current image orientation matrix

  int wcsAlignment;          // flag to indicate wcs alignment
  double wcsZoom;            // wcs image zoom factor
  double wcsRotation;        // wcs image rotation angle
  Orientation wcsOrientation; // wcs image orientation
  Matrix wcsOrientationMatrix; // wcs image orientation matrix

  Pixmap panPM;              // tmp pixmap for pan motion
  Vector panCursor;          // tmp cursor for interactive panning
  GC panGCXOR;               // GC for interactive panning

  ClipScope clipScope;       // color scale clip scope
  ClipMode clipMode;         // color scale clip mode
  MinMaxMode mmMode;         // method to use for determining minmax

  double low;                // low cut level for all data
  double high;               // high cut level for all data

  double min;                // min for all data
  double max;                // max for all data

  double uLow;               // low cut via user for all data
  double uHigh;              // high cut via user for all data

  float zContrast;           // zscale slope transfer function
  int zSampleSize;           // zscale optimal sample size
  int zSampleLine;           // zscale number of lines to sample
			     // for median scaling

  int mmIncr;                // minmax sampling incr

  float autoCutPer;          // autoCut percentage

  ScanMode scanMode;       // use keyword DATASEC			  

  Vector crosshair;          // crosshair location
  int useCrosshair;          // flag to use crosshair
  GC crosshairGC;            // crosshair gc
  GC crosshairGCXOR;         // crosshair gcxor

  GC highliteGC;             // gc for frame highlite

  int mosaicFast;            // flag for fast mosaic render

  List<Marker> markers;				  
  GC markerGC;               // marker gc
  GC selectGCXOR;            // select gc
  XRectangle rectWidget[1];  // clip rect
  XRectangle rectWindow[1];  // clip rect

  CoordSystem markerDefaultSystem; // default marker sys (RAWXY)
  Vector markerBegin;        // marker move begin in image coords
  Marker* editMarker;        // pointer to marker currently being edited
  Marker* rotateMarker;      // pointer to marker currently being rotated
  List<Marker> undoMarkers;  // remember marker to undo
  UndoMarkerType undoMarkerType; // type
  List<Marker> pasteMarkers;  // cut/copy/paste
  Vector regionBegin;        // select region begin in canvas coords
  Vector regionEnd;          // select region end in canvas coords

  Pixmap pannerPixmap;       // pixmap for panner
  XImage* pannerXImage;      // ximage for panner
  int pannerWidth;           // panner width
  int pannerHeight;          // panner height
  char pannerName[32];       // panner widget name
  int usePanner;             // panner status

  Pixmap magnifierPixmap;    // pixmap for magnifier
  char magnifierName[32];    // magnifer widget name
  MagnifierMode magnifierMode; // magnifier algorithm
  int useMagnifier;          // magnifer status
  int useMagnifierGraphics;    // magnifer marker status
  int useMagnifierCursor;    // magnifier cursor status
  GC magnifierGC;            // GC for magnifier
				  
  int useHighlite;           // flag to draw highlite box

  Contour* contour;          // main contour
  List<Contour> auxcontours; // aux contours
  GC contourGC;

  Grid* grid;
  GC gridGC;

  int psLevel;               // postscript level
  PSColorSpace psColorSpace; // postscript color space

  FitsHist::Function binFunction; // bining operation
  double binFactor;        // bining factor
  int binBufferSize;       // size of bin buffer

protected:
  void update(UpdateType);         // schedule update

  virtual void updateColorScale() =0;
  virtual void fillXImage(XImage*, int, int, int, int,
			  double* (FitsImage::*)()) =0;
  virtual void fillXImageMosaic(XImage*, int, int, int, int,
				double* (FitsImage::*)()) =0;
  virtual void fillXImageMosaicFast(XImage*, int, int, int, int,
				    double* (FitsImage::*)()) =0;
  virtual void updateMagnifierReplicate(const Vector&) =0;
  virtual void rotateMotion() =0;

  void updateColorCells(unsigned short*, unsigned char*, int);

  void psLevel1(PSColorSpace, int, int, float);
  virtual void psLevel2(PSColorSpace, int, int, float) =0;
  Matrix psMatrix(float, int, int);
  void dvipsFix(char*, int);

  int isIIS();

  void updateHistEqu();
  void initHistogram();

  ContourScale* contourScale(int, float, float, ColorScaleType);

  void saveFitsResample(OutFitsStream&);
  void saveFitsResampleSingle(OutFitsStream&);
  void saveFitsResampleMosaic(OutFitsStream&);
  void saveFitsResampleKeyword(OutFitsStream&, FitsHead&);

private:
  void load(MemType, const char*, FitsImage*);
  void loadMosaic(FitsImage*);
  void loadMosaicImage(MemType, const char*, FitsImage*, MosaicType);
  void loadWFPC2(MemType, const char*, FitsImage*);

  int postscriptProc(int);   // generate postscript

  void update(UpdateType,BBox);    // schedule update
  void updateNow(UpdateType);      // schedule update
  void updateNow(UpdateType,BBox); // schedule update

  void invalidPixmap();      // update all pixmaps
  int updatePixmap(const BBox&); // renders image/graphics into pixmap
  void updateMatrices();     // recalculate matrices
  void updateBase(const BBox&);    // generate base pixmap
  void updatePM(const BBox&);      // update part of pixmap
  void updateGCs();          // update gcs

  void ximageToPixmap(Pixmap, XImage*, int, int, int, int, 
		      double* (FitsImage::*)());

  void updatePanner();       // render panner
  double calcPannerZoom();  // calculate the zoom factor for the panner
  void pannerPush();         // send panner and update message
  void updateMagnifier(); // render magnifier
  void updateMagnifier(const Vector&); // render magnifier
  void updateMagnifierMatrix();
  void updateMagnifierZoom(const Vector&); // render magnifier
  void renderMagnifierCursor(const Vector&); // render magnifier cursor

  void renderContours(Pixmap, const Matrix&, const BBox&);
  void psContours(PSColorSpace);
  void updateContourCoords(const Matrix&);

  void getInfo(const Vector&, CoordSystem,
	       Matrix& (FitsImage::*)(), Matrix& (FitsImage::*)(),
	       CoordSystem, CoordFormat, char*);
  void getInfoSingle(const Vector&, CoordSystem,
		     Matrix& (FitsImage::*)(), Matrix& (FitsImage::*)(),
		     CoordSystem, CoordFormat, char*);
  void getInfoMosaic(const Vector&, CoordSystem,
		     Matrix& (FitsImage::*)(), Matrix& (FitsImage::*)(),
		     CoordSystem, CoordFormat, char*);

  void printIntegerTcl(int);
  void printDoubleTcl(double, Precision);
  void printVectorTcl(const Vector&, Precision);
  void printCoordTcl(FitsImage*, const Vector&, CoordSystem, CoordSystem, 
		     CoordFormat, Precision);
  void coordToTclArray(FitsImage*, const Vector&, CoordSystem, CoordSystem, 
		       Precision, const char*, const char*);
  void doubleToTclArray(double, Precision, const char*, const char*, 
			const char*);

  void markerPrintCoordTcl(const Vector&, const Vector&,
			   CoordSystem, CoordFormat);
  void markerPrintVectorTcl(const Vector&, const Vector&,
			    CoordSystem, CoordFormat);
  void markerPrintDoubleTcl(const Vector&, double, CoordSystem, CoordFormat);
  void markerPrintDoubleTcl(const Vector&, const Vector&, const Vector&, 
			    CoordSystem, CoordFormat);

  void centerImage();
  void reset();              // resets current position,zoom,rotation,flip
  void wcsAlign();           // sets wcs values

  void renderHighLite();     // render a hightlite bbox

  void renderCrosshair();    // render the crosshair
  void psCrosshair(PSColorSpace);
  void drawCrosshair();
  void eraseCrosshair();

  void bltProjection(char*, char*, const Vector&, const Vector&, double, int);
  void bltProjSingle(double*, double*, int, 
		     const Vector&, const Vector&, double, int);
  void bltProjMosaic(double*, double*, int, 
		     const Vector&, const Vector&, double, int);
  void bltHist(char*, char*, int);
  void bltCut(char*, char*, Orientation, const Vector&);
  void bltCutSingle(double*, double*, int, Vector&, Orientation);
  void bltCutMosaic(double*, double*, int, Vector&, Orientation);

  void unloadFits();         // unload all fits images
  double calcZoom(int,int,Vector);  // calculate the zoom factor
  void updateClip();         // set clip values
  Vector getClip(ClipMode,float);    // get clip values
  void updateLowHigh(double*, double*);  // calculate global low high

  void deleteMarkers();             // delete all markers
  void psMarkers(int);              // generate ps for all markers
  void renderMarkers(const BBox&);  // render the markers into pixmap
  void renderMagnifierMarkers(const Matrix&, const BBox& bb);
  void unselectMarkers();           // unselect all markers
  void updateMarkers();             // update marker bbox
  void updateMarkerCoords(const Matrix&); // update marker coords
  void markerUndo(Marker*, UndoMarkerType);

  void markerListHeader(ostream&, CoordSystem, CoordFormat, char);
  void markerListCiaoHeader(ostream&, CoordSystem, CoordFormat, char);
  void markerListSAOtngHeader(ostream&, CoordSystem, CoordFormat, char);
  void markerListSAOimageHeader(ostream&, CoordSystem, CoordFormat, char) {}
  void markerListProsHeader(ostream&, CoordSystem, CoordFormat, char) {}
  void markerListXYHeader(ostream&, CoordSystem, CoordFormat, char) {}

  void createMarker(Marker*);       // process new marker
  void parseMarker(MarkerFormat,istream&);

  void psLevel1Head(PSColorSpace, int, int);
  void psLevel1Single(PSColorSpace, int, int, float);
  void psLevel1Mosaic(PSColorSpace, int, int, float);
  void psLevel1MosaicFast(PSColorSpace, int, int, float);
  Vector psOrigin();

  FitsHead* parseWCS(istream&);

  void updateBin(const Matrix&);

  void setScanModeIncr(LoadMethod);
  ScanMode getScanMode() {return scanMode;}

public:
  Frame(Tcl_Interp*, Tk_Canvas, Tk_Item*);
  virtual ~Frame();

  int parse(istrstream&);

  DisplayMode getDisplayMode() {return displayMode;}
  double getZoom() {return zoom*wcsZoom;}
  double getMagnifierZoom() {return zoom*wcsZoom*magnifierZoom;}
  double getRotation() {return rotation+wcsRotation;}
  Vector getCenter();
  double getWCSRotation();
  Orientation getWCSOrientation();

  int hasLTMV();
  int hasATMV();
  int hasDTMV();

  int hasWCS();
  int hasWCSEqu();
  int hasWCSLinear();

  Matrix& getCanvasToRef() {return canvasToRef;}
  Matrix& getCanvasToWidget() {return canvasToWidget;}
  Matrix& getRefToCanvas() {return refToCanvas;}
  Matrix& getRefToWidget() {return refToWidget;}
  Matrix& getRefToWindow() {return refToWindow;}

  CoordSystem getMarkerDefaultSystem() {return markerDefaultSystem;}

  double mapAngleToRef(double, CoordSystem);
  double mapAngleFromRef(double, CoordSystem);

  FitsImage* isInFits(const Vector&, Matrix& (FitsImage::*)(), Vector*);
  FitsImage* findFits(const Vector&);
  FitsImage* findFits(int);
  int findFits(FitsImage*);

  // Subcommand Functions

  void clearCmd();
  void colorScaleCmd(ColorScaleType);
  void getWidthCmd();
  void getHeightCmd();
  void getBitpixCmd();
  void getColormapCmd();
  void getColorScaleCmd();
  void getInfoCmd(char*);
  void getInfoCmd(const Vector&, CoordSystem, CoordSystem, CoordFormat, char*);
  void getNameCmd();
  void getValueCmd(const Vector&, CoordSystem);
  void highliteCmd(int);
  virtual void colormapCmd(int, float, float, int, unsigned short*, 
			   unsigned char*, int) =0;
  void colormapAreaCmd(int s, ColormapMode m) {cmapArea = s; cmapMode = m;}
  virtual void colormapBeginCmd() =0;
  virtual void colormapBeginCmd(const Vector&) =0;
  virtual void colormapMotionCmd(int, float, float, int, unsigned short*,
				 unsigned char*, int) =0;
  virtual void colormapEndCmd(int, float, float, int, unsigned short*,
			      unsigned char*, int) =0;
  void mosaicFastCmd(int);
  void setNameCmd(char*);
  void warpCmd(const Vector&);
  void warpToCmd(const Vector&);

  // Clip Commands

  void clipAutoCutCmd(float);
  void clipMinMaxModeCmd(Frame::MinMaxMode);
  void clipMinMaxParamCmd(int);
  void clipModeCmd(float);
  void clipModeCmd(ClipMode);
  void clipScopeCmd(ClipScope);
  void clipUserCmd(double, double);
  void clipUserLowCmd(double);
  void clipUserHighCmd(double);
  void clipZScaleParamCmd(float, int, int);
  void getClipCmd();
  void getClipCmd(float);
  void getClipCmd(ClipMode);
  void getClipMinMaxModeCmd();
  void getClipMinMaxSampleCmd() {printIntegerTcl(mmIncr);}
  void getClipModeCmd();
  void getClipScopeCmd();
  void getClipUserCmd();
  void getClipZScaleParamCmd();
  void getHistogramCmd(char*, char*, int);
  void getMinMaxCmd();
  void hasDATAMINCmd();
  void hasIRAFMINCmd();

  // CrossHair Commands

  void crosshairCmd(int);
  void crosshairCmd(const Vector&, CoordSystem);
  void crosshairBeginCmd(const Vector&, CoordSystem);
  void crosshairMotionCmd(const Vector&, CoordSystem);
  void crosshairWarpCmd(const Vector&);
  void getCrosshairCmd(CoordSystem, CoordFormat, Precision);
  void getCrosshairStatusCmd();

  void getHorzCutCmd(char*, char*);
  void getHorzCutCmd(char*, char*, const Vector&);
  void getVertCutCmd(char*, char*);
  void getVertCutCmd(char*, char*, const Vector&);

  // DATASEC Commands

  void DATASECCmd(int);
  void getDATASECCmd();
  void hasDATASECCmd();

  // Fits Commands

  void getFitsCountCmd() {printIntegerTcl(fitsCount);}
  void getFitsHeaderCmd(int);
  void getFitsHeaderKeywordCmd(int,const char*);
  void getFitsFileNameCmd(int,FileNameType);
  void getFitsFileNameCmd(const Vector&, CoordSystem,FileNameType);
  void getFitsFileNamesCmd(FileNameType);
  void getFitsObjectNameCmd(const Vector&, CoordSystem);
  void getFitsObjectNameCmd(int);
  void getFitsMasterCmd();
  void getFitsSizeCmd(CoordSystem, CoordFormat);
  void getFitsDepthCmd();
  void getFitsSliceCmd();

  void hasFitsCmd();
  void hasFitsCubeCmd();
  void hasFitsExtCmd(int);
  void hasFitsImageCmd();
  void hasFitsMosaicCmd();
  void hasFitsTableCmd();

  void loadFitsAllocCmd(const char*);
  void loadFitsAllocGZCmd(const char*);
  void loadFitsChannelCmd(const char*, const char*);
  void loadFitsMMapCmd(const char* fn, LoadMethod);
  void loadFitsShareCmd(ShmType, int, int, const char*, LoadMethod);
  void loadFitsSocketCmd(int, const char* fn);
  void loadFitsSocketGZCmd(int, const char* fn);
  void loadFitsVarCmd(const char*, const char*, LoadMethod);

  void loadArrAllocCmd(const char*);
  void loadArrAllocGZCmd(const char*);
  void loadArrChannelCmd(const char*, const char*);
  void loadArrMMapCmd(const char* fn);
  void loadArrShareCmd(ShmType, int, int, const char*);
  void loadArrSocketCmd(int, const char* fn);
  void loadArrSocketGZCmd(int, const char* fn);
  void loadArrVarCmd(const char*, const char*);

  void loadMosaicAllocCmd(const char*, MosaicType);
  void loadMosaicAllocGZCmd(const char*, MosaicType);
  void loadMosaicChannelCmd(const char*, const char*, MosaicType);
  void loadMosaicMMapCmd(const char*, MosaicType, LoadMethod);
  void loadMosaicShareCmd(ShmType, int, int, const char*, MosaicType,
			  LoadMethod);
  void loadMosaicSocketCmd(int, const char*, MosaicType);
  void loadMosaicSocketGZCmd(int, const char*, MosaicType);
  void loadMosaicVarCmd(const char*, const char*, MosaicType, LoadMethod);

  void loadMosaicImageAllocCmd(const char*, MosaicType);
  void loadMosaicImageAllocGZCmd(const char*, MosaicType);
  void loadMosaicImageChannelCmd(const char*, const char*, MosaicType);
  void loadMosaicImageMMapCmd(const char*, MosaicType, LoadMethod);
  void loadMosaicImageShareCmd(ShmType, int, int, const char*,
			     MosaicType, LoadMethod);
  void loadMosaicImageSocketCmd(int, const char*, MosaicType);
  void loadMosaicImageSocketGZCmd(int, const char*, MosaicType);
  void loadMosaicImageVarCmd(const char*, const char*, MosaicType, LoadMethod);

  void loadMosaicWFPC2AllocCmd(const char*);
  void loadMosaicWFPC2AllocGZCmd(const char*);
  void loadMosaicWFPC2ChannelCmd(const char*, const char*);
  void loadMosaicWFPC2MMapCmd(const char* fn, LoadMethod);
  void loadMosaicWFPC2ShareCmd(ShmType, int, int, const char*, LoadMethod);
  void loadMosaicWFPC2SocketCmd(int, const char* fn);
  void loadMosaicWFPC2SocketGZCmd(int, const char* fn);
  void loadMosaicWFPC2VarCmd(const char*, const char*, LoadMethod);

  void loadIncrDataCmd(int, int, int, int, int);
  void loadIncrMinMaxCmd(int, int, int, int, int);
  void loadIncrEndCmd();

  void loadFitsSlaveCmd(const char*, FitsFile*);

  void saveFitsImageFileCmd(const char*, int);
  void saveFitsTableFileCmd(const char*, int);
  void saveFitsResampleFileCmd(const char*, int);

  void saveFitsImageChannelCmd(const char*, int);
  void saveFitsTableChannelCmd(const char*, int);
  void saveFitsResampleChannelCmd(const char*, int);

  void saveFitsImageSocketCmd(int, int);
  void saveFitsTableSocketCmd(int, int);
  void saveFitsResampleSocketCmd(int, int);

  void sliceCmd(int);

  void updateFitsCmd(int);
  void updateFitsCmd(int, BBox, int);
  void unloadFitsCmd();

  // Fitsy Commands

  void fitsyHasExtCmd(const char*);

  // Coordinate Commands

  void getCoordCmd(const Vector& v, CoordSystem in, CoordSystem out, 
		   CoordFormat format, Precision p);
  void getCursorCmd(CoordSystem out, CoordFormat format, Precision p);

  void getWCSFlipCmd();
  void getWCSSystemCmd();
  void hasAmplifierCmd();
  void hasDetectorCmd();
  void hasPhysicalCmd();
  void hasWCSCmd();
  void hasWCSEquCmd();
  void hasWCSLinearCmd();
  void mapCmd(const Vector&, CoordSystem, CoordSystem);

  // Marker Commands

  void createCircleCmd(const Vector&, double,
		       const char*, int, const char*, 
		       const char*, unsigned short, const char*);
  void createAnnulusCmd(const Vector&, double, double, int,
			const char*, int, const char*, 
			const char*, unsigned short, const char*);
  void createAnnulusCmd(const Vector&, double*, int,
			const char*, int, const char*, 
			const char*, unsigned short, const char*);

  void createEllipseCmd(const Vector&, const Vector&, 
			double, const char*, int, const char*, 
			const char*, unsigned short, const char*);
  void createEllipseAnnulusCmd(const Vector&, const Vector&, 
			       const Vector&, int, double,
			       const char*, int, const char*, 
			       const char*, unsigned short, const char*);
  void createEllipseAnnulusCmd(const Vector&, double, int, Vector*,
			       const char*, int,  const char*, 
			       const char*, unsigned short, const char*);

  void createBoxCmd(const Vector&, const Vector&, 
		    double, const char*, int, const char*, 
		    const char*, unsigned short, const char*);
  void createBoxAnnulusCmd(const Vector&, const Vector&, 
			   const Vector&, int, double,
			   const char*, int, const char*, 
			   const char*, unsigned short, const char*);
  void createBoxAnnulusCmd(const Vector&, double, int, Vector*,
			   const char*, int, const char*, 
			   const char*, unsigned short, const char*);

  void createPandaCmd(const Vector&, double, double, int, double, double, int,
			const char*, int, const char*, 
			const char*, unsigned short, const char*);
  void createPandaCmd(const Vector&, double*, int, double*, int,
			const char*, int, const char*, 
			const char*, unsigned short, const char*);

  void createPolygonCmd(const Vector&, const Vector&,
			const char*, int, const char*, 
			const char*, unsigned short, const char*);
  void createPolygonCmd(const List<Vertex>&,
			const char*, int, const char*, 
			const char*, unsigned short, const char*);
  void createContourPolygonCmd(const char*, int, const char*, 
			       const char*, unsigned short, const char*);
  void createLineCmd(const Vector&, const Vector&, int, int,
		     const char*, int, const char*,
		     const char*, unsigned short, const char*);
  void createTextCmd(const Vector&, double,
		     const char*, int, const char*, 
		     const char*, unsigned short, const char*);
  void createRulerCmd(const Vector&, const Vector&, 
		      CoordSystem, CoordSystem, CoordFormat,
		      const char*, int, const char*,
		      const char*, unsigned short, const char*);
  void createProjectionCmd(const Vector&, const Vector&,
			   double, const char*, const char*,
			   const char*, int, const char*,
			   const char*, unsigned short, const char*);
  void createCirclePointCmd(const Vector&,
			    const char*, int, const char*,
			    const char*, unsigned short, const char*);
  void createBoxPointCmd(const Vector&,
			 const char*, int, const char*,
			 const char*, unsigned short, const char*);
  void createDiamondPointCmd(const Vector&,
			     const char*, int, const char*, 
			     const char*, unsigned short, const char*);
  void createCrossPointCmd(const Vector&, 
			   const char*, int, const char*,
			   const char*, unsigned short, const char*);
  void createExPointCmd(const Vector&,
			const char*, int, const char*,
			const char*, unsigned short, const char*);
  void createArrowPointCmd(const Vector&, 
			   const char*, int, const char*,
			   const char*, unsigned short, const char*);
  void createBoxCirclePointCmd(const Vector&,
			       const char*, int, const char*,
			       const char*, unsigned short, const char*);

  void getMarkerAngleCmd(int, CoordSystem);
  void getMarkerAnnulusRadiusCmd(int, CoordSystem, CoordFormat);
  void getMarkerBoxAnnulusSizeCmd(int, CoordSystem, CoordFormat);
  void getMarkerBoxSizeCmd(int, CoordSystem, CoordFormat);
  void getMarkerCenterCmd(int, CoordSystem, CoordFormat);
  void getMarkerCircleRadiusCmd(int, CoordSystem, CoordFormat);
  void getMarkerColorCmd();
  void getMarkerColorCmd(int);
  void getMarkerEllipseRadiusCmd(int, CoordSystem, CoordFormat);
  void getMarkerEllipseAnnulusRadiusCmd(int, CoordSystem, CoordFormat);
  void getMarkerFontCmd();
  void getMarkerFontCmd(int);
  void getMarkerHandleCmd(const Vector&);
  void getMarkerIdCmd(const Vector&);
  void getMarkerIdAllCmd();
  void getMarkerLineCmd(int, CoordSystem, CoordFormat);
  void getMarkerLineArrowCmd(int);
  void getMarkerLineDistanceCmd(int, CoordSystem, CoordFormat);
  void getMarkerLineWidthCmd();
  void getMarkerLineWidthCmd(int);
  void getMarkerPandaAnglesCmd(int, CoordSystem);
  void getMarkerPandaRadiusCmd(int, CoordSystem, CoordFormat);
  void getMarkerPolygonSegmentCmd(const Vector&);
  void getMarkerProjectionCmd(int, char*, char*);
  void getMarkerProjectionPointsCmd(int, CoordSystem, CoordFormat);
  void getMarkerProjectionDistanceCmd(int, CoordSystem, CoordFormat);
  void getMarkerProjectionMethodCmd(int);
  void getMarkerProjectionWidthCmd(int, CoordSystem, CoordFormat);
  void getMarkerPropertyCmd(unsigned short);
  void getMarkerPropertyCmd(int, unsigned short);
  void getMarkerRulerDistanceCmd(int, CoordSystem, CoordFormat);
  void getMarkerRulerPointCmd(int, CoordSystem, CoordFormat);
  void getMarkerSelectedCmd();
  void getMarkerSelectedCmd(const Vector&);
  void getMarkerSelectedNumberCmd();
  void getMarkerTextCmd(int);
  void getMarkerTypeCmd(int);

  void hasMarkerSelectedCmd();
  void hasMarkerPasteCmd();
  void hasMarkerUndoCmd();

  void markerAngleCmd(int, double, CoordSystem);
  void markerAnnulusCreateRadiusCmd(int, const Vector&);
  void markerAnnulusDeleteRadiusCmd(int, int);
  void markerAnnulusRadiusCmd(int, double, double, int, 
			      CoordSystem, CoordFormat);
  void markerAnnulusRadiusCmd(int, const char*, CoordSystem, CoordFormat);
  void markerBackCmd();
  void markerBackCmd(int);
  void markerBoxAnnulusSizeCmd(int, const Vector&, const Vector&, int, 
			       CoordSystem, CoordFormat);
  void markerBoxAnnulusSizeCmd(int, const char*, CoordSystem, CoordFormat);
  void markerBoxAnnulusCreateSizeCmd(int, const Vector&);
  void markerBoxAnnulusDeleteSizeCmd(int, int);
  void markerBoxSizeCmd(int, const Vector&, CoordSystem, CoordFormat);
  void markerCallBackCmd(int, Marker::Callback, const char*, const char*);
  void markerCircleRadiusCmd(int, double, CoordSystem, CoordFormat);
  void markerColorCmd(const char*);
  void markerColorCmd(int, const char*);
  void markerCommandCmd(MarkerFormat,const char*);
  void markerDeleteCmd();
  void markerDeleteCmd(int);
  void markerDeleteAllCmd();
  void markerDeleteLastCmd();
  void markerDeleteCallBackCmd(int, Marker::Callback, const char*);
  void markerEditBeginCmd(int, int);
  void markerEditBeginCmd(const Vector&, int);
  void markerEditMotionCmd(const Vector&, int h);
  void markerEditEndCmd();
  void markerEllipseRadiusCmd(int, const Vector&, CoordSystem, CoordFormat);
  void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     CoordSystem, CoordFormat);
  void markerEllipseAnnulusRadiusCmd(int, const char*, 
				     CoordSystem, CoordFormat);
  void markerEllipseAnnulusCreateRadiusCmd(int, const Vector&);
  void markerEllipseAnnulusDeleteRadiusCmd(int, int);
  void markerFontCmd(const char*);
  void markerFontCmd(int, const char*);
  void markerFrontCmd();
  void markerFrontCmd(int);
  void markerLineCmd(int, const Vector&, const Vector&, CoordSystem);
  void markerLineArrowCmd(int, int, int);
  void markerLineWidthCmd(int);
  void markerLineWidthCmd(int, int);
  void markerListCmd(MarkerFormat, CoordSystem, CoordFormat, int,
		     unsigned short, unsigned short);
  void markerLoadCmd(MarkerFormat,const char*);
  void markerLoadCmd(MarkerFormat,const char*,CoordSystem);
  void markerLoadCmd(MarkerFormat,int);
  void markerLoadCmd(MarkerFormat,int,CoordSystem);
  void markerLoadFitsCmd(const char*);
  void markerMoveCmd(const Vector&);
  void markerMoveCmd(int id, const Vector&);
  void markerMoveBeginCmd(const Vector&);
  void markerMoveBeginCmd(int, const Vector&);
  void markerMoveMotionCmd(const Vector&);
  void markerMoveMotionCmd(int, const Vector&);
  void markerMoveEndCmd();
  void markerMoveEndCmd(int);
  void markerMoveToCmd(const Vector&, CoordSystem);
  void markerMoveToCmd(int, const Vector&, CoordSystem);
  void markerPandaCreateAnglesCmd(int, const Vector&);
  void markerPandaCreateRadiusCmd(int, const Vector&);
  void markerPandaDeleteCmd(int, int);
  void markerPandaEditCmd(int, double, double, int, double, double, int,
			  CoordSystem, CoordFormat);
  void markerPandaEditCmd(int, const char*, const char*, CoordSystem,
			  CoordSystem, CoordFormat);
  void markerPolygonCreateVertexCmd(int, int, const Vector&);
  void markerPolygonDeleteVertexCmd(int, int);
  void markerPropertyCmd(unsigned short, int);
  void markerPropertyCmd(unsigned short, int, const Vector&);
  void markerPropertyCmd(int, unsigned short, int);
  void markerProjectionCmd(int, const Vector&, const Vector&, CoordSystem,
			   double, int, CoordSystem, CoordFormat);
  void markerRotateBeginCmd(int);
  void markerRotateBeginCmd(const Vector&);
  void markerRotateMotionCmd(const Vector&, int h);
  void markerRotateEndCmd();
  void markerRulerPointCmd(int, const Vector&, const Vector&, CoordSystem);
  void markerRulerSystemCmd(int, CoordSystem, CoordSystem, CoordFormat);
  void markerSaveCmd(const char*, MarkerFormat, CoordSystem, CoordFormat);
  void markerSelectCmd(int);
  void markerSelectCmd(const Vector&);
  void markerSelectAllCmd();
  void markerSelectToggleCmd(const Vector&);
  void markerTextCmd(int, const char*);
  void markerUnselectCmd(int);
  void markerUnselectAllCmd();

  void markerCopyCmd();
  void markerCutCmd();
  void markerPasteCmd();
  void markerUndoCmd();

  void regionSelectBeginCmd(const Vector&);
  void regionSelectMotionCmd(const Vector&);
  void regionSelectEndCmd();
  void regionSelectShiftEndCmd();

  // Pan Zoom Rotate Orient Commands

  void centerCmd();
  void getOrientCmd();
  void getRotateCmd(Precision);
  void getWCSAlignCmd();
  void getWCSZoomCmd(Precision);
  void getZoomCmd(Precision);
  void orientCmd(Orientation);
  void panCmd(const Vector&, const Vector&);
  void panCmd(const Vector&, CoordSystem);
  void panBBoxCmd(BBox);
  void panMotionBeginCmd(const Vector&);
  void panMotionCmd(const Vector&);
  void panMotionEndCmd(const Vector&);
  void panToCmd(const Vector&, CoordSystem);
  void rotateCmd(double);
  void rotateToCmd(double);
  void rotateMotionBeginCmd();
  void rotateMotionCmd(double);
  void rotateMotionEndCmd();
  void wcsAlignCmd(int);
  void wcsZoomCmd(double);
  void zoomCmd(double);
  void zoomAboutCmd(double, const Vector&, CoordSystem);
  void zoomToCmd(double);
  void zoomToAboutCmd(double, const Vector&, CoordSystem);
  void zoomToFitCmd();

  // Panner/Magnifier Commands

  void magnifierCmd(char*, int, int);
  void magnifierCmd(int);
  void magnifierGraphicsCmd(int);
  void magnifierCursorCmd(int);
  void magnifierModeCmd(MagnifierMode);
  void magnifierZoomCmd(double);

  void pannerCmd(char*, int, int);
  void pannerCmd(int);
  void updatePannerCmd();
  void updateMagnifierCmd();
  void updateMagnifierCmd(const Vector& v);

  // Postscript Commands
       
  void psInterpolateCmd(int w) {psInterpolate = w;}
  void psLevelCmd(int l) {psLevel = (l == 2) ? 2 : 1;}
  void psResolutionCmd(int r) {psResolution = r;}
  void psColorSpaceCmd(PSColorSpace c) {psColorSpace = c;}

  // Bin Commands

  void binCmd(double, const Vector&, CoordSystem, 
	      const char*, const char*, const char*);
  void binCmd(double, const char*, const char*, const char*);
  void binAboutCmd(const Vector&, CoordSystem);
  void binColsCmd(const char*, const char*);
  void binFactorCmd(double);
  void binFactorAboutCmd(double, const Vector&, CoordSystem);
  void binFactorToCmd(double);
  void binFactorToAboutCmd(double, const Vector&, CoordSystem);
  void binToFitCmd();
  void binFunctionCmd(FitsHist::Function);
  void binBufferSizeCmd(int);
  void binFilterCmd(const char*);
  void getBinFactorCmd();
  void getBinFunctionCmd();
  void getBinBufferSizeCmd() {printIntegerTcl(binBufferSize);}
  void getBinCursorCmd();
  void getBinFilterCmd();
  void getBinColsCmd();
  void getBinListCmd();
  void hasBinCmd();

  // Contour Commands

  void contourCreateCmd(const char*, int, int, float, int);
  void contourCreateCmd(const char*, int, int, float, int, ColorScaleType);
  void contourCreateCmd(const char*, int, int, float, int,
			float, float, ColorScaleType);
  void contourCreateCmd(const char*, int, float, int, const char*);
  void contourLoadCmd(const char*, int, const char*, CoordSystem);
  void contourCopyCmd(CoordSystem);
  void contourDeleteCmd();
  void contourPasteCmd(const char*, int, void*, CoordSystem);
  void contourSaveCmd(const char*, CoordSystem);
  void contourSetColorCmd(const char*);
  void contourSetLineWidthCmd(int);
  void getContourLevelCmd();
  void getContourLevelCmd(int, float, float, ColorScaleType);
  void getContourMethodCmd();
  void getContourColorNameCmd();
  void getContourLineWidthCmd();
  void hasContourCmd();

  // Grid Commands

  void gridCmd(CoordSystem, int, const char*);
  void gridDeleteCmd();
  void hasGridCmd();
  void getGridSystemCmd();
  void getGridTypeCmd();
  void getGridOptionCmd();

  // IIS Commands

  void iisCmd(int, int);
  void iisEraseCmd();
  void iisMessageCmd(const char*);
  void iisSetCmd(const char*, int, int, int, int);
  void iisWCSCmd(const Matrix&, const Vector&, int);
  void iisGetCmd(char*, int, int, int, int);
  void iisUpdateCmd() {updateNow(MATRIX);}
  void iisSetCursorCmd(const Vector&, CoordSystem);
  void iisGetCursorCmd();
  void iisCursorModeCmd(int);
  void hasIISCmd();

  // Pixel Table

  void getPixelTableCmd(const Vector&, CoordSystem, int, char*);
  void getPixelTable(const Vector&, Matrix& (FitsImage::*matrixToData)(),
		     int, char*);
  void getPixelTable(FitsImage*, const Vector&, int, char*);

  // Data Values

  void getDataValuesCmd(int, const Vector&, CoordSystem, const Vector&, char*);

  // WCS

  void wcsResetCmd();
  void wcsReplaceCmd(int);
  void wcsReplaceCmd(const char*);
  void wcsAppendCmd(int);
  void wcsAppendCmd(const char*);
};

#endif
