#include <Xlib.h>
#include <Xutil.h>

#include "frametrue.h"
#include "fitsimage.h"
#include "nan.h"
#include "ps.h"
#include "util.h"

// FrameTrueColor Member Functions

FrameTrueColor::FrameTrueColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : Frame(i, c, item)
{
  skipUpdate = 0;

  colormapData = NULL;
  colormapPM = 0;
  colormapXM = NULL;
  colormapGCXOR = 0;
}

void FrameTrueColor::fillXImage(XImage* xmap,
				int x0, int y0, int x1, int y1,
				double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render

  if (!colorScale)
    return;

  if (!currentFits)
    return;

  int& width = xmap->width;
  int& height = xmap->height;
  int bytesPerPixel = xmap->bits_per_pixel/8;

  double* m = (currentFits->*getMatrix)();
  int* params = currentFits->getDataParams(scanMode);
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;
  const unsigned char* table = colorScale->colors();

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  // clear ximage

  memset(xmap->data, 255, xmap->bytes_per_line * height);

  for (int j=y0; j<y1; j++) {
    // the line may be padded at the end
    char* dest = xmap->data + j*xmap->bytes_per_line + x0*bytesPerPixel;

    for (int i=x0; i<x1; i++, dest+=bytesPerPixel) {

      double x = i*m[0] + j*m[3] + m[6];
      double y = i*m[1] + j*m[4] + m[7];

      if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	double value = currentFits->getValueDouble(((int)y)*srcw + (int)x);
	
	if (!isNaNd(value)) {
	  if (value <= ll)
	    memcpy(dest, table, bytesPerPixel);
	  else if (value >= hh)
	    memcpy(dest, table+last, bytesPerPixel);
	  else
	    memcpy(dest, table+((int)(((value - ll)/diff * length) + .5))*
		   bytesPerPixel,bytesPerPixel);
	}
      }
    }
  }
}

void FrameTrueColor::fillXImageMosaic(XImage* xmap, 
				      int x0, int y0, int x1, int y1,
				      double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render

  if (!colorScale)
    return;

  int& width = xmap->width;
  int& height = xmap->height;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();
  int bytesPerPixel = xmap->bits_per_pixel/8;
  int last = length * bytesPerPixel;

  // clear ximage

  memset(xmap->data, 255, xmap->bytes_per_line * height);

  // test as doubles, convert to int's at last minute.
  // use double to maintain precision
  // This solves a host of problems.

  for (int j=y0; j<y1; j++) {

    // the line may be padded at the end
    char* dest = xmap->data + j*xmap->bytes_per_line + x0*bytesPerPixel;

    for (int i=x0; i<x1; i++, dest+=bytesPerPixel) {

      FitsImage* ptr = fits;
      while (ptr) {
	double* m = (ptr->*getMatrix)();
	double x = i*m[0] + j*m[3] + m[6];
	double y = i*m[1] + j*m[4] + m[7];

	int* params = ptr->getDataParams(scanMode);
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  double ll = ptr->getLowDouble();
	  double hh = ptr->getHighDouble();
	  double diff = hh - ll;

	  if (!isNaNd(value))
	    if (value <= ll)
	      memcpy(dest, table, bytesPerPixel);
	    else if (value >= hh)
	      memcpy(dest, table+last, bytesPerPixel);
	    else
	      memcpy(dest,
		     table+((int)(((value-ll)/diff*length)+.5))*bytesPerPixel,
		     bytesPerPixel);

	  break;
	}
	else 
	  ptr = ptr->next();
      }
    }
  }
}

void FrameTrueColor::fillXImageMosaicFast(XImage* xmap, 
					  int x0, int y0, int x1, int y1,
					  double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render

  if (!colorScale)
    return;

  if (!fits)
    return;

  int& width = xmap->width;
  int& height = xmap->height;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();
  int bytesPerPixel = xmap->bits_per_pixel/8;
  int last = length * bytesPerPixel;

  // clear ximage

  memset(xmap->data, 255, xmap->bytes_per_line * height);

  FitsImage* ptr = NULL;
  double* m;

  int* params;
  int srcw;
  int xmin;
  int xmax;
  int ymin;
  int ymax;

  double ll;
  double hh;
  double diff;

  // test as doubles, convert to int's at last minute.
  // use double to maintain precision
  // This solves a host of problems.

  for (int j=y0; j<y1; j++) {

    // the line may be padded at the end
    char* dest = xmap->data + j*xmap->bytes_per_line + x0*bytesPerPixel;

    for (int i=x0; i<x1; i++, dest+=bytesPerPixel) {
      double x,y;

      if (ptr) {
	x = i*m[0] + j*m[3] + m[6];
	y = i*m[1] + j*m[4] + m[7];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  if (!isNaNd(value))
	    if (value <= ll)
	      memcpy(dest, table, bytesPerPixel);
	    else if (value >= hh)
	      memcpy(dest, table+last, bytesPerPixel);
	    else
	      memcpy(dest,
		     table+((int)(((value-ll)/diff*length)+.5))*bytesPerPixel,
		     bytesPerPixel);
	}
	else
	  ptr = NULL;
      }

      if (!ptr) {
	ptr = fits;
	while (ptr) {
	  m = (ptr->*getMatrix)();
	  x = i*m[0] + j*m[3] + m[6];
	  y = i*m[1] + j*m[4] + m[7];

	  params = ptr->getDataParams(scanMode);
	  srcw = params[0];
	  xmin = params[1];
	  xmax = params[2];
	  ymin = params[3];
	  ymax = params[4];

	  if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	    double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	    ll = ptr->getLowDouble();
	    hh = ptr->getHighDouble();
	    diff = hh - ll;

	    if (!isNaNd(value))
	      if (value <= ll)
		memcpy(dest, table, bytesPerPixel);
	      else if (value >= hh)
		memcpy(dest, table+last, bytesPerPixel);
	      else
		memcpy(dest,
		       table+((int)(((value-ll)/diff*length)+.5))*
		       bytesPerPixel,
		       bytesPerPixel);
	    break;
	  }
	  else 
	    ptr = ptr->next();
	}
      }
    }
  }
}

void FrameTrueColor::updateMagnifierReplicate(const Vector& v)
{
  // v is in CANVAS Coords
  // we have a race condition here. It's possible that a magnifier command has
  // been generated right after the pixmap has been invalidated and before
  // a new XImage has been generated. So, bail, and try again.

  if (!baseXImage || !magnifierXImage)
    return;

  char* image = baseXImage->data;
  int imageBytesPerPixel = baseXImage->bits_per_pixel/8;
  int& imageBytesPerLine = baseXImage->bytes_per_line;

  char* mag = magnifierXImage->data;
  int magBytesPerPixel = magnifierXImage->bits_per_pixel/8;
  int& magBytesPerLine = magnifierXImage->bytes_per_line;

  int white = getWhiteColor();

  // map upper left of magnifier to image

  Matrix m = Translate(-(v * canvasToWidget)) * ximageToMagnifier;
  Matrix n = m.invert();

  Vector ll = Vector(0,0) * n;

  // and init image row and column counters

  int imageY = (int)ll[1];

  // loop thru each magnifier pixel

  char* row = new char[magBytesPerLine];
  for (int j=0; j<=magnifierHeight; j+=(int)magnifierZoom, imageY++) {

    // Clear tmp row

    memset(row, white, magBytesPerLine);

    // Build tmp row

    int imageX = (int)ll[0];
    for (int i=0; i<=magnifierWidth; i+=(int)magnifierZoom, imageX++) {

      // find pixel value

      unsigned char* value = new unsigned char[magBytesPerPixel];
      if (imageX>=0 && imageX<options->width && imageY>=0 
	  && imageY<options->height)
	memcpy(value, image + imageY*imageBytesPerLine + 
	       imageX*imageBytesPerPixel, magBytesPerPixel);
      else
	memset(value, white, magBytesPerPixel);

      // replicate pixel

      int start = i - (int)(magnifierZoom/2.);
      if (start < 0)
	start = 0;
      int stop = i + (int)(magnifierZoom/2.);
      if (stop > magnifierWidth)
	stop = magnifierWidth;

      for (int k=start; k<stop; k++)
	memcpy(row+k*magBytesPerPixel, value, magBytesPerPixel);

      delete value;
    }

    // and replicate temp row
    
    int start = j - (int)(magnifierZoom/2.);
    if (start < 0)
      start = 0;
    int stop = j + (int)(magnifierZoom/2.);
    if (stop > magnifierHeight)
      stop = magnifierHeight;
    for (int k=start; k<stop; k++)
	memcpy(mag+(k*magBytesPerLine), row, magBytesPerLine);
  }
  delete row;

  XPutImage(display, magnifierPixmap, gc, magnifierXImage, 0, 0, 0, 0, 
	    magnifierWidth, magnifierHeight);
}

void FrameTrueColor::colormapCmd(int id, float b, float c, int i, 
				 unsigned short* index, unsigned char* cells, 
				 int cnt)
{
  // first check for change

  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change

  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  if (fitsCount) {
    update(BASE); // always update
    skipUpdate = 0;
  }
  else {
    if (!skipUpdate) {
      update(BASE); // update once
      skipUpdate = 1;
    }      
  }
}

void FrameTrueColor::colormapBeginCmd()
{
  colormapBeginCmd(Vector(options->width,options->height)/2*widgetToCanvas);
}

void FrameTrueColor::colormapBeginCmd(const Vector& w)
{
  // we need some fits data
  // we assume the colorScale length will not change during motion calls

  if (!currentFits)
    return;

  int width = options->width;
  int height = options->height;
  cmapOffset = Vector();
  if (cmapArea) {
    width = height = cmapArea;

    switch (cmapMode) {
    case CENTER:
      cmapOffset = (Vector(options->width,options->height) - 
		    Vector(width,height))/2;
      break;
    case CLICK: 
      {
	Vector where = w * canvasToWidget;
	if (where[0] >= 0 && where[0] < options->width &&
	    where[1] >= 0 && where[1] < options->height)
	  cmapOffset = where - Vector(width,height)/2;
	else
	  cmapOffset = (Vector(options->width,options->height) - 
			Vector(width,height))/2;
      }
      break;
    }

    if (cmapOffset[0] < 0) {
      width += (int)cmapOffset[0];
      cmapOffset[0] = 0;
    }
    if (cmapOffset[0]+width >= options->width)
      width = options->width - (int)cmapOffset[0];

    if (cmapOffset[1] < 0) {
      height += (int)cmapOffset[1];
      cmapOffset[1] = 0;
    }

    if (cmapOffset[1]+height >= options->height)
      height = options->height - (int)cmapOffset[1];
  }

  // Create XImage

  if (!(colormapXM = XGETIMAGE(display, pixmap, 0, 0, 
			       width, height, AllPlanes, ZPixmap))) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap XImage" 
	 << endl;
    exit(1);
  }

  // Create Pixmap

  colormapPM = Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height, depth);
  if (!colormapPM) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap Pixmap" 
	 << endl;
    exit(1);
  }

  // colormapGCXOR

  colormapGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  // Create table index array

  if (colormapData)
    delete [] colormapData;
  colormapData = new long[width*height];

  if (!colormapData) {
    cerr << "FrameTrueColor Internal Error: Unable to alloc tmp data array" 
	 << endl;
    exit(1);
  }

  // fill data array

  switch (displayMode) {
  case SINGLE:
    colormapBeginSingle(width, height);
    break;
  case MOSAIC:
    if (mosaicFast && fits->mosaicType() == IRAF)
      colormapBeginMosaicFast(width, height);
    else
      colormapBeginMosaic(width, height);

    break;
  }
}

void FrameTrueColor::colormapBeginSingle(int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  double* mm = currentFits->getmWidgetToData();
  int* params = currentFits->getDataParams(scanMode);
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  long* dest = colormapData;
  for (int j=yoff; j<yoff+height; j++) {
    for (int i=xoff; i<xoff+width; i++, dest++) {
      *dest = -1; // initialize

      double x = i*mm[0] + j*mm[3] + mm[6];
      double y = i*mm[1] + j*mm[4] + mm[7];

      if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	double value = currentFits->getValueDouble(((int)y)*srcw + (int)x);
	
	if (!isNaNd(value))
	  if (value <= ll)
	    *dest = 0;
	  else if (value >= hh)
	    *dest = last;
	  else
	    *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
      }
    }
  }
}

void FrameTrueColor::colormapBeginMosaic(int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  // scan fits

  long* dest = colormapData;
  for (int j=yoff; j<yoff+height; j++) {
    for (int i=xoff; i<xoff+width; i++, dest++) {
      *dest = -1; // initialize

      FitsImage* ptr = fits;
      while (ptr) {
	double* m = ptr->getmWidgetToData();
	double x = i*m[0] + j*m[3] + m[6];
	double y = i*m[1] + j*m[4] + m[7];

	int* params = ptr->getDataParams(scanMode);
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  double ll = ptr->getLowDouble();
	  double hh = ptr->getHighDouble();
	  double diff = hh - ll;

	  if (!isNaNd(value))
	    if (value <= ll)
	      *dest = 0;
	    else if (value >= hh)
	      *dest = last;
	    else
	      *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;

	  break;
	}
	else 
	  ptr = ptr->next();
      }
    }
  }
}

void FrameTrueColor::colormapBeginMosaicFast(int width, int height)
{
  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  FitsImage* ptr = NULL;
  double* m;

  int* params;
  int srcw;
  int xmin;
  int xmax;
  int ymin;
  int ymax;

  double ll;
  double hh;
  double diff;

  // scan fits

  long* dest = colormapData;
  for (int j=yoff; j<yoff+height; j++) {
    for (int i=xoff; i<xoff+width; i++, dest++) {
      double x,y;
      *dest = -1; // initialize

      if (ptr) {
	x = i*m[0] + j*m[3] + m[6];
	y = i*m[1] + j*m[4] + m[7];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	  if (!isNaNd(value))
	    if (value <= ll)
	      *dest = 0;
	    else if (value >= hh)
	      *dest = last;
	    else
	      *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	}
	else
	  ptr = NULL;
      }

      if (!ptr) {
	ptr = fits;
	while (ptr) {
	  m = ptr->getmWidgetToData();
	  x = i*m[0] + j*m[3] + m[6];
	  y = i*m[1] + j*m[4] + m[7];

	  params = ptr->getDataParams(scanMode);
	  srcw = params[0];
	  xmin = params[1];
	  xmax = params[2];
	  ymin = params[3];
	  ymax = params[4];

	  if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	    double value = ptr->getValueDouble(((int)y)*srcw + (int)x);

	    ll = ptr->getLowDouble();
	    hh = ptr->getHighDouble();
	    diff = hh - ll;

	    if (!isNaNd(value))
	      if (value <= ll)
		*dest = 0;
	      else if (value >= hh)
		*dest = last;
	      else
		*dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;

	    break;
	  }
	  else 
	    ptr = ptr->next();
	}
      }
    }
  }
}

void FrameTrueColor::colormapMotionCmd(int id, float b, float c, int i, 
				       unsigned short* index, 
				       unsigned char* cells, int cnt)
{
  // first check for change

  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change

  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  // if we have no data, stop now

  if (!currentFits) 
    return;

  // clear ximage

  int& width = colormapXM->width;
  int& height = colormapXM->height;

  memset(colormapXM->data, 255, colormapXM->bytes_per_line * height);

  int bytesPerPixel = colormapXM->bits_per_pixel/8;
  int& bytesPerLine = colormapXM->bytes_per_line;

  const unsigned char* table = colorScale->colors();

  long* src = colormapData;
  for (int j=0; j<height; j++) {
    // line may be padded at end
    char* dest = colormapXM->data + j*bytesPerLine;

    for (int i=0; i<width; i++, src++, dest+=bytesPerPixel)
      if (*src >=0)
	memcpy(dest, table+(*src), bytesPerPixel);
  }

  // XImage to Pixmap

  XPutImage(display, colormapPM, gc, colormapXM, 0, 0, 0, 0, width, height);

  // Display Pixmap

  Vector d = cmapOffset * widgetToWindow;
  XCopyArea(display, colormapPM, Tk_WindowId(tkwin), colormapGCXOR, 0, 0, 
	    width, height, (int)d[0], (int)d[1]);
}

void FrameTrueColor::colormapEndCmd(int id, float b, float c, int i, 
				    unsigned short* index, 
				    unsigned char* cells, int cnt)
{
  if (colormapXM) {
    XDestroyImage(colormapXM);
    colormapXM = NULL;
  }

  if (colormapPM) {
    Tk_FreePixmap(display, colormapPM);
    colormapPM = 0;
  }

  if (colormapData) {
    delete [] colormapData;
    colormapData = NULL;
  }

  if (colormapGCXOR) {
    XFreeGC(display, colormapGCXOR);
    colormapGCXOR = 0;
  }

  if (currentFits)
    update(BASE); // always update
}

void FrameTrueColor::rotateMotion()
{
  // Rotate from src to dest

  Vector c = Vector(options->width,options->height)/2;
  Matrix m = (Translate(-c) * Rotate(rotation-rotateRotation) * 
	      Translate(c)).invert();
  double* mm = m.mm();

  int& width = options->width;
  int& height = options->height;
  char* src = rotateSrcXM->data;

  int bytesPerPixel = rotateDestXM->bits_per_pixel/8;

  for (int j=0; j<height; j++) {
    // the line may be padded at the end
    char* dest = rotateDestXM->data + j*rotateDestXM->bytes_per_line;

    for (int i=0; i<width; i++, dest+=bytesPerPixel) {
      double x = i*mm[0] + j*mm[3] + mm[6];
      double y = i*mm[1] + j*mm[4] + mm[7];

      if (x >= 0 && x < width && y >= 0 && y < height)
	memcpy(dest, 
	       src + ((int)y)*rotateDestXM->bytes_per_line+
	       ((int)x)*bytesPerPixel, 
	       bytesPerPixel);
      else
	memset(dest, 255, bytesPerPixel);
    }
  }

  // XImage to Pixmap

  XPutImage(display, rotatePM, gc, rotateDestXM,
	    0, 0, 0, 0, options->width, options->height);

  // Display Pixmap

  Vector d = Vector(0,0) * widgetToWindow;
  XCopyArea(display, rotatePM, Tk_WindowId(tkwin), rotateGCXOR, 0, 0, 
	    options->width, options->height, (int)d[0], (int)d[1]);
}
