#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include <unistd.h>

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <iomanip.h>

#include <Xlib.h>
#include <Xutil.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "mmap.h"

#include "fvcontour.h"
#include "grid.h"

// Frame Command Member Functions

void Frame::clearCmd()
{
  unloadFits();

  reset();

  updateClip();
  update(MATRIX);
}

void Frame::colorScaleCmd(ColorScaleType s)
{
  colorScaleType = s;
  updateColorScale();

  update(BASE);
}

void Frame::getColormapCmd()
{
  char buf[64];
  ostrstream str(buf,64);
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getColorScaleCmd()
{
  switch (colorScaleType) {
  case LINEARSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case LOGSCALE:
    Tcl_AppendResult(interp, "log", NULL);
    break;
  case SQUAREDSCALE:
    Tcl_AppendResult(interp, "squared", NULL);
    break;
  case SQRTSCALE:
    Tcl_AppendResult(interp, "sqrt", NULL);
    break;
  case IISSCALE:
    Tcl_AppendResult(interp, "linear", NULL);
    break;
  case HISTEQUSCALE:
    Tcl_AppendResult(interp, "histequ", NULL);
    break;
  }
}

void Frame::getCoordCmd(const Vector& v, CoordSystem in, CoordSystem out, 
			CoordFormat format, Precision p)
{
  switch (in) {
  case CANVAS:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getCanvasToData, NULL))
      printCoordTcl(ptr, v, in, out, format, p);
    break;
  case PANNER:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getPannerToData, NULL))
      printCoordTcl(ptr, v, in, out, format, p);
    break;
  }
}

void Frame::getCursorCmd(CoordSystem out, CoordFormat format, Precision p)
{
  if (currentFits)
    printCoordTcl(currentFits, cursor, USER, out, format, p);
  else
    printVectorTcl(Vector(), DEFAULT);
} 

void Frame::getInfoCmd(char* var)
{
  if (currentFits) {
    Tcl_SetVar2(interp,var,"filename",(char*)currentFits->getRootFileName(),0);
    Tcl_SetVar2(interp,var,"objectname",(char*)currentFits->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)currentFits->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)currentFits->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)currentFits->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)currentFits->getHigh(),0);
  }
  else {
    Tcl_SetVar2(interp,var,"objectname","",0);
    Tcl_SetVar2(interp,var,"filename","",0);
    Tcl_SetVar2(interp,var,"min","",0);
    Tcl_SetVar2(interp,var,"max","",0);
    Tcl_SetVar2(interp,var,"low","",0);
    Tcl_SetVar2(interp,var,"high","",0);
  }

  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

  Tcl_SetVar2(interp,var,"xWCS","",0);
  Tcl_SetVar2(interp,var,"yWCS","",0);
  Tcl_SetVar2(interp,var,"sysWCS","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS","",0);
}

void Frame::getInfoCmd(const Vector& v, CoordSystem ref,
		       CoordSystem wcs, CoordFormat format, char* var)
{
  switch (ref) {
  case CANVAS:
    getInfo(v, ref, &FitsImage::getCanvasToData, &FitsImage::getCanvasToImage,
	    wcs, format, var);
    break;
  case PANNER:
    getInfo(v, ref, &FitsImage::getPannerToData, &FitsImage::getPannerToImage,
	    wcs, format, var);
    break;
  }
}

void Frame::getInfo(const Vector& v, CoordSystem ref,
		    Matrix& (FitsImage::*matrixToData)(),
		    Matrix& (FitsImage::*matrixToImage)(),
		    CoordSystem wcs, CoordFormat format, char* var)
{
  switch (displayMode) {
  case SINGLE:
    getInfoSingle(v,ref,matrixToData,matrixToImage,wcs,format,var);
    break;
  case MOSAIC:
    getInfoMosaic(v,ref,matrixToData,matrixToImage,wcs,format,var);
    break;
  }
}

void Frame::getInfoSingle(const Vector& v, CoordSystem ref,
			  Matrix& (FitsImage::*matrixToData)(),
			  Matrix& (FitsImage::*matrixToImage)(),
			  CoordSystem wcs, CoordFormat format, char* var)
{
  if (currentFits) {
    Vector img = v * (currentFits->*matrixToData)();
    int* params = currentFits->getDataParams(scanMode);
    int& xmin = params[1];
    int& xmax = params[2];
    int& ymin = params[3];
    int& ymax = params[4];

    Tcl_SetVar2(interp,var,"filename",(char*)currentFits->getRootFileName(),0);
    Tcl_SetVar2(interp,var,"objectname",(char*)currentFits->getObjectName(),0);
    Tcl_SetVar2(interp,var,"min",(char*)currentFits->getMin(),0);
    Tcl_SetVar2(interp,var,"max",(char*)currentFits->getMax(),0);
    Tcl_SetVar2(interp,var,"low",(char*)currentFits->getLow(),0);
    Tcl_SetVar2(interp,var,"high",(char*)currentFits->getHigh(),0);

    if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
      Tcl_SetVar2(interp,var,"value",(char*)currentFits->getValue(img),0);

      coordToTclArray(currentFits, v, ref, IMAGE, FIXED, var, "Image");
      coordToTclArray(currentFits, v, ref, PHYSICAL, FIXED, var, "Physical");

      if (hasATMV())
	coordToTclArray(currentFits,v,ref, AMPLIFIER, FIXED, var, "Amplifier");
      else {
	Tcl_SetVar2(interp,var,"xAmplifier","",0);
	Tcl_SetVar2(interp,var,"yAmplifier","",0);
      }

      if (hasDTMV())
	coordToTclArray(currentFits, v, ref, DETECTOR, FIXED, var, "Detector");
      else {
	Tcl_SetVar2(interp,var,"xDetector","",0);
	Tcl_SetVar2(interp,var,"yDetector","",0);
      }

      if (hasWCS()) {
	char buf[64];
	currentFits->pix2wcs(v*(currentFits->*matrixToImage)(), wcs, 
			     format, buf, 64);
	int argc;
	char** argv;
	Tcl_SplitList(interp, buf, &argc, &argv);

	if (argv) {
	  if (argc > 0 && argv[0])
	    Tcl_SetVar2(interp,var,"xWCS",argv[0],0);
	  else
	    Tcl_SetVar2(interp,var,"xWCS","",0);

	  if (argc > 1 && argv[1])
	    Tcl_SetVar2(interp,var,"yWCS",argv[1],0);
	  else
	    Tcl_SetVar2(interp,var,"yWCS","",0);

	  if (argc > 2 && argv[2])
	    Tcl_SetVar2(interp,var,"sysWCS",argv[2],0);
	  else
	    Tcl_SetVar2(interp,var,"sysWCS","",0);

	  if (argc > 3 && argv[3])
	    Tcl_SetVar2(interp,var,"xUnitWCS",argv[3],0);
	  else
	    Tcl_SetVar2(interp,var,"xUnitWCS","",0);

	  if (argc > 4 && argv[4])
	    Tcl_SetVar2(interp,var,"yUnitWCS",argv[4],0);
	  else
	    Tcl_SetVar2(interp,var,"yUnitWCS","",0);

	  Tcl_Free((char*)argv);
	}
	else
	  goto noWCS;
      }
      else 
	goto noWCS;
    }
    else
      goto noImage;

    return;
  }

  // else, return blanks

  Tcl_SetVar2(interp,var,"objectname","",0);
  Tcl_SetVar2(interp,var,"filename","",0);
  Tcl_SetVar2(interp,var,"min","",0);
  Tcl_SetVar2(interp,var,"max","",0);
  Tcl_SetVar2(interp,var,"low","",0);
  Tcl_SetVar2(interp,var,"high","",0);

noImage:
  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

noWCS:
  Tcl_SetVar2(interp,var,"xWCS","",0);
  Tcl_SetVar2(interp,var,"yWCS","",0);
  Tcl_SetVar2(interp,var,"sysWCS","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS","",0);
}

void Frame::getInfoMosaic(const Vector& v, CoordSystem ref,
			  Matrix& (FitsImage::*matrixToData)(),
			  Matrix& (FitsImage::*matrixToImage)(),
			  CoordSystem wcs, CoordFormat format, char* var)
{
  if (fits) {
    FitsImage* ptr = fits;
    while (ptr) {
      Vector img = v * (ptr->*matrixToData)();
      int* params = ptr->getDataParams(scanMode);
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	Tcl_SetVar2(interp,var,"filename",(char*)ptr->getRootFileName(),0);
	Tcl_SetVar2(interp,var,"objectname",(char*)ptr->getObjectName(),0);
	Tcl_SetVar2(interp,var,"min",(char*)ptr->getMin(),0);
	Tcl_SetVar2(interp,var,"max",(char*)ptr->getMax(),0);
	Tcl_SetVar2(interp,var,"low",(char*)ptr->getLow(),0);
	Tcl_SetVar2(interp,var,"high",(char*)ptr->getHigh(),0);
	Tcl_SetVar2(interp,var,"value",(char*)ptr->getValue(img),0);

	coordToTclArray(ptr, v, ref, IMAGE, FIXED, var, "Image");
	coordToTclArray(ptr, v, ref, PHYSICAL, FIXED, var, "Physical");

	if (hasATMV())
	  coordToTclArray(ptr, v, ref, AMPLIFIER, FIXED, var, "Amplifier");
	else {
	  Tcl_SetVar2(interp,var,"xAmplifier","",0);
	  Tcl_SetVar2(interp,var,"yAmplifier","",0);
	}

	if (hasDTMV())
	  coordToTclArray(ptr, v, ref, DETECTOR, FIXED, var, "Detector");
	else {
	  Tcl_SetVar2(interp,var,"xDetector","",0);
	  Tcl_SetVar2(interp,var,"yDetector","",0);
	}

	if (ptr->hasWCS()) {
	  char buf[64];
	  ptr->pix2wcs(v*(ptr->*matrixToImage)(), wcs, format, buf, 64);

	  int argc;
	  char** argv;
	  Tcl_SplitList(interp, buf, &argc, &argv);

	  if (argv) {
	    if (argc > 0 && argv[0])
	      Tcl_SetVar2(interp,var,"xWCS",argv[0],0);
	    else
	      Tcl_SetVar2(interp,var,"xWCS","",0);

	    if (argc > 1 && argv[1])
	      Tcl_SetVar2(interp,var,"yWCS",argv[1],0);
	    else
	      Tcl_SetVar2(interp,var,"yWCS","",0);

	    if (argc > 2 && argv[2])
	      Tcl_SetVar2(interp,var,"sysWCS",argv[2],0);
	    else
	      Tcl_SetVar2(interp,var,"sysWCS","",0);

	    if (argc > 3 && argv[3])
	      Tcl_SetVar2(interp,var,"xUnitWCS",argv[3],0);
	    else
	      Tcl_SetVar2(interp,var,"xUnitWCS","",0);

	    if (argc > 4 && argv[4])
	      Tcl_SetVar2(interp,var,"yUnitWCS",argv[4],0);
	    else
	      Tcl_SetVar2(interp,var,"yUnitWCS","",0);

	    Tcl_Free((char*)argv);
	  }
	  else
	    goto noWCS;
	}
	else 
	  goto noWCS;

	return;
      }
      ptr = ptr->next();
    }
  }

  // else, return blanks

  Tcl_SetVar2(interp,var,"objectname","",0);
  Tcl_SetVar2(interp,var,"filename","",0);
  Tcl_SetVar2(interp,var,"min","",0);
  Tcl_SetVar2(interp,var,"max","",0);
  Tcl_SetVar2(interp,var,"low","",0);
  Tcl_SetVar2(interp,var,"high","",0);
  Tcl_SetVar2(interp,var,"value","",0);
  Tcl_SetVar2(interp,var,"xImage","",0);
  Tcl_SetVar2(interp,var,"yImage","",0);
  Tcl_SetVar2(interp,var,"xPhysical","",0);
  Tcl_SetVar2(interp,var,"yPhysical","",0);
  Tcl_SetVar2(interp,var,"xAmplifier","",0);
  Tcl_SetVar2(interp,var,"yAmplifier","",0);
  Tcl_SetVar2(interp,var,"xDetector","",0);
  Tcl_SetVar2(interp,var,"yDetector","",0);

noWCS:
  Tcl_SetVar2(interp,var,"xWCS","",0);
  Tcl_SetVar2(interp,var,"yWCS","",0);
  Tcl_SetVar2(interp,var,"sysWCS","",0);
  Tcl_SetVar2(interp,var,"xUnitWCS","",0);
  Tcl_SetVar2(interp,var,"yUnitWCS","",0);
}

void Frame::getNameCmd()
{
  Tcl_AppendResult(interp, name, NULL);
}

void Frame::getValueCmd(const Vector& v, CoordSystem ref)
{
  Vector r;

  switch (ref) {
  case CANVAS:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getCanvasToData, &r))
      Tcl_AppendResult(interp, ptr->getValue(r), NULL);
    break;
  case PANNER:
    if (FitsImage* ptr=isInFits(v, &FitsImage::getPannerToData, &r))
      Tcl_AppendResult(interp, ptr->getValue(r), NULL);
    break;
  }
}

void Frame::getWidthCmd()
{
  if (currentFits)
    printIntegerTcl(currentFits->width());
}

void Frame::getHeightCmd()
{
  if (currentFits)
    printIntegerTcl(currentFits->height());
}

void Frame::getBitpixCmd()
{
  if (currentFits)
    printIntegerTcl(currentFits->bitpix());
}

void Frame::highliteCmd(int which)
{
  useHighlite = which ? 1 : 0;
  update(PIXMAP);
}

void Frame::mosaicFastCmd(int which)
{
  mosaicFast = which ? 1 : 0;
}

void Frame::setNameCmd(char* n)
{
  if (name)
    delete name;

  name = dupstr(n);
}

void Frame::warpCmd(const Vector& v)
{
  Vector a=v; // can't use a const here
  XWARPPOINTER(display, None, None, 0, 0, 0, 0, (int)a[0], (int)a[1]);
}

void Frame::warpToCmd(const Vector& v)
{
  Vector a =v*canvasToWindow;
  XWARPPOINTER(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)a[0],(int)a[1]);
}

// Clip Commands

void Frame::clipMinMaxModeCmd(MinMaxMode m)
{
  mmMode = m;
  updateClip();
  initHistogram();
  update(BASE);
}

void Frame::clipMinMaxParamCmd(int i)
{
  mmIncr = i;
  updateClip();
  initHistogram();
  update(BASE);
}

void Frame::clipModeCmd(float per)
{
  if (per == 100)
    clipMode = MINMAX;
  else {
    autoCutPer = per;
    clipMode = AUTOCUT;
  }
  updateClip();
  update(BASE);
}

void Frame::clipModeCmd(ClipMode m)
{
  clipMode = m;
  updateClip();
  update(BASE);
}

void Frame::clipScopeCmd(ClipScope s)
{
  clipScope = s;
  updateClip();
  update(BASE);
}

void Frame::clipUserCmd(double l, double h)
{
  uLow = l;
  uHigh = h;
  updateClip();
  update(BASE);
}

void Frame::clipUserLowCmd(double l)
{
  uLow = l;
  updateClip();
  update(BASE);
}

void Frame::clipUserHighCmd(double h)
{
  uHigh = h;
  updateClip();
  update(BASE);
}

void Frame::clipZScaleParamCmd(float c, int s, int l)
{
  zContrast = c;
  zSampleSize = s;
  zSampleLine = l;
  updateClip();
  update(BASE);
}

void Frame::getClipCmd(float per)
{
  ClipMode cm = (per == 100) ? MINMAX : AUTOCUT;
  float ac = per;

  char buf[32];
  ostrstream str(buf,32);
  str << getClip(cm, ac) << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getClipCmd(ClipMode cm)
{
  char buf[32];
  ostrstream str(buf,32);
  str << getClip(cm, autoCutPer) << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getClipCmd()  // Get Frame Low High
{
  printVectorTcl(Vector(low, high), DEFAULT);
}

void Frame::getClipMinMaxModeCmd()
{
  switch (mmMode) {
  case SCAN:
    Tcl_AppendResult(interp, "scan", NULL);
    return;
  case SAMPLE:
    Tcl_AppendResult(interp, "sample", NULL);
    return;
  case DATAMIN:
    Tcl_AppendResult(interp, "datamin", NULL);
    return;
  case IRAFMIN:
    Tcl_AppendResult(interp, "irafmin", NULL);
    return;
  }
}

void Frame::getClipModeCmd()
{
  switch (clipMode) {
  case MINMAX:
    Tcl_AppendResult(interp, "minmax", NULL);
    break;
  case ZSCALE:
    Tcl_AppendResult(interp, "zscale", NULL);
    break;
  case ZMAX:
    Tcl_AppendResult(interp, "zmax", NULL);
    break;
  case AUTOCUT:
    printDoubleTcl(autoCutPer, DEFAULT);
    break;
  case USERCLIP:
    Tcl_AppendResult(interp, "user", NULL);
    break;
  }
}

void Frame::getClipScopeCmd()
{
  switch (clipScope) {
  case GLOBAL:
    Tcl_AppendResult(interp, "global", NULL);
    break;
  case LOCAL:
    Tcl_AppendResult(interp, "local", NULL);
    break;
  }
}

void Frame::getClipUserCmd()
{
  printVectorTcl(Vector(uLow, uHigh), DEFAULT);
}

void Frame::getClipZScaleParamCmd()
{
  char buf[64];
  ostrstream str(buf,64);
  str << zContrast << ' ' << zSampleSize << ' ' << zSampleLine << ends;

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::getHistogramCmd(char* xName, char* yName, int num)
{
  bltHist(xName,yName,num);
  printDoubleTcl((max-min)/num, DEFAULT);
}

void Frame::getMinMaxCmd()  // Get Frame Min Max
{
  printVectorTcl(Vector(min, max), DEFAULT);
}

void Frame::hasDATAMINCmd()
{
  if (currentFits && currentFits->hasDATAMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasIRAFMINCmd()
{
  if (currentFits && currentFits->hasIRAFMIN())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Crosshair Commands

void Frame::crosshairCmd(int which)
{
  useCrosshair = which ? 1 : 0;
  update(PIXMAP);
}

void Frame::crosshairCmd(const Vector& v, CoordSystem sys)
{
  useCrosshair = 1;
  switch (sys) {
  case CANVAS:
    crosshair = v * canvasToRef;
    break;
  case PANNER:
    crosshair = v * pannerToRef;
    break;
  default:
    crosshair = currentFits ? currentFits->map(v, sys, REF) : v;
  }
  update(PIXMAP);
}

void Frame::crosshairWarpCmd(const Vector& v)
{
  useCrosshair = 1;

  Vector r = crosshair * refToCanvas;
  r += v;
  crosshair = r * canvasToRef;

  update(PIXMAP);
  updateMagnifier();
}

void Frame::crosshairBeginCmd(const Vector& v, CoordSystem sys)
{
  // This routine is optimalized for rendering an interactive cursor. First,
  // a cursorless pixmap is generated. Then as the cursor is moved about, the 
  // cursor is written directly to the display, and erase by XCopyArea from the
  // widget pixmap.

  // turn cursor off and generate pixmap

  useCrosshair = 0;
  updateNow(PIXMAP);

  // turn cursor on, and draw directly to display

  useCrosshair = 1;
  switch (sys) {
  case CANVAS:
    crosshair = v * canvasToRef;
    break;
  case PANNER:
    crosshair = v * pannerToRef;
    break;
  default:
    crosshair = currentFits ? currentFits->map(v, sys, REF) : v;
  }
  drawCrosshair();
}

void Frame::crosshairMotionCmd(const Vector& v, CoordSystem sys)
{
  eraseCrosshair();
  switch (sys) {
  case CANVAS:
    crosshair = v * canvasToRef;
    break;
  case PANNER:
    crosshair = v * pannerToRef;
    break;
  default:
    crosshair = currentFits ? currentFits->map(v, sys, REF) : v;
  }
  drawCrosshair();
}

void Frame::getCrosshairCmd(CoordSystem sys, CoordFormat format, Precision p)
{
  switch (sys) {
  case CANVAS:
    printVectorTcl(crosshair*refToCanvas, p);
    break;
  case PANNER:
    printVectorTcl(crosshair*refToPanner, p);
    break;
  default:
    if (currentFits)
      printCoordTcl(currentFits, crosshair, REF, sys, format, p);
  }
}

void Frame::getCrosshairStatusCmd()
{
  if (useCrosshair)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getHorzCutCmd(char* x, char* y)
{
  bltCut(x, y, XX, crosshair * refToWidget);
}

void Frame::getHorzCutCmd(char* x, char* y, const Vector& v)
{
  bltCut(x, y, XX, v * canvasToWidget);
}

void Frame::getVertCutCmd(char* x, char* y)
{
  bltCut(x, y, YY, crosshair * refToWidget);
}

void Frame::getVertCutCmd(char* x, char* y, const Vector& v)
{
  bltCut(x, y, YY, v * canvasToWidget);
}

// DATASEC Commands

void Frame::DATASECCmd(int which)
{
  switch (scanMode) {
  case NODATASEC:
  case DATASEC:
  case RESET:
    scanMode = which ? DATASEC : NODATASEC;
    break;
  case UNODATASEC:
  case UDATASEC:
    scanMode = which ? UDATASEC : UNODATASEC;
    break;
  }

  updateClip();
  initHistogram();
  update(BASE);
}

void Frame::getDATASECCmd()
{
  if (scanMode)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasDATASECCmd()
{
  if (currentFits && currentFits->hasDATASEC())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Fits Commands

void Frame::getFitsHeaderCmd(int which)
{
  FitsImage* ptr = fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }
  }
  else if (which < 0) {
    for (int i=0; i<((-which)-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }
  }

  if (ptr) {
    char* h = (which > 0) ? ptr->displayHeader() : ptr->displayPrimary();
    Tcl_AppendResult(interp, h, NULL);
    delete h;
  }
  else
    result = TCL_ERROR;
}

void Frame::getFitsHeaderKeywordCmd(int which, const char* key)
{
  FitsImage* ptr = fits;
  for (int i=0; i<(which-1); i++)
    if (ptr)
      ptr = ptr->next();

  if (ptr) {
    char* value = ptr->getKeyword(key);
    if (value) {
      Tcl_AppendResult(interp, value, NULL);
      delete value;
      return;
    }
  }

  result = TCL_ERROR;
}

void Frame::getFitsFileNameCmd(int which, FileNameType type)
{
  if (which > 0) {
    FitsImage* ptr = fits;
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }

    if (ptr) {
      if (type == Frame::BASIC)
	Tcl_AppendResult(interp, ptr->getRootFileName(), NULL);
      else
	Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
    }
  }
  else {
    Tcl_AppendResult(interp, " file index must be greater than zero.", NULL);
    result = TCL_ERROR;
  }
}

void Frame::getFitsFileNameCmd(const Vector& v, CoordSystem ref, 
			       FileNameType type)
{
  switch (ref) {
  case CANVAS:
    if (FitsImage* ptr = isInFits(v, &FitsImage::getCanvasToData, NULL)) {
      if (type == Frame::BASIC)
	Tcl_AppendResult(interp, ptr->getRootFileName(), NULL);
      else
	Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
    }
    break;
  case PANNER:
    if (FitsImage* ptr = isInFits(v, &FitsImage::getPannerToData, NULL)) {
      if (type == Frame::BASIC)
	Tcl_AppendResult(interp, ptr->getRootFileName(), NULL);
      else
	Tcl_AppendResult(interp, ptr->getFullFileName(), NULL);
    }
    break;
  }
}

void Frame::getFitsFileNamesCmd(FileNameType type)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      if (type == Frame::BASIC)
	Tcl_AppendResult(interp, (char*)currentFits->getRootFileName(), NULL);
      else
	Tcl_AppendResult(interp, (char*)currentFits->getFullFileName(), NULL);
    }
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      if (type == Frame::BASIC)
	Tcl_AppendElement(interp, (char*)ptr->getRootFileName());
      else
	Tcl_AppendElement(interp, (char*)ptr->getFullFileName());

      ptr = ptr->next();
    }
    break;
  }
}

void Frame::getFitsObjectNameCmd(int which)
{
  if (which > 0) {
    FitsImage* ptr = fits;
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }

    if (ptr)
      Tcl_AppendResult(interp, ptr->getObjectName(), NULL);
    else {
      char buf[4];
      ostrstream str(buf,4);
      str << which << ends;

      Tcl_AppendResult(interp, " file ", buf, " not loaded.", NULL);
      result = TCL_ERROR;
    }
  }
  else {
    Tcl_AppendResult(interp, " file index must be greater than zero.", NULL);
    result = TCL_ERROR;
  }
}

void Frame::getFitsObjectNameCmd(const Vector& v, CoordSystem ref)
{
  switch (ref) {
  case CANVAS:
    if (FitsImage* ptr = isInFits(v, &FitsImage::getCanvasToData, NULL)) {
      const char* n = ptr->getObjectName();
      if (n)
	Tcl_AppendResult(interp, n, NULL);
    }
    break;
  case PANNER:
    if (FitsImage* ptr = isInFits(v, &FitsImage::getPannerToData, NULL)) {
      const char* n = ptr->getObjectName();
      if (n)
	Tcl_AppendResult(interp, n, NULL);
    }
    break;
  }
}

void Frame::getFitsMasterCmd()
{
  if (currentFits) {
    char buf[64];
    ostrstream str(buf,64);
    str << currentFits->getFullFileName() << ' ' 
	<< currentFits->getFitsFile() << ends;

    Tcl_AppendResult(interp, buf, NULL);
  }
}

void Frame::getFitsSizeCmd(CoordSystem sys, CoordFormat format)
{
  if (currentFits)
    printVectorTcl(currentFits->mapLenFromRef(
      currentFits->getSize(), sys, format), FIXED);
  else
    printVectorTcl(Vector(), DEFAULT);
}

void Frame::getFitsDepthCmd()
{
  if (fits)
    printIntegerTcl(fits->depth());
  else
    printIntegerTcl(1);
}

void Frame::getFitsSliceCmd()
{
  if (fits)
    printIntegerTcl(slice);
  else
    printIntegerTcl(1);
}

void Frame::hasFitsCmd()
{
  if (fitsCount)
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasFitsCubeCmd()
{
  if (fits && fits->isCube())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasFitsExtCmd(int which)
{
  if (which > 0) {
    FitsImage* ptr = fits;
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }

    if (ptr && (ptr->ext() > 0)) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasFitsImageCmd()
{
  if (currentFits && currentFits->isImage())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasFitsMosaicCmd()
{
  if (currentFits && currentFits->isMosaic())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasFitsTableCmd()
{
  if (currentFits && currentFits->isTable())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::loadIncrDataCmd(int which, int x0, int y0, int x1, int y1)
{
  FitsImage* ptr = fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }
  }

  if (ptr)
    ptr->setDataParams(x0,y0,x1,y1);
}

void Frame::loadIncrMinMaxCmd(int which, int x0, int y0, int x1, int y1)
{
  FitsImage* ptr = fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }
  }

  if (ptr)
    ptr->setMinMaxParams(x0,y0,x1,y1);

  updateClip();
  initHistogram();
}

void Frame::loadIncrEndCmd()
{
  switch (scanMode) {
  case NODATASEC:
  case UNODATASEC:
  case RESET:
    scanMode = NODATASEC;
    break;
  case DATASEC:
  case UDATASEC:
    scanMode = DATASEC;
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX);
}

void Frame::sliceCmd(int s)
{
  if (fits && fits->isCube() && (s>0) && (s<=fitsCount)) {
    slice = s;
    currentFits = fits;
    for (int i=1; i<s; i++)
      currentFits = currentFits->next();
  }

  update(BASE);
}

void Frame::updateFitsCmd(int now)
{
  if (now)
    updateNow(BASE);
  else
    update(BASE);
}

void Frame::updateFitsCmd(int which, BBox bb, int now)
{
  // Note: bb is in IMAGE coords

  FitsImage* ptr = fits;
  if (which > 0) {
    for (int i=0; i<(which-1); i++) {
      if (ptr)
	ptr = ptr->next();
    }
  }

  if (ptr) {
    Vector ll = bb.ll * ptr->getImageToCanvas();
    Vector lr = bb.lr() * ptr->getImageToCanvas();
    Vector ur = bb.ur * ptr->getImageToCanvas();
    Vector ul = bb.ul() * ptr->getImageToCanvas();
    BBox rr(ll);
    rr.bound(lr);
    rr.bound(ur);
    rr.bound(ul);

    if (now)
      updateNow(BASE, rr);
    else
      update(BASE, rr);
  }
}

void Frame::unloadFitsCmd()
{
  unloadFits();

  // reset some values

  centerImage();
  wcsAlign();
  crosshair = cursor;

  update(MATRIX);
}

// Fitsy Commands

void Frame::fitsyHasExtCmd(const char* fn)
{
  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    Tcl_AppendResult(interp, "0", NULL);
    return;
  }

  FitsFile* ext = new FitsFitsMMap(fn, FitsFile::EXACT);
  if (ext->isValid())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);

  delete ext;
  return;
}

// Coordinate Commands

void Frame::getWCSFlipCmd()
{
  if (currentFits && currentFits->getWCSFlip())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getWCSSystemCmd()
{
  if (currentFits)
    Tcl_AppendResult(interp, currentFits->getWCSSystem(), NULL);
}

void Frame::hasPhysicalCmd()
{
  if (hasLTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasDetectorCmd()
{
  if (hasDTMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasAmplifierCmd()
{
  if (hasATMV())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasWCSCmd()
{
  if (hasWCS())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasWCSEquCmd()
{
  if (hasWCSEqu())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasWCSLinearCmd()
{
  if (hasWCSLinear())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Pan Zoom Commands

void Frame::centerCmd()
{
  centerImage();
  update(MATRIX);
}

void Frame::getOrientCmd()
{
  switch (orientation) {
  case NORMAL:
    Tcl_AppendResult(interp, "none", NULL);
    return;
  case XX:
    Tcl_AppendResult(interp, "x", NULL);
    return;
  case YY:
    Tcl_AppendResult(interp, "y", NULL);
    return;
  case XY:
    Tcl_AppendResult(interp, "xy", NULL);
    return;
  }
}

void Frame::getRotateCmd(Precision p)
{
  printDoubleTcl(radToDeg(rotation), p);
}

void Frame::getWCSAlignCmd()
{
  if (wcsAlignment)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getWCSZoomCmd(Precision p)
{
  if (hasWCSEqu()) {
    Vector v = currentFits->getWCScdelt();
    printDoubleTcl(v[0], p);
  }
  else
    Tcl_AppendResult(interp, "1", NULL);
}

void Frame::getZoomCmd(Precision p)
{
  printDoubleTcl(zoom, p);
}

void Frame::orientCmd(Orientation which)
{
  orientation = which;

  switch (orientation) {
  case NORMAL:
    orientationMatrix.identity();
    break;
  case XX:
    orientationMatrix = FlipX();
    break;
  case YY:
    orientationMatrix = FlipY();
    break;
  case XY:
    orientationMatrix = FlipXY();
    break;
  }

  update(MATRIX);
}

void Frame::panCmd(const Vector& v1, const Vector& v2)
{
  if (!currentFits)
    return;

  Vector start = v1 * canvasToUser;
  Vector stop = v2 * canvasToUser;
  cursor -= stop - start;
  update(MATRIX);
}

void Frame::panCmd(const Vector& v, CoordSystem sys)
{
  if (!currentFits)
    return;

  Vector u = currentFits->map(cursor, USER, sys);
  u += v;
  cursor = currentFits->map(u, sys, USER);
  currentFits->setHistCursor(cursor);
  update(MATRIX);
}

void Frame::panBBoxCmd(BBox bb)
{
  if (!currentFits)
    return;

  // we want to round to nearest pixel center
  Vector r = bb.center() * currentFits->getPannerToData();
  cursor = (r.floor() + Vector(.5,.5)) * currentFits->getDataToUser();
  currentFits->setHistCursor(cursor);
  update(MATRIX);
}

void Frame::panMotionBeginCmd(const Vector& v)
{
  if (!currentFits)
    return;

  // save the current cursor

  panCursor = v;

  // copy tmp pixmap

  panPM = Tk_GetPixmap(display, Tk_WindowId(tkwin),
		       options->width, options->height, depth);
  if (!panPM) {
    cerr << "Frame Internal Error: Unable to Create Pan Motion Pixmap" << endl;
    exit(1);
  }
  XCopyArea(display, pixmap, panPM, gc, 0, 0, options->width, 
	    options->height, 0,0);
}

void Frame::panMotionCmd(const Vector& v)
{
  if (!currentFits)
    return;

  if (visible) {

    // Clear

    Vector diff = ((v * canvasToWidget) - (panCursor * canvasToWidget));

    BBox hh,vv;
    if (diff[0]>0 && diff[1]>0) {
      hh = BBox(Vector(0,0), Vector(options->width, diff[1]));
      vv = BBox(Vector(0,0), Vector(diff[0], options->height));
    } else if (diff[0]>0 && diff[1]<0) {
      hh = BBox(Vector(options->width,options->height), 
		Vector(0,options->height+diff[1]));
      vv = BBox(Vector(0,0), Vector(diff[0], options->height));
    } else if (diff[0]<0 && diff[1]>0) {
      hh = BBox(Vector(0,0), Vector(options->width, diff[1]));
      vv = BBox(Vector(options->width,options->height), 
		Vector(options->width+diff[0], 0));
    } else if (diff[0]<0 && diff[1]<0) {
      hh = BBox(Vector(options->width,options->height), 
		Vector(0,options->height+diff[1]));
      vv = BBox(Vector(options->width,options->height), 
		Vector(options->width+diff[0], 0));
    }

    hh = hh * widgetToWindow;
    vv = vv * widgetToWindow;
    
    XSetForeground(display, gc, getWhiteColor());

    Vector hs = hh.size();
    XFillRectangle(display, Tk_WindowId(tkwin), gc, 
		   (int)hh.ll[0], (int)hh.ll[1], (int)hs[0], (int)hs[1]);

    Vector vs = vv.size();
    XFillRectangle(display, Tk_WindowId(tkwin), gc, 
		   (int)vv.ll[0], (int)vv.ll[1], (int)vs[0], (int)vs[1]);

    // display tmp pixmap at new location

    Vector d = ((v * canvasToWidget) - (panCursor * canvasToWidget)) * 
      widgetToWindow;
    XCopyArea(display, panPM, Tk_WindowId(tkwin), panGCXOR, 
	      0, 0, options->width, options->height, (int)d[0], (int)d[1]);
  }
}

void Frame::panMotionEndCmd(const Vector& v)
{
  if (!currentFits)
    return;

  // delete tmp pixmap

  if (panPM)
    Tk_FreePixmap(display, panPM);
  panPM = 0;

  Vector start = panCursor * canvasToUser;
  Vector stop = v * canvasToUser;
  cursor -= stop - start;
  currentFits->setHistCursor(cursor);
  update(MATRIX);
}

void Frame::panToCmd(const Vector& v, CoordSystem sys)
{
  if (!currentFits)
    return;

  Vector vv = currentFits->map(v, sys, USER);
  cursor = vv;
  currentFits->setHistCursor(cursor);
  update(MATRIX);
}

void Frame::rotateCmd(double r)
{
  rotation += r;
  update(MATRIX);
}

void Frame::rotateToCmd(double r)
{
  rotation = r;
  update(MATRIX);
}

void Frame::rotateMotionBeginCmd()
{
  // save the current rotation

  rotateRotation = rotation;

  // Create src XImage

  if (!(rotateSrcXM = XGETIMAGE(display, pixmap, 0, 0,
				options->width, options->height, 
				AllPlanes, ZPixmap))) {
    cerr << "Frame Internal Error: Unable to Create Rotate XImage" << endl;
    exit(1);
  }

  // Create dest XImage

  if (!(rotateDestXM = XGETIMAGE(display, pixmap, 0, 0,
				 options->width, options->height, 
				 AllPlanes, ZPixmap))) {
    cerr << "Frame Internal Error: Unable to Create Rotate XImage" << endl;
    exit(1);
  }

  // Create dest Pixmap

  rotatePM = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
			  options->width, options->height, depth);
  if (!rotatePM) {
    cerr << "Frame Internal Error: Unable to Create Rotate Motion Pixmap" 
	 << endl;
    exit(1);
  }
}

void Frame::rotateMotionCmd(double angle)
{
  rotation = rotateRotation + angle;
  rotateMotion();
}

void Frame::rotateMotionEndCmd()
{
  // Clean up

  if (rotateSrcXM)
    XDestroyImage(rotateSrcXM);
 
  if (rotateDestXM)
    XDestroyImage(rotateDestXM);

  if (rotatePM)
    Tk_FreePixmap(display, rotatePM);

  update(MATRIX);
}

void Frame::wcsAlignCmd(int which)
{
  wcsAlignment = which;
  wcsAlign();
  update(MATRIX);
}

void Frame::wcsZoomCmd(double a)
{
  if (hasWCSEqu()) {
    Vector b = currentFits->getWCScdelt();
    wcsZoom = fabs(b[0]/a);
  }
  else
    wcsZoom = 1;

  update(MATRIX);
}

void Frame::zoomCmd(double z)
{
  zoom *= z;
  update(MATRIX);
}

void Frame::zoomAboutCmd(double z, const Vector& v, CoordSystem sys)
{
  if (!currentFits)
    return;

  zoom *= z;
  Vector vv = currentFits->map(v, sys, USER);
  cursor = vv;
  currentFits->setHistCursor(cursor);
  update(MATRIX);
}

void Frame::zoomToCmd(double z)
{
  zoom = z;
  update(MATRIX);
}

void Frame::zoomToAboutCmd(double z, const Vector& v, CoordSystem sys)
{
  if (!currentFits)
    return;

  zoom = z;
  Vector vv = currentFits->map(v, sys, USER);
  cursor = vv;
  currentFits->setHistCursor(cursor);
  update(MATRIX);
}

void Frame::zoomToFitCmd()
{
  if (!currentFits)
    return;

  centerImage();
  zoom = calcZoom(options->width,options->height,currentFits->getSize());
  update(MATRIX);
}

// Panner/Magnifer Commands

void Frame::magnifierCmd(char* n, int w, int h)
{
  strcpy(magnifierName,n);
  magnifierWidth = w;
  magnifierHeight = h;

  if (magnifierPixmap)
    Tk_FreePixmap(display, magnifierPixmap);
  magnifierPixmap = 0;

  if (magnifierXImage)
    XDestroyImage(magnifierXImage);
  magnifierXImage = NULL;

  if (magnifierWidth > 0 && magnifierHeight > 0) {
    magnifierPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin),
			     magnifierWidth, magnifierHeight, depth);
    if (!magnifierPixmap) {
      cerr << "Frame Internal Error: Unable to Create Magnifier Pixmap" 
	   << endl;
      exit(1);
    }

    if (!magnifierXImage)
      if (!(magnifierXImage = XGETIMAGE(display, magnifierPixmap, 0, 0, 
					magnifierWidth, magnifierHeight,
					AllPlanes, ZPixmap))){
	cerr << "Frame Internal Error: Unable to Create Magnifier XImage" 
	     << endl;
	exit(1);
      }
  }

  if (!magnifierGC)
    magnifierGC = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  updateMagnifierMatrix();
}

void Frame::magnifierCmd(int s)
{
  useMagnifier = s;
}

void Frame::magnifierGraphicsCmd(int which)
{
  useMagnifierGraphics = which;
  updateMagnifier();
}

void Frame::magnifierCursorCmd(int which)
{
  useMagnifierCursor = which;
  updateMagnifier();
}

void Frame::magnifierModeCmd(MagnifierMode m)
{
  magnifierMode = m;
  updateMagnifier();
}

void Frame::magnifierZoomCmd(double z)
{
  magnifierZoom = z;
  updateMagnifierMatrix();
  updateMagnifier();
}

void Frame::pannerCmd(char* n, int w, int h)
{
  strcpy(pannerName,n);
  pannerWidth = w;
  pannerHeight = h;

  if (pannerPixmap)
    Tk_FreePixmap(display, pannerPixmap);
  pannerPixmap = 0;

  if (pannerXImage)
    XDestroyImage(pannerXImage);
  pannerXImage = NULL;

  if (pannerWidth > 0 && pannerHeight > 0) {
    if (!(pannerPixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), 
				      pannerWidth, pannerHeight, depth))) {
      cerr << "Frame Internal Error: Unable to Create Panner Pixmap" << endl;
      exit(1);
    }

    if (!(pannerXImage = XGETIMAGE(display, pannerPixmap, 0, 0,
				   pannerWidth, pannerHeight,
				   AllPlanes, ZPixmap))){
      cerr << "Frame Internal Error: Unable to Create Panner XImage" << endl;
      exit(1);
    }
  }

  update(MATRIX); // update panner matrices
}

void Frame::pannerCmd(int s)
{
  usePanner = s;
  updatePanner();
}

void Frame::updatePannerCmd()
{
  pannerPush();
}

void Frame::updateMagnifierCmd()
{
  updateMagnifier();
}

void Frame::updateMagnifierCmd(const Vector& v)
{
  updateMagnifier(v);
}

// Bin Commands

void Frame::binCmd(double b, const Vector& v, CoordSystem sys, 
		   const char* x, const char* y, const char* filter)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;

  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      currentFits->setHistX(x);
      currentFits->setHistY(y);
      currentFits->setHistFilter(filter);
      updateBin(currentFits->updateHist(currentFits->map(v, sys, USER)));
    }
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->setHistX(x);
      ptr->setHistY(y);
      ptr->setHistFilter(filter);
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binCmd(double b, const char* x, const char* y, const char* filter)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;

  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      currentFits->setHistX(x);
      currentFits->setHistY(y);
      currentFits->setHistFilter(filter);
      updateBin(currentFits->updateHistCenter());
    }
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->setHistX(x);
      ptr->setHistY(y);
      ptr->setHistFilter(filter);
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binAboutCmd(const Vector& v, CoordSystem sys)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHist(currentFits->map(v, sys, USER)));
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      //      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binColsCmd(const char* x, const char* y)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      currentFits->setHistX(x);
      currentFits->setHistY(y);
      updateBin(currentFits->updateHistCursor());
    }
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->setHistX(x);
      ptr->setHistY(y);
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX);
}

void Frame::binFactorCmd(double b)
{
  binFactor *= b;
  if (binFactor <= 0)
    binFactor = 1;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHistCursor());
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFactorAboutCmd(double b, const Vector& v, CoordSystem sys)
{
  binFactor *= b;
  if (binFactor <= 0)
    binFactor = 1;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHist(currentFits->map(v, sys, USER)));
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFactorToCmd(double b)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHistCursor());
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFactorToAboutCmd(double b, const Vector& v, CoordSystem sys)
{
  binFactor = b;
  if (binFactor <= 0)
    binFactor = 1;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHist(currentFits->map(v, sys, USER)));
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binToFitCmd()
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      double bf = 1/calcZoom(options->width,options->height,
			     currentFits->getHistDim());

      // round up to next power of 2

      if (bf < 1)
	binFactor = 1;
      else {
	int p=1;
	while (p<bf)
	  p*=2;
	binFactor = p;
      }
  
      updateBin(currentFits->updateHistCursor());
    }
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      double bf = 1/calcZoom(options->width,options->height,
			       ptr->getHistDim());

      // round up to next power of 2

      if (bf < 1)
	binFactor = 1;
      else {
	int p=1;
	while (p<bf)
	  p*=2;
	binFactor = p;
      }
      
      updateBin(ptr->updateHistCenter());

      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX); // because we have changed zoom
}

void Frame::binFunctionCmd(FitsHist::Function func)
{
  binFunction = func;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHistCursor());
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX);
}

void Frame::binBufferSizeCmd(int size)
{
  binBufferSize = size;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      updateBin(currentFits->updateHistCursor());
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX);
}

void Frame::binFilterCmd(const char* filter)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits) {
      currentFits->setHistFilter(filter);
      updateBin(currentFits->updateHistCursor());
    }
    break;
  case MOSAIC:
    FitsImage* ptr = fits;
    while (ptr) {
      ptr->setHistFilter(filter);
      updateBin(ptr->updateHistCenter());
      ptr = ptr->next();
    }
    break;
  }

  updateClip();
  initHistogram();
  update(MATRIX);
}

void Frame::getBinCursorCmd()
{
  if (currentFits)
    printVectorTcl(currentFits->getHistCursor(), DEFAULT);
  else
    printVectorTcl(Vector(), DEFAULT);
}

void Frame::getBinFactorCmd()
{
  printDoubleTcl(binFactor, DEFAULT);
}

void Frame::getBinFunctionCmd()
{
  switch (binFunction) {
  case FitsHist::AVERAGE:
    Tcl_AppendResult(interp, "average", NULL);
    return;
  case FitsHist::SUM:
    Tcl_AppendResult(interp, "sum", NULL);
    return;
  }
}

void Frame::getBinFilterCmd()
{
  if (currentFits && currentFits->isHist())
    Tcl_AppendResult(interp, currentFits->getHistFilter(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getBinColsCmd()
{
  if (currentFits && currentFits->isHist())
    Tcl_AppendResult(interp, currentFits->getHistX(), " ", 
		     currentFits->getHistY(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getBinListCmd()
{
  if (currentFits && currentFits->isHist()) {
    char* cols = currentFits->getHistList();
    Tcl_AppendResult(interp, cols, NULL);
    delete cols;
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::hasBinCmd()
{
  if (currentFits && currentFits->isHist())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::contourCreateCmd(const char* color, int w,
			     int num, float res, int al)
{
  if (contour)
    delete contour;
  contour = NULL;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      contour = new FVContour(this, currentFits, color, w, res, al,
			      contourScale(num, low, high, colorScaleType));
    break;
  case MOSAIC:
    if (fits) {
      contour = new FVContour(this, fits, color, w, res, al,
			      contourScale(num, low, high, colorScaleType));
      FitsImage* ptr = fits->next();
      while (ptr) {
	contour->append(ptr);
	ptr = ptr->next();
      }
    }
    break;
  }

  update(BASE);
}

void Frame::contourCreateCmd(const char* color, int w, 
			     int num, float res, int al, ColorScaleType s)
{
  if (contour)
    delete contour;
  contour = NULL;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      contour = new FVContour(this, currentFits, color, w, res, al,
			      contourScale(num, low, high, s));
    break;
  case MOSAIC:
    if (fits) {
      contour = new FVContour(this, fits, color, w, res, al,
			      contourScale(num, low, high, s));
      FitsImage* ptr = fits->next();
      while (ptr) {
	contour->append(ptr);
	ptr = ptr->next();
      }
    }
    break;
  }

  update(BASE);
}

void Frame::contourCreateCmd(const char* color, int w,
			     int num, float res, int al,
			     float l, float h, ColorScaleType s)
{
  if (contour)
    delete contour;
  contour = NULL;

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      contour = new FVContour(this, currentFits, color, w, res, al,
			      contourScale(num, low, high, s));
    break;
  case MOSAIC:
    if (fits) {
      contour = new FVContour(this, fits, color, w, res, al,
			      contourScale(num, low, high, s));
      FitsImage* ptr = fits->next();
      while (ptr) {
	contour->append(ptr);
	ptr = ptr->next();
      }
    }
    break;
  }

  update(BASE);
}

void Frame::contourCreateCmd(const char* color, int w, 
			     float res, int al, const char* lev)
{
  if (contour)
    delete contour;
  contour = NULL;

  int cnt = 0;
  float levels[100];
  istrstream str(lev);
  while ((cnt<100) && (str >> levels[cnt]))
    cnt++;

  cnt -= 1; // -1 for diffs in levels

  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      contour = new FVContour(this, currentFits, color, w, res, al, 
			      new ContourScale(cnt, levels));
    break;
  case MOSAIC:
    if (fits) {
      contour = new FVContour(this, fits, color, w, res, al, 
			      new ContourScale(cnt, levels));
      FitsImage* ptr = fits->next();
      while (ptr) {
	contour->append(ptr);
	ptr = ptr->next();
      }
    }
    break;
  }

  update(BASE);
}

void Frame::contourLoadCmd(const char* color, int w, const char* fn,
			   CoordSystem sys)
{
  if (!currentFits)
    return;

  ifstream str(fn);

  if (str) {
    List<Vertex> c;
 
    while (!str.eof()) {
      Vertex* vv;

      char buf[64];
      str.getline(buf,64,'\n');
      if (strlen(buf) > 0) {
	Vector v;
	istrstream sstr(buf);
	sstr >> v[0] >> v[1];
	vv = new Vertex(currentFits->map(v, sys, REF));
      }
      else
	vv = new Vertex(Vector(DBL_MAX, DBL_MAX));

      c.append(vv);
    }

    auxcontours.append(new Contour(this, color, w, c));
  }

  update(BASE);
}

void Frame::contourDeleteCmd()
{
  // main contour
  if (contour)
    delete contour;
  contour = NULL;

  // aux contours
  auxcontours.deleteAll();

  update(BASE);
}

void Frame::contourCopyCmd(CoordSystem sys)
{
  if (currentFits && contour) {
    List<Vertex>* v = new List<Vertex>(contour->contours());
    if (v->head())
      do {
	Vector& w = v->current()->vector;
	if (w[0] != DBL_MAX)
	  w = currentFits->map(w, REF, sys);
      } while (v->next());

    char buf[32];
    ostrstream str(buf,32);
    str << hex << v << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::contourPasteCmd(const char* c, int w, void* u, CoordSystem sys)
{
  if (!currentFits)
    return;

  List<Vertex>* v = (List<Vertex>*)u;

  if (v->head())
      do {
	Vector& w = v->current()->vector;
	if (w[0] != DBL_MAX)
	  w = currentFits->map(w, sys, REF);
      } while (v->next());
  
  auxcontours.append(new Contour(this, c, w, *v));
  delete v;

  update(BASE);
}

void Frame::contourSaveCmd(const char* fn, CoordSystem sys)
{
  if (currentFits && contour) {
    ofstream str(fn);

    if (str) {
      List<Vertex>& c = (List<Vertex>&)(contour->contours());

      if (c.head())
	do {
	  Vector v = (c.current())->vector;
	  if (v[0] != DBL_MAX && v[1] != DBL_MAX)
	    str << setiosflags(ios::scientific) 
		<< setprecision(8) 
		<< currentFits->map(v, REF, sys) 
		<< endl;
	  else
	    str << endl;
	}
	while (c.next());
    }
  }
}

void Frame::contourSetColorCmd(const char* clr)
{
  if (contour) {
    contour->setColor(clr);
    update(BASE);
  }
}

void Frame::contourSetLineWidthCmd(int w)
{
  if (contour) {
    contour->setLineWidth(w);
    update(BASE);
  }
}

void Frame::getContourLevelCmd()
{
  if (contour) {
    ContourScale* scale = contour->getScale();
    if (scale) {
      char buf[2048];
      ostrstream str(buf,2048);
      str << *scale << ends;

      Tcl_AppendResult(interp, buf, NULL);
    }
  }
}

void Frame::getContourLevelCmd(int count, float ll, float hh,
			       ColorScaleType scaleType)
{
  ContourScale* scale;
  switch (scaleType) {
  case Frame::LINEARSCALE:
  case Frame::IISSCALE:
    scale = new LinearContourScale(count-1, ll, hh);
    break;
  case Frame::LOGSCALE:
    scale = new LogContourScale(count-1, ll, hh);
    break;
  case Frame::SQUAREDSCALE:
    scale = new SquaredContourScale(count-1, ll, hh);
    break;
  case Frame::SQRTSCALE:
    scale = new SqrtContourScale(count-1, ll, hh);
    break;
  case Frame::HISTEQUSCALE:
    updateHistEqu();
    scale = new HistEquContourScale(count-1, min, max, histequ, HISTEQUSIZE);
    break;
  }
 
  char buf[2048];
  ostrstream str(buf,2048);
  str << *scale << ends;
  Tcl_AppendResult(interp, buf, NULL);
 
  delete scale;
}

void Frame::getContourMethodCmd()
{
  if (contour)
    Tcl_AppendResult(interp, contour->getMethodName(), NULL);
}

void Frame::getContourColorNameCmd()
{
  if (contour)
    Tcl_AppendResult(interp, contour->getColorName(), NULL);
}

void Frame::getContourLineWidthCmd()
{
  if (contour) {
    char buf[8];
    ostrstream str(buf,8);
    str << contour->getLineWidth() << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }
}

void Frame::hasContourCmd()
{
  if (contour)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// IIS Commands

void Frame::iisCmd(int width, int height)
{
  unloadFits();

  // force clip mode to user between 1 to 200

  clipMode = USERCLIP;
  mmMode = SAMPLE;
  mmIncr = 1;
  uLow = IISMIN;
  uHigh = IISMAX;
  
  colorScaleType = IISSCALE;
  updateColorScale();

  FitsImage* img = new FitsImageIIS(this, width, height);

  load(ALLOC,"",img);
}

void Frame::iisEraseCmd()
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisErase();
}

void Frame::iisMessageCmd(const char* ptr)
{
  if (!currentFits)
    return;

  // note: the file name is the second value passed
  char fn[64];
  istrstream str(ptr);
  str >> fn >> fn;
  currentFits->setFileName(fn);
}

void Frame::iisGetCmd(char* ptr, int x, int y, int dx, int dy)
{
  if (currentFits)
    memcpy(ptr, ((FitsImageIIS*)currentFits)->iisGet(x,y), dx*dy);
}

void Frame::iisSetCmd(const char* data, int x, int y, int dx, int dy)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisSet(x, y, data, dx*dy);
}

void Frame::iisWCSCmd(const Matrix& m, const Vector& z, int zt)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisWCS(m, z, zt);
}

void Frame::iisSetCursorCmd(const Vector& v, CoordSystem sys)
{
  if (!currentFits)
    return;

  iisLastCursor = currentFits->map(v,sys,WIDGET);

  Vector r = iisLastCursor * widgetToWindow;
  XWARPPOINTER(display, None, Tk_WindowId(tkwin), 0,0,0,0,(int)r[0],(int)r[1]);
}

void Frame::iisGetCursorCmd()
{
  printVectorTcl(iisLastCursor,DEFAULT);
}

void Frame::iisCursorModeCmd(int state)
{
  if (state) {
    // save cursor pos

    Window root, child;
    int rootx,rooty,winx,winy;
    unsigned int mask;
    XQueryPointer(display,Tk_WindowId(tkwin),&root,&child,
		  &rootx,&rooty,&winx,&winy,&mask);

    if (!iisInteractive) {
      iisSaveRoot = root;
      iisSaveCursor = Vector(rootx,rooty);
    }
    iisInteractive++;

    // set iisLastCursor if first time thru
    if (iisLastCursor[0] == 0 && iisLastCursor[1] == 0)
      iisLastCursor = Vector(options->width,options->height)/2;

    // and move to last cursor position
    Vector r = iisLastCursor * widgetToWindow;
    XWARPPOINTER(display, None,Tk_WindowId(tkwin),0,0,0,0,(int)r[0],(int)r[1]);
  }
  else {
    // restore cursor pos

    if (!(--iisInteractive)) {
      XWARPPOINTER (display, None, iisSaveRoot, 0, 0, 0, 0, 
		    (int)iisSaveCursor[0], (int)iisSaveCursor[1]);
      iisSaveRoot = 0;
      iisSaveCursor = Vector();
    }
  }
}

void Frame::hasIISCmd()
{
  if (isIIS())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

// Grid Commands

void Frame::gridCmd(CoordSystem sys, int type, const char* ops)
{
  if (grid)
    delete grid;

  grid = new Grid(this, sys, type, ops);

  update(BASE);
}

void Frame::gridDeleteCmd()
{
  if (grid)
    delete grid;
  grid = NULL;

  update(BASE);
}

void Frame::hasGridCmd()
{
  if (grid)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getGridTypeCmd()
{
  if (grid)
    printIntegerTcl(grid->type());
}

void Frame::getGridSystemCmd()
{
  if (grid) {
    switch(grid->system()) {
    case EQUATORIAL_FK4:
      Tcl_AppendResult(interp, "fk4", NULL);
      return;
    case EQUATORIAL_FK5:
      Tcl_AppendResult(interp, "fk5", NULL);
      return;
    case ICRS:
      Tcl_AppendResult(interp, "icrs", NULL);
      return;
    case GALACTIC:
      Tcl_AppendResult(interp, "galactic", NULL);
      return;
    case ECLIPTIC:
      Tcl_AppendResult(interp, "ecliptic", NULL);
      return;
    case LINEAR:
      Tcl_AppendResult(interp, "linear", NULL);
      return;
    case DETECTOR:
      Tcl_AppendResult(interp, "detector", NULL);
      return;
    case AMPLIFIER:
      Tcl_AppendResult(interp, "amplifier", NULL);
      return;
    case PHYSICAL:
      Tcl_AppendResult(interp, "physical", NULL);
      return;
    case IMAGE:
      Tcl_AppendResult(interp, "image", NULL);
      return;
    }
  }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::getGridOptionCmd()
{
  if (grid)
    Tcl_AppendResult(interp, grid->option(), NULL);
  else
    Tcl_AppendResult(interp, "", NULL);
}

// Pixel Table commands

void Frame::getPixelTableCmd(const Vector& v, CoordSystem ref,
			     int size, char* var)
{
  switch (ref) {
  case CANVAS:
    getPixelTable(v, &FitsImage::getCanvasToData, size, var);
    break;
  case PANNER:
    getPixelTable(v, &FitsImage::getPannerToData, size, var);
    break;
  }
}

void Frame::getPixelTable(const Vector& v, 
			  Matrix& (FitsImage::*getMatrix)(),
			  int size, char* var)
{
  Vector r;
  if (FitsImage* ptr = isInFits(v, getMatrix, &r)) {
    getPixelTable(ptr, r, size, var);
    return;
  }

  // else return blank

  for (int i=0; i<=size; i++)
    for (int j=0; j<=size; j++) {
      char index[8];
      ostrstream str(index,8);
      str << i << j << ends;

      Tcl_SetVar2(interp,var,index,"",0);
    }
}

void Frame::getPixelTable(FitsImage* ptr, const Vector& img, 
			  int size, char* var)
{
  Vector half((size-1)/2,(size-1)/2);
  Vector ur = img+half;
  Vector ll = img-half;
  int i,j;

  int* params = ptr->getDataParams(scanMode);
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  // x (columns)
  for (i=0,j=0; i<size; i++) {
    char index[8];
    ostrstream str(index,8);
    str << i+1 << j << ends;

    if (ur[0]>=xmin && ll[0]<xmax && ur[1]>=ymin && ll[1]<ymax) {
      Vector pt = ((ll+Vector(i,j)) * Translate(.5,.5)).round();
      if (pt[0]>xmin && pt[0]<=xmax) {
	char lbuf[8];
	ostrstream lstr(lbuf,8);
	lstr << pt[0] << ends;
	Tcl_SetVar2(interp,var,index,lbuf,0);
      }
      else
	Tcl_SetVar2(interp,var,index,"",0);
    }
    else
      Tcl_SetVar2(interp,var,index,"",0);
  }

  // y (rows)
  for (i=0,j=0; j<size; j++) {
    char index[8];
    ostrstream str(index,8);
    str << i << j+1 << ends;

    if (ur[0]>=xmin && ll[0]<xmax && ur[1]>=ymin && ll[1]<ymax) {
      Vector pt = ((ll+Vector(i,j)) * Translate(.5,.5)).round();
      if (pt[1]>ymin && pt[1]<=ymax) {
	char lbuf[8];
	ostrstream lstr(lbuf,8);
	lstr << pt[1] << ends;
	Tcl_SetVar2(interp,var,index,lbuf,0);
      }
      else
	Tcl_SetVar2(interp,var,index,"",0);
    }
    else
      Tcl_SetVar2(interp,var,index,"",0);
  }

  // body
  for (i=0; i<size; i++)
    for (j=0; j<size; j++) {
      Vector pt = ll+Vector(i,j);

      char index[8];
      ostrstream str(index,8);
      str << i+1 << j+1 << ends;

      if (pt[0]>=xmin && pt[0]<xmax && pt[1]>=ymin && pt[1]<ymax)
	Tcl_SetVar2(interp,var,index,(char*)ptr->getValue(pt),0);
      else
	Tcl_SetVar2(interp,var,index,"",0);
    }
}

// Data Values

void Frame::getDataValuesCmd(int which, const Vector& v, CoordSystem sys, 
			     const Vector& dd, char* var)
{
  // clear an previous values
  Tcl_UnsetVar(interp,var,0);

  // find number of bbox in image pixels to return
  Vector d = dd;

  // find anchor point

  FitsImage* base = findFits(which);
  if (!base) {
    Tcl_SetVar(interp,var,"",0);
    result = TCL_ERROR;
    return;
  }
  Vector ll = base->map(v,sys,REF);

  for (int i=0; i<d[0]; i++)
    for (int j=0; j<d[1]; j++) {
      Vector rr = ll+Vector(i,j);

      // index is in terms of native coords
      Vector in = base->map(rr,REF,sys);
      char index[16];
      ostrstream str(index,16);
      str << in[0] << ',' << in[1] << ends;

      int found = 0;
      FitsImage* ptr = fits;
      while (ptr) {
	Vector dd = rr * ptr->getRefToData();

	int* params = ptr->getDataParams(scanMode);
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (dd[0]>=xmin && dd[0]<xmax && dd[1]>=ymin && dd[1]<ymax) {
	  Tcl_SetVar2(interp,var,index,(char*)ptr->getValue(dd),0);
	  found = 1;
	  break;
	}
	ptr = ptr->next();
      }

      if (!found)
	Tcl_SetVar2(interp,var,index,"",0);
    }
}

// WCS commands

void Frame::wcsResetCmd()
{
  if (currentFits)
    currentFits->resetWCS();
}

void Frame::wcsReplaceCmd(int fd)
{
  if (!currentFits)
    return;

  ifstream str(fd);
  if (!str) {
    Tcl_AppendResult(interp, " unable to read wcs infomation", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->replaceWCS(r);
  delete r;
}

void Frame::wcsReplaceCmd(const char* fn)
{
  if (!currentFits)
    return;

  ifstream str(fn);
  if (!str) {
    Tcl_AppendResult(interp, " unable to load wcs file ", fn, NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->replaceWCS(r);
  delete r;
}

void Frame::wcsAppendCmd(int fd)
{
  if (!currentFits)
    return;

  ifstream str(fd);
  if (!str) {
    Tcl_AppendResult(interp, " unable to read wcs infomation", NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->appendWCS(r);
  delete r;
}

void Frame::wcsAppendCmd(const char* fn)
{
  if (!currentFits)
    return;

  ifstream str(fn);
  if (!str) {
    Tcl_AppendResult(interp, " unable to load wcs file ", fn, NULL);
    result = TCL_ERROR;
    return;
  }

  FitsHead* r = parseWCS(str);
  currentFits->appendWCS(r);
  delete r;
}

FitsHead* Frame::parseWCS(istream& str)
{
  if (!currentFits)
    return NULL;

  FitsHead* head = currentFits->getHead();
  FitsHead* r = new FitsHead(head->naxis(1),head->naxis(2),head->bitpix());

  while (!str.eof()) {
    char buf[256];
    str.getline(buf,256,'\n');

    if (strlen(buf) > 0) {
      istrstream sstr(buf);

      char keyword[256];
      sstr >> keyword;

      if (strchr(buf,'=')) {
	char dummy;
	sstr >> dummy;
      }

      if (strchr(buf,'\'')) {
	char string[32];
	sstr >> string;

	// remove both '
	string[strlen(string)-1] = '\0';
	r->appendString(keyword, string+1, "");
      }
      else {
	double val;
	sstr >> val;
	r->appendReal(keyword, val, 9, "");
      }
    }
    else
      break;
  }

  return r;
}

// utility member functions for cmds

void Frame::printIntegerTcl(int i)
{
  char buf[16];
  ostrstream str(buf,16);
  str << i << ends;
  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::printDoubleTcl(double d, Precision p)
{
  char buf[64];
  ostrstream str(buf,64);

  switch (p) {
  case DEFAULT:
    str << d << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) << d << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << d  << ends;
    break;
  case INTEGER:
    str << (int)d << ends;
    break;
  }

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::printVectorTcl(const Vector& v, Precision p)
{
  char buf[64];
  ostrstream str(buf,64);

  switch (p) {
  case DEFAULT:
    str << v << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) << v << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << v << ends;
    break;
  case INTEGER:
    {
      Vector z = v;
      str << z.round() << ends;
      break;
    }
  }

  Tcl_AppendResult(interp, buf, NULL);
}

void Frame::printCoordTcl(FitsImage* ptr, const Vector& v, CoordSystem in, 
			  CoordSystem out, CoordFormat format, Precision p)
{
  switch (out) {
  case CANVAS:
  case PANNER:
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    printVectorTcl(ptr->map(v, in, out), p);
    return;
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ICRS:
  case GALACTIC:
  case ECLIPTIC:
  case LINEAR: 
    {
      char buf[64];
      buf[0] = '\0';
      ptr->mapToWCSStr(v, in, out, format, buf, 64);

      // grap only the first two items
      char* ptr = buf;
      while (*ptr)
	ptr++;
      while (*ptr != ' ' && ptr >= buf)
	ptr--;
      *ptr = '\0';

      Tcl_AppendResult(interp, buf, " ", NULL);
    }
    return;
  }
}

void Frame::coordToTclArray(FitsImage* ptr, const Vector& v, 
			    CoordSystem in, CoordSystem out, 
			    Precision p, const char* var, const char* index)
{
  Vector r = ptr->map(v, in, out);

  doubleToTclArray(r[0], p, var, index, "x");
  doubleToTclArray(r[1], p, var, index, "y");
}

void Frame::doubleToTclArray(double d, Precision p, const char* var, 
			     const char* base, const char* mod)
{
  char index[64];
  strcpy(index,mod);
  strcat(index,base);

  char buf[64];
  ostrstream str(buf,64);

  switch (p) {
  case DEFAULT:
    str << d << ends;
    break;
  case FIXED:
    str << setiosflags(ios::fixed) << setw(9) << setprecision(3) 
	<< d << ends;
    break;
  case SCIENTIFIC:
    str << setiosflags(ios::scientific) << setprecision(8) << d << ends;
    break;
  case INTEGER:
    str << (int)d << ends;
    break;
  }

  Tcl_SetVar2(interp, (char*)var, index, buf, 0);
}

