#include <stdlib.h>
#include <string.h>

#include <iostream.h>
#include <strstream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "mmap.h"
#include "contour.h"

#include "marker.h"
#include "basemarker.h"
#include "basecircle.h"
#include "annulus.h"
#include "panda.h"
#include "circle.h"
#include "ellipse.h"
#include "ellipseannulus.h"
#include "box.h"
#include "boxannulus.h"
#include "polygon.h"
#include "line.h"
#include "text.h"
#include "ruler.h"
#include "projection.h"
#include "point.h"

#define LISTBUFSIZE 8192

// NOTE: all marker traversal routines use a local ptr as opposed to the
// list current() because marker call backs may invoke another traversal 
// routine or the event loop, which may process pending events

// DS9/Funtools Marker Parser Stuff
extern Frame* fr;              // used by mkerror,ciaoerror,otherror

#undef yyFlexLexer
#define yyFlexLexer mkFlexLexer
#include <FlexLexer.h>

mkFlexLexer* mklexx = NULL;    // used by mklex
extern int mkparse(void*);

int mklex()
{
  return (mklexx ? mklexx->yylex() : 0);
}

void mkerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = mklexx ? mklexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// CIAO Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer ciaoFlexLexer
#include <FlexLexer.h>

ciaoFlexLexer* ciaolexx = NULL;    // used by ciaolex
extern int ciaoparse(void*);

int ciaolex()
{
  return (ciaolexx ? ciaolexx->yylex() : 0);
}

void ciaoerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = ciaolexx ? ciaolexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// Pros Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer prosFlexLexer
#include <FlexLexer.h>

prosFlexLexer* proslexx = NULL;    // used by proslex
extern int prosparse(void*);

int proslex()
{
  return (proslexx ? proslexx->yylex() : 0);
}

void proserror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = proslexx ? proslexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// SAOtng Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer tngFlexLexer
#include <FlexLexer.h>

tngFlexLexer* tnglexx = NULL;    // used by tnglex
extern int tngparse(void*);

int tnglex()
{
  return (tnglexx ? tnglexx->yylex() : 0);
}

void tngerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = tnglexx ? tnglexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// SAOimage Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer saoFlexLexer
#include <FlexLexer.h>

saoFlexLexer* saolexx = NULL;    // used by saolex
extern int saoparse(void*);

int saolex()
{
  return (saolexx ? saolexx->yylex() : 0);
}

void saoerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = saolexx ? saolexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// XY Marker Parser Stuff

#undef yyFlexLexer
#define yyFlexLexer xyFlexLexer
#include <FlexLexer.h>

xyFlexLexer* xylexx = NULL;    // used by xylex
extern int xyparse(void*);

int xylex()
{
  return (xylexx ? xylexx->yylex() : 0);
}

void xyerror(const char* m)
{
  if (fr) {
    fr->error(m);
    const char* cmd = xylexx ? xylexx->YYText() : (const char*)NULL;
    if (cmd && cmd[0] != '\n') {
      fr->error(": ");
      fr->error(cmd);
    }
  }
}

// Marker Commands

void Frame::createCircleCmd(const Vector& center, double radius,
			    const char* color, int width, const char* font,
			    const char* text, unsigned short prop, 
			    const char* comment)
{
  createMarker(new Circle(this, center, radius, 
			  color, width, font, text, prop, comment));
}

void Frame::createAnnulusCmd(const Vector& center,
			     double start, double stop, int num,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)
{
  createMarker(new Annulus(this, center, start, stop, num, 
			   color, width, font, text, prop, comment));
}

void Frame::createAnnulusCmd(const Vector& center,
			     double* radii, int num,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)
{
  createMarker(new Annulus(this, center, radii, num, 
			   color, width, font, text, prop, comment));
}

void Frame::createBoxCmd(const Vector& center, const Vector& size, 
			 double angle,
			 const char* color, int width, const char* font,
			 const char* text, unsigned short prop,
			 const char* comment)
{
  createMarker(new Box(this, center, size, angle, 
		       color, width, font, text, prop, comment));
}

void Frame::createBoxAnnulusCmd(const Vector& center,
				const Vector& inner, const Vector& outer,
				int num,
				double angle,
				const char* color, int width,
				const char* font, 
				const char* text, unsigned short prop,
				const char* comment)
{
  createMarker(new BoxAnnulus(this, center, inner, outer, num, angle, 
			      color, width, font, text, prop, comment));
}

void Frame::createBoxAnnulusCmd(const Vector& center, double angle, int num,
			       Vector* size,
			       const char* color, int width, const char* font, 
			       const char* text, unsigned short prop,
				const char* comment)
{
  createMarker(new BoxAnnulus(this, center, angle, num, size, 
			      color, width, font, text, prop, comment));
}

void Frame::createEllipseCmd(const Vector& center, const Vector& radius,
			     double angle,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)

{
  createMarker(new Ellipse(this,center, radius, angle, 
			   color, width, font, text, prop, comment));
}

void Frame::createEllipseAnnulusCmd(const Vector& center,
				    const Vector& inner, const Vector& outer,
				    int num, double angle,
				    const char* color, int width,
				    const char* font, 
				    const char* text, unsigned short prop,
				    const char* comment)
{
  createMarker(new EllipseAnnulus(this, center, inner, outer, num, angle, 
				 color, width, font, text, prop, comment));
}

void Frame::createEllipseAnnulusCmd(const Vector& center, double angle,
				    int num, Vector* radii,
				    const char* color, int width,
				    const char* font, 
				    const char* text, unsigned short prop,
				    const char* comment)
{
  createMarker(new EllipseAnnulus(this, center, angle, num, radii, 
				 color, width, font, text, prop, comment));
}

void Frame::createPandaCmd(const Vector& center,
			     double ang1, double ang2, int an,
			     double rad1, double rad2, int rn,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)
{
  createMarker(new Panda(this, center, ang1, ang2, an, rad1, rad2, rn, 
			 color, width, font, text, prop, comment));
}

void Frame::createPandaCmd(const Vector& center,
			   double* a, int an,
			   double* r, int rn,
			   const char* color, int width, const char* font,
			   const char* text, unsigned short prop,
			   const char* comment)
{
  createMarker(new Panda(this, center, a, an, r, rn,
			 color, width, font, text, prop, comment));
}

void Frame::createPolygonCmd(const Vector& center, const Vector& bb,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)
{
  createMarker(new Polygon(this, center, bb, color, width, font, 
			   text, prop, comment));
}

void Frame::createPolygonCmd(const List<Vertex>& list,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)
{
  createMarker(new Polygon(this, list, color, width, font, 
			   text, prop, comment));
}

void Frame::createContourPolygonCmd(const char* color, int width, 
				    const char* font,
				    const char* text, unsigned short prop,
				    const char* comment)

{
  if (contour) {
    List<Vertex>& c = (List<Vertex>&)(contour->contours());

    // build from back to front
    if (c.tail()) {
      List<Vertex> list;
      while (c.previous()) {
	if (c.current()->vector[0] != DBL_MAX)
	  list.append(new Vertex(c.current()->vector));
	else {
	  createMarker(new Polygon(this, list, 
				   color, width, font, text, prop, NULL));
	  list.deleteAll();
	}
      }
      createMarker(new Polygon(this, list, color, width, font, 
			       text, prop, comment));
    }
    delete contour;
    contour = NULL;
  }
}

void Frame::createLineCmd(const Vector& center, const Vector& p2, 
			  int arrow1, int arrow2,
			  const char* color, int width, const char* font,
			  const char* text, unsigned short prop,
			  const char* comment)
{
  createMarker(new Line(this, center, p2, arrow1, arrow2,
		       color, width, font, text, prop, comment));
}

void Frame::createTextCmd(const Vector& center, double angle,
			  const char* color, int width, const char* font,
			  const char* text, unsigned short prop,
			  const char* comment)
{
  createMarker(new Text(this, center, angle, 
			color, width, font, text, prop, comment));
}

void Frame::createRulerCmd(const Vector& center, const Vector& p2,
			   CoordSystem coordsys, 
			   CoordSystem distsys, CoordFormat distformat,
			   const char* color, int width, const char* font,
			   const char* text, unsigned short prop,
			   const char* comment)
{
  createMarker(new Ruler(this, center, p2, 
			coordsys, distsys, distformat, 
			color, width, font, text, prop, comment));
}

void Frame::createProjectionCmd(const Vector& center, const Vector& p2, 
				double w, const char* mvcb, const char* delcb,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment)
{
  createMarker(new Projection(this, center, p2, w, mvcb, delcb,
			      color, width, font, text, prop, comment));
}

void Frame::createCirclePointCmd(const Vector& center,
				 const char* color, int width,
				 const char* font,
				 const char* text, unsigned short prop,
				 const char* comment)
{
  createMarker(new CirclePoint(this, center, color, width, font, 
			       text, prop, comment));
}

void Frame::createBoxPointCmd(const Vector& center,
			      const char* color, int width, const char* font,
			      const char* text, unsigned short prop,
			      const char* comment)
{
  createMarker(new BoxPoint(this, center, color, width, font, 
			    text, prop, comment));
}

void Frame::createDiamondPointCmd(const Vector& center,
				  const char* color, int width,
				  const char* font,
				  const char* text, unsigned short prop,
				  const char* comment)
{
  createMarker(new DiamondPoint(this, center, color, width, font, 
				text, prop, comment));
}

void Frame::createCrossPointCmd(const Vector& center,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment)
{
  createMarker(new CrossPoint(this, center, color, width, font, 
			      text, prop, comment));
}

void Frame::createExPointCmd(const Vector& center,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment)
{
  createMarker(new ExPoint(this, center, color, width, font, 
			   text, prop, comment));
}

void Frame::createArrowPointCmd(const Vector& center,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment)
{
  createMarker(new ArrowPoint(this, center, color, width, font, 
			      text, prop, comment));
}

void Frame::createBoxCirclePointCmd(const Vector& center,
				    const char* color, int width,
				    const char* font,
				    const char* text, unsigned short prop,
				    const char* comment)
{
  createMarker(new BoxCirclePoint(this, center, 
				  color, width, font, text, prop, comment));
}

void Frame::createMarker(Marker* m)
{
  markers.append(m);

  // now update new marker

  update(PIXMAP, m->getAllBBox());

  // and return id

  printIntegerTcl(m->getId());
}

void Frame::getMarkerAngleCmd(int id, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(radToDeg(mapAngleFromRef(m->getAngle(), sys)),
		     DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerAnnulusRadiusCmd(int id, CoordSystem sys, 
				      CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const double* r = ((Annulus*)m)->getRadii();
      int cnt = ((Annulus*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	markerPrintDoubleTcl(m->getCenter(), r[i], sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerBoxAnnulusSizeCmd(int id, CoordSystem sys,
				       CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const Vector* r = ((BoxAnnulus*)m)->getSize();
      int cnt = ((BoxAnnulus*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	markerPrintVectorTcl(m->getCenter(), r[i], sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerBoxSizeCmd(int id, CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintVectorTcl(m->getCenter(), ((Box*)m)->getSize(), sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerCenterCmd(int id, CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoordTcl(m->getCenter(), m->getCenter(), sys, format);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerCircleRadiusCmd(int id, CoordSystem sys, 
				     CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDoubleTcl(m->getCenter(), ((Circle*)m)->getRadius(),
			   sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerColorCmd()
{
  // return first found

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerColorCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseRadiusCmd(int id, CoordSystem sys, 
				      CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintVectorTcl(m->getCenter(), ((Ellipse*)m)->getRadius(), 
			   sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerEllipseAnnulusRadiusCmd(int id, CoordSystem sys, 
					     CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const Vector* r = ((EllipseAnnulus*)m)->getRadii();
      int cnt = ((EllipseAnnulus*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	markerPrintVectorTcl(m->getCenter(), r[i], sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerFontCmd()
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerFontCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerHandleCmd(const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      int h = m->onHandle(v);
      if (h) {
	char buf [16];
	ostrstream str(buf,16);
	str << m->getId() << ' ' << h << ends;
	Tcl_AppendResult(interp, buf, NULL);
	return;
      }
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0 0", NULL);
}

void Frame::getMarkerIdCmd(const Vector& v)
{
  // v is in canvas coords

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v)) {
      printIntegerTcl(m->getId());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerIdAllCmd()
{
  char buf[8];
  Marker* m=markers.head();
  while (m) {
    ostrstream str(buf,8);
    str << m->getId() << ' ' << ends;
    Tcl_AppendResult(interp, buf, NULL);
    m=m->next();
  }
}

void Frame::getMarkerLineCmd(int id, CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoordTcl(m->getCenter(), ((Line*)m)->getP1(), sys, format);
      markerPrintCoordTcl(m->getCenter(), ((Line*)m)->getP2(), sys, format);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerLineArrowCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (((Line*)m)->getP1Arrow())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);

      if (((Line*)m)->getP2Arrow())
	Tcl_AppendResult(interp, " 1", NULL);
      else
	Tcl_AppendResult(interp, " 0", NULL);

      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineDistanceCmd(int id, CoordSystem sys, 
				     CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDoubleTcl(m->getCenter(), 
			   ((Line*)m)->getP2(), ((Line*)m)->getP1(),
			   sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineWidthCmd()
{
  // return first found

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      char buf[8];
      ostrstream str(buf,8);
      str << m->getLineWidth() << ends;
      Tcl_AppendResult(interp,  buf, NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerLineWidthCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      char buf[8];
      ostrstream str(buf,8);
      str << m->getLineWidth() << ends;
      Tcl_AppendResult(interp,  buf, NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPandaAnglesCmd(int id, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const double* a = ((Panda*)m)->getAngles();
      int cnt = ((Panda*)m)->getAnglesCount();
      for (int i=0; i<cnt; i++) {
	// use panda version of radToDeg
	printDoubleTcl(((Panda*)m)->radToDeg(mapAngleFromRef(a[i],sys)), 
		       DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPandaRadiusCmd(int id, CoordSystem sys, 
				    CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      const double* r = ((Panda*)m)->getRadii();
      int cnt = ((Panda*)m)->getAnnuliCount();
      for (int i=0; i<cnt; i++) {
	markerPrintDoubleTcl(m->getCenter(), r[i], sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPolygonSegmentCmd(const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->isPolygon()) {
      int s = ((Polygon*)(m))->getSegment(v);
      if (s) {
	char buf [16];
	ostrstream str(buf,16);
	str << m->getId() << ' ' << s << ends;
	Tcl_AppendResult(interp, buf, NULL);
	return;
      }
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0 0", NULL);
}

void Frame::getMarkerProjectionCmd(int id, char* xName, char* yName)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Projection* p = (Projection*)m;
      bltProjection(xName, yName, p->getP1(), p->getP2(), 
		    p->getWidth(), p->getMethod());
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerProjectionPointsCmd(int id, CoordSystem sys, 
					 CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoordTcl(m->getCenter(), ((Projection*)m)->getP1(), 
			  sys, format);
      markerPrintCoordTcl(m->getCenter(), ((Projection*)m)->getP2(), 
			  sys, format);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerProjectionDistanceCmd(int id, CoordSystem sys,
					   CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDoubleTcl(m->getCenter(), 
			   ((Projection*)m)->getP2(),((Projection*)m)->getP1(),
			   sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerProjectionMethodCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printIntegerTcl(((Projection*)m)->getMethod());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerProjectionWidthCmd(int id, 
					CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDoubleTcl(m->getCenter(), ((Projection*)m)->getWidth(), 
			   sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerPropertyCmd(unsigned short which)
{
  // return first selected found

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL): 
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerPropertyCmd(int id, unsigned short which)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL):
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerRulerDistanceCmd(int id, 
				      CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      printDoubleTcl(radToDeg(((Ruler*)m)->getAngle()), DEFAULT);

      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDoubleTcl(m->getCenter(), 
			   ((Ruler*)m)->getP2(), ((Ruler*)m)->getP1(),
			   sys, format);
      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDoubleTcl(m->getCenter(), 
			   ((Ruler*)m)->getP3(), ((Ruler*)m)->getP1(),
			   sys, format);
      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDoubleTcl(m->getCenter(), 
			   ((Ruler*)m)->getP3(), ((Ruler*)m)->getP2(),
			   sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void Frame::getMarkerRulerPointCmd(int id, CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoordTcl(m->getCenter(), ((Ruler*)m)->getP1(), sys, format);
      markerPrintCoordTcl(m->getCenter(), ((Ruler*)m)->getP2(), sys, format);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerSelectedCmd()
{
  char buf[8];
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      ostrstream str(buf,8);
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, buf, NULL);
    }
    m=m->next();
  }
}

void Frame::getMarkerSelectedCmd(const Vector& v)
{
  char buf[8];
  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v) && m->isSelected()) {
      ostrstream str(buf,8);
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, buf, NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::getMarkerSelectedNumberCmd()
{
  int count=0;
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected())
      count++;
    m=m->next();
  }

  printIntegerTcl(count);
}

void Frame::getMarkerTextCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getText(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::getMarkerTypeCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getType(), NULL);
      return;
    }
    m=m->next();
  }
}

void Frame::hasMarkerSelectedCmd()
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasMarkerPasteCmd()
{
  if (!pasteMarkers.isEmpty())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::hasMarkerUndoCmd()
{
  if (!undoMarkers.isEmpty()) 
    switch (undoMarkerType) {
    case MOVE:
      Tcl_AppendResult(interp, "move", NULL);
      return;
    case EDIT:
      Tcl_AppendResult(interp, "edit", NULL);
      return;
    case DELETE:
      Tcl_AppendResult(interp, "delete", NULL);
      return;
    default:
      Tcl_AppendResult(interp, "", NULL);
      return;
    }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void Frame::markerAngleCmd(int id, double angle, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      m->setAngle(mapAngleToRef(angle,sys));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      printIntegerTcl(((Annulus*)m)->addRadii(v*canvasToRef));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Annulus*)m)->deleteRadii(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerAnnulusRadiusCmd(int id, double inner, double outer, int num,
				   CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      double r1 = ptr->mapLenToRef(inner, sys, format);
      double r2 = ptr->mapLenToRef(outer, sys, format);
      ((Annulus*)m)->setRadii(r1,r2,num);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerAnnulusRadiusCmd(int id, const char* lev, 
				   CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int cnt = 0;
      double radii[MAXANNULI];
      istrstream str(lev);
      while ((cnt<MAXANNULI) && (str >> radii[cnt]))
	++cnt;
	
      FitsImage* ptr = findFits(m->getCenter());
      for (int i=0; i<cnt; i++)
	radii[i] = ptr->mapLenToRef(radii[i], sys, format);

      ((Annulus*)m)->setRadii(radii,cnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBackCmd()
{
  Marker* m = markers.tail();
  while (m) {
    if (m->isSelected()) {
      Marker* prev = markers.extractPrev(m);
      markers.append(m);
      update(PIXMAP, m->getAllBBox());
      m = prev;
    }
    else
      m=m->previous();
  }
}

void Frame::markerBackCmd(int id)
{
  Marker* m = markers.tail();
  while (m) {
    if (m->getId() == id) {
      markers.extractPrev(m);
      markers.append(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m=m->previous();
  }
}

void Frame::markerBoxAnnulusSizeCmd(int id, const Vector& inner,
				    const Vector& outer, int num,
				    CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      Vector s1 = ptr->mapLenToRef(inner, sys, format);
      Vector s2 = ptr->mapLenToRef(outer, sys, format);
      ((BoxAnnulus*)(m))->setSize(s1,s2,num);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAnnulusSizeCmd(int id,const char* lev, 
				    CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int cnt = 0;
      Vector size[MAXANNULI];
      istrstream str(lev);
      while ((cnt<MAXANNULI) && (str >> size[cnt][0]))
	str >> size[cnt++][1];
	
      // verify proper ratios
      {
	for (int i=0; i<cnt; i++)
	  size[i][1] = size[i][0]*size[cnt-1][1]/size[cnt-1][0];
      }

      // map to ref coord sys
      {
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<cnt; i++)
	  size[i] = ptr->mapLenToRef(size[i], sys, format);
      }

      ((BoxAnnulus*)m)->setSize(size,cnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAnnulusCreateSizeCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      printIntegerTcl(((BoxAnnulus*)m)->addSize(v*canvasToRef));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerBoxAnnulusDeleteSizeCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((BoxAnnulus*)m)->deleteSize(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerBoxSizeCmd(int id, const Vector& size, 
			     CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      ((Box*)(m))->setSize(ptr->mapLenToRef(size, sys, format));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerCallBackCmd(int id, Marker::Callback cb, const char* p, 
			      const char* a)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      result = m->setCallBack(cb, p, a);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerCircleRadiusCmd(int id, double r, 
				  CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      double rr = ptr->mapLenToRef(r, sys, format);
      ((Circle*)m)->setRadius(rr);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerColorCmd(const char* clr)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerColorCmd(int id, const char* clr)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerCommandCmd(MarkerFormat fm, const char* ccmd)
{
  // only make command string as long as needed
  // or the rest will be processed as garbage
  int len = strlen(ccmd)+1;
  char* buf = new char[len];
  memcpy(buf, ccmd, len);

  // add terminator to make parser happy
  buf[len-1] = '\n';

  istrstream istr(buf,len);
  parseMarker(fm, istr);

  delete buf;
}
 
void Frame::markerDeleteCallBackCmd(int id, Marker::Callback cb, const char* p)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      result = m->deleteCallBack(cb, p);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerDeleteCmd()
{
  undoMarkers.deleteAll();
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected() && m->canDelete()) {
      Marker* next = markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      undoMarkers.append(m);
      m->doCallBack(Marker::DELETECB);
      m->clearCB();
      undoMarkerType = DELETE;
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerDeleteCmd(int id)
{
  undoMarkers.deleteAll();
  Marker* m = markers.head();
  while (m) {
    if (m->getId() == id) {
      markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      undoMarkers.append(m);
      m->doCallBack(Marker::DELETECB);
      m->clearCB();
      undoMarkerType = DELETE;
      return;
    }
    else
      m=m->next();
  }
}

void Frame::markerDeleteAllCmd()
{
  undoMarkers.deleteAll();
  Marker* m = markers.head();
  while (m) {
    if (m->canDelete()) {
      update(PIXMAP, m->getAllBBox());
      Marker* next = markers.extractNext(m);
      undoMarkers.append(m);
      m->doCallBack(Marker::DELETECB);
      m->clearCB();
      undoMarkerType = DELETE;
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerDeleteLastCmd()
{
  undoMarkers.deleteAll();
  Marker* m=markers.tail();
  if (m) {
    markers.extractNext(m);
    update(PIXMAP, m->getAllBBox());
    undoMarkers.append(m);
    m->doCallBack(Marker::DELETECB);
    m->clearCB();
    undoMarkerType = DELETE;
    return;
  }
}

void Frame::markerEditBeginCmd(int id, int h)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id && m->canEdit()) {
      markerUndo(m, EDIT);

      editMarker = m;
      editMarker->editBegin(h);
      return;
    }
    m=m->next();
  }

  editMarker = NULL;
}

void Frame::markerEditBeginCmd(const Vector& v, int h)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canEdit()) {
      markerUndo(m, EDIT);

      editMarker = m;
      editMarker->editBegin(h);
      return;
    }
    m=m->next();
  }

  editMarker = NULL;
}

void Frame::markerEditMotionCmd(const Vector& v, int h)
{
  if (editMarker) {

    // erase current marker now

    redraw(editMarker->getAllBBox());
    forceUpdate();

    editMarker->edit(v*canvasToRef, h);
    editMarker->drawXOR();
  }
}

void Frame::markerEditEndCmd()
{
  if (editMarker)
    editMarker->editEnd();
  editMarker = NULL;

  update(PIXMAP);
}

void Frame::markerEllipseRadiusCmd(int id, const Vector& radius, 
				   CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      ((Ellipse*)m)->setRadius(ptr->mapLenToRef(radius, sys, format));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusRadiusCmd(int id, const Vector& inner,
					  const Vector& outer,
					  int num,
					  CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      Vector r1 = ptr->mapLenToRef(inner, sys, format);
      Vector r2 = ptr->mapLenToRef(outer, sys, format);
      ((EllipseAnnulus*)(m))->setRadii(r1,r2,num);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusRadiusCmd(int id, const char* lev,
					  CoordSystem sys, CoordFormat format)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int cnt = 0;
      Vector radii[MAXANNULI];
      istrstream str(lev);
      while ((cnt<MAXANNULI) && (str >> radii[cnt][0]))
	str >> radii[cnt++][1];
	
      // verify proper ratios
      {
	for (int i=0; i<cnt; i++)
	  radii[i][1] = radii[i][0]*radii[cnt-1][1]/radii[cnt-1][0];
      }

      // map to ref coord sys
      {
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<cnt; i++)
	  radii[i] = ptr->mapLenToRef(radii[i], sys, format);
      }

      ((EllipseAnnulus*)(m))->setRadii(radii,cnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      printIntegerTcl(((EllipseAnnulus*)m)->addRadii(v*canvasToRef));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerEllipseAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((EllipseAnnulus*)m)->deleteRadii(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerFontCmd(const char* f)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerFontCmd(int id, const char* f)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerFrontCmd()
{
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected()) {
      Marker* next = markers.extractNext(m);
      markers.insertHead(m);
      update(PIXMAP, m->getAllBBox());
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerFrontCmd(int id)
{
  Marker* m = markers.head();
  while (m) {
    if (m->getId() == id) {
      markers.extractNext(m);
      markers.insertHead(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m=m->next();
  }
}

void Frame::markerLineCmd(int id, const Vector& p1, const Vector& p2, 
			  CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      ((Line*)(m))->setPoints(ptr->map(p1,sys,REF), ptr->map(p2,sys,REF));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerLineArrowCmd(int id, int p1, int p2)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      ((Line*)(m))->setArrows(p1, p2);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerLineWidthCmd(int w)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      m->setLineWidth(w);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerLineWidthCmd(int id, int w)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->setLineWidth(w);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerListCmd(MarkerFormat type, CoordSystem sys,
			  CoordFormat format, int strip, 
			  unsigned short mask, unsigned short value)
{
  char buf[LISTBUFSIZE];

  char delim;
  if (strip)
    delim = ';';
  else
    delim = '\n';

  void (Frame::*hproc)(ostream&, CoordSystem, CoordFormat, char);
  void (Marker::*mproc)(ostream&, CoordSystem, CoordFormat, char);

  switch (type) {
  case DS9:
    hproc = &Frame::markerListHeader;
    mproc = &Marker::list;
    break;
  case CIAO:
    hproc = &Frame::markerListCiaoHeader;
    mproc = &Marker::listCiao;
    break;
  case SAOTNG:
    hproc = &Frame::markerListSAOtngHeader;
    mproc = &Marker::listSAOtng;
    break;
  case SAOIMAGE:
    hproc = &Frame::markerListSAOimageHeader;
    mproc = &Marker::listSAOimage;
    break;
  case PROS:
    hproc = &Frame::markerListProsHeader;
    mproc = &Marker::listPros;
    break;
  case RAWXY:
    hproc = &Frame::markerListXYHeader;
    mproc = &Marker::listXY;
    break;
  }

  {
    ostrstream str(buf,LISTBUFSIZE);
    (this->*hproc)(str, sys, format, delim);
    str << ends;
    Tcl_AppendResult(interp, buf, NULL);
  }

  Marker* m=markers.head();
  while (m) {
    if (!mask || ((m->getProperty() & mask)  == value)) {
      ostrstream str(buf,LISTBUFSIZE);
      (m->*mproc)(str, sys, format, delim);
      str << ends;
      Tcl_AppendResult(interp, buf, NULL);
    }
    m=m->next();
  }
}

void Frame::markerListHeader(ostream& str, CoordSystem sys, 
			     CoordFormat format, char delim)
{
  // no comments for semicolons
  if (delim != ';') {
    str << "# Region file format: DS9 version 3.0" << delim;

    str << "global color=green font=\"helvetica 10 normal\" "
	<< "select=1 edit=1 move=1 delete=1 include=1 fixed=0 source" 
	<< delim;
  }
}

void Frame::markerListCiaoHeader(ostream& str, CoordSystem sys, 
			     CoordFormat format, char delim)
{
  // no comments for semicolons
  if (delim != ';')
    str << "# Region file format: CIAO version 1.0" << delim;
}

void Frame::markerListSAOtngHeader(ostream& str, CoordSystem sys, 
				   CoordFormat format, char delim)
{
  // no comments for semicolons
  if (delim == ';')
    return;

  if (currentFits)
    str << "# filename: " << currentFits->getRootFileName() << delim;

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
    str << "# format: pixels (physical)" << delim;
    break;
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ECLIPTIC:
  case GALACTIC:
  case LINEAR:
  case ICRS:
    str << "# format: ";
    switch (format) {
    case PIXELS:
    case DEGREES:
      str << "degrees (";
      break;
    case SEXAGESIMAL:
      str << "hms (";
      break;
    }

    switch (sys) {
    case EQUATORIAL_FK4:
      str << "fk4";
      break;
    case EQUATORIAL_FK5:
      str << "fk5";
      break;
    case ICRS:
      str << "icrs";
      break;
    case GALACTIC:
      str << "galactic";
      break;
    case ECLIPTIC:
      str << "ecliptic";
      break;
    case LINEAR:
      str << "linear";
      break;
    }

    str << ')' << delim;
    break;
  }
}

void Frame::markerLoadCmd(MarkerFormat fm, const char* fn)
{
  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(fm, str);
}

void Frame::markerLoadCmd(MarkerFormat fm, const char* fn, CoordSystem sys)
{
  markerDefaultSystem = sys;

  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(fm, str);
}

void Frame::markerLoadCmd(MarkerFormat fm, int fd)
{
  ifstream str(fd);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(fm, str);
}

void Frame::markerLoadCmd(MarkerFormat fm, int fd, CoordSystem sys)
{
  markerDefaultSystem = sys;

  ifstream str(fd);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  parseMarker(fm, str);
}

void Frame::markerLoadFitsCmd(const char* fn)
{
  if (!currentFits)
    return;

  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    result = TCL_ERROR;
    return;
  }

  FitsFile* mk = new FitsFitsMMap(fn, FitsFile::EXACT);
  if (!mk->isValid() || !mk->isTable()) {
    delete mk;
    result = TCL_ERROR;
    return;
  }

  FitsHead* mkh = mk->head();
  if (mkh->naxes() != 2) {
    delete mk;
    result = TCL_ERROR;
    return;
  }
  FitsTableHDU* mkhdu = (FitsTableHDU*)mkh->hdu();

  FitsColumn* shape = mkhdu->find("shape");
  FitsColumn* x = mkhdu->find("x");
  FitsColumn* y = mkhdu->find("y");
  FitsColumn* r = mkhdu->find("r");
  FitsColumn* ang = mkhdu->find("rotang");
  
  // manatory columns
  if (!x || !y) {
    delete mk;
    result = TCL_ERROR;
    return;
  }

  char* ptr = (char*)mk->data();
  int rows = mkhdu->rows();
  int rowlen = mkhdu->width();

  char color[] = "green";
  int width = 1;
  char font[] = "helvetica 10 normal";
  char text[] = "";

  for (int i=0; i<rows; i++, ptr+=rowlen) {
    char* s1;
    if (shape)
      s1 = toUpper(shape->str(ptr));
    else {
      s1 = new char[7];
      strcpy(s1,"POINT ");
    }

    // look for '!', which sets include/exclude
    char* s2 = s1;
    unsigned short props = Marker::SELECT | Marker::EDIT | Marker::MOVE |
      Marker::ROTATE | Marker::DELETE | Marker::SOURCE;
    if (s2[0]=='!')
      s2++;
    else
      props |= Marker::INCLUDE;

    if (!strncmp(s2, "CIRCLE", 6) && r)
      createCircleCmd(
	currentFits->map(Vector(x->value(ptr,0),y->value(ptr,0)),PHYSICAL,REF),
	currentFits->mapLenToRef(r->value(ptr), PHYSICAL, PIXELS), 
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "ANNULU", 6) && r)
      createAnnulusCmd(
	currentFits->map(Vector(x->value(ptr,0),y->value(ptr,0)),PHYSICAL,REF),
	currentFits->mapLenToRef(r->value(ptr,1),PHYSICAL,PIXELS),
	currentFits->mapLenToRef(r->value(ptr,0),PHYSICAL,PIXELS), 1,
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "BOX   ", 6) && r)
      createBoxCmd(
	currentFits->map(Vector(x->value(ptr,0),y->value(ptr,0)),PHYSICAL,REF),
	currentFits->mapLenToRef(Vector(r->value(ptr,0),r->value(ptr,1)),
				 PHYSICAL,PIXELS), 0,
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "ROTBOX", 6) && r && ang)
      createBoxCmd(
	currentFits->map(Vector(x->value(ptr,0), y->value(ptr,0)),
			 PHYSICAL,REF),
	currentFits->mapLenToRef(Vector(r->value(ptr,0),r->value(ptr,1)),
				 PHYSICAL,PIXELS), 
	degToRad(ang->value(ptr)),
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "RECTAN", 6)) {
      Vector v1(x->value(ptr,0), y->value(ptr,0));
      Vector v2(x->value(ptr,1), y->value(ptr,1));
      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createBoxCmd(currentFits->map(c,PHYSICAL,REF), 
	currentFits->mapLenToRef(d,PHYSICAL,PIXELS), 
	0,
	color, width, font, text, props, NULL);
    }

    else if (!strncmp(s2, "ROTREC", 6) && ang) {
      Vector v1(x->value(ptr,0), y->value(ptr,0));
      Vector v2(x->value(ptr,1), y->value(ptr,1));
      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createBoxCmd(currentFits->map(c,PHYSICAL,REF), 
	currentFits->mapLenToRef(d,PHYSICAL,PIXELS), 
	degToRad(ang->value(ptr)),
	color, width, font, text, props, NULL);
    }

    else if (!strncmp(s2, "ELLIPS", 6) && r && ang)
      createEllipseCmd(
        currentFits->map(Vector(x->value(ptr,0),y->value(ptr,0)),PHYSICAL,REF),
        currentFits->mapLenToRef(Vector(r->value(ptr,0),r->value(ptr,1)),
				 PHYSICAL,PIXELS),degToRad(ang->value(ptr)),
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "PIE   ", 6) && r && ang)
      createPandaCmd(
	currentFits->map(Vector(x->value(ptr,0),y->value(ptr,0)),PHYSICAL,REF),
	degToRad(ang->value(ptr,0)-90), degToRad(ang->value(ptr,1)-90), 1,
	currentFits->mapLenToRef(r->value(ptr,1),PHYSICAL,PIXELS),
	currentFits->mapLenToRef(r->value(ptr,0),PHYSICAL,PIXELS), 1,
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "POINT ", 6))
      createBoxCirclePointCmd(
	currentFits->map(Vector(x->value(ptr,0),y->value(ptr,0)),PHYSICAL,REF),
	color, width, font, text, props, NULL);

    else if (!strncmp(s2, "POLYGO", 6)) {
      List<Vertex> list;
      Vertex* first = new Vertex(
	currentFits->map(Vector(x->value(ptr,0), y->value(ptr,0)),
			 PHYSICAL,REF));
      list.append(first);

      // this could be a problem, if the end points are not the same

      int i=1;
      while (1) {
	Vertex* n = new Vertex(
	  currentFits->map(Vector(x->value(ptr,i), y->value(ptr,i)),
			   PHYSICAL,REF));
	if (n->vector[0] != first->vector[0] || 
	    n->vector[1] != first->vector[1])
	  list.append(n);
	else {
	  delete n;
	  break;
	}

	i++;
      }

      createPolygonCmd(list, color, width, font, text, props, NULL);
    }

    delete s1;
  }

  delete mk;
}

void Frame::markerMoveCmd(const Vector& v)
{
  undoMarkers.deleteAll();
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      undoMarkers.append(m->dup());
      undoMarkerType = MOVE;

      Vector c = m->getCenter() * refToCanvas;
      update(PIXMAP, m->getAllBBox());
      m->moveTo((c + v) * canvasToRef);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerMoveCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() && m->canMove()) {
      markerUndo(m, MOVE);

      Vector c = m->getCenter() * refToCanvas;
      update(PIXMAP, m->getAllBBox());
      m->moveTo((c + v) * canvasToRef);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerMoveBeginCmd(const Vector& v)
{
  markerBegin = v * canvasToRef;
  undoMarkers.deleteAll();
  Marker* m=markers.head();
  if (m)
    while (m) {
      if (m->isSelected()) {
	undoMarkers.append(m->dup());
	undoMarkerType = MOVE;
      }
      m=m->next();
    }
}

void Frame::markerMoveBeginCmd(int id, const Vector& v)
{
  markerBegin = v * canvasToRef;
  undoMarkers.deleteAll();
  Marker* m=markers.head();
  if (m) {
    while (m) {
      if (m->getId() == id) {
	undoMarkers.append(m->dup());
	undoMarkerType = MOVE;
      }
      m=m->next();
    }
  }
}

void Frame::markerMoveMotionCmd(const Vector& v)
{
  // first, accumulate erase markers

  Marker* m=markers.head();
  if (m) {
    while (m) {
      if (m->isSelected() && m->canMove())
	redraw(m->getAllBBox());
      m=m->next();
    }

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    m=markers.head();
    while (m) {
      if (m->isSelected() && m->canMove()) {
	m->move(diff);
	m->drawXOR();
      }
      m=m->next();
    }
  }
}

void Frame::markerMoveMotionCmd(int id, const Vector& v)
{
  // first, accumulate erase markers

  Marker* m=markers.head(); 
  if (m) {
    while (m) {
      if (m->getId() && m->canMove()) {
	redraw(m->getAllBBox());
	break;
      }
      m=m->next();
    }

    if (!m)
      return; // can't find it

    Marker *ptr = m;

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    ptr->move(diff);
    ptr->drawXOR();
  }
}

void Frame::markerMoveEndCmd()
{
  // update widget since we don't know where the selected markers came from
  update(PIXMAP);
}

void Frame::markerMoveEndCmd(int id)
{
  // update widget since we don't know where the selected markers came from
  update(PIXMAP);
}

void Frame::markerMoveToCmd(const Vector& v, CoordSystem sys)
{
  undoMarkers.deleteAll();
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      undoMarkers.append(m->dup());
      undoMarkerType = MOVE;

      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      m->moveTo(ptr->map(v,sys,REF));
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void Frame::markerMoveToCmd(int id, const Vector& v, CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if ((m->getId() == id) && m->canMove()) {
      markerUndo(m, MOVE);

      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      m->moveTo(ptr->map(v,sys,REF));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerPandaCreateAnglesCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      printIntegerTcl(((Panda*)m)->addAngles(v*canvasToRef));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPandaCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      printIntegerTcl(((Panda*)m)->addRadii(v*canvasToRef));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPandaDeleteCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Panda*)m)->deleteAnglesRadii(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerPandaEditCmd(int id, double a1, double a2, int an, 
			       double r1, double r2, int rn, 
			       CoordSystem sys, CoordFormat format)

{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());

      ((Panda*)m)->setAnglesRadii(mapAngleToRef(a1,sys),mapAngleToRef(a2,sys),
				  an,r1,r2,rn);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPandaEditCmd(int id, const char* a, const char* r,
			       CoordSystem sys,
			       CoordSystem rsys, CoordFormat rformat)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());

      int acnt = 0;
      double angles[MAXANGLES];
      istrstream astr(a);
      while ((acnt<MAXANGLES) && (astr >> angles[acnt]))
	++acnt;
	
      {
	for (int i=0; i<acnt; i++)
	  angles[i] = mapAngleToRef(degToRad(angles[i]),sys);
      }

      int rcnt = 0;
      double radii[MAXANNULI];
      istrstream rstr(r);
      while ((rcnt<MAXANNULI) && (rstr >> radii[rcnt]))
	++rcnt;
	
      {
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<rcnt; i++)
	  radii[i] = ptr->mapLenToRef(radii[i], rsys, rformat);
      }

      ((Panda*)m)->setAnglesRadii(angles,acnt,radii,rcnt);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPolygonCreateVertexCmd(int id, int seg, const Vector& v)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      ((Polygon*)(m))->createVertex(seg, v * canvasToRef);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPolygonDeleteVertexCmd(int id, int h)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Polygon*)(m))->deleteVertex(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void Frame::markerProjectionCmd(int id, const Vector& p1, const Vector& p2, 
				CoordSystem sys, double width, int method,
				CoordSystem wdsys, CoordFormat wdformat)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      ((Projection*)(m))->set(ptr->map(p1,sys,REF), ptr->map(p2,sys,REF), 
			      ptr->mapLenToRef(width, wdsys, wdformat),method);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerPropertyCmd(unsigned short prop, int value)
{
  Marker* m=markers.head();
  while (m) {
    if (m->isSelected()) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE || 
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);
    }
    m=m->next();
  }
}

void Frame::markerPropertyCmd(unsigned short prop, int value, const Vector& v)
{
  // v is in canvas coords

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v)) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE || 
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);
    }
    m=m->next();
  }
}

void Frame::markerPropertyCmd(int id, unsigned short prop, int value)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE ||
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);

      return;
    }
    m=m->next();
  }
}

void Frame::markerRotateBeginCmd(int id)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id && m->canRotate()) {
      markerUndo(m, EDIT);

      rotateMarker = m;
      return;
    }
    m=m->next();
  }

  rotateMarker = NULL;
}

void Frame::markerRotateBeginCmd(const Vector& v)
{
  // remember which marker is being edited

  Marker* m=markers.head();
  while (m) {
    if (m->isSelected() && m->canRotate()) {
      markerUndo(m, EDIT);

      rotateMarker = m;
      return;
    }
    m=m->next();
  }

  rotateMarker = NULL;
}

void Frame::markerRotateMotionCmd(const Vector& v, int h)
{
  if (rotateMarker) {

    // erase current marker now

    redraw(rotateMarker->getAllBBox());
    forceUpdate();

    rotateMarker->rotate(v*canvasToRef, h);
    rotateMarker->drawXOR();
  }
}

void Frame::markerRotateEndCmd()
{
  rotateMarker = NULL;
  update(PIXMAP);
}

void Frame::markerRulerPointCmd(int id, const Vector& p1, const Vector& p2,
				CoordSystem sys)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      markerUndo(m, EDIT);

      // it may shrink
      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      ((Ruler*)(m))->setPoints(ptr->map(p1,sys,REF), ptr->map(p2,sys,REF));
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerRulerSystemCmd(int id, CoordSystem coordsys, 
				 CoordSystem distsys, CoordFormat distformat)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Ruler*)(m))->setCoordSystem(coordsys, distsys, distformat);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void Frame::markerSaveCmd(const char* fileName, MarkerFormat type, 
			  CoordSystem sys, CoordFormat format)
{
  char delim = '\n';

  ofstream fn(fileName);
  if (fn) {

    void (Frame::*hproc)(ostream&, CoordSystem, CoordFormat, char);
    void (Marker::*mproc)(ostream&, CoordSystem, CoordFormat, char);

    switch (type) {
    case DS9:
      hproc = &Frame::markerListHeader;
      mproc = &Marker::list;
      break;
    case CIAO:
      hproc = &Frame::markerListCiaoHeader;
      mproc = &Marker::listCiao;
      break;
    case SAOTNG:
      hproc = &Frame::markerListSAOtngHeader;
      mproc = &Marker::listSAOtng;
      break;
    case SAOIMAGE:
      hproc = &Frame::markerListSAOimageHeader;
      mproc = &Marker::listSAOimage;
      break;
    case PROS:
      hproc = &Frame::markerListProsHeader;
      mproc = &Marker::listPros;
      break;
    case RAWXY:
      hproc = &Frame::markerListXYHeader;
      mproc = &Marker::listXY;
      break;
    }

    (this->*hproc)(fn, sys, format, delim);
    Marker* m=markers.head();
    while (m) {
      (m->*mproc)(fn, sys, format, delim);
      m=m->next();
    }
  }
  else {
    Tcl_AppendResult(interp, "Unable to open file ", fileName, NULL);
    result = TCL_ERROR;
  }
}

void Frame::markerSelectCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->select();
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerSelectCmd(const Vector& v)
{
  // first, check to see if we clicked on an already selected marker

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v) && m->isSelected()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  // ok, now select the first found, and unselect the rest

  int found = 0;

  m=markers.head();
  while (m) {
    if (m->isIn(v) && m->canSelect() && !found) {
      if (!m->isSelected()) {
	m->select();
	update(PIXMAP, m->getBBox());
      }
      found = 1;
    }
    else {
      if (m->isSelected()) {
	m->unselect();
	update(PIXMAP, m->getBBox());
      }
    }
    m=m->next();
  } 

  if (found)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void Frame::markerSelectAllCmd()
{
  Marker* m=markers.head();
  while (m) {
    if (m->canSelect()) {
      m->select();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void Frame::markerSelectToggleCmd(const Vector& v)
{
  // toggle the select of the first found

  Marker* m=markers.head();
  while (m) {
    if (m->isIn(v) && m->canSelect()) {
      m->toggleSelect();
      update(PIXMAP, m->getBBox());
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void Frame::markerTextCmd(int id, const char* text)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      // things can shrink, so do before and after
      update(PIXMAP,m->getAllBBox());
      m->setText(text);
      update(PIXMAP,m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerCopyCmd()
{
  undoMarkers.deleteAll();
  pasteMarkers.deleteAll();
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected())
      pasteMarkers.append(m->dup());
    m=m->next();
  }
}

void Frame::markerCutCmd()
{
  undoMarkers.deleteAll();
  pasteMarkers.deleteAll();
  Marker* m = markers.head();
  while (m) {
    if (m->isSelected() && m->canDelete()) {
      Marker* next = markers.extractNext(m);
      update(PIXMAP, m->getAllBBox());
      pasteMarkers.append(m);
      m->doCallBack(Marker::DELETECB);
      m->disableCB();
      m = next;
    }
    else
      m=m->next();
  }
}

void Frame::markerPasteCmd()
{
  unselectMarkers();

  undoMarkers.deleteAll();
  Marker* m=pasteMarkers.head();
  while (m) {
    Marker* n = m->dup();
    n->newIdentity();
    markers.append(n);

    m = m->next();
  }

  update(PIXMAP);
}

void Frame::markerUndoCmd()
{
  Marker* n=undoMarkers.head();
  while (n) {
    Marker* next = n->next();
    undoMarkers.extractPrev(n);
    
    switch (undoMarkerType) {
    case NONE:
      break;
    case DELETE:
      markers.append(n);
      n->updateBBox();
      update(PIXMAP,n->getAllBBox());
      break;
    case EDIT:
    case MOVE:
      {
	Marker* m=markers.head();
	while (m) {
	  if (m->getId() == n->getId()) {
	    n->enableCB();
	    m->updateBBox();
	    update(PIXMAP,m->getAllBBox());

	    markers.insertNext(m,n);
	    markers.extractNext(m);

	    n->updateBBox();
	    update(PIXMAP,n->getAllBBox());
	    switch (undoMarkerType) {
	    case EDIT:
	      n->doCallBack(Marker::EDITCB);
	      break;
	    case MOVE:
	      n->doCallBack(Marker::MOVECB);
	      break;
	    }

	    m->disableCB();
	    delete m;
	    break;
	  }
	  m=m->next();
	}
      }
      break;
    }

    n=next;
  }

  undoMarkerType = NONE;
}

void Frame::markerUnselectCmd(int id)
{
  Marker* m=markers.head();
  while (m) {
    if (m->getId() == id) {
      m->unselect();
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }
}

void Frame::markerUnselectAllCmd()
{
  Marker* m=markers.head();
  while (m) {
    m->unselect();
    update(PIXMAP, m->getBBox());
    m=m->next();
  } 
}

void Frame::regionSelectBeginCmd(const Vector& v)
{
  regionBegin = v;
  regionEnd = v;
}

void Frame::regionSelectMotionCmd(const Vector& v)
{
  if (visible) {

    // erase 

    redrawNow((BBox(regionBegin, regionEnd)).expand(2));

    // and draw to window

    regionEnd = v;
    BBox cc = BBox(regionBegin, regionEnd) * canvasToWindow;
    Vector size = cc.size();

    XDRAWRECTANGLE(display, Tk_WindowId(tkwin), selectGCXOR,
		   (int)cc.ll[0], (int)cc.ll[1], (int)size[0], (int)size[1]);
  }
}

void Frame::regionSelectEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers.head();
  while (m) {
    if (bb.isIn(m->getBBox())==4 && m->canSelect())
      m->select();
    else
      m->unselect();
    m=m->next();
  } 

  update(PIXMAP, bb.expand(2));
}

void Frame::regionSelectShiftEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers.head();
  while (m) {
    if (bb.isIn(m->getBBox())==4)
      m->select();
    m=m->next();
  }

  update(PIXMAP, bb.expand(2));
}

// Marker Support

void Frame::markerUndo(Marker* m, UndoMarkerType t)
{
  undoMarkers.deleteAll();
  undoMarkers.append(m->dup());
  undoMarkerType = t;
}

void Frame::parseMarker(MarkerFormat fm, istream& str)
{
  switch (fm) {
  case DS9:
    mklexx = new mkFlexLexer(&str);
    mkparse(this);
    delete mklexx;
    mklexx = NULL;
    break;

  case CIAO:
    ciaolexx = new ciaoFlexLexer(&str);
    ciaoparse(this);
    delete ciaolexx;
    ciaolexx = NULL;
    break;

  case PROS:
    proslexx = new prosFlexLexer(&str);
    prosparse(this);
    delete proslexx;
    proslexx = NULL;
    break;

  case SAOTNG:
    tnglexx = new tngFlexLexer(&str);
    tngparse(this);
    delete tnglexx;
    tnglexx = NULL;
    break;

  case SAOIMAGE:
    saolexx = new saoFlexLexer(&str);
    saoparse(this);
    delete saolexx;
    saolexx = NULL;
    break;

  case RAWXY:
    xylexx = new xyFlexLexer(&str);
    xyparse(this);
    delete xylexx;
    xylexx = NULL;
    break;
  }
}

void Frame::psMarkers(int mode)
{
  // render from back to front
  // bbox is in canvas coords

  const BBox bb = BBox(0, 0, options->width-1, options->height-1) * 
    widgetToCanvas;

  Marker* m=markers.tail();
  while (m) {
    if (m->isVisible(bb))
      m->ps(mode);
    m=m->previous();
  }
}

void Frame::renderMarkers(const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords

  Marker* m=markers.tail();
  while (m) {
    if (m->isVisible(bb))
      m->draw();
    m=m->previous();
  }
}

void Frame::renderMagnifierMarkers(const Matrix& matrix, const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords

  Marker* m=markers.tail();
  while (m) {
    if (m->isVisible(bb))
      m->drawMagnifier(matrix);
    m=m->previous();
  }
}

void Frame::unselectMarkers()
{
  Marker* m=markers.head();
  while (m) {
    m->unselect();
    m=m->next();
  }
}

void Frame::updateMarkers()
{
  Marker* m=markers.head();
  while (m) {
    m->updateBBox();
    m=m->next();
  }
}

void Frame::updateMarkerCoords(const Matrix& mm)
{
  Marker* m=markers.head();
  while (m) {
    m->updateCoords(mm);
    m=m->next();
  }
}

void Frame::markerPrintCoordTcl(const Vector& c, const Vector& v,
				CoordSystem sys, CoordFormat format)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      printCoordTcl(currentFits, v, REF, sys, format, DEFAULT);
    break;
  case MOSAIC:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(scanMode);
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printCoordTcl(ptr, v, REF, sys, format, DEFAULT);
	  return;
	}

	ptr = ptr->next();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case NATIVEWCS:
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case DETECTOR:
      if (currentFits)
	printCoordTcl(currentFits, v, REF, sys, format, DEFAULT);
      break;

    case AMPLIFIER:
    case PHYSICAL:
    case IMAGE:
      break;
    }

    break;
  }
}

void Frame::markerPrintVectorTcl(const Vector& c, const Vector& v,
				 CoordSystem sys, CoordFormat format)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      printVectorTcl(currentFits->mapLenFromRef(v, sys, format), DEFAULT);
    break;
  case MOSAIC:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(scanMode);
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printVectorTcl(ptr->mapLenFromRef(v, sys, format), DEFAULT);
	  return;
	}

	ptr = ptr->next();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case NATIVEWCS:
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case DETECTOR:
      if (currentFits)
	printVectorTcl(currentFits->mapLenFromRef(v, sys, format), DEFAULT);
      break;

    case AMPLIFIER:
    case PHYSICAL:
    case IMAGE:
      break;
    }

    break;
  }
}

void Frame::markerPrintDoubleTcl(const Vector& c, double d,
				 CoordSystem sys, CoordFormat format)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      printDoubleTcl(currentFits->mapLenFromRef(d, sys, format), DEFAULT);
    break;
  case MOSAIC:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(scanMode);
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printDoubleTcl(ptr->mapLenFromRef(d, sys, format), DEFAULT);
	  return;
	}

	ptr = ptr->next();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case NATIVEWCS:
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case DETECTOR:
      if (currentFits)
	printDoubleTcl(currentFits->mapLenFromRef(d, sys, format), DEFAULT);
      break;

    case AMPLIFIER:
    case PHYSICAL:
    case IMAGE:
      break;
    }

    break;
  }
}

void Frame::markerPrintDoubleTcl(const Vector& c, 
				 const Vector& p1, const Vector& p2,
				 CoordSystem sys, CoordFormat format)
{
  switch (displayMode) {
  case SINGLE:
    if (currentFits)
      printDoubleTcl(currentFits->mapLenFromRef(p1, p2 , sys, format),DEFAULT);
    break;
  case MOSAIC:
    if (fits) {
      FitsImage* ptr = fits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(scanMode);
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printDoubleTcl(ptr->mapLenFromRef(p1, p2, sys, format), DEFAULT);
	  return;
	}

	ptr = ptr->next();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case NATIVEWCS:
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
    case ICRS:
    case DETECTOR:
      if (currentFits)
	printDoubleTcl(currentFits->mapLenFromRef(p1,p2,sys,format), DEFAULT);
      break;

    case AMPLIFIER:
    case PHYSICAL:
    case IMAGE:
      break;
    }

    break;
  }
}

