#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "nan.h"
#include "outfile.h"
#include "outchannel.h"
#include "outsocket.h"

void Frame::saveFitsImageFileCmd(const char* fn, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsImageFile(fn, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsTableFileCmd(const char* fn, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsTableFile(fn, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsImageChannelCmd(const char* ch, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsImageChannel(interp, ch, compress)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsTableChannelCmd(const char* ch, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsTableChannel(interp, ch, compress)) {
      Tcl_AppendResult(interp, "Unable to save fits to channel ", ch, NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsImageSocketCmd(int s, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsImageSocket(s,compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsTableSocketCmd(int s, int compress)
{
  if (currentFits) {
    if (!currentFits->saveFitsTableSocket(s, compress)) {
      Tcl_AppendResult(interp, "Unable to save Fits", NULL);
      result = TCL_ERROR;
    }
  }
}

void Frame::saveFitsResampleFileCmd(const char* fn, int compress)
{
  if (!currentFits) {
    Tcl_AppendResult(interp, " no data loaded ", NULL);
    result = TCL_ERROR;
    return;
  }
    
  if (!compress) {
    OutFitsFile str(fn);
    if (str.valid())
      return saveFitsResample(str);
  }
  else {
    OutFitsFileGZ str(fn);
    if (str.valid())
      return saveFitsResample(str);
  }

  Tcl_AppendResult(interp, " unable to open fits output ", fn, NULL);
  result = TCL_ERROR;
}

void Frame::saveFitsResampleChannelCmd(const char* ch, int compress)
{
  if (!currentFits) {
    Tcl_AppendResult(interp, " no data loaded ", NULL);
    result = TCL_ERROR;
    return;
  }
    
  OutFitsChannel str(interp, ch);
  if (str.valid())
    return saveFitsResample(str);

  Tcl_AppendResult(interp, " unable to output to channel", NULL);
  result = TCL_ERROR;
}

void Frame::saveFitsResampleSocketCmd(int s, int compress)
{
  if (!currentFits) {
    Tcl_AppendResult(interp, " no data loaded ", NULL);
    result = TCL_ERROR;
    return;
  }
    
  if (!compress) {
    OutFitsSocket str(s);
    if (str.valid())
      return saveFitsResample(str);
  }
  else {
    OutFitsSocketGZ str(s);
    if (str.valid())
      return saveFitsResample(str);
  }

  Tcl_AppendResult(interp, " unable to output to socket", NULL);
  result = TCL_ERROR;
}

void Frame::saveFitsResample(OutFitsStream& str)
{
  int& width = options->width;
  int& height = options->height;

  int bitpix_ = -32;
  int datapixels_ = width*height;
  int realbytes_ = datapixels_ * (abs(bitpix_)/8);
  int datablocks_ = (realbytes_ + (FTY_BLOCK-1))/FTY_BLOCK;
  int databytes_ = datablocks_ * FTY_BLOCK;

  // create header
  FitsHead hd(width, height, bitpix_);

  // write keywords
  saveFitsResampleKeyword(str, hd);

  // write header
  str.write(hd.cards(), hd.headbytes());

  // write data
  switch (displayMode) {
  case SINGLE:
    saveFitsResampleSingle(str);
    break;
  case MOSAIC:
    saveFitsResampleMosaic(str);
    break;
  }

  // pad rest of block
  {
    int diff = databytes_ - realbytes_;
    char buf[diff];
    memset(buf,'\0',diff);
    str.write(buf, diff);
  }
}

void Frame::saveFitsResampleKeyword(OutFitsStream& str, FitsHead& dst)
{
  FitsHead* src = currentFits->getHead();
  Vector center = Vector(options->width, options->height)/2.;

  // OBJECT
  char* object = src->getString("OBJECT");
  if (object) {
    dst.appendString("OBJECT", object, NULL);
    delete object;
  }

  // DATE-OBS
  char* date = src->getString("DATE");
  if (date) {
    dst.appendString("DATE", date, NULL);
    delete date;
  }
  char* dateobs = src->getString("DATE-OBS");
  if (dateobs) {
    dst.appendString("DATE-OBS", dateobs, NULL);
    delete dateobs;
  }
  char* timeobs = src->getString("TIME-OBS");
  if (timeobs) {
    dst.appendString("TIME-OBS", timeobs, NULL);
    delete timeobs;
  }
  char* dateend = src->getString("DATE-END");
  if (dateend) {
    dst.appendString("DATE-END", dateend, NULL);
    delete dateend;
  }
  char* timeend = src->getString("TIME-END");
  if (timeend) {
    dst.appendString("TIME-END", timeend, NULL);
    delete timeend;
  }

  // LTMV,DTMV
  switch (displayMode) {
  case SINGLE:
    if (currentFits->hasLTMV()) {
      Matrix ltmv = 
	currentFits->getPhysicalToUser() *
	Translate(-cursor) *
	Rotate(rotation) *
	Rotate(wcsRotation) *
	orientationMatrix *
	wcsOrientationMatrix *
	Scale(zoom) *
	Scale(wcsZoom) *
	FlipY() * 
	Translate(center);

      dst.appendReal("LTM1_1", ltmv[0][0], 9, NULL);
      dst.appendReal("LTM1_2", ltmv[0][1], 9, NULL);
      dst.appendReal("LTM2_1", ltmv[1][0], 9, NULL);
      dst.appendReal("LTM2_2", ltmv[1][1], 9, NULL);
      dst.appendReal("LTV1",   ltmv[2][0], 9, NULL);
      dst.appendReal("LTV2",   ltmv[2][1], 9, NULL);
    }
    break;
  case MOSAIC:
    if (currentFits->hasDTMV()) {
      Matrix dtmv = 
	currentFits->getDetectorToUser() *
	Translate(-cursor) *
	Rotate(rotation) *
	Rotate(wcsRotation) *
	orientationMatrix *
	wcsOrientationMatrix *
	Scale(zoom) *
	Scale(wcsZoom) *
	FlipY() * 
	Translate(center);

      dst.appendReal("DTM1_1", dtmv[0][0], 9, NULL);
      dst.appendReal("DTM1_2", dtmv[0][1], 9, NULL);
      dst.appendReal("DTM2_1", dtmv[1][0], 9, NULL);
      dst.appendReal("DTM2_2", dtmv[1][1], 9, NULL);
      dst.appendReal("DTV1",   dtmv[2][0], 9, NULL);
      dst.appendReal("DTV2",   dtmv[2][1], 9, NULL);
    }
    break;
  }

  // WCS
  if (currentFits->hasWCS()) {
    WorldCoor* wcs = currentFits->getWCS();

    dst.appendString("RADECSYS", wcs->radecsys, NULL);
    dst.appendReal("EQUINOX", wcs->equinox, 9, NULL);

    dst.appendString("CTYPE1", wcs->ctype[0], NULL);
    dst.appendString("CTYPE2", wcs->ctype[1], NULL);
    dst.appendReal("CRVAL1", wcs->crval[0], 9, NULL);
    dst.appendReal("CRVAL2", wcs->crval[1], 9, NULL);

    char* cunit1 = src->getString("CUNIT1");
    if (cunit1) {
      dst.appendString("CUNIT1", cunit1, NULL);
      delete cunit1;
    }

    char* cunit2 = src->getString("CUNIT2");
    if (cunit2) {
      dst.appendString("CUNIT2", cunit2, NULL);
      delete cunit2;
    }

    // crpix
    Vector crpix = Vector(wcs->crpix[0],wcs->crpix[1]) * 
      currentFits->getImageToUser() * 
      Translate(-cursor) *
      Rotate(rotation) *
      Rotate(wcsRotation) *
      orientationMatrix *
      wcsOrientationMatrix *
      Scale(zoom) *
      Scale(wcsZoom) *
      FlipY() * 
      Translate(center);

    dst.appendReal("CRPIX1", crpix[0], 9, NULL);
    dst.appendReal("CRPIX2", crpix[1], 9, NULL);

    // cd matrix
    Matrix cd = Matrix(wcs->cd[0],wcs->cd[1],wcs->cd[2],wcs->cd[3],0,0) *
      currentFits->getImageToUser() *
      Translate(-cursor) *
      Rotate(rotation) *
      Rotate(wcsRotation) *
      orientationMatrix *
      wcsOrientationMatrix *
      Scale(1/zoom) *
      Scale(1/wcsZoom) *
      FlipY() * 
      Translate(center);

    dst.appendReal("CD1_1", cd.matrix(0,0), 9, NULL);
    dst.appendReal("CD1_2", cd.matrix(0,1), 9, NULL);
    dst.appendReal("CD2_1", cd.matrix(1,0), 9, NULL);
    dst.appendReal("CD2_2", cd.matrix(1,1), 9, NULL);
  }
}

void Frame::saveFitsResampleSingle(OutFitsStream& str)
{
  float nanf = getNaNf();
  int& width = options->width;
  int& height = options->height;

  double* m = currentFits->getmWidgetToData();
  int* params = currentFits->getDataParams(scanMode);
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  for (int j=height-1; j>=0; j--)
    for (int i=0; i<width; i++) {
      double x = i*m[0] + j*m[3] + m[6];
      double y = i*m[1] + j*m[4] + m[7];

      float v = nanf;
      if (x>=xmin && x<xmax && y>=ymin && y<ymax)
	v = currentFits->getValueFloat(((int)y)*srcw + (int)x);
      if (lsb())
	str.writeSwap((char*)(&v), 4, -32);
      else
	str.write((char*)(&v), 4);
    }
}

void Frame::saveFitsResampleMosaic(OutFitsStream& str)
{
  float nanf = getNaNf();
  int& width = options->width;
  int& height = options->height;

  for (int j=height-1; j>=0; j--)
    for (int i=0; i<width; i++) {

      float v = nanf;
      FitsImage* ptr = fits;
      while (ptr) {
	double* m = ptr->getmWidgetToData();
	int* params = ptr->getDataParams(scanMode);
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	double x = i*m[0] + j*m[3] + m[6];
	double y = i*m[1] + j*m[4] + m[7];

	if (x>=xmin && x<xmax && y>=ymin && y<ymax) {
	  v = ptr->getValueFloat(((int)y)*srcw + (int)x);
	  break;
	}
	else
	  ptr = ptr->next();
      }

      if (lsb())
	str.writeSwap((char*)(&v), 4, -32);
      else
	str.write((char*)(&v), 4);
    }
}
