#include <iostream.h>

#include "grid.h"
#include "util.h"
#include "frame.h"
#include "fitsimage.h"
#include "rotstr.h"

#include <Xutil.h>

extern "C" {
  #include "wcs.h"
  #include "ast.h"
  #include "grf.h"
}

Attribute::Attribute(Frame* p) : parent(p)
{
  style_ = SOLID;
  width_ = 1;
  size_ = 10;
  font_ = DEFAULT;
  tkfont_ = Tk_GetFont(parent->interp, parent->tkwin, "helvetica 10 normal");

  colour_ = WHITE;
  colorName_ = NULL;
  setColour(colour_);
}

Attribute::~Attribute()
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);

  if (colorName_)
    delete colorName_;
}

void Attribute::setStyle(double v)
{
  switch ((int)v) {
  case SOLID:
  case DASH:
    break;
  default:
    return;
  }

  style_ = (Style)v;
}

void Attribute::setWidth(double v)
{
  width_ = v;
}

void Attribute::setSize(double v)
{
  if (v >= 1) {
    size_ = (int)v;
    initFont();
  }
}

void Attribute::setFont(double v)
{
  switch ((Font)v) {
  case DEFAULT:
  case TIMES:
  case HELVETICA:
  case SYMBOL:
  case COURIER:

  case DEFAULTBOLD:
  case TIMESBOLD:
  case HELVETICABOLD:
  case SYMBOLBOLD:
  case COURIERBOLD:

  case DEFAULTITALIC:
  case TIMESITALIC:
  case HELVETICAITALIC:
  case SYMBOLITALIC:
  case COURIERITALIC:
    font_ = (Font)v;
    initFont();
    return;

  default:
    return;
  }
}

void Attribute::initFont()
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);
  tkfont_ = NULL;

  char buf[256];
  ostrstream str(buf,256);

  switch (font_) {
  case DEFAULT:
    str << "system " << size_ << " normal" << ends;
    break;
  case TIMES:
    str << "times " << size_ << " normal" << ends;
    break;
  case HELVETICA:
    str << "helvetica " << size_ << " normal" << ends;
    break;
  case SYMBOL:
    str << "symbol " << size_ << " normal" << ends;
    break;
  case COURIER:
    str << "courier " << size_ << " normal" << ends;
    break;

  case DEFAULTBOLD:
    str << "system " << size_ << " bold" << ends;
    break;
  case TIMESBOLD:
    str << "times " << size_ << " bold" << ends;
    break;
  case HELVETICABOLD:
    str << "helvetica " << size_ << " bold" << ends;
    break;
  case SYMBOLBOLD:
    str << "symbol " << size_ << " bold" << ends;
    break;
  case COURIERBOLD:
    str << "courier " << size_ << " bold" << ends;
    break;

  case DEFAULTITALIC:
    str << "system " << size_ << " italic" << ends;
    break;
  case TIMESITALIC:
    str << "times " << size_ << " italic" << ends;
    break;
  case HELVETICAITALIC:
    str << "helvetica " << size_ << " italic" << ends;
    break;
  case SYMBOLITALIC:
    str << "symbol " << size_ << " italic" << ends;
    break;
  case COURIERITALIC:
    str << "courier " << size_ << " italic" << ends;
    break;
  default:
    return;
  }

  tkfont_ = Tk_GetFont(parent->interp, parent->tkwin, buf);
}

void Attribute::setColour(double v)
{
  if (colorName_)
    delete colorName_;
  colorName_ = NULL;

  switch ((int)v) {
  case BLACK:
    colorName_ = dupstr("black");
    break;
  case WHITE:
    colorName_ = dupstr("white");
    break;
  case RED:
    colorName_ = dupstr("red");
    break;
  case GREEN:
    colorName_ = dupstr("green");
    break;
  case BLUE:
    colorName_ = dupstr("blue");
    break;
  case CYAN:
    colorName_ = dupstr("cyan");
    break;
  case MAGENTA:
    colorName_ = dupstr("magenta");
    break;
  case YELLOW:
    colorName_ = dupstr("yellow");
    break;
  default:
    return;
  }

  color_ = parent->getColor(colorName_);
  colour_ = (Colour)v;
}

Grid* gr = NULL;

Grid::Grid(Frame* p, int sys, int t, const char* o) 
  : parent(p), system_(sys), type_(t)
{
  option_ = dupstr(o);

  line = new Attribute(parent);
  mark = new Attribute(parent);
  text = new Attribute(parent);
  psMode = Frame::RGB;
  fits = NULL;
}

Grid::~Grid()
{
  if (option_)
    delete option_;

  if (line)
    delete line;
  if (mark)
    delete mark;
  if (text)
    delete text;
}

int Grid::render()
{
  switch (parent->displayMode) {
  case Frame::SINGLE:
    fits = parent->fits;
    switch (type_) {
    case 1:
    case 2:
      return doit(X11,WIDGETBB);
    case 3:
    case 4:
      return doit(X11,IMAGEBB);
    }

  case Frame::MOSAIC:
    switch (system_) {
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case ICRS:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
      fits = parent->fits;
      switch (type_) {
      case 1:
      case 2:
	return doit(X11,WIDGETBB);
      case 3:
      case 4:
	return doit(X11,MOSAICBB);
      }
      break;

    default:
      FitsImage* ptr = parent->fits;
      int status = 1;
      while (ptr) {
	fits = ptr;
	status &= doit(X11,IMAGEBB);
	ptr = ptr->next();
      }
      return status;
    }
  }
}

int Grid::ps(int mode)
{
  psMode = mode;

  switch (parent->displayMode) {
  case Frame::SINGLE:
    fits = parent->fits;
    switch (type_) {
    case 1:
    case 2:
      return doit(PS,WIDGETBB);
    case 3:
    case 4:
      return doit(PS,IMAGEBB);
    }

  case Frame::MOSAIC:
    switch (system_) {
    case EQUATORIAL_FK4:
    case EQUATORIAL_FK5:
    case ICRS:
    case GALACTIC:
    case ECLIPTIC:
    case LINEAR:
      fits = parent->fits;
      switch (type_) {
      case 1:
      case 2:
	return doit(PS,WIDGETBB);
      case 3:
      case 4:
	return doit(PS,MOSAICBB);
      }

    default:
      FitsImage* ptr = parent->fits;
      int status = 1;
      while (ptr) {
	fits = ptr;
	status &= doit(PS,IMAGEBB);
	ptr = ptr->next();
      }
      return status;
    }
  }
}

int Grid::doit(RenderMode rm, RenderBound bound)
{
  int width = fits->width();
  int height = fits->height();

  astClearStatus; // just to make sure
  astBegin; // start memory management

  AstFrameSet* frameSet = NULL;
  AstPlot* plot = NULL;

  switch (system_) {
  case EQUATORIAL_FK4:
  case EQUATORIAL_FK5:
  case ICRS:
  case GALACTIC:
  case ECLIPTIC:
    {
      if (!fits->hasWCSEqu())
	goto error;

      WorldCoor* wcs = fits->getWCS();
      if (wcs->prjcode>=WCS_AZP && 
	  wcs->prjcode<=WCS_NCP && 
	  wcs->prjcode!=WCS_PAR) {

	if (!(frameSet = (AstFrameSet*)wcs2ast()))
	  goto error;
      }
      else {
	if (!(frameSet = (AstFrameSet*)fits2ast()))
	  goto error;
      }

      if (!strcmp(astGetC(frameSet,"DOMAIN"), "CMP"))
	goto error;

      // set desired WCS system

      switch (system_) {
      case EQUATORIAL_FK4:
	astSet(frameSet, "System=FK4, Equinox=B1950");
	break;
      case EQUATORIAL_FK5:
	astSet(frameSet, "System=FK5, Equinox=J2000");
	break;
      case ICRS:
	astSet(frameSet, "System=FK5, Equinox=J2000");
	break;
      case GALACTIC:
	astSet(frameSet, "System=Galactic");
	break;
      case ECLIPTIC:
	astSet(frameSet, "System=Ecliptic");
	// this is a kludge to get AST to agree with wcssubs
	astSetD(frameSet, "Equinox", astGetD(frameSet, "Epoch"));
	break;
      }
    }
    break;

  case LINEAR:
    {
      if (!fits->hasWCSLinear())
	goto error;

      WorldCoor* wcs = fits->getWCS();

      Vector t = Vector(wcs->crval[0] - 
			(wcs->cd[0]*wcs->crpix[0]) - 
			(wcs->cd[2]*wcs->crpix[1]),
			wcs->crval[1] - 
			(wcs->cd[1]*wcs->crpix[0]) -
			(wcs->cd[3]*wcs->crpix[1]));

      Vector imgll = Vector(0,0);
      Vector imgur = Vector(1,1);
      Vector linll = imgll+t;
      Vector linur = imgur+t;

      if (!(frameSet = astFrameSet(astFrame(2, ""), "")))
	goto error;

      AstMatrixMap* mm;
      if (!(mm= astMatrixMap(2, 2, 0, wcs->cd, "")))
	goto error;

      AstWinMap* wm;
      if (!(wm = astWinMap(2, imgll.v, imgur.v, linll.v, linur.v, "")))
	goto error;

      AstCmpMap* cmp;
      if (!(cmp = astCmpMap(mm, wm, 1, "")))
	goto error;

      astAddFrame(frameSet, 1, cmp, astFrame(2, ""));
    }
    break;

  case DETECTOR:
    {
      Matrix m = fits->getImageToPhysical() * fits->getPhysicalToDetector();
      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector detll = imgll * m;
      Vector detur = imgur * m;

      if (!(frameSet = astFrameSet(astFrame(2, ""), "")))
	goto error;

      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, detll.v, detur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case AMPLIFIER:
    {
      Matrix m = fits->getImageToPhysical() * fits->getPhysicalToAmplifier();
      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector ampll = imgll * m;
      Vector ampur = imgur * m;

      if (!(frameSet = astFrameSet(astFrame(2, ""), "")))
	goto error;

      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, ampll.v, ampur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case PHYSICAL:
    {
      Vector imgll(1,1);
      Vector imgur = Vector(width, height);
      Vector phyll = imgll * fits->getImageToPhysical();
      Vector phyur = imgur * fits->getImageToPhysical();

      if (!(frameSet = astFrameSet(astFrame(2, ""), "")))
	goto error;

      astAddFrame(frameSet, 1, 
		  astWinMap(2, imgll.v, imgur.v, phyll.v, phyur.v, ""),
		  astFrame(2, ""));
    }
    break;

  case IMAGE:
    if (!(frameSet = astFrameSet(astFrame(2,""),"")))
      goto error;
    break;
  }

  astSet(frameSet,"Title=%s", " ");

  if (DebugAST) {
    int status = astStatus;
    astClearStatus;
    astShow(frameSet);
    astSetStatus(status);
  }

  // create astPlot

  float gbox[4];
  double pbox[4];

  switch (bound) {
  case WIDGETBB:
    {
      Matrix m = fits->getWidgetToImage();
      Vector ll = Vector(0,0) * m;
      Vector lr = Vector(parent->options->width,0) * m;
      Vector ur = Vector(parent->options->width, parent->options->height) * m;
      Vector ul = Vector(0, parent->options->height) * m;

      BBox bb(ll,ll);
      bb.bound(lr);
      bb.bound(ur);
      bb.bound(ul);

      gbox[0] = pbox[0] = (int)bb.ll[0];
      gbox[1] = pbox[1] = (int)bb.ll[1];
      gbox[2] = pbox[2] = (int)bb.ur[0];
      gbox[3] = pbox[3] = (int)bb.ur[1];
    }
    break;
  case IMAGEBB:
    gbox[0] = pbox[0] = 1;
    gbox[1] = pbox[1] = 1;
    gbox[2] = pbox[2] = width;
    gbox[3] = pbox[3] = height;
    break;
  case MOSAICBB:
    {
      BBox bb = fits->getWCSbb();

      gbox[0] = pbox[0] = (int)bb.ll[0];
      gbox[1] = pbox[1] = (int)bb.ll[1];
      gbox[2] = pbox[2] = (int)bb.ur[0];
      gbox[3] = pbox[3] = (int)bb.ur[1];
    }
    break;
  }

  plot = astPlot(frameSet, gbox, pbox, option_);

  // and now create astGrid

  gr = this;
  renderMode = rm;
  astGrid(plot);

  astEnd; // now, clean up memory
  return 1;

error:
  astEnd;
  return 0;
}

// Generic

void Grid::gAttr(int which, double value, double* old, int prim)
{
  if (value == AST__BAD) {
    if (old)
      *old = 0;
    return;
  }

  Attribute* attr;
  switch (prim) {
  case GRF__TEXT:
    attr = text;
    break;
  case GRF__LINE:
    attr = line;
    break;
  case GRF__MARK:
    attr = mark;
    break;
  }

  switch (which) {
  case GRF__STYLE:
    if (old)
      *old = attr->style();
    attr->setStyle(value);
    break;
  case GRF__WIDTH:
    if (old)
      *old = attr->width();
    attr->setWidth(value);
    break;
  case GRF__SIZE:
    if (old)
      *old = attr->size();
    attr->setSize(value);
    break;
  case GRF__FONT:
    if (old)
      *old = attr->font();
    attr->setFont(value);
    break;
  case GRF__COLOUR:
    if (old)
      *old = attr->colour();
    attr->setColour(value);
    break;
  }
}

void Grid::gQch(float* chv, float* chh)
{
  //  cerr << "gQch" << endl;
  if (text->tkfont()) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);

    *chv = (float)metrics.linespace;
    *chh = (float)metrics.linespace;
  }
  else
    *chv = *chh = 0;
}

void Grid::gTxExt(const char* txt, float x, float y, const char* just,
		  float upx, float upy, float* xb, float* yb)
{
  if (txt && text->tkfont()) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1,m2;
    if (just) {
      switch (just[0]) {
      case 'T':
	m1 = Translate(0,metrics.ascent);
	break;
      case 'C':
	break;
      case 'B':
	m1 = Translate(0,metrics.descent);
	break;
      }

      switch (just[1]) {
      case 'L':
	break;
      case 'C':
	m2 = Translate(-width/2,0);
	break;
      case 'R':
	m2 = Translate(-width,0);
	break;
      }
    }

    Vector c = Vector(x,y) * fits->getImageToWidget() *  m1 * m2;

    BBox bb = BBox(c*Translate(-width/2,-metrics.descent), 
		   c*Translate(width/2,metrics.ascent));

    xb[0] = xb[1] = bb.ll[0];
    xb[2] = xb[3] = bb.ur[0];

    yb[0] = yb[3] = bb.ll[1];
    yb[1] = yb[2] = bb.ur[1];
  }
  else {
    xb[0] = xb[1] = xb[2] = xb[3] = 0;
    yb[0] = yb[1] = yb[2] = yb[3] = 0;
  }
}

// X11 Render functions

void Grid::x11Line(int n, const float* x, const float* y)
{
  if (n<2 || !x || !y)
    return;

  XSetForeground(parent->display, parent->gridGC, line->color());
  int w = line->width();
  if (w<1)
    w = 1;
  switch (line->style()) {
  case Attribute::SOLID:
    XSetLineAttributes(parent->display, parent->gridGC, w, 
		       LineSolid, CapButt, JoinMiter);
    break;
  case Attribute::DASH:
    XSetLineAttributes(parent->display, parent->gridGC, w, 
		       LineOnOffDash, CapButt, JoinMiter);
    char dlist[] = {8,3};
    XSetDashes(parent->display, parent->gridGC, 0, dlist, 2);
    break;
  }

  Matrix m = fits->getImageToWidget();

  Vector s = Vector(x[0],y[0]) * m;
  Vector e;
  for (int i=0; i<n; i++) {
    e = Vector(x[i],y[i]) * m;
    XDRAWLINE(parent->display, parent->basePixmap, parent->gridGC, 
	      (int)s[0],(int)s[1],(int)e[0],(int)e[1]);
    s = e;
  }
}

void Grid::x11Text(const char* txt, float x, float y, 
		   const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    XSetFont(parent->display, parent->gridGC, Tk_FontId(text->tkfont()));
    XSetForeground(parent->display, parent->gridGC, text->color());

    Vector v = Vector(x,y) * fits->getImageToWidget();

    int flip = 0;
    double angle = calcTextAngle(Vector(upx,upy).angle(), &flip);

    if ((angle>-.01 && angle<+.01) && !flip)
      drawString(txt, v, just);
    else
      drawRotString(txt, v, just, angle, flip);
  }
}

void Grid::drawString(const char* txt, Vector v, const char* j)
{
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(text->tkfont(), &metrics);
  int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

  switch (j[0]) {
  case 'T':
    v[1] += metrics.ascent;
    break;
  case 'C':
    break;
  case 'B':
    v[1] -= metrics.descent;
    break;
  }

  switch (j[1]) {
  case 'L':
    break;
  case 'C':
    v[0] -= width/2.;
    break;
  case 'R':
    v[0] -= width;
    break;
  }

  XDrawString(parent->display, parent->basePixmap, parent->gridGC,
	      (int)v[0], (int)v[1], txt, strlen(txt));
}

void Grid::drawRotString(const char* txt, Vector v, 
			 const char* just, double angle, int flip)
{
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(text->tkfont(), &metrics);
  int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));
  int height = metrics.linespace;

  Matrix m1;
  switch (just[0]) {
  case 'T':
    m1 = !flip ? Translate(0, height/2) : Translate(0, -height/2);
    break;
  case 'C':
    break;
  case 'B':
    m1 = !flip ? Translate(0, -height/2) : Translate(0, height/2);
    break;
  }

  Matrix m2;
  switch (just[1]) {
  case 'L':
    m2 = !flip ? Translate(width/2, 0) : Translate(-width/2, 0);
    break;
  case 'C':
    break;
  case 'R':
    m2 = !flip ? Translate(-width/2, 0) : Translate(width/2, 0);
    break;
  }

  Matrix m = Translate(-v) * Rotate(-angle) * 
    m1 * m2 *
    Rotate(angle) * Translate(v);

  XDrawRotString(parent->display, parent->basePixmap, parent->gridGC,
		 parent->baseXImage->byte_order, 
		 parent->getBlackColor(), parent->getWhiteColor(),
		 txt, text->tkfont(), angle, v*m);
}

// PS Render functions

void Grid::psLine(int n, const float* x, const float* y)
{
  if (n<2 || !x || !y)
    return;

  char buf[256];
  Matrix m = fits->getImageToCanvas();

  psColor(line);
  psWidth(line);
  psStyle(line);

  {
    ostrstream str(buf,256);
    Vector v = Vector(x[0],y[0]) * m;
    str << "newpath " << endl
	<< v.TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }

  for (int i=1; i<n; i++) {
    ostrstream str(buf,256);
    Vector v = Vector(x[i],y[i]) * m;
    str << v.TkCanvasPs(parent->canvas) << " lineto" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }

  {
    ostrstream str(buf,256);
    str << "stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, buf, NULL);
  }
}

void Grid::psText(const char* txt, float x, float y, 
		  const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    int flip = 0;
    double angle = calcTextAngle(Vector(upx,upy).angle(),&flip);

    Tcl_DString psFont;
    Tcl_DStringInit(&psFont);
    int psSize = Tk_PostscriptFontName(text->tkfont(), &psFont);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1;
    switch (just[0]) {
    case 'T':
      m1 = !flip ? Translate(0,metrics.ascent) : Translate(0,-metrics.descent);
      break;
    case 'C':
      break;
    case 'B':
      m1 = !flip ? Translate(0,-metrics.descent) : Translate(0,metrics.ascent);
      break;
    }

    Matrix m2;
    switch (just[1]) {
    case 'L':
      m2 = !flip ? Translate() : Translate(-width,0);
      break;
    case 'C':
      m2 = Translate(-width/2.,0);
      break;
    case 'R':
      m2 = !flip ? Translate(-width,0) : Translate();
      break;
    }

    Vector v = Vector(x,y) * fits->getImageToCanvas();
    Matrix m = Translate(-v) * Rotate(-angle) *
      m1 * m2 *
      Rotate(angle) * Translate(v);
    
    char buf[512];
    ostrstream str(buf,512);
    str << '/' << Tcl_DStringValue(&psFont) 
	<< " findfont " << psSize << " scalefont setfont" << endl;

    Tcl_DStringFree(&psFont);

    psColor(text);

    str << "gsave " 
	<< (v*m).TkCanvasPs(parent->canvas) << " moveto" << endl
      	<< radToDeg(angle) << " rotate " 
	<< '(' << psQuote(txt) << ')' << " show"
	<< " grestore" << endl << ends;

    Tcl_AppendResult(parent->interp, buf, NULL);
  }
}

// PS Support

void Grid::psColor(Attribute* attr)
{
  char buf[256];

  ostrstream str(buf,256);
  switch ((Frame::PSColorSpace)psMode) {
  case Frame::BW:
  case Frame::GRAY:
    psColorGray(attr->colorName(), str);
    str << " setgray";
    break;
  case Frame::RGB:
    psColorRGB(attr->colorName(), str);
    str << " setrgbcolor";
    break;
  case Frame::CMYK:
    psColorCMYK(attr->colorName(), str);
    str << " setcmykcolor";
    break;
  }
  str << endl << ends;

  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Grid::psWidth(Attribute* attr)
{
  char buf[256];
  ostrstream str(buf,256);
  str << attr->width() << " setlinewidth" << endl << ends;
  Tcl_AppendResult(parent->interp, buf, NULL);
}

void Grid::psStyle(Attribute* attr)
{
  char buf[256];
  ostrstream str(buf,256);

  switch (attr->style()) {
  case Attribute::SOLID:
    str << "[] 0 setdash" << endl << ends;
    break;
  case Attribute::DASH:
    str << "[8 3] 0 setdash" << endl << ends;
    break;
  }

  Tcl_AppendResult(parent->interp, buf, NULL);
}

double Grid::calcTextAngle(double a, int* flip)
{
  if (flip)
    *flip = 0;

  // our angle is 90 off from ast's
  double r = a - M_PI/2;

  // and add any image rotation
  r += parent->getRotation();

  // normalize
  if (r>0)
    while (r>2*M_PI)
      r -= 2*M_PI;
  else 
    while (r<0)
      r += 2*M_PI;

  // and flip if upside down
  // we only want to do this if the number is really, really upside down
  // you might have the case of a curved vertical line
  if (r > M_PI/2+.1 && r < 3*M_PI/2-.1) {
    r -= M_PI;
    *flip = 1;
  }

  return r;
}

// AST support

void* Grid::fits2ast() 
{
  // read FITS header into astChannel

  // we may have an error, just reset
  astClearStatus;

  // new fitschan
  AstFitsChan* chan = astFitsChan(NULL, NULL, "");
  if (!astOK || chan == AST__NULL)
    return NULL;

  // no warning messages
  astClear(chan,"Warnings");

  // fill up chan
  char* header = fits->getCards();
  int ncards = fits->getNCards();
  for (int i=0; i<ncards; i++) {
    char buf[81];
    strncpy(buf,header+(i*80),80);
    buf[80] = '\0';

    astPutFits(chan, buf, 0);
    // sometimes, we get a bad parse, just ignore
    if (!astOK)
      astClearStatus;
  }

  // parse FITS header for WCS

  // we may have an error, just reset
  astClearStatus;
  astClear(chan, "Card");

  // parse header
  AstFrameSet* frameSet = (AstFrameSet*)astRead(chan);

  // do we have anything?
  if (!astOK || frameSet == AST__NULL || 
      strcmp(astGetC(frameSet,"Class"), "FrameSet"))
    return NULL;

  return frameSet;
}

void* Grid::wcs2ast() 
{
  // read wcs struct into astChannel

  // we may have an error, just reset
  astClearStatus;

  // new fitschan
  AstFitsChan* chan = astFitsChan(NULL, NULL, "");
  if (!astOK || chan == AST__NULL)
    return NULL;

  // no warning messages
  astClear(chan,"Warnings");

  // fill up chan
  WorldCoor* wcs = fits->getWCS();

  if (!(strncmp("RA---TNX", wcs->ctype[0], 8))) {
    putFitsCard(chan, "CTYPE1  ", "RA---TAN");
    putFitsCard(chan, "CTYPE2  ", "DEC--TAN");
  }
  else {
    putFitsCard(chan, "CTYPE1  ", wcs->ctype[0]);
    putFitsCard(chan, "CTYPE2  ", wcs->ctype[1]);
  }

  if (!strncmp("FK4",wcs->radecsys,3) ||
      !strncmp("FK5",wcs->radecsys,3) ||
      !strncmp("ICRS",wcs->radecsys,4))
    putFitsCard(chan, "RADESYS ", wcs->radecsys);
  
  putFitsCard(chan, "CD1_1   ", wcs->cd[0]);
  putFitsCard(chan, "CD1_2   ", wcs->cd[1]);
  putFitsCard(chan, "CD2_1   ", wcs->cd[2]);
  putFitsCard(chan, "CD2_2   ", wcs->cd[3]);

  putFitsCard(chan, "CRPIX1  ", wcs->crpix[0]);
  putFitsCard(chan, "CRPIX2  ", wcs->crpix[1]);

  putFitsCard(chan, "CRVAL1  ", wcs->crval[0]);
  putFitsCard(chan, "CRVAL2  ", wcs->crval[1]);

  putFitsCard(chan, "EQUINOX ", wcs->equinox);

  // from wcssub/wcsinit.c line 800
  // wcs->epoch = 1900.0 + (mjd - 15019.81352) / 365.242198781;
  putFitsCard(chan, "MJD-OBS ", (wcs->epoch-1900)*365.242198781+15019.81352);

  if (wcs->latpole != 999)
    putFitsCard(chan, "LATPOLE ", wcs->latpole * AST__DD2R);
  if (wcs->longpole != 999)
    putFitsCard(chan, "LONPOLE ", wcs->longpole * AST__DD2R);

  // rewind chan
  astClear(chan, "Card");

  // parse header
  AstFrameSet* frameSet = (AstFrameSet*)astRead(chan);

  // do we have anything?
  if (!astOK || frameSet == AST__NULL || 
      strcmp(astGetC(frameSet,"Class"), "FrameSet"))
    return NULL;

  if (wcs->coorflip) {
    int orr[] = {2,1};
    astPermAxes(frameSet,orr);
  }

  return frameSet;
}

void Grid::putFitsCard(void* chan, const char* key, const char* value)
{
  char buf[80];
  ostrstream str(buf,80);
  memset(buf,'\0', 80);
  str << key << "= '" << value << "'";
  astPutFits(chan, buf, 0);
  astClearStatus;

  if (DebugAST)
    cerr << buf << endl;
}

void Grid::putFitsCard(void* chan, const char* key, double value)
{
  char buf[80];
  ostrstream str(buf,80);
  memset(buf,'\0', 80);
  str.setf(ios::scientific,ios::floatfield);
  str << key << "= " << setprecision(12) << value;
  astPutFits(chan, buf, 0);
  astClearStatus;

  if (DebugAST)
    cerr << buf << endl;
}

/*
static void printStore(FitsStore* store)
{
  if (store->ctype)
    printf("ctype: %s %s\n",(*store->ctype)[0],(*store->ctype)[1]);
  if (store->ctype_com)
    printf("ctype_com: %s %s\n",(*store->ctype_com)[0],(*store->ctype_com)[1]);
  if (store->cunit)
    printf("cunit: %s %s\n",(*store->cunit)[0],(*store->cunit)[1]);

  if (store->radesys)
    printf("radesys: %s\n",**store->radesys);
  if (store->wcsname)
    printf("wcsname: %s\n",**store->wcsname);

  if (store->cd)
    printf("cd: %g %g %g %g\n",(*store->cd)[0][0],(*store->cd)[0][1],
	   (*store->cd)[1][0],(*store->cd)[1][1]);
  if (store->crpix)
    printf("crpix: %f %f\n",(*store->crpix)[0][0],(*store->crpix)[0][1]);
  if (store->crval)
    printf("crval: %f %f\n",(*store->crval)[0][0],(*store->crval)[1][0]);

  if (store->equinox)
    printf("equinox: %f\n",(*store->equinox)[0][0]);

  if (store->latpole)
    printf("latpole: %f\n",(*store->latpole)[0][0]);
  if (store->lonpole)
    printf("lonpole: %f\n",(*store->lonpole)[0][0]);

  if (store->mjdobs)
    printf("mjdobs: %f\n",(*store->mjdobs)[0][0]);

  if (store->pv)
    printf("pv: %g %g %g %g\n",(*store->pv)[0][0],(*store->pv)[0][1],
	   (*store->pv)[1][0],(*store->pv)[1][1]);

  fflush(stdout);
}
*/
