#ifndef __marker_h__
#define __marker_h__

#include <iostream.h>
#include <strstream.h>
#include <iomanip.h>

#include <tk.h>

#include "vector.h"
#include "callback.h"
#include "list.h"
#include "coord.h"

class Frame;
class FitsImage;

static int sequenceID = 1;

class Marker {
public:
  enum Property {NONE=0, SELECT=1, EDIT=2, MOVE=4, ROTATE=8, DELETE=16, 
		 FIXED=32, INCLUDE=64, SOURCE=128};
  // Select-- user may select the marker
  // Edit-- user may edit the marker
  // Move-- user may move the marker
  // Rotate-- user may rotate the marker
  // Delete-- user may delete the marker
  // Fixed-- marker is fixed in size (not scaled based on zoom)
  // Include-- marker is 'included' or 'excluded' ie '+' or '-'
  // Source-- marker is a 'source' or 'background' marker

  enum Callback {SELECTCB, EDITCB, MOVECB, ROTATECB, DELETECB, 
		 TEXTCB, COLORCB, LINEWIDTHCB, PROPERTYCB, FONTCB};
  enum RenderMode {SRC,XOR};

protected:
  int id;
  char type[16];
  Vector center;                // ref coordinates
  BBox bbox;                    // canvas coordinates
  BBox allBBox;                 // canvas coordinates

  double angle;                 // radians

  Vector* handle;               // canvas coordinates 
  int numHandle;

  int color;
  char* colorName;
  int lineWidth;
  unsigned short properties;
  int selected;

  char* text;
  Tk_Font font;              // Tk font
  char* comment;

  Display* display;
  GC gc;

  int doCB;

  List<CallBack> selectCB;
  List<CallBack> moveCB;
  List<CallBack> editCB;
  List<CallBack> rotateCB;
  List<CallBack> deleteCB;
  List<CallBack> textCB;
  List<CallBack> colorCB;
  List<CallBack> lineWidthCB;
  List<CallBack> propertyCB;
  List<CallBack> fontCB;

  Frame* parent;
  Marker* previous_;
  Marker* next_;

protected:
  virtual void render(Drawable, const Matrix&, double, RenderMode) =0;
  
  Vector* arrow(const Vector&, const Vector&);
  void renderArrow(Drawable, const Vector&, const Vector&);
  void psArrow(const Vector&, const Vector&);

  void setGC(RenderMode);
  virtual void setLineDash();
  void setLineNoDash();
  virtual void psLineDash();
  void psLineNoDash();

  void doCallBack(List<CallBack>*);
  void psColor(int, const char*);
  void calcAllBBox();
  void renderHandles();
  void renderText(Drawable, const Matrix&, RenderMode);
  virtual void listProperties(ostream&, char);

  void listProps(ostream&);
  void listProsCoordSystem(ostream&, CoordSystem);

  void listPre(ostream&, CoordSystem,FitsImage*);
  void listPre2(ostream&, CoordSystem,FitsImage*);
  void listPost(ostream&, char);

  void listCiaoPre(ostream&);

  void listSAOtngPre(ostream&, char);
  void listSAOtngPost(ostream&, char);

public:
  Marker(const Marker&);
  Marker(Frame* p, const Vector& v, 
	 const char* clr, int w, const char* f, 
	 const char* t, unsigned short prop, const char* c);
  virtual ~Marker();

  virtual Marker* dup() =0;
  void newIdentity();
  virtual void updateBBox() =0;
  virtual void updateCoords(const Matrix&);

  virtual void draw();
  virtual void drawXOR();
  virtual void drawMagnifier(const Matrix&);

  virtual void move(const Vector& v);
  virtual void moveTo(const Vector& v);
  virtual void editBegin(int) {}
  virtual void edit(const Vector& v, int h) {}
  virtual void editEnd() {}
  virtual void rotate(const Vector& v, int h);
  virtual void ps(int);
  virtual int isIn(const Vector& v) {return bbox.isIn(v);}

  void setAngle(double);
  double getAngle() {return angle;}

  virtual int isPolygon() {return 0;}

  virtual void list(ostream&, CoordSystem, CoordFormat, char) =0;
  virtual void listCiao(ostream&, CoordSystem, CoordFormat, char) {}
  virtual void listSAOtng(ostream&, CoordSystem, CoordFormat, char) {}
  virtual void listSAOimage(ostream&, CoordSystem, CoordFormat, char) {}
  virtual void listPros(ostream&, CoordSystem, CoordFormat, char) {}
  virtual void listXY(ostream&, CoordSystem, CoordFormat, char);

  void listCoordSystem(ostream&, CoordSystem);
  void listCoordFormat(ostream&, CoordFormat);

  const BBox& getBBox() {return bbox;}
  const BBox& getAllBBox() {return allBBox;}
  int isVisible(const BBox&);

  void select();
  void unselect();
  void toggleSelect();
  int isSelected() {return selected;}

  void setColor(const char*);
  const char* getColorName() {return colorName;}

  void setLineWidth(int);
  int getLineWidth() {return lineWidth;}

  void setText(const char*);
  const char* getText() {return text;}
  void setFont(const char*);
  const char* getFont();

  int onHandle(const Vector& v);

  int getId() {return id;}
  const char* getType() {return type;}
  virtual Vector getCenter() {return center;}

  int canSelect() {return (properties & SELECT) ? 1:0;}
  int canEdit() {return (properties & EDIT) ? 1:0;}
  int canMove() {return (properties & MOVE) ? 1:0;}
  int canRotate() {return (properties & ROTATE) ? 1:0;}
  int canDelete() {return (properties & DELETE) ? 1:0;}
  int isFixed() {return (properties & FIXED) ? 1:0;}
  void setProperty(unsigned short, int);
  int getProperty(unsigned short);
  unsigned short getProperty() {return properties;}

  int setCallBack(Callback, const char*, const char*);
  int deleteCallBack(Callback, const char*);
  void deleteAllCallBack(Callback);
  void doCallBack(Callback);

  Marker* previous() {return previous_;}
  void setPrevious(Marker* m) {previous_ = m;}
  Marker* next() {return next_;}
  void setNext(Marker* m) {next_ = m;}

  void enableCB() {doCB = 1;}
  void disableCB() {doCB = 0;}
  void clearCB();
};

#endif
