%{
#define YYPARSE_PARAM fr
#define YYDEBUG 1

#define FR ((Frame*)fr)
#define FITS (FR->findFits(1))

#define DISCARD_(x) {yyclearin; saoDiscard(x);}

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "vector.h"
#include "list.h"
#include "basemarker.h"
#include "coord.h"

extern int saolex(void);
extern void saoerror(const char*);
extern void saoDiscard(int);

static unsigned short globalProps;
static unsigned short localProps;

static char *color = "green";
static char *font = "helvetica 10 normal";
static char *text = "";

static char localComment[80];

static List<Vertex> polylist;

static double aAnnuli[MAXANNULI];
static Vector aVector[MAXANNULI];
static int aNum;
static int aStatus;
static Vector aCenter;
static double aAngle;
static unsigned short aProps;
static char aComment[80];

static void setProps(unsigned short* props, unsigned short prop, int value);

%}

%union {
  double real;
  int integer;
  char str[256];
  double vector[3];
}

%type <real> numeric

%type <real> angle
%type <real> optangle
%type <real> value
%type <vector> coord
%type <integer> numberof

%token <integer> INT
%token <real> REAL
%token <str> STRING

%token ANNULUS_
%token BOX_
%token CIRCLE_
%token DEBUG_
%token ELLIPSE_
%token N_
%token OFF_
%token ON_
%token POINT_
%token POLYGON_
%token ROTBOX_
%token VERSION_

%%

start	: initGlobal commands processAnnulus
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "SAOimage" << endl;}
	| initLocal include shape
	| generalComment
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

optangle: /* empty */ {$$ = 0;}
	| angle {$$ = $1;}
	;

angle	: numeric {$$ = degToRad($1);} /* assume degree */
	;

value	: numeric
	  {$$ = FITS->mapLenToRef($1, IMAGE, PIXELS);}
	;

numberof: N_ '=' INT {$$ = $3;}
	;

coord	: numeric sp numeric 
	{
	  Vector r = FITS->map(Vector($1,$3), IMAGE, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	;

initGlobal:{
	  globalProps =
	    Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE | Marker::SOURCE;
	} 
	;

initLocal : {localProps = globalProps;}
	;

include	: /* empty */ {setProps(&localProps, Marker::INCLUDE, 1);}
	| '+' {setProps(&localProps, Marker::INCLUDE, 1);}
	| '-' {setProps(&localProps, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep shapeComment
	  {FR->createCircleCmd(Vector($3),$5,
	    color,1,font,text,localProps,localComment);}

	| ELLIPSE_ bp coord sp value sp value sp optangle ep shapeComment
	{
	  // for ellipse annulus
	  aStatus = 1;
	  aCenter = Vector($3);
	  aAngle = $9;
	  aVector[0] = Vector($5,$7);
	  aNum = 1;
	  strncpy(aComment,localComment,80);
	  aProps = localProps;

	  FR->createEllipseCmd(Vector($3),Vector($5,$7),$9,
	  color,1,font,text,localProps,localComment);
	}

	| BOX_ bp coord sp value sp value sp optangle ep shapeComment
	{
	  // for box annulus
	  aStatus = 3;
	  aCenter = Vector($3);
	  aAngle = $9;
	  aVector[0] = Vector($5,$7);
	  aNum = 1;
	  strncpy(aComment,localComment,80);
	  aProps = localProps;

	  FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    color,1,font,text,localProps,localComment);
	}
	| ROTBOX_ bp coord sp value sp value sp optangle ep shapeComment
	  {FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    color,1,font,text,localProps,localComment);}

	| POLYGON_ {polylist.deleteAll();} bp polyNodes ep shapeComment
	  {FR->createPolygonCmd(polylist,color,1,font,
	      text,localProps,localComment);} 

	| POINT_ bp coord ep shapeComment
	  {FR->createBoxCirclePointCmd(Vector($3),
	    color,1,font,text,localProps,localComment);}

	| ANNULUS_ bp coord sp value sp value ep shapeComment
	  {FR->createAnnulusCmd(Vector($3),$5,$7,1,
	    color,1,font,text,localProps,localComment);}
	| ANNULUS_ bp coord sp value sp value sp {aNum=2;} aRads ep
	   shapeComment
	{
	  aAnnuli[0] = $5;
	  aAnnuli[1] = $7;
	  FR->createAnnulusCmd(Vector($3),aAnnuli,aNum,
	  color,1,font,text,localProps,localComment);
	}
	| ANNULUS_ bp coord sp value sp value sp numberof ep shapeComment
	  {FR->createAnnulusCmd(Vector($3),$5,$7,$9,
	    color,1,font,text,localProps,localComment);}

	| ELLIPSE_ bp coord sp value sp value sp optangle ep '&' '!' 
	  ELLIPSE_ bp coord sp value sp value sp optangle ep
	{	
	  aStatus = 2;
	  aVector[aNum++] = Vector($5,$7);
	}

	| BOX_ bp coord sp value sp value sp optangle ep '&' '!' 
	  BOX_ bp coord sp value sp value sp optangle ep
	{	
	  aStatus = 4;
	  aVector[aNum++] = Vector($5,$7);
	}
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist.append(new Vertex($1));}
	;

aRads	: aRads sp aRad
	| aRad
	;

aRad	: value {aAnnuli[aNum++] = $1;}
	;

processAnnulus : /* empty */
	{
	  switch (aStatus) {
	  case 0: // do nothing
	    break;
	  case 1: // we found just an ellipse, do nothing
	    break;
	  case 2: // ok we have an ellipse annulus
	    FR->markerDeleteLastCmd(); // delete the previous ellipse
	    FR->createEllipseAnnulusCmd(aCenter,aAngle,aNum,aVector,
	      color,1,font,text,aProps,aComment);
	    break;
	  case 3: // we found just a box, do nothing
	    break;
	  case 4: // ok, we have a box annulus
	    FR->markerDeleteLastCmd(); // delete the previous box
	    FR->createBoxAnnulusCmd(aCenter,aAngle,aNum,aVector,
	      color,1,font,text,aProps,aComment);
	    break;
	  }
	  aStatus = 0;
	}

generalComment : '#' {DISCARD_(1)} STRING 
	;

shapeComment : /* empty */ processAnnulus
	| '#' {DISCARD_(0)} STRING processAnnulus 
	  {strncpy(localComment,$3,80);}
	;

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}
