%{
#define YYPARSE_PARAM fr
#define YYDEBUG 1

#define FR ((Frame*)fr)
#define FITS (FR->findFits(1))

#define DISCARD_(x) {yyclearin; tngDiscard(x);}

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "vector.h"
#include "list.h"
#include "basemarker.h"
#include "coord.h"

extern int tnglex(void);
extern void tngerror(const char*);
extern void tngDiscard(int);

int tngsign;
int tngsign2;

static CoordSystem globalSystem;
static CoordSystem localSystem;

static unsigned short globalProps;
static unsigned short localProps;

static char globalColor[16];
static char localColor[16];

static char globalFont[32];

static char globalText[80];
static char localText[80];

static char localComment[80];

static List<Vertex> polylist;

static void setProps(unsigned short* props, unsigned short prop, int value);
static CoordSystem checkWCSSystem();

%}

%union {
  double real;
  int integer;
  char str[256];
  double vector[3];
}

%type <real> numeric

%type <real> angle
%type <real> optangle
%type <real> value
%type <real> sexagesimal
%type <vector> coord
%type <integer> wcsSystem

%token <integer> INT
%token <real> REAL
%token <str> STRING

%token <real> DEGREE

%token <str> SEXSTR

%token BACKGROUND_
%token BOX_
%token BLACK_
%token BLUE_
%token CIRCLE_
%token CYAN_
%token DATE_
%token DEBUG_
%token DEGREES_
%token ECLIPTIC_
%token ELLIPSE_
%token FILENAME_
%token FK4_
%token FK5_
%token FORMAT_
%token GALACTIC_
%token GREEN_
%token HMS_
%token ICRS_
%token LINE_
%token LINEAR_
%token MAGENTA_
%token OFF_
%token ON_
%token PHYSICAL_
%token PIXELS_
%token POINT_
%token POLYGON_
%token RED_
%token SOURCE_
%token TEXT_
%token VERSION_
%token WHITE_
%token YELLOW_

%%

start	: initGlobal commands
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "SAOtng" << endl;}
	| initLocal include shape
	| generalComment
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

sp	: /* empty */
	| ','
	;

bp	: /* empty */
	| '('
	;

ep	: /* emtpy */
	| ')'
	;

optangle: /* empty */ {$$ = 0;}
	| angle {$$ = $1;}
	;

angle	: numeric {$$ = degToRad($1);} /* assume degree */
	| DEGREE {$$ = degToRad($1);}
	;

value	: numeric
	  {$$ = FITS->mapLenToRef($1, IMAGE, PIXELS);}
	| DEGREE
	  {$$ = FITS->mapLenToRef($1, checkWCSSystem(), DEGREES);}
	;

sexagesimal: SEXSTR {$$ = parseDMS($1);}

coord	: sexagesimal sp sexagesimal
	{
	  localSystem = checkWCSSystem();
	  Vector r;
	  if (localSystem == GALACTIC ||
	      localSystem == ECLIPTIC) 
	    r = FITS->map(Vector($1,$3), localSystem, REF);
	  else
	    r = FITS->map(Vector($1*360./24.,$3), localSystem, REF);

	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| numeric sp numeric 
	{
	  Vector r = FITS->map(Vector($1,$3), localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| DEGREE sp DEGREE
	{
	  Vector r = FITS->map(Vector($1,$3), checkWCSSystem(), REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	;

wcsSystem : FK4_ {$$ = EQUATORIAL_FK4;}
	| FK5_ {$$ = EQUATORIAL_FK5;}
	| GALACTIC_ {$$ = GALACTIC;}
	| ECLIPTIC_ {$$ = ECLIPTIC;}
	| LINEAR_ {$$ = LINEAR;}
	| ICRS_ {$$ = ICRS;}
	;

initGlobal:{
	  // global properties
	  globalSystem = IMAGE;
	  globalProps =
	    Marker::SELECT | Marker::EDIT | Marker::MOVE |
	    Marker::ROTATE | Marker::DELETE | Marker::INCLUDE | Marker::SOURCE;
	  strcpy(globalColor,"green");
	  strcpy(globalFont,"helvetica 10 normal");
	  strcpy(globalText,"");
	} 
	;

initLocal : {
	  // global properties
	  localSystem = globalSystem;
	  localProps = globalProps;
	  strcpy(localColor,globalColor);
	  strcpy(localText,globalText);
	  strcpy(localComment,"");

	  strcpy(globalText,"");
	}
	;

include	: /* empty */ {setProps(&localProps, Marker::INCLUDE, 1);}
	| '+' {setProps(&localProps, Marker::INCLUDE, 1);}
	| '-' {setProps(&localProps, Marker::INCLUDE, 0);}
	;

shape	: CIRCLE_ bp coord sp value ep shapeComment
	  {FR->createCircleCmd(Vector($3),$5,
	    localColor,1,globalFont,localText,localProps,localComment);}

	| ELLIPSE_ bp coord sp value sp value sp optangle ep shapeComment
	  {FR->createEllipseCmd(Vector($3),Vector($5,$7),$9,
	    localColor,1,globalFont,localText,localProps,localComment);}

	| BOX_ bp coord sp value sp value sp optangle ep shapeComment
	  {FR->createBoxCmd(Vector($3),Vector($5,$7),$9,
	    localColor,1,globalFont,localText,localProps,localComment);}

	| POLYGON_ {polylist.deleteAll();} bp polyNodes ep shapeComment
	  {FR->createPolygonCmd(polylist,localColor,1,globalFont,
	      localText,localProps,localComment);} 

	| LINE_ bp coord sp coord ep shapeComment
	  {FR->createLineCmd(Vector($3),Vector($5),
	    0,0,localColor,1,globalFont,localText,localProps,localComment);}

	| TEXT_ bp coord sp STRING ep {strncpy(localText,$5,80);} shapeComment
	  {FR->createTextCmd(Vector($3),0,
	    localColor,1,globalFont,localText,localProps,localComment);}

	| POINT_ bp coord ep shapeComment
	  {FR->createBoxCirclePointCmd(Vector($3),
	    localColor,1,globalFont,localText,localProps,localComment);}
	;

polyNodes : polyNodes sp polyNode
	| polyNode
	;

polyNode : coord {polylist.append(new Vertex($1));}
	;

generalComment : '#' FORMAT_ ':' tngFormat
	| '#' FILENAME_ ':' {DISCARD_(0)} STRING
	| '#' DATE_ ':' {DISCARD_(0)} STRING
	| '#' {DISCARD_(1)} STRING {strncpy(globalText,$3,80);}
	;

shapeComment : /* empty */
	| '#' WHITE_ {strcpy(localColor,"white");}
	| '#' BLACK_ {strcpy(localColor,"black");}
	| '#' RED_ {strcpy(localColor,"red");}
	| '#' GREEN_ {strcpy(localColor,"green");}
	| '#' BLUE_ {strcpy(localColor,"blue");}
	| '#' CYAN_ {strcpy(localColor,"cyan");}
	| '#' MAGENTA_ {strcpy(localColor,"magenta");}
	| '#' YELLOW_ {strcpy(localColor,"yellow");}
	| '#' SOURCE_ 
	{
	  strcpy(localColor,"green");
	  setProps(&localProps,Marker::SOURCE,1);
	}
	| '#' BACKGROUND_ 
	{
	  strcpy(localColor,"red");
	  setProps(&localProps,Marker::SOURCE,0);
	}
	| '#' {DISCARD_(1)} STRING {strncpy(localComment,$3,80);}
	;

tngFormat : HMS_ tngWCS
	| DEGREES_ tngWCS
	| PIXELS_ {globalSystem = localSystem = IMAGE;}
	| PIXELS_ '(' PHYSICAL_ ')' {globalSystem = localSystem = IMAGE;}
	;

tngWCS	: /* empty */
	  {globalSystem = localSystem = NATIVEWCS;}
	| '(' wcsSystem ')' 
	  {globalSystem = localSystem = (CoordSystem)$2;}
	;

%%

static void setProps(unsigned short* props, unsigned short prop, int value)
{
  if (value)
    *props |= prop;
  else
    *props &= ~prop;
}

static CoordSystem checkWCSSystem()
{
  // if we have not defined a specific WCS, set to Native
  switch (localSystem) {
    case (IMAGE):
    case (PHYSICAL):
      return (NATIVEWCS);
    default:
      return localSystem;
  }
}
