%{
#define YYPARSE_PARAM fr
#define YYDEBUG 1

#define FR ((Frame*)fr)
#define FITS (FR->findFits(1))

#include <math.h>
#include <string.h>

#include <iostream.h>

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "vector.h"
#include "list.h"
#include "basemarker.h"
#include "coord.h"

extern int xylex(void);
extern void xyerror(const char*);

int xysign;
int xysign2;

static CoordSystem globalSystem;
static CoordSystem localSystem;

static CoordSystem checkWCSSystem();
	
%}

%union {
  double real;
  int integer;
  char str[256];
  double vector[3];
}

%type <real> numeric

%type <real> sexagesimal
%type <vector> coord
%type <integer> coordSystem

%token <integer> INT
%token <real> REAL

%token <integer> HOUR
%token <integer> MINUTE
%token <real> SECOND
%token <real> DEGREE
%token <real> ARCMINUTE
%token <real> ARCSECOND

%token <str> SEXSTR

%token AMPLIFIER_
%token B1950_
%token CCD_
%token DEBUG_
%token DETECTOR_
%token ECLIPTIC_
%token FK4_
%token FK5_
%token GALACTIC_
%token ICRS_
%token IMAGE_
%token J2000_
%token LINEAR_
%token LOGICAL_
%token OFF_
%token ON_
%token PHYSICAL_
%token VERSION_

%%

start	: {globalSystem = FR->getMarkerDefaultSystem();} commands
	;

commands: commands command terminator
	| command terminator
	;

command : /* empty */
	| DEBUG_ debug
	| VERSION_ {cerr << "X Y Format 1.0" << endl;}
	| coordSystem {globalSystem = (CoordSystem)$1;}
	| {localSystem = globalSystem;} shape comment
	;

comment	: /* empty */
	| junks
	;

junks	: junks junk
	| junk
	;

junk	:  numeric {}
	;

terminator: '\n'
	| ';'
	;

numeric	: REAL {$$=$1;}
	| INT {$$=$1;}
	;

debug	: ON_ {yydebug=1;}
	| OFF_ {yydebug=0;}
	;

sp	: /* empty */
	| ','
	;

sexagesimal: SEXSTR {$$ = parseDMS($1);}

coord	: sexagesimal sp sexagesimal
	{
	  localSystem = checkWCSSystem();
	  Vector r;
	  if (localSystem == GALACTIC ||
	      localSystem == ECLIPTIC) 
	    r = FITS->map(Vector($1,$3), localSystem, REF);
	  else
	    r = FITS->map(Vector($1*360./24.,$3), localSystem, REF);

	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| HOUR MINUTE SECOND {xysign2 = xysign;} sp INT ARCMINUTE ARCSECOND
        {
	  localSystem = checkWCSSystem();
	  Vector r = FITS->map(
	    Vector(hmsToDegree(xysign2,$1,$2,$3), 
	    dmsToDegree(xysign,$6,$7,$8)),
	    localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| INT ARCMINUTE ARCSECOND {xysign2 = xysign;} sp 
	  INT ARCMINUTE ARCSECOND
        {
	  localSystem = checkWCSSystem();
	  Vector r = FITS->map(
	    Vector(dmsToDegree(xysign2,$1,$2,$3),
	    dmsToDegree(xysign,$6,$7,$8)),
	    localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| numeric sp numeric 
	{
	  Vector r = FITS->map(Vector($1,$3), localSystem, REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	| DEGREE sp DEGREE
	{
	  Vector r = FITS->map(Vector($1,$3), checkWCSSystem(), REF);
	  $$[0] = r[0];
	  $$[1] = r[1];
	  $$[2] = r[2];
	}
	;

coordSystem : IMAGE_ {$$ = IMAGE;}
	| LOGICAL_ {$$ = IMAGE;}
	| PHYSICAL_ {$$ = PHYSICAL;}
	| CCD_ {$$ = PHYSICAL;}
	| AMPLIFIER_ {$$ = AMPLIFIER;}
	| DETECTOR_ {$$ = DETECTOR;}
	| FK4_ {$$ = EQUATORIAL_FK4;}
	| B1950_  {$$ = EQUATORIAL_FK4;}
	| FK5_ {$$ = EQUATORIAL_FK5;}
	| J2000_  {$$ = EQUATORIAL_FK5;}	
	| ICRS_ {$$ = ICRS;}
	| GALACTIC_ {$$ = GALACTIC;}
	| ECLIPTIC_ {$$ = ECLIPTIC;}
	| LINEAR_ {$$ = LINEAR;}
	;

shape	: coord {FR->createBoxCirclePointCmd(Vector($1),
	    "green",1,"helvetica 10 normal","",
	    Marker::SELECT | Marker::EDIT | Marker::MOVE | Marker::ROTATE | 
	    Marker::DELETE | Marker::INCLUDE | Marker::SOURCE,
	    "");
	}
	;

%%

static CoordSystem checkWCSSystem()
{
  // if we have not defined a specific WCS, set to Native
  switch (localSystem) {
    case (IMAGE):
    case (PHYSICAL):
    case (DETECTOR):
    case (AMPLIFIER):
      return (NATIVEWCS);
    default:
      return localSystem;
  }
}
