/*
 *      Copyrights:
 *
 *    	  Copyright (c) 1996, 1998, 2002 Smithsonian Astrophysical Observatory
 *
 *        Permission to use, copy, modify, distribute, and  sell  this
 *        software  and  its  documentation  for any purpose is hereby
 *        granted without  fee,  provided  that  the  above  copyright
 *        notice  appear  in  all  copies and that both that copyright
 *        notice and this permission notice appear in supporting docu-
 *        mentation,  and  that  the  name  of the  Smithsonian Astro-
 *	  physical Observatory not be used in advertising or publicity
 *  	  pertaining to distribution of the software without specific,
 *	  written  prior  permission.   The Smithsonian  Astrophysical 
 *        Observatory makes no representations about  the  suitability
 *	  of  this  software for any purpose.  It is provided  "as is"
 *	  without express or implied warranty.
 *        THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  DISCLAIMS  ALL
 *	  WARRANTIES  WITH  REGARD  TO  THIS  SOFTWARE,  INCLUDING ALL
 *	  IMPLIED  WARRANTIES  OF  MERCHANTABILITY AND FITNESS, IN  NO
 *	  EVENT SHALL THE  SMITHSONIAN  ASTROPHYSICAL  OBSERVATORY  BE 
 *	  LIABLE FOR  ANY SPECIAL, INDIRECT  OR  CONSEQUENTIAL DAMAGES
 *	  OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS  OF USE,  DATA
 *	  OR  PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 *	  OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
 *	  THE  USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <xpap.h>

/*
 *----------------------------------------------------------------------------
 *
 *
 * 			Private Routines and Data
 *
 *
 *----------------------------------------------------------------------------
 */

static int alarm_flag=0;
static struct sigaction act1, oact1;
static struct sigaction act2, oact2;

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAlarmFunc
 *
 * Purpose:	function to call when alarm goes off
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
XPAAlarmFunc (int signo)
#else
static void XPAAlarmFunc(signo)
	int signo;
#endif
{
  alarm_flag = 1;
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAlarmCheck
 *
 * Purpose:	function to call when alarm goes off
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int
XPAAlarmCheck (void)
#else
static int XPAAlarmCheck()
#endif
{
  return(alarm_flag);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAlarmSetup
 *
 * Purpose:	set up the AlarmSetup handler
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static int
XPAAlarmSetup (Sigfunc *func1, Sigfunc *func2)
#else
static int XPAAlarmSetup(func1, func2)
	Sigfunc *func1;
	Sigfunc *func2;
#endif
{
  alarm_flag = 0;
  if( func1 != NULL ){
    /* set up the signal handler */
    act1.sa_handler = func1;
    sigemptyset(&act1.sa_mask);
    act1.sa_flags = 0;
#ifdef SA_INTERRUPT
    act1.sa_flags |= SA_INTERRUPT;
#endif
    if( sigaction(SIGALRM, &act1, &oact1) < 0 )
      return(-1);
  }

  if( func2 != NULL ){
    if( XPASigusr1() ){
      act2.sa_handler = func2;
      sigemptyset(&act2.sa_mask);
      act2.sa_flags = 0;
#ifdef SA_INTERRUPT
      act2.sa_flags |= SA_INTERRUPT;
#endif
      if( sigaction(SIGUSR1, &act2, &oact2) < 0 )
	return(-1);
    }
  }
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAlarmStart
 *
 * Purpose:	set up sigaction and start alarm
 *
 * Returns:	0 for success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
XPAAlarmStart (int seconds)
#else
static void XPAAlarmStart(seconds)
     int seconds;
#endif
{
  /* Disable SIGPIPE so we do not die if the client dies.
   * Rather, we will get an EOF on reading or writing.
   */
  signal(SIGPIPE, SIG_IGN);

  /* set up the signal callbacks */
  XPAAlarmSetup(XPAAlarmFunc, XPAAlarmFunc);

  /* start the alarm, if necessary */
  if( seconds > 0 )
    alarm(seconds);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAAlarmEnd
 *
 * Purpose:	stop the alarm and restore
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
static void
XPAAlarmEnd (void)
#else
static void XPAAlarmEnd()
#endif
{
  /* turn off the alarm */
  alarm(0);

  /* restore old sigaction */
  XPAAlarmSetup(oact1.sa_handler, oact2.sa_handler);
}

/*
 *----------------------------------------------------------------------------
 *
 *
 * 			Public Routines and Data
 *
 *
 *----------------------------------------------------------------------------
 */

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	rdl
 *
 * Purpose:	read characters up a new-line
 *
 * Returns:	number of characters read
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
rdl (int fd, char *buf, int len)
#else
int rdl(fd, buf, len)
     int fd;
     char *buf;
     int len;
#endif
{
  int i=0;
  int got;

  /* start out clean */
  *buf = '\0';

  /* make sure we have a valid channel */
  if( fd < 0 )
    return(-1);

  /* grab characters up to a new-line or max len */
  while( i < (len-1) ){
    got = read(fd, &(buf[i]), 1);
    if( got < 1 ){
#ifdef XPA_DEBUG
	if( (got == 0) )
	  fprintf(stderr, "rdl: read EOF\n");
	else
	  perror("rdl");
#endif
      break;
    }
    else if( buf[i++] == '\n' ){
      break;
    }
  }
  buf[i] = '\0';
  return(i);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAGets
 *
 * Purpose:	read a line of characters up to a newline -- sockets only
 *
 * Returns:	number of characters read
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAGets (int fd, char *buf, int len, int timeout)
#else
int XPAGets(fd, buf, len, timeout)
     int fd;
     char *buf;
     int len;
     int timeout;
#endif
{
  int got;

  /* start out pessimistically */
  *buf = '\0';

  /* make sure we have a valid channel */
  if( fd <0 )
    return(-1);

  /* start the alarm */
  /* this is done in XPARdl */
  /* XPAAlarmStart(timeout); */

  /* read the line */
  /* can't use stdio with sockets under windows ... */
  got = XPARdl(fd, buf, len, timeout);

  /* stop the alarm */
  /* this is done in XPARdl */
  /* XPAAlarmEnd(); */

  /* return error or length of string */
  if( XPAAlarmCheck() )
    return(-1);
  else
    return(got);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAPuts
 *
 * Purpose:	write a line of characters up to a newline -- sockets only
 *
 * Returns:	number of characters written
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAPuts (int fd, char *buf, int timeout)
#else
int XPAPuts(fd, buf, timeout)
     int fd;
     char *buf;
     int timeout;
#endif
{
  int n;

  /* make sure we have a valid channel */
  if( fd < 0 )
    return(-1);

  /* start the alarm */
  XPAAlarmStart(timeout);

  /* write the line */
  n=send(fd, buf, strlen(buf), 0);

  /* stop the alarm */
  XPAAlarmEnd();

  /* return error or length of string */
  if( XPAAlarmCheck() )
    return(-1);
  else
    return(n);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAGetBuf
 *
 * Purpose:	read data from the xpa data socket until EOF -- sockets only
 *
 * Results:	0 on success, -1 on failure
 *		also allocates memory for buf and returns len
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAGetBuf (int fd, char **buf, int *len, int timeout)
#else
int XPAGetBuf(fd, buf, len, timeout)
     int fd;
     char **buf;
     int *len;
     int timeout;
#endif
{
  int get;
  int got;
  int slen;
  int total;
  int status;
  char *s;

  /* make sure we have a valid data channel */
  if( fd < 0 )
    return(-1);

  /* make sure we have a valid buf ptr and len ptr */
  if( (len == NULL) || (buf == NULL) ){
    return(-1);
  }

#ifdef XPA_DEBUG
  fprintf(stderr, "XPAGetBuf: entering fd %d\n", fd);
#endif

  /* allocate the first buffer, if necessary */
  if( (*len != 0) && (*buf != NULL) ){
    slen = *len;
    total = *len;
    s = *buf;
  }
  else{
    slen = XPA_IOSIZE;
    total = 0;
    if( (s=(char *)malloc(slen)) == NULL )
      return(0);
  }
  get = XPA_IOSIZE;

  /* start the alarm */
  XPAAlarmStart(timeout);

  /* read in batches until done */
  while( 1 ){
    while( (total + get) > slen ){
      slen += (get*10);
      if( (s = (char *)realloc(s, slen)) == NULL ){
	status = -1;
	break;
      }
    }
    /* read all the bytes in this chunk */
#ifdef XPA_DEBUG
    fprintf(stderr, "XPAGetBuf: trying to read %d bytes from %d\n", get, fd);
#endif
    got = recv(fd, &(s[total]), get, 0);
#ifdef XPA_DEBUG
    fprintf(stderr, "XPAGetBuf: read %d bytes\n", got);
    if( got == -1 )
      perror("XPAGetBuf");
#endif
    /* check for error or timeout */
    if( (got < 0) || XPAAlarmCheck() ){
      status = -1;
      break;
    }
    /* end of file, all done */
    else if( got == 0){
      status = 0;
      break;
    }
    /* got data */
    else{
      total += got;
    }
  }

  /* stop the alarm */
  XPAAlarmEnd();

  /* null terminate after last byte as a courtesy */
  if( status == 0 ){
    s = (char *)realloc(s, total+1);
    s[total] = '\0';
    *buf = s;
    *len = total;
  }
  else{
    if( s )
      free(s);
    *buf = NULL;
    *len = 0;
  }
  return(status);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPAPutBuf
 *
 * Purpose:	write data to the xpa data socket -- sockets only
 *
 * Results:	0 on success, -1 for failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPAPutBuf (int fd, char *buf, int len, int timeout)
#else
int XPAPutBuf(fd, buf, len, timeout)
     int fd;
     char *buf;
     int len;
     int timeout;
#endif
{
  int cur;
  int left;
  int got;
  int n;
  int status=0;

  /* make sure we have a valid data channel */
  if( fd < 0 )
    return(-1);

  /* make sure we have a valid buf ptr and len ptr */
  if( (len == 0) || (buf == NULL) ){
    return(-1);
  }

  /* start the alarm */
  XPAAlarmStart(timeout);

#ifdef XPA_DEBUG
    fprintf(stderr, "XPAPutBuf: writing %d bytes to %d\n", len, fd);
#endif

  /* write in batches until done */
  for(cur=0, left=len; left > 0; cur += got, left -= got){
    /* calculate size of next chunk */
    if( (n = MIN(left, XPA_IOSIZE)) <=0 )
      break;
    /* write the data */
    got = send(fd, &(buf[cur]), n, 0);
    /* check for error or timeout */
    if( (got < 0) || XPAAlarmCheck() ){
#ifdef XPA_DEBUG
      perror("XPAPutBuf");
#endif
      status = -1;
      break;
    }
  }

  /* stop the alarm */
  XPAAlarmEnd();

  return(status);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPASetBuf
 *
 * Purpose:	set the return buffer in a server send callback
 *		(use mainly by tcl to transfer tcl buffers to C)
 *
 * Returns:	0 on success, -1 on failure
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPASetBuf (XPA xpa, char *buf, int len, int xcopy)
#else
int XPASetBuf(xpa, buf, len, xcopy)
     XPA xpa;
     char *buf;
     int len;
     int xcopy;
#endif
{
  /* xcopy >0 => make a copy of the data, else just assign */
  if( xcopy ){
    xpa->len = len;
    if( (xpa->buf = (char *)malloc(len)) == NULL )
      return(-1);
    memcpy(xpa->buf, buf, len);
  }
  else{
    xpa->len = len;
    xpa->buf = buf;
  }
  return(0);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPARdl
 *
 * Purpose:	read characters up a new-line -- sockets only
 *
 * Returns:	number of characters read
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
int 
XPARdl (int fd, char *buf, int len, int timeout)
#else
int XPARdl(fd, buf, len, timeout)
     int fd;
     char *buf;
     int len;
     int timeout;
#endif
{
  int i=0;
  int got;

  *buf = '\0';

  /* make sure we have a valid channel */
  if( fd < 0 )
    return(-1);

  /* start the alarm */
  XPAAlarmStart(timeout);

  /* grab characters up to a new-line or max len */
  while( i < (len-1) ){
    got = recv(fd, &(buf[i]), 1, 0);
    if( (got < 1) || XPAAlarmCheck() ){
#ifdef XPA_DEBUG
      if( (got == 0) )
	fprintf(stderr, "XPARdl: read %d EOF\n", fd);
      else if( XPAAlarmCheck() )
	fprintf(stderr, "XPARdl: %d timed out\n", fd);
      else{
	char tbuf[SZ_LINE];
	sprintf(tbuf, "XPARdl: error %d", fd);
	perror(tbuf);
      }
#endif
      break;
    }
    else if( buf[i++] == '\n' ){
      break;
    }
  }

  /* stop the alarm */
  XPAAlarmEnd();

  buf[i] = '\0';
#ifdef XPA_DEBUG
  if( i >= 1 )
    fprintf(stderr, "XPARdl: %d read %d bytes:\n%.79s", fd, i, buf);
  if( i >=79 )
    fprintf(stderr, "\n");
#endif
  return(i);
}

/*
 *----------------------------------------------------------------------------
 *
 * Routine:	XPADefaultSigaction
 *
 * Purpose:	set up default actions for signals
 *
 * Returns:	none
 *
 *----------------------------------------------------------------------------
 */
#ifdef ANSI_FUNC
void 
XPADefaultSigaction (void)
#else
void XPADefaultSigaction()
#endif
{

  /* Disable SIGPIPE so we do not die if the client dies.
   * Rather, we will get an EOF on reading or writing.
   */
  signal(SIGPIPE, SIG_IGN);

  /* set up usr1 action */
  XPAAlarmSetup(NULL, XPAAlarmFunc);
}

